use core::fmt::{self, Debug, Display, Formatter};
use core::str::FromStr;

use crate::timezone_impl::{TimeSpans, FixedTimespanSet, FixedTimespan};

/// TimeZones built at compile time from the tz database
///
/// This implements [`chrono::TimeZone`] so that it may be used in and to
/// construct chrono's DateTime type. See the root module documentation
/// for details.
#[derive(Clone, Copy, PartialEq, Eq, Hash)]
#[cfg_attr(feature = "arbitrary", derive(arbitrary::Arbitrary))]
pub enum Tz {
    /// Africa/Abidjan
    Africa__Abidjan,
    /// Africa/Accra
    Africa__Accra,
    /// Africa/Addis_Ababa
    Africa__Addis_Ababa,
    /// Africa/Algiers
    Africa__Algiers,
    /// Africa/Asmara
    Africa__Asmara,
    /// Africa/Asmera
    Africa__Asmera,
    /// Africa/Bamako
    Africa__Bamako,
    /// Africa/Bangui
    Africa__Bangui,
    /// Africa/Banjul
    Africa__Banjul,
    /// Africa/Bissau
    Africa__Bissau,
    /// Africa/Blantyre
    Africa__Blantyre,
    /// Africa/Brazzaville
    Africa__Brazzaville,
    /// Africa/Bujumbura
    Africa__Bujumbura,
    /// Africa/Cairo
    Africa__Cairo,
    /// Africa/Casablanca
    Africa__Casablanca,
    /// Africa/Ceuta
    Africa__Ceuta,
    /// Africa/Conakry
    Africa__Conakry,
    /// Africa/Dakar
    Africa__Dakar,
    /// Africa/Dar_es_Salaam
    Africa__Dar_es_Salaam,
    /// Africa/Djibouti
    Africa__Djibouti,
    /// Africa/Douala
    Africa__Douala,
    /// Africa/El_Aaiun
    Africa__El_Aaiun,
    /// Africa/Freetown
    Africa__Freetown,
    /// Africa/Gaborone
    Africa__Gaborone,
    /// Africa/Harare
    Africa__Harare,
    /// Africa/Johannesburg
    Africa__Johannesburg,
    /// Africa/Juba
    Africa__Juba,
    /// Africa/Kampala
    Africa__Kampala,
    /// Africa/Khartoum
    Africa__Khartoum,
    /// Africa/Kigali
    Africa__Kigali,
    /// Africa/Kinshasa
    Africa__Kinshasa,
    /// Africa/Lagos
    Africa__Lagos,
    /// Africa/Libreville
    Africa__Libreville,
    /// Africa/Lome
    Africa__Lome,
    /// Africa/Luanda
    Africa__Luanda,
    /// Africa/Lubumbashi
    Africa__Lubumbashi,
    /// Africa/Lusaka
    Africa__Lusaka,
    /// Africa/Malabo
    Africa__Malabo,
    /// Africa/Maputo
    Africa__Maputo,
    /// Africa/Maseru
    Africa__Maseru,
    /// Africa/Mbabane
    Africa__Mbabane,
    /// Africa/Mogadishu
    Africa__Mogadishu,
    /// Africa/Monrovia
    Africa__Monrovia,
    /// Africa/Nairobi
    Africa__Nairobi,
    /// Africa/Ndjamena
    Africa__Ndjamena,
    /// Africa/Niamey
    Africa__Niamey,
    /// Africa/Nouakchott
    Africa__Nouakchott,
    /// Africa/Ouagadougou
    Africa__Ouagadougou,
    /// Africa/Porto-Novo
    Africa__PortoNovo,
    /// Africa/Sao_Tome
    Africa__Sao_Tome,
    /// Africa/Timbuktu
    Africa__Timbuktu,
    /// Africa/Tripoli
    Africa__Tripoli,
    /// Africa/Tunis
    Africa__Tunis,
    /// Africa/Windhoek
    Africa__Windhoek,
    /// America/Adak
    America__Adak,
    /// America/Anchorage
    America__Anchorage,
    /// America/Anguilla
    America__Anguilla,
    /// America/Antigua
    America__Antigua,
    /// America/Araguaina
    America__Araguaina,
    /// America/Argentina/Buenos_Aires
    America__Argentina__Buenos_Aires,
    /// America/Argentina/Catamarca
    America__Argentina__Catamarca,
    /// America/Argentina/ComodRivadavia
    America__Argentina__ComodRivadavia,
    /// America/Argentina/Cordoba
    America__Argentina__Cordoba,
    /// America/Argentina/Jujuy
    America__Argentina__Jujuy,
    /// America/Argentina/La_Rioja
    America__Argentina__La_Rioja,
    /// America/Argentina/Mendoza
    America__Argentina__Mendoza,
    /// America/Argentina/Rio_Gallegos
    America__Argentina__Rio_Gallegos,
    /// America/Argentina/Salta
    America__Argentina__Salta,
    /// America/Argentina/San_Juan
    America__Argentina__San_Juan,
    /// America/Argentina/San_Luis
    America__Argentina__San_Luis,
    /// America/Argentina/Tucuman
    America__Argentina__Tucuman,
    /// America/Argentina/Ushuaia
    America__Argentina__Ushuaia,
    /// America/Aruba
    America__Aruba,
    /// America/Asuncion
    America__Asuncion,
    /// America/Atikokan
    America__Atikokan,
    /// America/Atka
    America__Atka,
    /// America/Bahia
    America__Bahia,
    /// America/Bahia_Banderas
    America__Bahia_Banderas,
    /// America/Barbados
    America__Barbados,
    /// America/Belem
    America__Belem,
    /// America/Belize
    America__Belize,
    /// America/Blanc-Sablon
    America__BlancSablon,
    /// America/Boa_Vista
    America__Boa_Vista,
    /// America/Bogota
    America__Bogota,
    /// America/Boise
    America__Boise,
    /// America/Buenos_Aires
    America__Buenos_Aires,
    /// America/Cambridge_Bay
    America__Cambridge_Bay,
    /// America/Campo_Grande
    America__Campo_Grande,
    /// America/Cancun
    America__Cancun,
    /// America/Caracas
    America__Caracas,
    /// America/Catamarca
    America__Catamarca,
    /// America/Cayenne
    America__Cayenne,
    /// America/Cayman
    America__Cayman,
    /// America/Chicago
    America__Chicago,
    /// America/Chihuahua
    America__Chihuahua,
    /// America/Ciudad_Juarez
    America__Ciudad_Juarez,
    /// America/Coral_Harbour
    America__Coral_Harbour,
    /// America/Cordoba
    America__Cordoba,
    /// America/Costa_Rica
    America__Costa_Rica,
    /// America/Creston
    America__Creston,
    /// America/Cuiaba
    America__Cuiaba,
    /// America/Curacao
    America__Curacao,
    /// America/Danmarkshavn
    America__Danmarkshavn,
    /// America/Dawson
    America__Dawson,
    /// America/Dawson_Creek
    America__Dawson_Creek,
    /// America/Denver
    America__Denver,
    /// America/Detroit
    America__Detroit,
    /// America/Dominica
    America__Dominica,
    /// America/Edmonton
    America__Edmonton,
    /// America/Eirunepe
    America__Eirunepe,
    /// America/El_Salvador
    America__El_Salvador,
    /// America/Ensenada
    America__Ensenada,
    /// America/Fort_Nelson
    America__Fort_Nelson,
    /// America/Fort_Wayne
    America__Fort_Wayne,
    /// America/Fortaleza
    America__Fortaleza,
    /// America/Glace_Bay
    America__Glace_Bay,
    /// America/Godthab
    America__Godthab,
    /// America/Goose_Bay
    America__Goose_Bay,
    /// America/Grand_Turk
    America__Grand_Turk,
    /// America/Grenada
    America__Grenada,
    /// America/Guadeloupe
    America__Guadeloupe,
    /// America/Guatemala
    America__Guatemala,
    /// America/Guayaquil
    America__Guayaquil,
    /// America/Guyana
    America__Guyana,
    /// America/Halifax
    America__Halifax,
    /// America/Havana
    America__Havana,
    /// America/Hermosillo
    America__Hermosillo,
    /// America/Indiana/Indianapolis
    America__Indiana__Indianapolis,
    /// America/Indiana/Knox
    America__Indiana__Knox,
    /// America/Indiana/Marengo
    America__Indiana__Marengo,
    /// America/Indiana/Petersburg
    America__Indiana__Petersburg,
    /// America/Indiana/Tell_City
    America__Indiana__Tell_City,
    /// America/Indiana/Vevay
    America__Indiana__Vevay,
    /// America/Indiana/Vincennes
    America__Indiana__Vincennes,
    /// America/Indiana/Winamac
    America__Indiana__Winamac,
    /// America/Indianapolis
    America__Indianapolis,
    /// America/Inuvik
    America__Inuvik,
    /// America/Iqaluit
    America__Iqaluit,
    /// America/Jamaica
    America__Jamaica,
    /// America/Jujuy
    America__Jujuy,
    /// America/Juneau
    America__Juneau,
    /// America/Kentucky/Louisville
    America__Kentucky__Louisville,
    /// America/Kentucky/Monticello
    America__Kentucky__Monticello,
    /// America/Knox_IN
    America__Knox_IN,
    /// America/Kralendijk
    America__Kralendijk,
    /// America/La_Paz
    America__La_Paz,
    /// America/Lima
    America__Lima,
    /// America/Los_Angeles
    America__Los_Angeles,
    /// America/Louisville
    America__Louisville,
    /// America/Lower_Princes
    America__Lower_Princes,
    /// America/Maceio
    America__Maceio,
    /// America/Managua
    America__Managua,
    /// America/Manaus
    America__Manaus,
    /// America/Marigot
    America__Marigot,
    /// America/Martinique
    America__Martinique,
    /// America/Matamoros
    America__Matamoros,
    /// America/Mazatlan
    America__Mazatlan,
    /// America/Mendoza
    America__Mendoza,
    /// America/Menominee
    America__Menominee,
    /// America/Merida
    America__Merida,
    /// America/Metlakatla
    America__Metlakatla,
    /// America/Mexico_City
    America__Mexico_City,
    /// America/Miquelon
    America__Miquelon,
    /// America/Moncton
    America__Moncton,
    /// America/Monterrey
    America__Monterrey,
    /// America/Montevideo
    America__Montevideo,
    /// America/Montreal
    America__Montreal,
    /// America/Montserrat
    America__Montserrat,
    /// America/Nassau
    America__Nassau,
    /// America/New_York
    America__New_York,
    /// America/Nipigon
    America__Nipigon,
    /// America/Nome
    America__Nome,
    /// America/Noronha
    America__Noronha,
    /// America/North_Dakota/Beulah
    America__North_Dakota__Beulah,
    /// America/North_Dakota/Center
    America__North_Dakota__Center,
    /// America/North_Dakota/New_Salem
    America__North_Dakota__New_Salem,
    /// America/Nuuk
    America__Nuuk,
    /// America/Ojinaga
    America__Ojinaga,
    /// America/Panama
    America__Panama,
    /// America/Pangnirtung
    America__Pangnirtung,
    /// America/Paramaribo
    America__Paramaribo,
    /// America/Phoenix
    America__Phoenix,
    /// America/Port-au-Prince
    America__PortauPrince,
    /// America/Port_of_Spain
    America__Port_of_Spain,
    /// America/Porto_Acre
    America__Porto_Acre,
    /// America/Porto_Velho
    America__Porto_Velho,
    /// America/Puerto_Rico
    America__Puerto_Rico,
    /// America/Punta_Arenas
    America__Punta_Arenas,
    /// America/Rainy_River
    America__Rainy_River,
    /// America/Rankin_Inlet
    America__Rankin_Inlet,
    /// America/Recife
    America__Recife,
    /// America/Regina
    America__Regina,
    /// America/Resolute
    America__Resolute,
    /// America/Rio_Branco
    America__Rio_Branco,
    /// America/Rosario
    America__Rosario,
    /// America/Santa_Isabel
    America__Santa_Isabel,
    /// America/Santarem
    America__Santarem,
    /// America/Santiago
    America__Santiago,
    /// America/Santo_Domingo
    America__Santo_Domingo,
    /// America/Sao_Paulo
    America__Sao_Paulo,
    /// America/Scoresbysund
    America__Scoresbysund,
    /// America/Shiprock
    America__Shiprock,
    /// America/Sitka
    America__Sitka,
    /// America/St_Barthelemy
    America__St_Barthelemy,
    /// America/St_Johns
    America__St_Johns,
    /// America/St_Kitts
    America__St_Kitts,
    /// America/St_Lucia
    America__St_Lucia,
    /// America/St_Thomas
    America__St_Thomas,
    /// America/St_Vincent
    America__St_Vincent,
    /// America/Swift_Current
    America__Swift_Current,
    /// America/Tegucigalpa
    America__Tegucigalpa,
    /// America/Thule
    America__Thule,
    /// America/Thunder_Bay
    America__Thunder_Bay,
    /// America/Tijuana
    America__Tijuana,
    /// America/Toronto
    America__Toronto,
    /// America/Tortola
    America__Tortola,
    /// America/Vancouver
    America__Vancouver,
    /// America/Virgin
    America__Virgin,
    /// America/Whitehorse
    America__Whitehorse,
    /// America/Winnipeg
    America__Winnipeg,
    /// America/Yakutat
    America__Yakutat,
    /// America/Yellowknife
    America__Yellowknife,
    /// Antarctica/Casey
    Antarctica__Casey,
    /// Antarctica/Davis
    Antarctica__Davis,
    /// Antarctica/DumontDUrville
    Antarctica__DumontDUrville,
    /// Antarctica/Macquarie
    Antarctica__Macquarie,
    /// Antarctica/Mawson
    Antarctica__Mawson,
    /// Antarctica/McMurdo
    Antarctica__McMurdo,
    /// Antarctica/Palmer
    Antarctica__Palmer,
    /// Antarctica/Rothera
    Antarctica__Rothera,
    /// Antarctica/South_Pole
    Antarctica__South_Pole,
    /// Antarctica/Syowa
    Antarctica__Syowa,
    /// Antarctica/Troll
    Antarctica__Troll,
    /// Antarctica/Vostok
    Antarctica__Vostok,
    /// Arctic/Longyearbyen
    Arctic__Longyearbyen,
    /// Asia/Aden
    Asia__Aden,
    /// Asia/Almaty
    Asia__Almaty,
    /// Asia/Amman
    Asia__Amman,
    /// Asia/Anadyr
    Asia__Anadyr,
    /// Asia/Aqtau
    Asia__Aqtau,
    /// Asia/Aqtobe
    Asia__Aqtobe,
    /// Asia/Ashgabat
    Asia__Ashgabat,
    /// Asia/Ashkhabad
    Asia__Ashkhabad,
    /// Asia/Atyrau
    Asia__Atyrau,
    /// Asia/Baghdad
    Asia__Baghdad,
    /// Asia/Bahrain
    Asia__Bahrain,
    /// Asia/Baku
    Asia__Baku,
    /// Asia/Bangkok
    Asia__Bangkok,
    /// Asia/Barnaul
    Asia__Barnaul,
    /// Asia/Beirut
    Asia__Beirut,
    /// Asia/Bishkek
    Asia__Bishkek,
    /// Asia/Brunei
    Asia__Brunei,
    /// Asia/Calcutta
    Asia__Calcutta,
    /// Asia/Chita
    Asia__Chita,
    /// Asia/Choibalsan
    Asia__Choibalsan,
    /// Asia/Chongqing
    Asia__Chongqing,
    /// Asia/Chungking
    Asia__Chungking,
    /// Asia/Colombo
    Asia__Colombo,
    /// Asia/Dacca
    Asia__Dacca,
    /// Asia/Damascus
    Asia__Damascus,
    /// Asia/Dhaka
    Asia__Dhaka,
    /// Asia/Dili
    Asia__Dili,
    /// Asia/Dubai
    Asia__Dubai,
    /// Asia/Dushanbe
    Asia__Dushanbe,
    /// Asia/Famagusta
    Asia__Famagusta,
    /// Asia/Gaza
    Asia__Gaza,
    /// Asia/Harbin
    Asia__Harbin,
    /// Asia/Hebron
    Asia__Hebron,
    /// Asia/Ho_Chi_Minh
    Asia__Ho_Chi_Minh,
    /// Asia/Hong_Kong
    Asia__Hong_Kong,
    /// Asia/Hovd
    Asia__Hovd,
    /// Asia/Irkutsk
    Asia__Irkutsk,
    /// Asia/Istanbul
    Asia__Istanbul,
    /// Asia/Jakarta
    Asia__Jakarta,
    /// Asia/Jayapura
    Asia__Jayapura,
    /// Asia/Jerusalem
    Asia__Jerusalem,
    /// Asia/Kabul
    Asia__Kabul,
    /// Asia/Kamchatka
    Asia__Kamchatka,
    /// Asia/Karachi
    Asia__Karachi,
    /// Asia/Kashgar
    Asia__Kashgar,
    /// Asia/Kathmandu
    Asia__Kathmandu,
    /// Asia/Katmandu
    Asia__Katmandu,
    /// Asia/Khandyga
    Asia__Khandyga,
    /// Asia/Kolkata
    Asia__Kolkata,
    /// Asia/Krasnoyarsk
    Asia__Krasnoyarsk,
    /// Asia/Kuala_Lumpur
    Asia__Kuala_Lumpur,
    /// Asia/Kuching
    Asia__Kuching,
    /// Asia/Kuwait
    Asia__Kuwait,
    /// Asia/Macao
    Asia__Macao,
    /// Asia/Macau
    Asia__Macau,
    /// Asia/Magadan
    Asia__Magadan,
    /// Asia/Makassar
    Asia__Makassar,
    /// Asia/Manila
    Asia__Manila,
    /// Asia/Muscat
    Asia__Muscat,
    /// Asia/Nicosia
    Asia__Nicosia,
    /// Asia/Novokuznetsk
    Asia__Novokuznetsk,
    /// Asia/Novosibirsk
    Asia__Novosibirsk,
    /// Asia/Omsk
    Asia__Omsk,
    /// Asia/Oral
    Asia__Oral,
    /// Asia/Phnom_Penh
    Asia__Phnom_Penh,
    /// Asia/Pontianak
    Asia__Pontianak,
    /// Asia/Pyongyang
    Asia__Pyongyang,
    /// Asia/Qatar
    Asia__Qatar,
    /// Asia/Qostanay
    Asia__Qostanay,
    /// Asia/Qyzylorda
    Asia__Qyzylorda,
    /// Asia/Rangoon
    Asia__Rangoon,
    /// Asia/Riyadh
    Asia__Riyadh,
    /// Asia/Saigon
    Asia__Saigon,
    /// Asia/Sakhalin
    Asia__Sakhalin,
    /// Asia/Samarkand
    Asia__Samarkand,
    /// Asia/Seoul
    Asia__Seoul,
    /// Asia/Shanghai
    Asia__Shanghai,
    /// Asia/Singapore
    Asia__Singapore,
    /// Asia/Srednekolymsk
    Asia__Srednekolymsk,
    /// Asia/Taipei
    Asia__Taipei,
    /// Asia/Tashkent
    Asia__Tashkent,
    /// Asia/Tbilisi
    Asia__Tbilisi,
    /// Asia/Tehran
    Asia__Tehran,
    /// Asia/Tel_Aviv
    Asia__Tel_Aviv,
    /// Asia/Thimbu
    Asia__Thimbu,
    /// Asia/Thimphu
    Asia__Thimphu,
    /// Asia/Tokyo
    Asia__Tokyo,
    /// Asia/Tomsk
    Asia__Tomsk,
    /// Asia/Ujung_Pandang
    Asia__Ujung_Pandang,
    /// Asia/Ulaanbaatar
    Asia__Ulaanbaatar,
    /// Asia/Ulan_Bator
    Asia__Ulan_Bator,
    /// Asia/Urumqi
    Asia__Urumqi,
    /// Asia/Ust-Nera
    Asia__UstNera,
    /// Asia/Vientiane
    Asia__Vientiane,
    /// Asia/Vladivostok
    Asia__Vladivostok,
    /// Asia/Yakutsk
    Asia__Yakutsk,
    /// Asia/Yangon
    Asia__Yangon,
    /// Asia/Yekaterinburg
    Asia__Yekaterinburg,
    /// Asia/Yerevan
    Asia__Yerevan,
    /// Atlantic/Azores
    Atlantic__Azores,
    /// Atlantic/Bermuda
    Atlantic__Bermuda,
    /// Atlantic/Canary
    Atlantic__Canary,
    /// Atlantic/Cape_Verde
    Atlantic__Cape_Verde,
    /// Atlantic/Faeroe
    Atlantic__Faeroe,
    /// Atlantic/Faroe
    Atlantic__Faroe,
    /// Atlantic/Jan_Mayen
    Atlantic__Jan_Mayen,
    /// Atlantic/Madeira
    Atlantic__Madeira,
    /// Atlantic/Reykjavik
    Atlantic__Reykjavik,
    /// Atlantic/South_Georgia
    Atlantic__South_Georgia,
    /// Atlantic/St_Helena
    Atlantic__St_Helena,
    /// Atlantic/Stanley
    Atlantic__Stanley,
    /// Australia/ACT
    Australia__ACT,
    /// Australia/Adelaide
    Australia__Adelaide,
    /// Australia/Brisbane
    Australia__Brisbane,
    /// Australia/Broken_Hill
    Australia__Broken_Hill,
    /// Australia/Canberra
    Australia__Canberra,
    /// Australia/Currie
    Australia__Currie,
    /// Australia/Darwin
    Australia__Darwin,
    /// Australia/Eucla
    Australia__Eucla,
    /// Australia/Hobart
    Australia__Hobart,
    /// Australia/LHI
    Australia__LHI,
    /// Australia/Lindeman
    Australia__Lindeman,
    /// Australia/Lord_Howe
    Australia__Lord_Howe,
    /// Australia/Melbourne
    Australia__Melbourne,
    /// Australia/NSW
    Australia__NSW,
    /// Australia/North
    Australia__North,
    /// Australia/Perth
    Australia__Perth,
    /// Australia/Queensland
    Australia__Queensland,
    /// Australia/South
    Australia__South,
    /// Australia/Sydney
    Australia__Sydney,
    /// Australia/Tasmania
    Australia__Tasmania,
    /// Australia/Victoria
    Australia__Victoria,
    /// Australia/West
    Australia__West,
    /// Australia/Yancowinna
    Australia__Yancowinna,
    /// Brazil/Acre
    Brazil__Acre,
    /// Brazil/DeNoronha
    Brazil__DeNoronha,
    /// Brazil/East
    Brazil__East,
    /// Brazil/West
    Brazil__West,
    /// CET
    CET,
    /// CST6CDT
    CST6CDT,
    /// Canada/Atlantic
    Canada__Atlantic,
    /// Canada/Central
    Canada__Central,
    /// Canada/Eastern
    Canada__Eastern,
    /// Canada/Mountain
    Canada__Mountain,
    /// Canada/Newfoundland
    Canada__Newfoundland,
    /// Canada/Pacific
    Canada__Pacific,
    /// Canada/Saskatchewan
    Canada__Saskatchewan,
    /// Canada/Yukon
    Canada__Yukon,
    /// Chile/Continental
    Chile__Continental,
    /// Chile/EasterIsland
    Chile__EasterIsland,
    /// Cuba
    Cuba,
    /// EET
    EET,
    /// EST
    EST,
    /// EST5EDT
    EST5EDT,
    /// Egypt
    Egypt,
    /// Eire
    Eire,
    /// Etc/GMT
    Etc__GMT,
    /// Etc/GMT+0
    Etc__GMTPlus0,
    /// Etc/GMT+1
    Etc__GMTPlus1,
    /// Etc/GMT+10
    Etc__GMTPlus10,
    /// Etc/GMT+11
    Etc__GMTPlus11,
    /// Etc/GMT+12
    Etc__GMTPlus12,
    /// Etc/GMT+2
    Etc__GMTPlus2,
    /// Etc/GMT+3
    Etc__GMTPlus3,
    /// Etc/GMT+4
    Etc__GMTPlus4,
    /// Etc/GMT+5
    Etc__GMTPlus5,
    /// Etc/GMT+6
    Etc__GMTPlus6,
    /// Etc/GMT+7
    Etc__GMTPlus7,
    /// Etc/GMT+8
    Etc__GMTPlus8,
    /// Etc/GMT+9
    Etc__GMTPlus9,
    /// Etc/GMT-0
    Etc__GMTMinus0,
    /// Etc/GMT-1
    Etc__GMTMinus1,
    /// Etc/GMT-10
    Etc__GMTMinus10,
    /// Etc/GMT-11
    Etc__GMTMinus11,
    /// Etc/GMT-12
    Etc__GMTMinus12,
    /// Etc/GMT-13
    Etc__GMTMinus13,
    /// Etc/GMT-14
    Etc__GMTMinus14,
    /// Etc/GMT-2
    Etc__GMTMinus2,
    /// Etc/GMT-3
    Etc__GMTMinus3,
    /// Etc/GMT-4
    Etc__GMTMinus4,
    /// Etc/GMT-5
    Etc__GMTMinus5,
    /// Etc/GMT-6
    Etc__GMTMinus6,
    /// Etc/GMT-7
    Etc__GMTMinus7,
    /// Etc/GMT-8
    Etc__GMTMinus8,
    /// Etc/GMT-9
    Etc__GMTMinus9,
    /// Etc/GMT0
    Etc__GMT0,
    /// Etc/Greenwich
    Etc__Greenwich,
    /// Etc/UCT
    Etc__UCT,
    /// Etc/UTC
    Etc__UTC,
    /// Etc/Universal
    Etc__Universal,
    /// Etc/Zulu
    Etc__Zulu,
    /// Europe/Amsterdam
    Europe__Amsterdam,
    /// Europe/Andorra
    Europe__Andorra,
    /// Europe/Astrakhan
    Europe__Astrakhan,
    /// Europe/Athens
    Europe__Athens,
    /// Europe/Belfast
    Europe__Belfast,
    /// Europe/Belgrade
    Europe__Belgrade,
    /// Europe/Berlin
    Europe__Berlin,
    /// Europe/Bratislava
    Europe__Bratislava,
    /// Europe/Brussels
    Europe__Brussels,
    /// Europe/Bucharest
    Europe__Bucharest,
    /// Europe/Budapest
    Europe__Budapest,
    /// Europe/Busingen
    Europe__Busingen,
    /// Europe/Chisinau
    Europe__Chisinau,
    /// Europe/Copenhagen
    Europe__Copenhagen,
    /// Europe/Dublin
    Europe__Dublin,
    /// Europe/Gibraltar
    Europe__Gibraltar,
    /// Europe/Guernsey
    Europe__Guernsey,
    /// Europe/Helsinki
    Europe__Helsinki,
    /// Europe/Isle_of_Man
    Europe__Isle_of_Man,
    /// Europe/Istanbul
    Europe__Istanbul,
    /// Europe/Jersey
    Europe__Jersey,
    /// Europe/Kaliningrad
    Europe__Kaliningrad,
    /// Europe/Kiev
    Europe__Kiev,
    /// Europe/Kirov
    Europe__Kirov,
    /// Europe/Kyiv
    Europe__Kyiv,
    /// Europe/Lisbon
    Europe__Lisbon,
    /// Europe/Ljubljana
    Europe__Ljubljana,
    /// Europe/London
    Europe__London,
    /// Europe/Luxembourg
    Europe__Luxembourg,
    /// Europe/Madrid
    Europe__Madrid,
    /// Europe/Malta
    Europe__Malta,
    /// Europe/Mariehamn
    Europe__Mariehamn,
    /// Europe/Minsk
    Europe__Minsk,
    /// Europe/Monaco
    Europe__Monaco,
    /// Europe/Moscow
    Europe__Moscow,
    /// Europe/Nicosia
    Europe__Nicosia,
    /// Europe/Oslo
    Europe__Oslo,
    /// Europe/Paris
    Europe__Paris,
    /// Europe/Podgorica
    Europe__Podgorica,
    /// Europe/Prague
    Europe__Prague,
    /// Europe/Riga
    Europe__Riga,
    /// Europe/Rome
    Europe__Rome,
    /// Europe/Samara
    Europe__Samara,
    /// Europe/San_Marino
    Europe__San_Marino,
    /// Europe/Sarajevo
    Europe__Sarajevo,
    /// Europe/Saratov
    Europe__Saratov,
    /// Europe/Simferopol
    Europe__Simferopol,
    /// Europe/Skopje
    Europe__Skopje,
    /// Europe/Sofia
    Europe__Sofia,
    /// Europe/Stockholm
    Europe__Stockholm,
    /// Europe/Tallinn
    Europe__Tallinn,
    /// Europe/Tirane
    Europe__Tirane,
    /// Europe/Tiraspol
    Europe__Tiraspol,
    /// Europe/Ulyanovsk
    Europe__Ulyanovsk,
    /// Europe/Uzhgorod
    Europe__Uzhgorod,
    /// Europe/Vaduz
    Europe__Vaduz,
    /// Europe/Vatican
    Europe__Vatican,
    /// Europe/Vienna
    Europe__Vienna,
    /// Europe/Vilnius
    Europe__Vilnius,
    /// Europe/Volgograd
    Europe__Volgograd,
    /// Europe/Warsaw
    Europe__Warsaw,
    /// Europe/Zagreb
    Europe__Zagreb,
    /// Europe/Zaporozhye
    Europe__Zaporozhye,
    /// Europe/Zurich
    Europe__Zurich,
    /// GB
    GB,
    /// GB-Eire
    GBEire,
    /// GMT
    GMT,
    /// GMT+0
    GMTPlus0,
    /// GMT-0
    GMTMinus0,
    /// GMT0
    GMT0,
    /// Greenwich
    Greenwich,
    /// HST
    HST,
    /// Hongkong
    Hongkong,
    /// Iceland
    Iceland,
    /// Indian/Antananarivo
    Indian__Antananarivo,
    /// Indian/Chagos
    Indian__Chagos,
    /// Indian/Christmas
    Indian__Christmas,
    /// Indian/Cocos
    Indian__Cocos,
    /// Indian/Comoro
    Indian__Comoro,
    /// Indian/Kerguelen
    Indian__Kerguelen,
    /// Indian/Mahe
    Indian__Mahe,
    /// Indian/Maldives
    Indian__Maldives,
    /// Indian/Mauritius
    Indian__Mauritius,
    /// Indian/Mayotte
    Indian__Mayotte,
    /// Indian/Reunion
    Indian__Reunion,
    /// Iran
    Iran,
    /// Israel
    Israel,
    /// Jamaica
    Jamaica,
    /// Japan
    Japan,
    /// Kwajalein
    Kwajalein,
    /// Libya
    Libya,
    /// MET
    MET,
    /// MST
    MST,
    /// MST7MDT
    MST7MDT,
    /// Mexico/BajaNorte
    Mexico__BajaNorte,
    /// Mexico/BajaSur
    Mexico__BajaSur,
    /// Mexico/General
    Mexico__General,
    /// NZ
    NZ,
    /// NZ-CHAT
    NZCHAT,
    /// Navajo
    Navajo,
    /// PRC
    PRC,
    /// PST8PDT
    PST8PDT,
    /// Pacific/Apia
    Pacific__Apia,
    /// Pacific/Auckland
    Pacific__Auckland,
    /// Pacific/Bougainville
    Pacific__Bougainville,
    /// Pacific/Chatham
    Pacific__Chatham,
    /// Pacific/Chuuk
    Pacific__Chuuk,
    /// Pacific/Easter
    Pacific__Easter,
    /// Pacific/Efate
    Pacific__Efate,
    /// Pacific/Enderbury
    Pacific__Enderbury,
    /// Pacific/Fakaofo
    Pacific__Fakaofo,
    /// Pacific/Fiji
    Pacific__Fiji,
    /// Pacific/Funafuti
    Pacific__Funafuti,
    /// Pacific/Galapagos
    Pacific__Galapagos,
    /// Pacific/Gambier
    Pacific__Gambier,
    /// Pacific/Guadalcanal
    Pacific__Guadalcanal,
    /// Pacific/Guam
    Pacific__Guam,
    /// Pacific/Honolulu
    Pacific__Honolulu,
    /// Pacific/Johnston
    Pacific__Johnston,
    /// Pacific/Kanton
    Pacific__Kanton,
    /// Pacific/Kiritimati
    Pacific__Kiritimati,
    /// Pacific/Kosrae
    Pacific__Kosrae,
    /// Pacific/Kwajalein
    Pacific__Kwajalein,
    /// Pacific/Majuro
    Pacific__Majuro,
    /// Pacific/Marquesas
    Pacific__Marquesas,
    /// Pacific/Midway
    Pacific__Midway,
    /// Pacific/Nauru
    Pacific__Nauru,
    /// Pacific/Niue
    Pacific__Niue,
    /// Pacific/Norfolk
    Pacific__Norfolk,
    /// Pacific/Noumea
    Pacific__Noumea,
    /// Pacific/Pago_Pago
    Pacific__Pago_Pago,
    /// Pacific/Palau
    Pacific__Palau,
    /// Pacific/Pitcairn
    Pacific__Pitcairn,
    /// Pacific/Pohnpei
    Pacific__Pohnpei,
    /// Pacific/Ponape
    Pacific__Ponape,
    /// Pacific/Port_Moresby
    Pacific__Port_Moresby,
    /// Pacific/Rarotonga
    Pacific__Rarotonga,
    /// Pacific/Saipan
    Pacific__Saipan,
    /// Pacific/Samoa
    Pacific__Samoa,
    /// Pacific/Tahiti
    Pacific__Tahiti,
    /// Pacific/Tarawa
    Pacific__Tarawa,
    /// Pacific/Tongatapu
    Pacific__Tongatapu,
    /// Pacific/Truk
    Pacific__Truk,
    /// Pacific/Wake
    Pacific__Wake,
    /// Pacific/Wallis
    Pacific__Wallis,
    /// Pacific/Yap
    Pacific__Yap,
    /// Poland
    Poland,
    /// Portugal
    Portugal,
    /// ROC
    ROC,
    /// ROK
    ROK,
    /// Singapore
    Singapore,
    /// Turkey
    Turkey,
    /// UCT
    UCT,
    /// US/Alaska
    US__Alaska,
    /// US/Aleutian
    US__Aleutian,
    /// US/Arizona
    US__Arizona,
    /// US/Central
    US__Central,
    /// US/East-Indiana
    US__EastIndiana,
    /// US/Eastern
    US__Eastern,
    /// US/Hawaii
    US__Hawaii,
    /// US/Indiana-Starke
    US__IndianaStarke,
    /// US/Michigan
    US__Michigan,
    /// US/Mountain
    US__Mountain,
    /// US/Pacific
    US__Pacific,
    /// US/Samoa
    US__Samoa,
    /// UTC
    UTC,
    /// Universal
    Universal,
    /// W-SU
    WSU,
    /// WET
    WET,
    /// Zulu
    Zulu,
}
static TIMEZONES: ::phf::Map<&'static str, Tz> = 
::phf::Map {
    key: 12913932095322966823,
    disps: &[
        (0, 42),
        (0, 6),
        (0, 0),
        (0, 53),
        (1, 24),
        (0, 0),
        (0, 451),
        (0, 16),
        (0, 0),
        (0, 28),
        (0, 576),
        (0, 13),
        (0, 41),
        (0, 269),
        (0, 0),
        (0, 15),
        (0, 545),
        (0, 1),
        (0, 99),
        (0, 5),
        (0, 93),
        (0, 32),
        (0, 14),
        (0, 0),
        (0, 260),
        (0, 10),
        (0, 7),
        (0, 156),
        (0, 1),
        (5, 192),
        (1, 183),
        (0, 23),
        (0, 53),
        (0, 95),
        (0, 236),
        (0, 27),
        (0, 60),
        (0, 383),
        (0, 2),
        (0, 0),
        (0, 111),
        (0, 4),
        (0, 215),
        (0, 54),
        (0, 2),
        (0, 1),
        (0, 22),
        (1, 438),
        (0, 248),
        (1, 187),
        (0, 9),
        (0, 3),
        (0, 32),
        (0, 50),
        (2, 123),
        (0, 42),
        (1, 20),
        (0, 1),
        (6, 551),
        (0, 128),
        (1, 111),
        (0, 553),
        (0, 0),
        (0, 0),
        (0, 231),
        (0, 25),
        (0, 127),
        (0, 95),
        (0, 512),
        (8, 167),
        (4, 54),
        (0, 20),
        (0, 191),
        (13, 23),
        (1, 32),
        (0, 478),
        (0, 256),
        (0, 12),
        (0, 40),
        (0, 432),
        (0, 499),
        (0, 2),
        (0, 100),
        (0, 25),
        (5, 124),
        (0, 258),
        (0, 255),
        (0, 467),
        (0, 154),
        (0, 450),
        (0, 528),
        (0, 105),
        (0, 8),
        (1, 439),
        (0, 12),
        (19, 343),
        (0, 0),
        (2, 411),
        (0, 95),
        (36, 54),
        (0, 115),
        (0, 184),
        (1, 406),
        (1, 417),
        (0, 0),
        (0, 337),
        (7, 360),
        (1, 518),
        (6, 41),
        (1, 1),
        (0, 494),
        (0, 33),
        (0, 0),
        (10, 249),
        (6, 381),
        (0, 458),
        (2, 235),
        (12, 336),
        (19, 503),
        (0, 180),
    ],
    entries: &[
        ("America/Guyana", Tz::America__Guyana),
        ("Asia/Krasnoyarsk", Tz::Asia__Krasnoyarsk),
        ("America/Merida", Tz::America__Merida),
        ("Asia/Tel_Aviv", Tz::Asia__Tel_Aviv),
        ("America/Tortola", Tz::America__Tortola),
        ("America/Santarem", Tz::America__Santarem),
        ("Australia/Brisbane", Tz::Australia__Brisbane),
        ("America/Caracas", Tz::America__Caracas),
        ("America/Nassau", Tz::America__Nassau),
        ("America/Chicago", Tz::America__Chicago),
        ("America/Hermosillo", Tz::America__Hermosillo),
        ("Africa/Asmara", Tz::Africa__Asmara),
        ("America/Godthab", Tz::America__Godthab),
        ("Europe/Vatican", Tz::Europe__Vatican),
        ("NZ-CHAT", Tz::NZCHAT),
        ("Pacific/Kosrae", Tz::Pacific__Kosrae),
        ("America/Montreal", Tz::America__Montreal),
        ("America/Dominica", Tz::America__Dominica),
        ("Australia/Sydney", Tz::Australia__Sydney),
        ("US/East-Indiana", Tz::US__EastIndiana),
        ("America/Catamarca", Tz::America__Catamarca),
        ("Pacific/Wallis", Tz::Pacific__Wallis),
        ("NZ", Tz::NZ),
        ("Asia/Damascus", Tz::Asia__Damascus),
        ("Asia/Manila", Tz::Asia__Manila),
        ("America/Noronha", Tz::America__Noronha),
        ("Europe/Zagreb", Tz::Europe__Zagreb),
        ("Europe/Mariehamn", Tz::Europe__Mariehamn),
        ("UCT", Tz::UCT),
        ("Europe/Busingen", Tz::Europe__Busingen),
        ("Canada/Mountain", Tz::Canada__Mountain),
        ("Arctic/Longyearbyen", Tz::Arctic__Longyearbyen),
        ("America/Bogota", Tz::America__Bogota),
        ("Europe/Belgrade", Tz::Europe__Belgrade),
        ("US/Alaska", Tz::US__Alaska),
        ("Asia/Pyongyang", Tz::Asia__Pyongyang),
        ("America/Campo_Grande", Tz::America__Campo_Grande),
        ("MET", Tz::MET),
        ("Asia/Sakhalin", Tz::Asia__Sakhalin),
        ("Etc/GMT-3", Tz::Etc__GMTMinus3),
        ("Pacific/Ponape", Tz::Pacific__Ponape),
        ("Cuba", Tz::Cuba),
        ("America/Ojinaga", Tz::America__Ojinaga),
        ("Asia/Tehran", Tz::Asia__Tehran),
        ("America/Tegucigalpa", Tz::America__Tegucigalpa),
        ("Pacific/Honolulu", Tz::Pacific__Honolulu),
        ("America/Aruba", Tz::America__Aruba),
        ("America/Rio_Branco", Tz::America__Rio_Branco),
        ("America/Guayaquil", Tz::America__Guayaquil),
        ("Etc/Zulu", Tz::Etc__Zulu),
        ("Europe/Monaco", Tz::Europe__Monaco),
        ("Africa/Kinshasa", Tz::Africa__Kinshasa),
        ("America/St_Kitts", Tz::America__St_Kitts),
        ("Asia/Yekaterinburg", Tz::Asia__Yekaterinburg),
        ("Africa/Freetown", Tz::Africa__Freetown),
        ("Africa/El_Aaiun", Tz::Africa__El_Aaiun),
        ("America/Anguilla", Tz::America__Anguilla),
        ("Africa/Abidjan", Tz::Africa__Abidjan),
        ("Singapore", Tz::Singapore),
        ("GMT-0", Tz::GMTMinus0),
        ("America/Blanc-Sablon", Tz::America__BlancSablon),
        ("PST8PDT", Tz::PST8PDT),
        ("Pacific/Efate", Tz::Pacific__Efate),
        ("America/Indiana/Vincennes", Tz::America__Indiana__Vincennes),
        ("Europe/Jersey", Tz::Europe__Jersey),
        ("America/Lower_Princes", Tz::America__Lower_Princes),
        ("Africa/Sao_Tome", Tz::Africa__Sao_Tome),
        ("Asia/Irkutsk", Tz::Asia__Irkutsk),
        ("Europe/Chisinau", Tz::Europe__Chisinau),
        ("Asia/Taipei", Tz::Asia__Taipei),
        ("America/Indiana/Winamac", Tz::America__Indiana__Winamac),
        ("Africa/Blantyre", Tz::Africa__Blantyre),
        ("America/Swift_Current", Tz::America__Swift_Current),
        ("America/Boa_Vista", Tz::America__Boa_Vista),
        ("Africa/Brazzaville", Tz::Africa__Brazzaville),
        ("Antarctica/Rothera", Tz::Antarctica__Rothera),
        ("Asia/Macau", Tz::Asia__Macau),
        ("Etc/GMT-14", Tz::Etc__GMTMinus14),
        ("Australia/Tasmania", Tz::Australia__Tasmania),
        ("Pacific/Wake", Tz::Pacific__Wake),
        ("Asia/Seoul", Tz::Asia__Seoul),
        ("Turkey", Tz::Turkey),
        ("Asia/Tokyo", Tz::Asia__Tokyo),
        ("America/Recife", Tz::America__Recife),
        ("Atlantic/Madeira", Tz::Atlantic__Madeira),
        ("America/Paramaribo", Tz::America__Paramaribo),
        ("Etc/GMT0", Tz::Etc__GMT0),
        ("Asia/Srednekolymsk", Tz::Asia__Srednekolymsk),
        ("America/Mendoza", Tz::America__Mendoza),
        ("America/St_Johns", Tz::America__St_Johns),
        ("Europe/Podgorica", Tz::Europe__Podgorica),
        ("America/La_Paz", Tz::America__La_Paz),
        ("Asia/Ujung_Pandang", Tz::Asia__Ujung_Pandang),
        ("Indian/Cocos", Tz::Indian__Cocos),
        ("Asia/Brunei", Tz::Asia__Brunei),
        ("Asia/Oral", Tz::Asia__Oral),
        ("Europe/Vaduz", Tz::Europe__Vaduz),
        ("America/Argentina/San_Juan", Tz::America__Argentina__San_Juan),
        ("America/Lima", Tz::America__Lima),
        ("America/Porto_Velho", Tz::America__Porto_Velho),
        ("Etc/GMT-6", Tz::Etc__GMTMinus6),
        ("America/Belize", Tz::America__Belize),
        ("Australia/Victoria", Tz::Australia__Victoria),
        ("Pacific/Enderbury", Tz::Pacific__Enderbury),
        ("Europe/Minsk", Tz::Europe__Minsk),
        ("Asia/Aden", Tz::Asia__Aden),
        ("Asia/Yakutsk", Tz::Asia__Yakutsk),
        ("Europe/Sofia", Tz::Europe__Sofia),
        ("Africa/Nouakchott", Tz::Africa__Nouakchott),
        ("America/Detroit", Tz::America__Detroit),
        ("Europe/Paris", Tz::Europe__Paris),
        ("Etc/GMT+2", Tz::Etc__GMTPlus2),
        ("America/Manaus", Tz::America__Manaus),
        ("Europe/Isle_of_Man", Tz::Europe__Isle_of_Man),
        ("Australia/ACT", Tz::Australia__ACT),
        ("Atlantic/South_Georgia", Tz::Atlantic__South_Georgia),
        ("Etc/GMT+8", Tz::Etc__GMTPlus8),
        ("America/Havana", Tz::America__Havana),
        ("Asia/Istanbul", Tz::Asia__Istanbul),
        ("Etc/GMT+3", Tz::Etc__GMTPlus3),
        ("Pacific/Pohnpei", Tz::Pacific__Pohnpei),
        ("America/Argentina/Salta", Tz::America__Argentina__Salta),
        ("Pacific/Bougainville", Tz::Pacific__Bougainville),
        ("America/Coral_Harbour", Tz::America__Coral_Harbour),
        ("America/Bahia", Tz::America__Bahia),
        ("Pacific/Fakaofo", Tz::Pacific__Fakaofo),
        ("Africa/Asmera", Tz::Africa__Asmera),
        ("Asia/Baghdad", Tz::Asia__Baghdad),
        ("America/Boise", Tz::America__Boise),
        ("Antarctica/McMurdo", Tz::Antarctica__McMurdo),
        ("Pacific/Kiritimati", Tz::Pacific__Kiritimati),
        ("Africa/Accra", Tz::Africa__Accra),
        ("Europe/Malta", Tz::Europe__Malta),
        ("Etc/GMT-12", Tz::Etc__GMTMinus12),
        ("Mexico/BajaSur", Tz::Mexico__BajaSur),
        ("Asia/Bangkok", Tz::Asia__Bangkok),
        ("Indian/Chagos", Tz::Indian__Chagos),
        ("Etc/Greenwich", Tz::Etc__Greenwich),
        ("America/Fort_Wayne", Tz::America__Fort_Wayne),
        ("Pacific/Kwajalein", Tz::Pacific__Kwajalein),
        ("Europe/Lisbon", Tz::Europe__Lisbon),
        ("America/Regina", Tz::America__Regina),
        ("America/Marigot", Tz::America__Marigot),
        ("Etc/GMT+7", Tz::Etc__GMTPlus7),
        ("America/Halifax", Tz::America__Halifax),
        ("America/Indianapolis", Tz::America__Indianapolis),
        ("US/Samoa", Tz::US__Samoa),
        ("America/Thunder_Bay", Tz::America__Thunder_Bay),
        ("Japan", Tz::Japan),
        ("Africa/Harare", Tz::Africa__Harare),
        ("America/Resolute", Tz::America__Resolute),
        ("America/Moncton", Tz::America__Moncton),
        ("Asia/Harbin", Tz::Asia__Harbin),
        ("Asia/Rangoon", Tz::Asia__Rangoon),
        ("Europe/Gibraltar", Tz::Europe__Gibraltar),
        ("Canada/Central", Tz::Canada__Central),
        ("Europe/Saratov", Tz::Europe__Saratov),
        ("Antarctica/Macquarie", Tz::Antarctica__Macquarie),
        ("America/Indiana/Marengo", Tz::America__Indiana__Marengo),
        ("Pacific/Gambier", Tz::Pacific__Gambier),
        ("Asia/Ulaanbaatar", Tz::Asia__Ulaanbaatar),
        ("Africa/Maseru", Tz::Africa__Maseru),
        ("Europe/Helsinki", Tz::Europe__Helsinki),
        ("America/St_Vincent", Tz::America__St_Vincent),
        ("MST7MDT", Tz::MST7MDT),
        ("Portugal", Tz::Portugal),
        ("Asia/Bishkek", Tz::Asia__Bishkek),
        ("Indian/Antananarivo", Tz::Indian__Antananarivo),
        ("America/Edmonton", Tz::America__Edmonton),
        ("Antarctica/DumontDUrville", Tz::Antarctica__DumontDUrville),
        ("Australia/Melbourne", Tz::Australia__Melbourne),
        ("America/Ensenada", Tz::America__Ensenada),
        ("Pacific/Majuro", Tz::Pacific__Majuro),
        ("America/Knox_IN", Tz::America__Knox_IN),
        ("Etc/GMT+11", Tz::Etc__GMTPlus11),
        ("Asia/Kuching", Tz::Asia__Kuching),
        ("Jamaica", Tz::Jamaica),
        ("US/Aleutian", Tz::US__Aleutian),
        ("America/Buenos_Aires", Tz::America__Buenos_Aires),
        ("America/Dawson_Creek", Tz::America__Dawson_Creek),
        ("US/Central", Tz::US__Central),
        ("America/Argentina/San_Luis", Tz::America__Argentina__San_Luis),
        ("America/Monterrey", Tz::America__Monterrey),
        ("Universal", Tz::Universal),
        ("Etc/GMT+10", Tz::Etc__GMTPlus10),
        ("Pacific/Niue", Tz::Pacific__Niue),
        ("Asia/Chita", Tz::Asia__Chita),
        ("America/Port_of_Spain", Tz::America__Port_of_Spain),
        ("Etc/GMT-11", Tz::Etc__GMTMinus11),
        ("Asia/Magadan", Tz::Asia__Magadan),
        ("Africa/Lubumbashi", Tz::Africa__Lubumbashi),
        ("Asia/Thimphu", Tz::Asia__Thimphu),
        ("Africa/Kampala", Tz::Africa__Kampala),
        ("America/Panama", Tz::America__Panama),
        ("Asia/Chongqing", Tz::Asia__Chongqing),
        ("Africa/Conakry", Tz::Africa__Conakry),
        ("Asia/Dushanbe", Tz::Asia__Dushanbe),
        ("Europe/Oslo", Tz::Europe__Oslo),
        ("Asia/Qyzylorda", Tz::Asia__Qyzylorda),
        ("Atlantic/Faeroe", Tz::Atlantic__Faeroe),
        ("America/Jujuy", Tz::America__Jujuy),
        ("Europe/Dublin", Tz::Europe__Dublin),
        ("Asia/Aqtau", Tz::Asia__Aqtau),
        ("Asia/Dacca", Tz::Asia__Dacca),
        ("America/Antigua", Tz::America__Antigua),
        ("America/New_York", Tz::America__New_York),
        ("Australia/North", Tz::Australia__North),
        ("US/Mountain", Tz::US__Mountain),
        ("America/Argentina/Rio_Gallegos", Tz::America__Argentina__Rio_Gallegos),
        ("Europe/Madrid", Tz::Europe__Madrid),
        ("Asia/Kashgar", Tz::Asia__Kashgar),
        ("America/Punta_Arenas", Tz::America__Punta_Arenas),
        ("Asia/Barnaul", Tz::Asia__Barnaul),
        ("America/North_Dakota/New_Salem", Tz::America__North_Dakota__New_Salem),
        ("America/Virgin", Tz::America__Virgin),
        ("Iran", Tz::Iran),
        ("Europe/Kaliningrad", Tz::Europe__Kaliningrad),
        ("PRC", Tz::PRC),
        ("Europe/Volgograd", Tz::Europe__Volgograd),
        ("America/Guatemala", Tz::America__Guatemala),
        ("Asia/Amman", Tz::Asia__Amman),
        ("America/Indiana/Tell_City", Tz::America__Indiana__Tell_City),
        ("Atlantic/Jan_Mayen", Tz::Atlantic__Jan_Mayen),
        ("Asia/Ho_Chi_Minh", Tz::Asia__Ho_Chi_Minh),
        ("Brazil/East", Tz::Brazil__East),
        ("Etc/GMT-4", Tz::Etc__GMTMinus4),
        ("America/Pangnirtung", Tz::America__Pangnirtung),
        ("Canada/Newfoundland", Tz::Canada__Newfoundland),
        ("Africa/Douala", Tz::Africa__Douala),
        ("GMT", Tz::GMT),
        ("America/Glace_Bay", Tz::America__Glace_Bay),
        ("Africa/Nairobi", Tz::Africa__Nairobi),
        ("Kwajalein", Tz::Kwajalein),
        ("America/Guadeloupe", Tz::America__Guadeloupe),
        ("Pacific/Funafuti", Tz::Pacific__Funafuti),
        ("Etc/GMT-1", Tz::Etc__GMTMinus1),
        ("Africa/Tunis", Tz::Africa__Tunis),
        ("America/Anchorage", Tz::America__Anchorage),
        ("US/Eastern", Tz::US__Eastern),
        ("Africa/Dar_es_Salaam", Tz::Africa__Dar_es_Salaam),
        ("Pacific/Yap", Tz::Pacific__Yap),
        ("Asia/Karachi", Tz::Asia__Karachi),
        ("Asia/Dhaka", Tz::Asia__Dhaka),
        ("America/Argentina/ComodRivadavia", Tz::America__Argentina__ComodRivadavia),
        ("Etc/GMT+6", Tz::Etc__GMTPlus6),
        ("Australia/Perth", Tz::Australia__Perth),
        ("CST6CDT", Tz::CST6CDT),
        ("Europe/Luxembourg", Tz::Europe__Luxembourg),
        ("Pacific/Chuuk", Tz::Pacific__Chuuk),
        ("Europe/Rome", Tz::Europe__Rome),
        ("US/Michigan", Tz::US__Michigan),
        ("America/Menominee", Tz::America__Menominee),
        ("Antarctica/Vostok", Tz::Antarctica__Vostok),
        ("Chile/EasterIsland", Tz::Chile__EasterIsland),
        ("ROC", Tz::ROC),
        ("Asia/Aqtobe", Tz::Asia__Aqtobe),
        ("Greenwich", Tz::Greenwich),
        ("Asia/Riyadh", Tz::Asia__Riyadh),
        ("Antarctica/Davis", Tz::Antarctica__Davis),
        ("America/Indiana/Indianapolis", Tz::America__Indiana__Indianapolis),
        ("Africa/Tripoli", Tz::Africa__Tripoli),
        ("Hongkong", Tz::Hongkong),
        ("America/Mazatlan", Tz::America__Mazatlan),
        ("America/Montserrat", Tz::America__Montserrat),
        ("EET", Tz::EET),
        ("Europe/Copenhagen", Tz::Europe__Copenhagen),
        ("Australia/Darwin", Tz::Australia__Darwin),
        ("Asia/Ulan_Bator", Tz::Asia__Ulan_Bator),
        ("Etc/GMT-13", Tz::Etc__GMTMinus13),
        ("Pacific/Guam", Tz::Pacific__Guam),
        ("Atlantic/Canary", Tz::Atlantic__Canary),
        ("Europe/Belfast", Tz::Europe__Belfast),
        ("Chile/Continental", Tz::Chile__Continental),
        ("America/Scoresbysund", Tz::America__Scoresbysund),
        ("Asia/Omsk", Tz::Asia__Omsk),
        ("Asia/Hong_Kong", Tz::Asia__Hong_Kong),
        ("Africa/Lagos", Tz::Africa__Lagos),
        ("Europe/Guernsey", Tz::Europe__Guernsey),
        ("America/Iqaluit", Tz::America__Iqaluit),
        ("Asia/Shanghai", Tz::Asia__Shanghai),
        ("Asia/Yerevan", Tz::Asia__Yerevan),
        ("America/Atikokan", Tz::America__Atikokan),
        ("America/Argentina/Catamarca", Tz::America__Argentina__Catamarca),
        ("America/Maceio", Tz::America__Maceio),
        ("Canada/Yukon", Tz::Canada__Yukon),
        ("Africa/Monrovia", Tz::Africa__Monrovia),
        ("America/Nome", Tz::America__Nome),
        ("Europe/Kyiv", Tz::Europe__Kyiv),
        ("America/Eirunepe", Tz::America__Eirunepe),
        ("Europe/Skopje", Tz::Europe__Skopje),
        ("Asia/Beirut", Tz::Asia__Beirut),
        ("America/Yakutat", Tz::America__Yakutat),
        ("GMT+0", Tz::GMTPlus0),
        ("Europe/Astrakhan", Tz::Europe__Astrakhan),
        ("Asia/Tbilisi", Tz::Asia__Tbilisi),
        ("Asia/Nicosia", Tz::Asia__Nicosia),
        ("Indian/Kerguelen", Tz::Indian__Kerguelen),
        ("Etc/GMT+9", Tz::Etc__GMTPlus9),
        ("Indian/Mayotte", Tz::Indian__Mayotte),
        ("Indian/Mahe", Tz::Indian__Mahe),
        ("Europe/Athens", Tz::Europe__Athens),
        ("Asia/Kabul", Tz::Asia__Kabul),
        ("Pacific/Kanton", Tz::Pacific__Kanton),
        ("Africa/Djibouti", Tz::Africa__Djibouti),
        ("America/Cambridge_Bay", Tz::America__Cambridge_Bay),
        ("America/Mexico_City", Tz::America__Mexico_City),
        ("Europe/Brussels", Tz::Europe__Brussels),
        ("America/Martinique", Tz::America__Martinique),
        ("Asia/Kuwait", Tz::Asia__Kuwait),
        ("Australia/Eucla", Tz::Australia__Eucla),
        ("Asia/Khandyga", Tz::Asia__Khandyga),
        ("US/Indiana-Starke", Tz::US__IndianaStarke),
        ("Asia/Kolkata", Tz::Asia__Kolkata),
        ("Pacific/Pitcairn", Tz::Pacific__Pitcairn),
        ("America/Nipigon", Tz::America__Nipigon),
        ("Antarctica/South_Pole", Tz::Antarctica__South_Pole),
        ("Mexico/General", Tz::Mexico__General),
        ("Europe/Tirane", Tz::Europe__Tirane),
        ("Asia/Samarkand", Tz::Asia__Samarkand),
        ("Canada/Eastern", Tz::Canada__Eastern),
        ("Antarctica/Mawson", Tz::Antarctica__Mawson),
        ("Europe/Berlin", Tz::Europe__Berlin),
        ("America/Cancun", Tz::America__Cancun),
        ("Asia/Vientiane", Tz::Asia__Vientiane),
        ("Africa/Windhoek", Tz::Africa__Windhoek),
        ("Asia/Anadyr", Tz::Asia__Anadyr),
        ("Africa/Khartoum", Tz::Africa__Khartoum),
        ("Africa/Gaborone", Tz::Africa__Gaborone),
        ("Europe/Bucharest", Tz::Europe__Bucharest),
        ("Africa/Kigali", Tz::Africa__Kigali),
        ("America/Inuvik", Tz::America__Inuvik),
        ("Australia/Canberra", Tz::Australia__Canberra),
        ("America/Denver", Tz::America__Denver),
        ("America/Shiprock", Tz::America__Shiprock),
        ("Europe/Nicosia", Tz::Europe__Nicosia),
        ("Europe/London", Tz::Europe__London),
        ("Europe/Ljubljana", Tz::Europe__Ljubljana),
        ("Antarctica/Casey", Tz::Antarctica__Casey),
        ("Atlantic/Cape_Verde", Tz::Atlantic__Cape_Verde),
        ("Australia/LHI", Tz::Australia__LHI),
        ("US/Hawaii", Tz::US__Hawaii),
        ("Libya", Tz::Libya),
        ("Asia/Novosibirsk", Tz::Asia__Novosibirsk),
        ("Europe/Tallinn", Tz::Europe__Tallinn),
        ("Asia/Calcutta", Tz::Asia__Calcutta),
        ("Pacific/Palau", Tz::Pacific__Palau),
        ("Africa/Ceuta", Tz::Africa__Ceuta),
        ("Brazil/Acre", Tz::Brazil__Acre),
        ("Pacific/Guadalcanal", Tz::Pacific__Guadalcanal),
        ("Asia/Ashkhabad", Tz::Asia__Ashkhabad),
        ("America/Jamaica", Tz::America__Jamaica),
        ("Pacific/Tarawa", Tz::Pacific__Tarawa),
        ("GB-Eire", Tz::GBEire),
        ("Indian/Reunion", Tz::Indian__Reunion),
        ("America/Argentina/La_Rioja", Tz::America__Argentina__La_Rioja),
        ("Australia/West", Tz::Australia__West),
        ("America/Belem", Tz::America__Belem),
        ("America/Porto_Acre", Tz::America__Porto_Acre),
        ("Asia/Chungking", Tz::Asia__Chungking),
        ("EST5EDT", Tz::EST5EDT),
        ("America/Louisville", Tz::America__Louisville),
        ("Asia/Qostanay", Tz::Asia__Qostanay),
        ("America/Cayman", Tz::America__Cayman),
        ("Pacific/Johnston", Tz::Pacific__Johnston),
        ("Asia/Urumqi", Tz::Asia__Urumqi),
        ("Africa/Mbabane", Tz::Africa__Mbabane),
        ("Africa/Casablanca", Tz::Africa__Casablanca),
        ("America/Sao_Paulo", Tz::America__Sao_Paulo),
        ("Navajo", Tz::Navajo),
        ("America/Winnipeg", Tz::America__Winnipeg),
        ("ROK", Tz::ROK),
        ("America/Fort_Nelson", Tz::America__Fort_Nelson),
        ("Asia/Tomsk", Tz::Asia__Tomsk),
        ("Africa/Maputo", Tz::Africa__Maputo),
        ("Europe/Samara", Tz::Europe__Samara),
        ("Pacific/Nauru", Tz::Pacific__Nauru),
        ("Europe/Istanbul", Tz::Europe__Istanbul),
        ("Africa/Dakar", Tz::Africa__Dakar),
        ("America/Yellowknife", Tz::America__Yellowknife),
        ("Asia/Novokuznetsk", Tz::Asia__Novokuznetsk),
        ("America/Sitka", Tz::America__Sitka),
        ("HST", Tz::HST),
        ("Antarctica/Palmer", Tz::Antarctica__Palmer),
        ("Australia/Lord_Howe", Tz::Australia__Lord_Howe),
        ("America/Rosario", Tz::America__Rosario),
        ("W-SU", Tz::WSU),
        ("Australia/NSW", Tz::Australia__NSW),
        ("CET", Tz::CET),
        ("America/Kralendijk", Tz::America__Kralendijk),
        ("US/Pacific", Tz::US__Pacific),
        ("Pacific/Port_Moresby", Tz::Pacific__Port_Moresby),
        ("Asia/Ashgabat", Tz::Asia__Ashgabat),
        ("Etc/GMT+12", Tz::Etc__GMTPlus12),
        ("America/Managua", Tz::America__Managua),
        ("Pacific/Marquesas", Tz::Pacific__Marquesas),
        ("Mexico/BajaNorte", Tz::Mexico__BajaNorte),
        ("America/Barbados", Tz::America__Barbados),
        ("Africa/Luanda", Tz::Africa__Luanda),
        ("Brazil/West", Tz::Brazil__West),
        ("Asia/Choibalsan", Tz::Asia__Choibalsan),
        ("America/Whitehorse", Tz::America__Whitehorse),
        ("America/Argentina/Tucuman", Tz::America__Argentina__Tucuman),
        ("Asia/Dili", Tz::Asia__Dili),
        ("Canada/Pacific", Tz::Canada__Pacific),
        ("Asia/Katmandu", Tz::Asia__Katmandu),
        ("Pacific/Noumea", Tz::Pacific__Noumea),
        ("Atlantic/Bermuda", Tz::Atlantic__Bermuda),
        ("Indian/Mauritius", Tz::Indian__Mauritius),
        ("Asia/Vladivostok", Tz::Asia__Vladivostok),
        ("Etc/GMT+0", Tz::Etc__GMTPlus0),
        ("America/Grenada", Tz::America__Grenada),
        ("America/Cayenne", Tz::America__Cayenne),
        ("Etc/GMT-2", Tz::Etc__GMTMinus2),
        ("Atlantic/Azores", Tz::Atlantic__Azores),
        ("UTC", Tz::UTC),
        ("America/Argentina/Mendoza", Tz::America__Argentina__Mendoza),
        ("Pacific/Tahiti", Tz::Pacific__Tahiti),
        ("Europe/Kirov", Tz::Europe__Kirov),
        ("Europe/Vilnius", Tz::Europe__Vilnius),
        ("Pacific/Midway", Tz::Pacific__Midway),
        ("America/Juneau", Tz::America__Juneau),
        ("America/Cordoba", Tz::America__Cordoba),
        ("America/North_Dakota/Center", Tz::America__North_Dakota__Center),
        ("Asia/Saigon", Tz::Asia__Saigon),
        ("America/Matamoros", Tz::America__Matamoros),
        ("Etc/GMT+4", Tz::Etc__GMTPlus4),
        ("America/Los_Angeles", Tz::America__Los_Angeles),
        ("Europe/Moscow", Tz::Europe__Moscow),
        ("Africa/Ndjamena", Tz::Africa__Ndjamena),
        ("Africa/Bangui", Tz::Africa__Bangui),
        ("Pacific/Galapagos", Tz::Pacific__Galapagos),
        ("Canada/Saskatchewan", Tz::Canada__Saskatchewan),
        ("Australia/South", Tz::Australia__South),
        ("Etc/GMT", Tz::Etc__GMT),
        ("Africa/Bissau", Tz::Africa__Bissau),
        ("Etc/GMT-7", Tz::Etc__GMTMinus7),
        ("Pacific/Pago_Pago", Tz::Pacific__Pago_Pago),
        ("America/Araguaina", Tz::America__Araguaina),
        ("America/Ciudad_Juarez", Tz::America__Ciudad_Juarez),
        ("Asia/Jayapura", Tz::Asia__Jayapura),
        ("Europe/Budapest", Tz::Europe__Budapest),
        ("America/Creston", Tz::America__Creston),
        ("Europe/Warsaw", Tz::Europe__Warsaw),
        ("Etc/GMT+1", Tz::Etc__GMTPlus1),
        ("Pacific/Rarotonga", Tz::Pacific__Rarotonga),
        ("Etc/GMT-5", Tz::Etc__GMTMinus5),
        ("Asia/Pontianak", Tz::Asia__Pontianak),
        ("Asia/Phnom_Penh", Tz::Asia__Phnom_Penh),
        ("Australia/Currie", Tz::Australia__Currie),
        ("Australia/Queensland", Tz::Australia__Queensland),
        ("Atlantic/Reykjavik", Tz::Atlantic__Reykjavik),
        ("Pacific/Auckland", Tz::Pacific__Auckland),
        ("Asia/Jerusalem", Tz::Asia__Jerusalem),
        ("Pacific/Chatham", Tz::Pacific__Chatham),
        ("Asia/Ust-Nera", Tz::Asia__UstNera),
        ("America/Tijuana", Tz::America__Tijuana),
        ("Pacific/Apia", Tz::Pacific__Apia),
        ("Asia/Kathmandu", Tz::Asia__Kathmandu),
        ("Canada/Atlantic", Tz::Canada__Atlantic),
        ("Pacific/Saipan", Tz::Pacific__Saipan),
        ("America/Argentina/Cordoba", Tz::America__Argentina__Cordoba),
        ("Asia/Macao", Tz::Asia__Macao),
        ("Europe/Zaporozhye", Tz::Europe__Zaporozhye),
        ("America/Kentucky/Louisville", Tz::America__Kentucky__Louisville),
        ("America/Argentina/Ushuaia", Tz::America__Argentina__Ushuaia),
        ("WET", Tz::WET),
        ("Europe/Uzhgorod", Tz::Europe__Uzhgorod),
        ("Etc/UCT", Tz::Etc__UCT),
        ("Etc/GMT-8", Tz::Etc__GMTMinus8),
        ("Poland", Tz::Poland),
        ("Africa/Cairo", Tz::Africa__Cairo),
        ("Asia/Singapore", Tz::Asia__Singapore),
        ("Asia/Bahrain", Tz::Asia__Bahrain),
        ("Europe/Andorra", Tz::Europe__Andorra),
        ("America/Bahia_Banderas", Tz::America__Bahia_Banderas),
        ("Indian/Comoro", Tz::Indian__Comoro),
        ("Asia/Gaza", Tz::Asia__Gaza),
        ("America/Santo_Domingo", Tz::America__Santo_Domingo),
        ("Europe/Vienna", Tz::Europe__Vienna),
        ("America/Dawson", Tz::America__Dawson),
        ("America/Indiana/Vevay", Tz::America__Indiana__Vevay),
        ("Brazil/DeNoronha", Tz::Brazil__DeNoronha),
        ("Asia/Almaty", Tz::Asia__Almaty),
        ("Europe/Stockholm", Tz::Europe__Stockholm),
        ("Europe/Riga", Tz::Europe__Riga),
        ("Africa/Addis_Ababa", Tz::Africa__Addis_Ababa),
        ("Asia/Muscat", Tz::Asia__Muscat),
        ("America/Indiana/Knox", Tz::America__Indiana__Knox),
        ("Europe/Sarajevo", Tz::Europe__Sarajevo),
        ("America/Danmarkshavn", Tz::America__Danmarkshavn),
        ("America/Rankin_Inlet", Tz::America__Rankin_Inlet),
        ("Africa/Ouagadougou", Tz::Africa__Ouagadougou),
        ("Africa/Porto-Novo", Tz::Africa__PortoNovo),
        ("America/Toronto", Tz::America__Toronto),
        ("America/Vancouver", Tz::America__Vancouver),
        ("Africa/Johannesburg", Tz::Africa__Johannesburg),
        ("Europe/Tiraspol", Tz::Europe__Tiraspol),
        ("Africa/Lusaka", Tz::Africa__Lusaka),
        ("America/Argentina/Jujuy", Tz::America__Argentina__Jujuy),
        ("GMT0", Tz::GMT0),
        ("America/Rainy_River", Tz::America__Rainy_River),
        ("Australia/Hobart", Tz::Australia__Hobart),
        ("EST", Tz::EST),
        ("Europe/Prague", Tz::Europe__Prague),
        ("Etc/GMT-0", Tz::Etc__GMTMinus0),
        ("Antarctica/Troll", Tz::Antarctica__Troll),
        ("Africa/Mogadishu", Tz::Africa__Mogadishu),
        ("America/Argentina/Buenos_Aires", Tz::America__Argentina__Buenos_Aires),
        ("America/Adak", Tz::America__Adak),
        ("MST", Tz::MST),
        ("Asia/Dubai", Tz::Asia__Dubai),
        ("Pacific/Easter", Tz::Pacific__Easter),
        ("Africa/Lome", Tz::Africa__Lome),
        ("Pacific/Norfolk", Tz::Pacific__Norfolk),
        ("Atlantic/Faroe", Tz::Atlantic__Faroe),
        ("Atlantic/St_Helena", Tz::Atlantic__St_Helena),
        ("Etc/GMT-9", Tz::Etc__GMTMinus9),
        ("America/St_Barthelemy", Tz::America__St_Barthelemy),
        ("Africa/Niamey", Tz::Africa__Niamey),
        ("Africa/Libreville", Tz::Africa__Libreville),
        ("Asia/Famagusta", Tz::Asia__Famagusta),
        ("America/Goose_Bay", Tz::America__Goose_Bay),
        ("Zulu", Tz::Zulu),
        ("Australia/Yancowinna", Tz::Australia__Yancowinna),
        ("Europe/San_Marino", Tz::Europe__San_Marino),
        ("Asia/Jakarta", Tz::Asia__Jakarta),
        ("America/Costa_Rica", Tz::America__Costa_Rica),
        ("Asia/Kamchatka", Tz::Asia__Kamchatka),
        ("Europe/Amsterdam", Tz::Europe__Amsterdam),
        ("Asia/Yangon", Tz::Asia__Yangon),
        ("Antarctica/Syowa", Tz::Antarctica__Syowa),
        ("Etc/GMT+5", Tz::Etc__GMTPlus5),
        ("America/Montevideo", Tz::America__Montevideo),
        ("Africa/Algiers", Tz::Africa__Algiers),
        ("America/Thule", Tz::America__Thule),
        ("Europe/Bratislava", Tz::Europe__Bratislava),
        ("Indian/Christmas", Tz::Indian__Christmas),
        ("Asia/Baku", Tz::Asia__Baku),
        ("Etc/GMT-10", Tz::Etc__GMTMinus10),
        ("Africa/Malabo", Tz::Africa__Malabo),
        ("America/Kentucky/Monticello", Tz::America__Kentucky__Monticello),
        ("America/Atka", Tz::America__Atka),
        ("America/Asuncion", Tz::America__Asuncion),
        ("America/North_Dakota/Beulah", Tz::America__North_Dakota__Beulah),
        ("Atlantic/Stanley", Tz::Atlantic__Stanley),
        ("Israel", Tz::Israel),
        ("America/Miquelon", Tz::America__Miquelon),
        ("Asia/Hovd", Tz::Asia__Hovd),
        ("Europe/Zurich", Tz::Europe__Zurich),
        ("America/Curacao", Tz::America__Curacao),
        ("America/Phoenix", Tz::America__Phoenix),
        ("America/Indiana/Petersburg", Tz::America__Indiana__Petersburg),
        ("Europe/Simferopol", Tz::Europe__Simferopol),
        ("America/St_Lucia", Tz::America__St_Lucia),
        ("Asia/Tashkent", Tz::Asia__Tashkent),
        ("Pacific/Truk", Tz::Pacific__Truk),
        ("Asia/Hebron", Tz::Asia__Hebron),
        ("Pacific/Tongatapu", Tz::Pacific__Tongatapu),
        ("Eire", Tz::Eire),
        ("Europe/Ulyanovsk", Tz::Europe__Ulyanovsk),
        ("America/Port-au-Prince", Tz::America__PortauPrince),
        ("Africa/Timbuktu", Tz::Africa__Timbuktu),
        ("Indian/Maldives", Tz::Indian__Maldives),
        ("Asia/Colombo", Tz::Asia__Colombo),
        ("Africa/Juba", Tz::Africa__Juba),
        ("Australia/Lindeman", Tz::Australia__Lindeman),
        ("Egypt", Tz::Egypt),
        ("Asia/Kuala_Lumpur", Tz::Asia__Kuala_Lumpur),
        ("GB", Tz::GB),
        ("America/Puerto_Rico", Tz::America__Puerto_Rico),
        ("America/Fortaleza", Tz::America__Fortaleza),
        ("America/El_Salvador", Tz::America__El_Salvador),
        ("US/Arizona", Tz::US__Arizona),
        ("America/Metlakatla", Tz::America__Metlakatla),
        ("Iceland", Tz::Iceland),
        ("Europe/Kiev", Tz::Europe__Kiev),
        ("America/Chihuahua", Tz::America__Chihuahua),
        ("Africa/Bamako", Tz::Africa__Bamako),
        ("Australia/Adelaide", Tz::Australia__Adelaide),
        ("Australia/Broken_Hill", Tz::Australia__Broken_Hill),
        ("Asia/Atyrau", Tz::Asia__Atyrau),
        ("Asia/Qatar", Tz::Asia__Qatar),
        ("America/Santiago", Tz::America__Santiago),
        ("America/St_Thomas", Tz::America__St_Thomas),
        ("Etc/UTC", Tz::Etc__UTC),
        ("Etc/Universal", Tz::Etc__Universal),
        ("Africa/Bujumbura", Tz::Africa__Bujumbura),
        ("America/Santa_Isabel", Tz::America__Santa_Isabel),
        ("Pacific/Fiji", Tz::Pacific__Fiji),
        ("Asia/Makassar", Tz::Asia__Makassar),
        ("Asia/Thimbu", Tz::Asia__Thimbu),
        ("America/Nuuk", Tz::America__Nuuk),
        ("Africa/Banjul", Tz::Africa__Banjul),
        ("America/Cuiaba", Tz::America__Cuiaba),
        ("America/Grand_Turk", Tz::America__Grand_Turk),
        ("Pacific/Samoa", Tz::Pacific__Samoa),
    ],
};
#[derive(Copy, Clone, Debug, PartialEq, Eq)]
pub struct ParseError(());

impl Display for ParseError {
    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
        f.write_str("failed to parse timezone")
    }
}

#[cfg(feature = "std")]
impl std::error::Error for ParseError {}

impl FromStr for Tz {
    type Err = ParseError;
    fn from_str(s: &str) -> Result<Self, Self::Err> {
        return TIMEZONES.get(s).cloned().ok_or(ParseError(()));
    }
}

impl Tz {
    pub fn name(self) -> &'static str {
        match self {
            Tz::Africa__Abidjan => "Africa/Abidjan",
            Tz::Africa__Accra => "Africa/Accra",
            Tz::Africa__Addis_Ababa => "Africa/Addis_Ababa",
            Tz::Africa__Algiers => "Africa/Algiers",
            Tz::Africa__Asmara => "Africa/Asmara",
            Tz::Africa__Asmera => "Africa/Asmera",
            Tz::Africa__Bamako => "Africa/Bamako",
            Tz::Africa__Bangui => "Africa/Bangui",
            Tz::Africa__Banjul => "Africa/Banjul",
            Tz::Africa__Bissau => "Africa/Bissau",
            Tz::Africa__Blantyre => "Africa/Blantyre",
            Tz::Africa__Brazzaville => "Africa/Brazzaville",
            Tz::Africa__Bujumbura => "Africa/Bujumbura",
            Tz::Africa__Cairo => "Africa/Cairo",
            Tz::Africa__Casablanca => "Africa/Casablanca",
            Tz::Africa__Ceuta => "Africa/Ceuta",
            Tz::Africa__Conakry => "Africa/Conakry",
            Tz::Africa__Dakar => "Africa/Dakar",
            Tz::Africa__Dar_es_Salaam => "Africa/Dar_es_Salaam",
            Tz::Africa__Djibouti => "Africa/Djibouti",
            Tz::Africa__Douala => "Africa/Douala",
            Tz::Africa__El_Aaiun => "Africa/El_Aaiun",
            Tz::Africa__Freetown => "Africa/Freetown",
            Tz::Africa__Gaborone => "Africa/Gaborone",
            Tz::Africa__Harare => "Africa/Harare",
            Tz::Africa__Johannesburg => "Africa/Johannesburg",
            Tz::Africa__Juba => "Africa/Juba",
            Tz::Africa__Kampala => "Africa/Kampala",
            Tz::Africa__Khartoum => "Africa/Khartoum",
            Tz::Africa__Kigali => "Africa/Kigali",
            Tz::Africa__Kinshasa => "Africa/Kinshasa",
            Tz::Africa__Lagos => "Africa/Lagos",
            Tz::Africa__Libreville => "Africa/Libreville",
            Tz::Africa__Lome => "Africa/Lome",
            Tz::Africa__Luanda => "Africa/Luanda",
            Tz::Africa__Lubumbashi => "Africa/Lubumbashi",
            Tz::Africa__Lusaka => "Africa/Lusaka",
            Tz::Africa__Malabo => "Africa/Malabo",
            Tz::Africa__Maputo => "Africa/Maputo",
            Tz::Africa__Maseru => "Africa/Maseru",
            Tz::Africa__Mbabane => "Africa/Mbabane",
            Tz::Africa__Mogadishu => "Africa/Mogadishu",
            Tz::Africa__Monrovia => "Africa/Monrovia",
            Tz::Africa__Nairobi => "Africa/Nairobi",
            Tz::Africa__Ndjamena => "Africa/Ndjamena",
            Tz::Africa__Niamey => "Africa/Niamey",
            Tz::Africa__Nouakchott => "Africa/Nouakchott",
            Tz::Africa__Ouagadougou => "Africa/Ouagadougou",
            Tz::Africa__PortoNovo => "Africa/Porto-Novo",
            Tz::Africa__Sao_Tome => "Africa/Sao_Tome",
            Tz::Africa__Timbuktu => "Africa/Timbuktu",
            Tz::Africa__Tripoli => "Africa/Tripoli",
            Tz::Africa__Tunis => "Africa/Tunis",
            Tz::Africa__Windhoek => "Africa/Windhoek",
            Tz::America__Adak => "America/Adak",
            Tz::America__Anchorage => "America/Anchorage",
            Tz::America__Anguilla => "America/Anguilla",
            Tz::America__Antigua => "America/Antigua",
            Tz::America__Araguaina => "America/Araguaina",
            Tz::America__Argentina__Buenos_Aires => "America/Argentina/Buenos_Aires",
            Tz::America__Argentina__Catamarca => "America/Argentina/Catamarca",
            Tz::America__Argentina__ComodRivadavia => "America/Argentina/ComodRivadavia",
            Tz::America__Argentina__Cordoba => "America/Argentina/Cordoba",
            Tz::America__Argentina__Jujuy => "America/Argentina/Jujuy",
            Tz::America__Argentina__La_Rioja => "America/Argentina/La_Rioja",
            Tz::America__Argentina__Mendoza => "America/Argentina/Mendoza",
            Tz::America__Argentina__Rio_Gallegos => "America/Argentina/Rio_Gallegos",
            Tz::America__Argentina__Salta => "America/Argentina/Salta",
            Tz::America__Argentina__San_Juan => "America/Argentina/San_Juan",
            Tz::America__Argentina__San_Luis => "America/Argentina/San_Luis",
            Tz::America__Argentina__Tucuman => "America/Argentina/Tucuman",
            Tz::America__Argentina__Ushuaia => "America/Argentina/Ushuaia",
            Tz::America__Aruba => "America/Aruba",
            Tz::America__Asuncion => "America/Asuncion",
            Tz::America__Atikokan => "America/Atikokan",
            Tz::America__Atka => "America/Atka",
            Tz::America__Bahia => "America/Bahia",
            Tz::America__Bahia_Banderas => "America/Bahia_Banderas",
            Tz::America__Barbados => "America/Barbados",
            Tz::America__Belem => "America/Belem",
            Tz::America__Belize => "America/Belize",
            Tz::America__BlancSablon => "America/Blanc-Sablon",
            Tz::America__Boa_Vista => "America/Boa_Vista",
            Tz::America__Bogota => "America/Bogota",
            Tz::America__Boise => "America/Boise",
            Tz::America__Buenos_Aires => "America/Buenos_Aires",
            Tz::America__Cambridge_Bay => "America/Cambridge_Bay",
            Tz::America__Campo_Grande => "America/Campo_Grande",
            Tz::America__Cancun => "America/Cancun",
            Tz::America__Caracas => "America/Caracas",
            Tz::America__Catamarca => "America/Catamarca",
            Tz::America__Cayenne => "America/Cayenne",
            Tz::America__Cayman => "America/Cayman",
            Tz::America__Chicago => "America/Chicago",
            Tz::America__Chihuahua => "America/Chihuahua",
            Tz::America__Ciudad_Juarez => "America/Ciudad_Juarez",
            Tz::America__Coral_Harbour => "America/Coral_Harbour",
            Tz::America__Cordoba => "America/Cordoba",
            Tz::America__Costa_Rica => "America/Costa_Rica",
            Tz::America__Creston => "America/Creston",
            Tz::America__Cuiaba => "America/Cuiaba",
            Tz::America__Curacao => "America/Curacao",
            Tz::America__Danmarkshavn => "America/Danmarkshavn",
            Tz::America__Dawson => "America/Dawson",
            Tz::America__Dawson_Creek => "America/Dawson_Creek",
            Tz::America__Denver => "America/Denver",
            Tz::America__Detroit => "America/Detroit",
            Tz::America__Dominica => "America/Dominica",
            Tz::America__Edmonton => "America/Edmonton",
            Tz::America__Eirunepe => "America/Eirunepe",
            Tz::America__El_Salvador => "America/El_Salvador",
            Tz::America__Ensenada => "America/Ensenada",
            Tz::America__Fort_Nelson => "America/Fort_Nelson",
            Tz::America__Fort_Wayne => "America/Fort_Wayne",
            Tz::America__Fortaleza => "America/Fortaleza",
            Tz::America__Glace_Bay => "America/Glace_Bay",
            Tz::America__Godthab => "America/Godthab",
            Tz::America__Goose_Bay => "America/Goose_Bay",
            Tz::America__Grand_Turk => "America/Grand_Turk",
            Tz::America__Grenada => "America/Grenada",
            Tz::America__Guadeloupe => "America/Guadeloupe",
            Tz::America__Guatemala => "America/Guatemala",
            Tz::America__Guayaquil => "America/Guayaquil",
            Tz::America__Guyana => "America/Guyana",
            Tz::America__Halifax => "America/Halifax",
            Tz::America__Havana => "America/Havana",
            Tz::America__Hermosillo => "America/Hermosillo",
            Tz::America__Indiana__Indianapolis => "America/Indiana/Indianapolis",
            Tz::America__Indiana__Knox => "America/Indiana/Knox",
            Tz::America__Indiana__Marengo => "America/Indiana/Marengo",
            Tz::America__Indiana__Petersburg => "America/Indiana/Petersburg",
            Tz::America__Indiana__Tell_City => "America/Indiana/Tell_City",
            Tz::America__Indiana__Vevay => "America/Indiana/Vevay",
            Tz::America__Indiana__Vincennes => "America/Indiana/Vincennes",
            Tz::America__Indiana__Winamac => "America/Indiana/Winamac",
            Tz::America__Indianapolis => "America/Indianapolis",
            Tz::America__Inuvik => "America/Inuvik",
            Tz::America__Iqaluit => "America/Iqaluit",
            Tz::America__Jamaica => "America/Jamaica",
            Tz::America__Jujuy => "America/Jujuy",
            Tz::America__Juneau => "America/Juneau",
            Tz::America__Kentucky__Louisville => "America/Kentucky/Louisville",
            Tz::America__Kentucky__Monticello => "America/Kentucky/Monticello",
            Tz::America__Knox_IN => "America/Knox_IN",
            Tz::America__Kralendijk => "America/Kralendijk",
            Tz::America__La_Paz => "America/La_Paz",
            Tz::America__Lima => "America/Lima",
            Tz::America__Los_Angeles => "America/Los_Angeles",
            Tz::America__Louisville => "America/Louisville",
            Tz::America__Lower_Princes => "America/Lower_Princes",
            Tz::America__Maceio => "America/Maceio",
            Tz::America__Managua => "America/Managua",
            Tz::America__Manaus => "America/Manaus",
            Tz::America__Marigot => "America/Marigot",
            Tz::America__Martinique => "America/Martinique",
            Tz::America__Matamoros => "America/Matamoros",
            Tz::America__Mazatlan => "America/Mazatlan",
            Tz::America__Mendoza => "America/Mendoza",
            Tz::America__Menominee => "America/Menominee",
            Tz::America__Merida => "America/Merida",
            Tz::America__Metlakatla => "America/Metlakatla",
            Tz::America__Mexico_City => "America/Mexico_City",
            Tz::America__Miquelon => "America/Miquelon",
            Tz::America__Moncton => "America/Moncton",
            Tz::America__Monterrey => "America/Monterrey",
            Tz::America__Montevideo => "America/Montevideo",
            Tz::America__Montreal => "America/Montreal",
            Tz::America__Montserrat => "America/Montserrat",
            Tz::America__Nassau => "America/Nassau",
            Tz::America__New_York => "America/New_York",
            Tz::America__Nipigon => "America/Nipigon",
            Tz::America__Nome => "America/Nome",
            Tz::America__Noronha => "America/Noronha",
            Tz::America__North_Dakota__Beulah => "America/North_Dakota/Beulah",
            Tz::America__North_Dakota__Center => "America/North_Dakota/Center",
            Tz::America__North_Dakota__New_Salem => "America/North_Dakota/New_Salem",
            Tz::America__Nuuk => "America/Nuuk",
            Tz::America__Ojinaga => "America/Ojinaga",
            Tz::America__Panama => "America/Panama",
            Tz::America__Pangnirtung => "America/Pangnirtung",
            Tz::America__Paramaribo => "America/Paramaribo",
            Tz::America__Phoenix => "America/Phoenix",
            Tz::America__PortauPrince => "America/Port-au-Prince",
            Tz::America__Port_of_Spain => "America/Port_of_Spain",
            Tz::America__Porto_Acre => "America/Porto_Acre",
            Tz::America__Porto_Velho => "America/Porto_Velho",
            Tz::America__Puerto_Rico => "America/Puerto_Rico",
            Tz::America__Punta_Arenas => "America/Punta_Arenas",
            Tz::America__Rainy_River => "America/Rainy_River",
            Tz::America__Rankin_Inlet => "America/Rankin_Inlet",
            Tz::America__Recife => "America/Recife",
            Tz::America__Regina => "America/Regina",
            Tz::America__Resolute => "America/Resolute",
            Tz::America__Rio_Branco => "America/Rio_Branco",
            Tz::America__Rosario => "America/Rosario",
            Tz::America__Santa_Isabel => "America/Santa_Isabel",
            Tz::America__Santarem => "America/Santarem",
            Tz::America__Santiago => "America/Santiago",
            Tz::America__Santo_Domingo => "America/Santo_Domingo",
            Tz::America__Sao_Paulo => "America/Sao_Paulo",
            Tz::America__Scoresbysund => "America/Scoresbysund",
            Tz::America__Shiprock => "America/Shiprock",
            Tz::America__Sitka => "America/Sitka",
            Tz::America__St_Barthelemy => "America/St_Barthelemy",
            Tz::America__St_Johns => "America/St_Johns",
            Tz::America__St_Kitts => "America/St_Kitts",
            Tz::America__St_Lucia => "America/St_Lucia",
            Tz::America__St_Thomas => "America/St_Thomas",
            Tz::America__St_Vincent => "America/St_Vincent",
            Tz::America__Swift_Current => "America/Swift_Current",
            Tz::America__Tegucigalpa => "America/Tegucigalpa",
            Tz::America__Thule => "America/Thule",
            Tz::America__Thunder_Bay => "America/Thunder_Bay",
            Tz::America__Tijuana => "America/Tijuana",
            Tz::America__Toronto => "America/Toronto",
            Tz::America__Tortola => "America/Tortola",
            Tz::America__Vancouver => "America/Vancouver",
            Tz::America__Virgin => "America/Virgin",
            Tz::America__Whitehorse => "America/Whitehorse",
            Tz::America__Winnipeg => "America/Winnipeg",
            Tz::America__Yakutat => "America/Yakutat",
            Tz::America__Yellowknife => "America/Yellowknife",
            Tz::Antarctica__Casey => "Antarctica/Casey",
            Tz::Antarctica__Davis => "Antarctica/Davis",
            Tz::Antarctica__DumontDUrville => "Antarctica/DumontDUrville",
            Tz::Antarctica__Macquarie => "Antarctica/Macquarie",
            Tz::Antarctica__Mawson => "Antarctica/Mawson",
            Tz::Antarctica__McMurdo => "Antarctica/McMurdo",
            Tz::Antarctica__Palmer => "Antarctica/Palmer",
            Tz::Antarctica__Rothera => "Antarctica/Rothera",
            Tz::Antarctica__South_Pole => "Antarctica/South_Pole",
            Tz::Antarctica__Syowa => "Antarctica/Syowa",
            Tz::Antarctica__Troll => "Antarctica/Troll",
            Tz::Antarctica__Vostok => "Antarctica/Vostok",
            Tz::Arctic__Longyearbyen => "Arctic/Longyearbyen",
            Tz::Asia__Aden => "Asia/Aden",
            Tz::Asia__Almaty => "Asia/Almaty",
            Tz::Asia__Amman => "Asia/Amman",
            Tz::Asia__Anadyr => "Asia/Anadyr",
            Tz::Asia__Aqtau => "Asia/Aqtau",
            Tz::Asia__Aqtobe => "Asia/Aqtobe",
            Tz::Asia__Ashgabat => "Asia/Ashgabat",
            Tz::Asia__Ashkhabad => "Asia/Ashkhabad",
            Tz::Asia__Atyrau => "Asia/Atyrau",
            Tz::Asia__Baghdad => "Asia/Baghdad",
            Tz::Asia__Bahrain => "Asia/Bahrain",
            Tz::Asia__Baku => "Asia/Baku",
            Tz::Asia__Bangkok => "Asia/Bangkok",
            Tz::Asia__Barnaul => "Asia/Barnaul",
            Tz::Asia__Beirut => "Asia/Beirut",
            Tz::Asia__Bishkek => "Asia/Bishkek",
            Tz::Asia__Brunei => "Asia/Brunei",
            Tz::Asia__Calcutta => "Asia/Calcutta",
            Tz::Asia__Chita => "Asia/Chita",
            Tz::Asia__Choibalsan => "Asia/Choibalsan",
            Tz::Asia__Chongqing => "Asia/Chongqing",
            Tz::Asia__Chungking => "Asia/Chungking",
            Tz::Asia__Colombo => "Asia/Colombo",
            Tz::Asia__Dacca => "Asia/Dacca",
            Tz::Asia__Damascus => "Asia/Damascus",
            Tz::Asia__Dhaka => "Asia/Dhaka",
            Tz::Asia__Dili => "Asia/Dili",
            Tz::Asia__Dubai => "Asia/Dubai",
            Tz::Asia__Dushanbe => "Asia/Dushanbe",
            Tz::Asia__Famagusta => "Asia/Famagusta",
            Tz::Asia__Gaza => "Asia/Gaza",
            Tz::Asia__Harbin => "Asia/Harbin",
            Tz::Asia__Hebron => "Asia/Hebron",
            Tz::Asia__Ho_Chi_Minh => "Asia/Ho_Chi_Minh",
            Tz::Asia__Hong_Kong => "Asia/Hong_Kong",
            Tz::Asia__Hovd => "Asia/Hovd",
            Tz::Asia__Irkutsk => "Asia/Irkutsk",
            Tz::Asia__Istanbul => "Asia/Istanbul",
            Tz::Asia__Jakarta => "Asia/Jakarta",
            Tz::Asia__Jayapura => "Asia/Jayapura",
            Tz::Asia__Jerusalem => "Asia/Jerusalem",
            Tz::Asia__Kabul => "Asia/Kabul",
            Tz::Asia__Kamchatka => "Asia/Kamchatka",
            Tz::Asia__Karachi => "Asia/Karachi",
            Tz::Asia__Kashgar => "Asia/Kashgar",
            Tz::Asia__Kathmandu => "Asia/Kathmandu",
            Tz::Asia__Katmandu => "Asia/Katmandu",
            Tz::Asia__Khandyga => "Asia/Khandyga",
            Tz::Asia__Kolkata => "Asia/Kolkata",
            Tz::Asia__Krasnoyarsk => "Asia/Krasnoyarsk",
            Tz::Asia__Kuala_Lumpur => "Asia/Kuala_Lumpur",
            Tz::Asia__Kuching => "Asia/Kuching",
            Tz::Asia__Kuwait => "Asia/Kuwait",
            Tz::Asia__Macao => "Asia/Macao",
            Tz::Asia__Macau => "Asia/Macau",
            Tz::Asia__Magadan => "Asia/Magadan",
            Tz::Asia__Makassar => "Asia/Makassar",
            Tz::Asia__Manila => "Asia/Manila",
            Tz::Asia__Muscat => "Asia/Muscat",
            Tz::Asia__Nicosia => "Asia/Nicosia",
            Tz::Asia__Novokuznetsk => "Asia/Novokuznetsk",
            Tz::Asia__Novosibirsk => "Asia/Novosibirsk",
            Tz::Asia__Omsk => "Asia/Omsk",
            Tz::Asia__Oral => "Asia/Oral",
            Tz::Asia__Phnom_Penh => "Asia/Phnom_Penh",
            Tz::Asia__Pontianak => "Asia/Pontianak",
            Tz::Asia__Pyongyang => "Asia/Pyongyang",
            Tz::Asia__Qatar => "Asia/Qatar",
            Tz::Asia__Qostanay => "Asia/Qostanay",
            Tz::Asia__Qyzylorda => "Asia/Qyzylorda",
            Tz::Asia__Rangoon => "Asia/Rangoon",
            Tz::Asia__Riyadh => "Asia/Riyadh",
            Tz::Asia__Saigon => "Asia/Saigon",
            Tz::Asia__Sakhalin => "Asia/Sakhalin",
            Tz::Asia__Samarkand => "Asia/Samarkand",
            Tz::Asia__Seoul => "Asia/Seoul",
            Tz::Asia__Shanghai => "Asia/Shanghai",
            Tz::Asia__Singapore => "Asia/Singapore",
            Tz::Asia__Srednekolymsk => "Asia/Srednekolymsk",
            Tz::Asia__Taipei => "Asia/Taipei",
            Tz::Asia__Tashkent => "Asia/Tashkent",
            Tz::Asia__Tbilisi => "Asia/Tbilisi",
            Tz::Asia__Tehran => "Asia/Tehran",
            Tz::Asia__Tel_Aviv => "Asia/Tel_Aviv",
            Tz::Asia__Thimbu => "Asia/Thimbu",
            Tz::Asia__Thimphu => "Asia/Thimphu",
            Tz::Asia__Tokyo => "Asia/Tokyo",
            Tz::Asia__Tomsk => "Asia/Tomsk",
            Tz::Asia__Ujung_Pandang => "Asia/Ujung_Pandang",
            Tz::Asia__Ulaanbaatar => "Asia/Ulaanbaatar",
            Tz::Asia__Ulan_Bator => "Asia/Ulan_Bator",
            Tz::Asia__Urumqi => "Asia/Urumqi",
            Tz::Asia__UstNera => "Asia/Ust-Nera",
            Tz::Asia__Vientiane => "Asia/Vientiane",
            Tz::Asia__Vladivostok => "Asia/Vladivostok",
            Tz::Asia__Yakutsk => "Asia/Yakutsk",
            Tz::Asia__Yangon => "Asia/Yangon",
            Tz::Asia__Yekaterinburg => "Asia/Yekaterinburg",
            Tz::Asia__Yerevan => "Asia/Yerevan",
            Tz::Atlantic__Azores => "Atlantic/Azores",
            Tz::Atlantic__Bermuda => "Atlantic/Bermuda",
            Tz::Atlantic__Canary => "Atlantic/Canary",
            Tz::Atlantic__Cape_Verde => "Atlantic/Cape_Verde",
            Tz::Atlantic__Faeroe => "Atlantic/Faeroe",
            Tz::Atlantic__Faroe => "Atlantic/Faroe",
            Tz::Atlantic__Jan_Mayen => "Atlantic/Jan_Mayen",
            Tz::Atlantic__Madeira => "Atlantic/Madeira",
            Tz::Atlantic__Reykjavik => "Atlantic/Reykjavik",
            Tz::Atlantic__South_Georgia => "Atlantic/South_Georgia",
            Tz::Atlantic__St_Helena => "Atlantic/St_Helena",
            Tz::Atlantic__Stanley => "Atlantic/Stanley",
            Tz::Australia__ACT => "Australia/ACT",
            Tz::Australia__Adelaide => "Australia/Adelaide",
            Tz::Australia__Brisbane => "Australia/Brisbane",
            Tz::Australia__Broken_Hill => "Australia/Broken_Hill",
            Tz::Australia__Canberra => "Australia/Canberra",
            Tz::Australia__Currie => "Australia/Currie",
            Tz::Australia__Darwin => "Australia/Darwin",
            Tz::Australia__Eucla => "Australia/Eucla",
            Tz::Australia__Hobart => "Australia/Hobart",
            Tz::Australia__LHI => "Australia/LHI",
            Tz::Australia__Lindeman => "Australia/Lindeman",
            Tz::Australia__Lord_Howe => "Australia/Lord_Howe",
            Tz::Australia__Melbourne => "Australia/Melbourne",
            Tz::Australia__NSW => "Australia/NSW",
            Tz::Australia__North => "Australia/North",
            Tz::Australia__Perth => "Australia/Perth",
            Tz::Australia__Queensland => "Australia/Queensland",
            Tz::Australia__South => "Australia/South",
            Tz::Australia__Sydney => "Australia/Sydney",
            Tz::Australia__Tasmania => "Australia/Tasmania",
            Tz::Australia__Victoria => "Australia/Victoria",
            Tz::Australia__West => "Australia/West",
            Tz::Australia__Yancowinna => "Australia/Yancowinna",
            Tz::Brazil__Acre => "Brazil/Acre",
            Tz::Brazil__DeNoronha => "Brazil/DeNoronha",
            Tz::Brazil__East => "Brazil/East",
            Tz::Brazil__West => "Brazil/West",
            Tz::CET => "CET",
            Tz::CST6CDT => "CST6CDT",
            Tz::Canada__Atlantic => "Canada/Atlantic",
            Tz::Canada__Central => "Canada/Central",
            Tz::Canada__Eastern => "Canada/Eastern",
            Tz::Canada__Mountain => "Canada/Mountain",
            Tz::Canada__Newfoundland => "Canada/Newfoundland",
            Tz::Canada__Pacific => "Canada/Pacific",
            Tz::Canada__Saskatchewan => "Canada/Saskatchewan",
            Tz::Canada__Yukon => "Canada/Yukon",
            Tz::Chile__Continental => "Chile/Continental",
            Tz::Chile__EasterIsland => "Chile/EasterIsland",
            Tz::Cuba => "Cuba",
            Tz::EET => "EET",
            Tz::EST => "EST",
            Tz::EST5EDT => "EST5EDT",
            Tz::Egypt => "Egypt",
            Tz::Eire => "Eire",
            Tz::Etc__GMT => "Etc/GMT",
            Tz::Etc__GMTPlus0 => "Etc/GMT+0",
            Tz::Etc__GMTPlus1 => "Etc/GMT+1",
            Tz::Etc__GMTPlus10 => "Etc/GMT+10",
            Tz::Etc__GMTPlus11 => "Etc/GMT+11",
            Tz::Etc__GMTPlus12 => "Etc/GMT+12",
            Tz::Etc__GMTPlus2 => "Etc/GMT+2",
            Tz::Etc__GMTPlus3 => "Etc/GMT+3",
            Tz::Etc__GMTPlus4 => "Etc/GMT+4",
            Tz::Etc__GMTPlus5 => "Etc/GMT+5",
            Tz::Etc__GMTPlus6 => "Etc/GMT+6",
            Tz::Etc__GMTPlus7 => "Etc/GMT+7",
            Tz::Etc__GMTPlus8 => "Etc/GMT+8",
            Tz::Etc__GMTPlus9 => "Etc/GMT+9",
            Tz::Etc__GMTMinus0 => "Etc/GMT-0",
            Tz::Etc__GMTMinus1 => "Etc/GMT-1",
            Tz::Etc__GMTMinus10 => "Etc/GMT-10",
            Tz::Etc__GMTMinus11 => "Etc/GMT-11",
            Tz::Etc__GMTMinus12 => "Etc/GMT-12",
            Tz::Etc__GMTMinus13 => "Etc/GMT-13",
            Tz::Etc__GMTMinus14 => "Etc/GMT-14",
            Tz::Etc__GMTMinus2 => "Etc/GMT-2",
            Tz::Etc__GMTMinus3 => "Etc/GMT-3",
            Tz::Etc__GMTMinus4 => "Etc/GMT-4",
            Tz::Etc__GMTMinus5 => "Etc/GMT-5",
            Tz::Etc__GMTMinus6 => "Etc/GMT-6",
            Tz::Etc__GMTMinus7 => "Etc/GMT-7",
            Tz::Etc__GMTMinus8 => "Etc/GMT-8",
            Tz::Etc__GMTMinus9 => "Etc/GMT-9",
            Tz::Etc__GMT0 => "Etc/GMT0",
            Tz::Etc__Greenwich => "Etc/Greenwich",
            Tz::Etc__UCT => "Etc/UCT",
            Tz::Etc__UTC => "Etc/UTC",
            Tz::Etc__Universal => "Etc/Universal",
            Tz::Etc__Zulu => "Etc/Zulu",
            Tz::Europe__Amsterdam => "Europe/Amsterdam",
            Tz::Europe__Andorra => "Europe/Andorra",
            Tz::Europe__Astrakhan => "Europe/Astrakhan",
            Tz::Europe__Athens => "Europe/Athens",
            Tz::Europe__Belfast => "Europe/Belfast",
            Tz::Europe__Belgrade => "Europe/Belgrade",
            Tz::Europe__Berlin => "Europe/Berlin",
            Tz::Europe__Bratislava => "Europe/Bratislava",
            Tz::Europe__Brussels => "Europe/Brussels",
            Tz::Europe__Bucharest => "Europe/Bucharest",
            Tz::Europe__Budapest => "Europe/Budapest",
            Tz::Europe__Busingen => "Europe/Busingen",
            Tz::Europe__Chisinau => "Europe/Chisinau",
            Tz::Europe__Copenhagen => "Europe/Copenhagen",
            Tz::Europe__Dublin => "Europe/Dublin",
            Tz::Europe__Gibraltar => "Europe/Gibraltar",
            Tz::Europe__Guernsey => "Europe/Guernsey",
            Tz::Europe__Helsinki => "Europe/Helsinki",
            Tz::Europe__Isle_of_Man => "Europe/Isle_of_Man",
            Tz::Europe__Istanbul => "Europe/Istanbul",
            Tz::Europe__Jersey => "Europe/Jersey",
            Tz::Europe__Kaliningrad => "Europe/Kaliningrad",
            Tz::Europe__Kiev => "Europe/Kiev",
            Tz::Europe__Kirov => "Europe/Kirov",
            Tz::Europe__Kyiv => "Europe/Kyiv",
            Tz::Europe__Lisbon => "Europe/Lisbon",
            Tz::Europe__Ljubljana => "Europe/Ljubljana",
            Tz::Europe__London => "Europe/London",
            Tz::Europe__Luxembourg => "Europe/Luxembourg",
            Tz::Europe__Madrid => "Europe/Madrid",
            Tz::Europe__Malta => "Europe/Malta",
            Tz::Europe__Mariehamn => "Europe/Mariehamn",
            Tz::Europe__Minsk => "Europe/Minsk",
            Tz::Europe__Monaco => "Europe/Monaco",
            Tz::Europe__Moscow => "Europe/Moscow",
            Tz::Europe__Nicosia => "Europe/Nicosia",
            Tz::Europe__Oslo => "Europe/Oslo",
            Tz::Europe__Paris => "Europe/Paris",
            Tz::Europe__Podgorica => "Europe/Podgorica",
            Tz::Europe__Prague => "Europe/Prague",
            Tz::Europe__Riga => "Europe/Riga",
            Tz::Europe__Rome => "Europe/Rome",
            Tz::Europe__Samara => "Europe/Samara",
            Tz::Europe__San_Marino => "Europe/San_Marino",
            Tz::Europe__Sarajevo => "Europe/Sarajevo",
            Tz::Europe__Saratov => "Europe/Saratov",
            Tz::Europe__Simferopol => "Europe/Simferopol",
            Tz::Europe__Skopje => "Europe/Skopje",
            Tz::Europe__Sofia => "Europe/Sofia",
            Tz::Europe__Stockholm => "Europe/Stockholm",
            Tz::Europe__Tallinn => "Europe/Tallinn",
            Tz::Europe__Tirane => "Europe/Tirane",
            Tz::Europe__Tiraspol => "Europe/Tiraspol",
            Tz::Europe__Ulyanovsk => "Europe/Ulyanovsk",
            Tz::Europe__Uzhgorod => "Europe/Uzhgorod",
            Tz::Europe__Vaduz => "Europe/Vaduz",
            Tz::Europe__Vatican => "Europe/Vatican",
            Tz::Europe__Vienna => "Europe/Vienna",
            Tz::Europe__Vilnius => "Europe/Vilnius",
            Tz::Europe__Volgograd => "Europe/Volgograd",
            Tz::Europe__Warsaw => "Europe/Warsaw",
            Tz::Europe__Zagreb => "Europe/Zagreb",
            Tz::Europe__Zaporozhye => "Europe/Zaporozhye",
            Tz::Europe__Zurich => "Europe/Zurich",
            Tz::GB => "GB",
            Tz::GBEire => "GB-Eire",
            Tz::GMT => "GMT",
            Tz::GMTPlus0 => "GMT+0",
            Tz::GMTMinus0 => "GMT-0",
            Tz::GMT0 => "GMT0",
            Tz::Greenwich => "Greenwich",
            Tz::HST => "HST",
            Tz::Hongkong => "Hongkong",
            Tz::Iceland => "Iceland",
            Tz::Indian__Antananarivo => "Indian/Antananarivo",
            Tz::Indian__Chagos => "Indian/Chagos",
            Tz::Indian__Christmas => "Indian/Christmas",
            Tz::Indian__Cocos => "Indian/Cocos",
            Tz::Indian__Comoro => "Indian/Comoro",
            Tz::Indian__Kerguelen => "Indian/Kerguelen",
            Tz::Indian__Mahe => "Indian/Mahe",
            Tz::Indian__Maldives => "Indian/Maldives",
            Tz::Indian__Mauritius => "Indian/Mauritius",
            Tz::Indian__Mayotte => "Indian/Mayotte",
            Tz::Indian__Reunion => "Indian/Reunion",
            Tz::Iran => "Iran",
            Tz::Israel => "Israel",
            Tz::Jamaica => "Jamaica",
            Tz::Japan => "Japan",
            Tz::Kwajalein => "Kwajalein",
            Tz::Libya => "Libya",
            Tz::MET => "MET",
            Tz::MST => "MST",
            Tz::MST7MDT => "MST7MDT",
            Tz::Mexico__BajaNorte => "Mexico/BajaNorte",
            Tz::Mexico__BajaSur => "Mexico/BajaSur",
            Tz::Mexico__General => "Mexico/General",
            Tz::NZ => "NZ",
            Tz::NZCHAT => "NZ-CHAT",
            Tz::Navajo => "Navajo",
            Tz::PRC => "PRC",
            Tz::PST8PDT => "PST8PDT",
            Tz::Pacific__Apia => "Pacific/Apia",
            Tz::Pacific__Auckland => "Pacific/Auckland",
            Tz::Pacific__Bougainville => "Pacific/Bougainville",
            Tz::Pacific__Chatham => "Pacific/Chatham",
            Tz::Pacific__Chuuk => "Pacific/Chuuk",
            Tz::Pacific__Easter => "Pacific/Easter",
            Tz::Pacific__Efate => "Pacific/Efate",
            Tz::Pacific__Enderbury => "Pacific/Enderbury",
            Tz::Pacific__Fakaofo => "Pacific/Fakaofo",
            Tz::Pacific__Fiji => "Pacific/Fiji",
            Tz::Pacific__Funafuti => "Pacific/Funafuti",
            Tz::Pacific__Galapagos => "Pacific/Galapagos",
            Tz::Pacific__Gambier => "Pacific/Gambier",
            Tz::Pacific__Guadalcanal => "Pacific/Guadalcanal",
            Tz::Pacific__Guam => "Pacific/Guam",
            Tz::Pacific__Honolulu => "Pacific/Honolulu",
            Tz::Pacific__Johnston => "Pacific/Johnston",
            Tz::Pacific__Kanton => "Pacific/Kanton",
            Tz::Pacific__Kiritimati => "Pacific/Kiritimati",
            Tz::Pacific__Kosrae => "Pacific/Kosrae",
            Tz::Pacific__Kwajalein => "Pacific/Kwajalein",
            Tz::Pacific__Majuro => "Pacific/Majuro",
            Tz::Pacific__Marquesas => "Pacific/Marquesas",
            Tz::Pacific__Midway => "Pacific/Midway",
            Tz::Pacific__Nauru => "Pacific/Nauru",
            Tz::Pacific__Niue => "Pacific/Niue",
            Tz::Pacific__Norfolk => "Pacific/Norfolk",
            Tz::Pacific__Noumea => "Pacific/Noumea",
            Tz::Pacific__Pago_Pago => "Pacific/Pago_Pago",
            Tz::Pacific__Palau => "Pacific/Palau",
            Tz::Pacific__Pitcairn => "Pacific/Pitcairn",
            Tz::Pacific__Pohnpei => "Pacific/Pohnpei",
            Tz::Pacific__Ponape => "Pacific/Ponape",
            Tz::Pacific__Port_Moresby => "Pacific/Port_Moresby",
            Tz::Pacific__Rarotonga => "Pacific/Rarotonga",
            Tz::Pacific__Saipan => "Pacific/Saipan",
            Tz::Pacific__Samoa => "Pacific/Samoa",
            Tz::Pacific__Tahiti => "Pacific/Tahiti",
            Tz::Pacific__Tarawa => "Pacific/Tarawa",
            Tz::Pacific__Tongatapu => "Pacific/Tongatapu",
            Tz::Pacific__Truk => "Pacific/Truk",
            Tz::Pacific__Wake => "Pacific/Wake",
            Tz::Pacific__Wallis => "Pacific/Wallis",
            Tz::Pacific__Yap => "Pacific/Yap",
            Tz::Poland => "Poland",
            Tz::Portugal => "Portugal",
            Tz::ROC => "ROC",
            Tz::ROK => "ROK",
            Tz::Singapore => "Singapore",
            Tz::Turkey => "Turkey",
            Tz::UCT => "UCT",
            Tz::US__Alaska => "US/Alaska",
            Tz::US__Aleutian => "US/Aleutian",
            Tz::US__Arizona => "US/Arizona",
            Tz::US__Central => "US/Central",
            Tz::US__EastIndiana => "US/East-Indiana",
            Tz::US__Eastern => "US/Eastern",
            Tz::US__Hawaii => "US/Hawaii",
            Tz::US__IndianaStarke => "US/Indiana-Starke",
            Tz::US__Michigan => "US/Michigan",
            Tz::US__Mountain => "US/Mountain",
            Tz::US__Pacific => "US/Pacific",
            Tz::US__Samoa => "US/Samoa",
            Tz::UTC => "UTC",
            Tz::Universal => "Universal",
            Tz::WSU => "W-SU",
            Tz::WET => "WET",
            Tz::Zulu => "Zulu",
        }
    }
}
impl Debug for Tz {
    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
        f.write_str(self.name().as_ref())
    }
}

impl Display for Tz {
    fn fmt(&self, f: &mut Formatter) -> fmt::Result {
        f.write_str(self.name().as_ref())
    }
}

impl TimeSpans for Tz {
    fn timespans(&self) -> FixedTimespanSet {
        match *self {
            Tz::Africa__Abidjan => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830383032, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -968,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Accra => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830383032, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -968,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Addis_Ababa => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1946168836, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1309746600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                    (-1261969200, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1041388200, FixedTimespan { utc_offset: 9900, dst_offset: 0, name: "+0245" }),
                    (-865305900, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8836,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Algiers => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2486592732, FixedTimespan { utc_offset: 561, dst_offset: 0, name: "PMT" }),
                    (-1855958961, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1689814800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1680397200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1665363600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1648342800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1635123600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1616893200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1604278800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1585443600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1574038800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1552266000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1539997200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1531443600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-956365200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-950486400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-942012000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796262400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766630800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-733280400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-439430400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-212029200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (41468400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (54774000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (231724800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (246236400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (259545600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (275274000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (309740400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (325468800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (341802000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (357523200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 732,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Asmara => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1946168836, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1309746600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                    (-1261969200, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1041388200, FixedTimespan { utc_offset: 9900, dst_offset: 0, name: "+0245" }),
                    (-865305900, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8836,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Asmera => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1946168836, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1309746600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                    (-1261969200, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1041388200, FixedTimespan { utc_offset: 9900, dst_offset: 0, name: "+0245" }),
                    (-865305900, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8836,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Bamako => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830383032, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -968,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Bangui => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2035584815, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1940889600, FixedTimespan { utc_offset: 815, dst_offset: 0, name: "LMT" }),
                    (-1767226415, FixedTimespan { utc_offset: 1800, dst_offset: 0, name: "+0030" }),
                    (-1588465800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "WAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 815,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Banjul => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830383032, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -968,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Bissau => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830376660, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (157770000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -3740,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Blantyre => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2109291020, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7820,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Brazzaville => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2035584815, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1940889600, FixedTimespan { utc_offset: 815, dst_offset: 0, name: "LMT" }),
                    (-1767226415, FixedTimespan { utc_offset: 1800, dst_offset: 0, name: "+0030" }),
                    (-1588465800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "WAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 815,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Bujumbura => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2109291020, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7820,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Cairo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2185409109, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-929844000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-923108400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-906170400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-892868400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-875844000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-857790000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-844308000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-825822000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-812685600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-794199600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-779853600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-762663600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-399088800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-386650800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-368330400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-355114800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-336790800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-323654400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-305168400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-292032000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-273632400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-260496000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-242096400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-228960000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-210560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-197424000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-178938000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-165801600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-147402000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-134265600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-115866000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-102643200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-84330000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-71107200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-52707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-39484800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-21171600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-7948800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (10364400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (23587200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (41900400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (55123200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (73522800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (86745600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (105058800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (118281600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (136594800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (149817600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (168130800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (181353600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (199753200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (212976000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (231289200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (244512000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (262825200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (276048000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (294361200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (307584000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (325983600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (339206400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (357519600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (370742400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (396399600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (402278400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (426812400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (433814400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (452214000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (465436800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (483750000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (496972800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (515286000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (528508800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (546822000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (560044800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (578444400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (591667200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (610412400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (623203200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (641516400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654739200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (673052400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686275200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (704674800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717897600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (736210800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749433600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (767746800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780969600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (799020000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (812322000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (830469600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (843771600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (861919200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (875221200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (893368800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (906670800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (925423200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (938725200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (956872800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (970174800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (988322400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1001624400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1019772000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1033074000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1051221600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1064523600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1083276000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1096578000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1114725600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1128027600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1146175200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1158872400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1177624800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1189112400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1209074400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1219957200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1240524000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1250802000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1272578400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1281474000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1284069600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1285880400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1400191200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1403816400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1406844000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1411678800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1682632800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698354000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1714082400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1730408400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1745532000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761858000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1776981600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1793307600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1809036000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824757200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1840485600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856206800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1871935200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887656400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1903384800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919710800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1934834400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1951160400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1966888800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982610000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1998338400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014059600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2029788000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045509200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2061237600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2076958800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2092687200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2109013200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2124136800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140462800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2156191200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2171912400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2187640800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203362000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2219090400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234811600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2250540000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266866000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2281989600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2298315600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2313439200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329765200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2345493600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361214800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2376943200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392664400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2408392800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424114000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2439842400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2456168400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2471292000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487618000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2503346400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519067600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2534796000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550517200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2566245600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2581966800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2597695200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2614021200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2629144800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645470800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2660594400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676920400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2692648800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708370000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2724098400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2739819600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2755548000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771269200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2786997600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2803323600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2818447200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834773200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2850501600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866222800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2881951200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897672400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2913400800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929122000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2944850400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960571600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2976300000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992626000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3007749600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3024075600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3039804000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055525200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3071253600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3086974800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3102703200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118424400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3134152800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150478800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3165602400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181928400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3197052000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3213378000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3229106400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3244827600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3260556000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3292005600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307726800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3323455200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3354904800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3371230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3386959200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3418408800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3449858400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465579600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3481308000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3512757600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3529083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3544207200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3576261600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3591982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3607711200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3639160800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3654882000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3670610400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3702060000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3718386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3734114400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3749835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3765564000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3797013600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3828463200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3859912800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3876238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3891362400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3923416800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3954866400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3986316000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002037200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4017765600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4034091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4049215200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4080664800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7509,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Casablanca => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1773012580, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (-956361600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (-950490000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (-942019200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (-761187600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (-617241600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (-605149200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (-81432000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (-71110800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (141264000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (147222000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (199756800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (207702000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (231292800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (244249200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (265507200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (271033200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (448243200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (504918000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1212278400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1220223600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1243814400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1250809200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1272758400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1281222000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1301788800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1312066800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1335664800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1342749600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1345428000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1348970400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1367114400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1373162400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1376100000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1382839200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1396144800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1403920800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1406944800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1414288800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1427594400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1434247200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1437271200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1445738400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1459044000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1465092000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1468116000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1477792800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1490493600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1495332000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1498960800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1509242400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1521943200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1526176800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1529200800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1540692000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1557021600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1560045600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1587261600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1590890400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1618106400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1621130400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1648346400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1651975200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1679191200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1682215200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1710036000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1713060000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1740276000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1743904800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1771120800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1774144800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1801965600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1804989600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1832205600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1835834400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1863050400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1866074400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1893290400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1896919200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1924135200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1927159200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1954980000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1958004000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1985220000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1988848800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2016064800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2019088800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2046304800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2049933600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2077149600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2080778400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2107994400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2111018400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2138234400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2141863200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2169079200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2172103200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2199924000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2202948000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2230164000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2233792800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2261008800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2264032800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2291248800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2294877600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2322093600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2325722400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2352938400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2355962400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2383178400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2386807200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2414023200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2417047200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2444868000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2447892000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2475108000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2478736800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2505952800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2508976800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2536192800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2539821600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2567037600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2570666400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2597882400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2600906400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2628122400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2631751200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2658967200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2661991200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2689812000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2692836000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2720052000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2723680800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2750896800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2753920800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2781136800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2784765600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2811981600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2815610400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2842826400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2845850400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2873066400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2876695200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2903911200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2906935200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2934756000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2937780000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2964996000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2968624800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2995840800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2998864800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3026080800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3029709600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3056925600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3060554400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3087770400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3090794400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3118010400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3121639200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3148855200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3151879200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3179700000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3182724000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3209940000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3213568800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3240784800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3243808800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3271024800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3274653600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3301869600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3305498400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3332714400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3335738400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3362954400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3366583200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3393799200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3396823200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3424644000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3427668000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3454884000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3458512800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3485728800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3488752800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3515968800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3519597600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3546813600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3549837600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3577658400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3580682400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3607898400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3611527200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3638743200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3641767200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3669588000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3672612000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3699828000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3703456800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -1820,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Ceuta => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177451524, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1630112400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1616810400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1442451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1427673600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1379293200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1364774400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1348448400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1333324800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1316390400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1301270400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-81432000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-71110800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (141264000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (147222000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (199756800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (207702000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (231292800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (244249200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (265507200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (271033200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (448243200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -1276,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Conakry => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830383032, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -968,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Dakar => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830383032, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -968,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Dar_es_Salaam => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1946168836, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1309746600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                    (-1261969200, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1041388200, FixedTimespan { utc_offset: 9900, dst_offset: 0, name: "+0245" }),
                    (-865305900, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8836,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Djibouti => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1946168836, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1309746600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                    (-1261969200, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1041388200, FixedTimespan { utc_offset: 9900, dst_offset: 0, name: "+0245" }),
                    (-865305900, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8836,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Douala => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2035584815, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1940889600, FixedTimespan { utc_offset: 815, dst_offset: 0, name: "LMT" }),
                    (-1767226415, FixedTimespan { utc_offset: 1800, dst_offset: 0, name: "+0030" }),
                    (-1588465800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "WAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 815,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__El_Aaiun => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1136070432, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (198291600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (199756800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (207702000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (231292800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (244249200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (265507200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (271033200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1212278400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1220223600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1243814400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1250809200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1272758400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1281222000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1301788800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1312066800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1335664800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1342749600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1345428000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1348970400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1367114400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1373162400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1376100000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1382839200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1396144800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1403920800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1406944800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1414288800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1427594400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1434247200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1437271200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1445738400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1459044000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1465092000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1468116000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1477792800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1490493600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1495332000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1498960800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1509242400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1521943200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1526176800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1529200800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "+01" }),
                    (1540692000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1557021600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1560045600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1587261600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1590890400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1618106400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1621130400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1648346400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1651975200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1679191200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1682215200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1710036000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1713060000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1740276000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1743904800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1771120800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1774144800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1801965600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1804989600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1832205600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1835834400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1863050400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1866074400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1893290400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1896919200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1924135200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1927159200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1954980000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1958004000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (1985220000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (1988848800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2016064800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2019088800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2046304800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2049933600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2077149600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2080778400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2107994400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2111018400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2138234400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2141863200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2169079200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2172103200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2199924000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2202948000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2230164000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2233792800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2261008800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2264032800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2291248800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2294877600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2322093600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2325722400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2352938400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2355962400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2383178400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2386807200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2414023200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2417047200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2444868000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2447892000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2475108000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2478736800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2505952800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2508976800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2536192800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2539821600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2567037600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2570666400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2597882400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2600906400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2628122400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2631751200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2658967200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2661991200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2689812000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2692836000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2720052000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2723680800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2750896800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2753920800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2781136800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2784765600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2811981600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2815610400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2842826400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2845850400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2873066400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2876695200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2903911200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2906935200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2934756000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2937780000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2964996000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2968624800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (2995840800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (2998864800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3026080800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3029709600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3056925600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3060554400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3087770400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3090794400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3118010400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3121639200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3148855200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3151879200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3179700000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3182724000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3209940000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3213568800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3240784800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3243808800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3271024800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3274653600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3301869600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3305498400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3332714400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3335738400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3362954400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3366583200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3393799200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3396823200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3424644000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3427668000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3454884000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3458512800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3485728800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3488752800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3515968800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3519597600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3546813600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3549837600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3577658400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3580682400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3607898400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3611527200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3638743200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3641767200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3669588000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3672612000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                    (3699828000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "+00" }),
                    (3703456800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "+01" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -3168,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Freetown => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830383032, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -968,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Gaborone => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2109291020, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7820,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Harare => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2109291020, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7820,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Johannesburg => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2458173120, FixedTimespan { utc_offset: 5400, dst_offset: 0, name: "SAST" }),
                    (-2109288600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "SAST" }),
                    (-860976000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "SAST" }),
                    (-845254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "SAST" }),
                    (-829526400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "SAST" }),
                    (-813805200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "SAST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 6720,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Juba => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1230775588, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (10360800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (24786000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (41810400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (56322000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (73432800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (87944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (104882400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (119480400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (136332000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (151016400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (167781600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (182552400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (199231200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (214174800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (230680800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (245710800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (262735200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (277246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (294184800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (308782800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (325634400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (340405200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (357084000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (371941200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (388533600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (403477200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (419983200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (435013200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (452037600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (466635600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (483487200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (498171600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (947930400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                    (1612126800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7588,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Kampala => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1946168836, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1309746600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                    (-1261969200, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1041388200, FixedTimespan { utc_offset: 9900, dst_offset: 0, name: "+0245" }),
                    (-865305900, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8836,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Khartoum => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1230775808, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (10360800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (24786000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (41810400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (56322000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (73432800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (87944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (104882400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (119480400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (136332000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (151016400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (167781600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (182552400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (199231200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (214174800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (230680800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (245710800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (262735200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (277246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (294184800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (308782800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (325634400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (340405200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (357084000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (371941200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (388533600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (403477200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (419983200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (435013200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (452037600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (466635600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (483487200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "CAST" }),
                    (498171600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (947930400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                    (1509483600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7808,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Kigali => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2109291020, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7820,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Kinshasa => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2035584815, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1940889600, FixedTimespan { utc_offset: 815, dst_offset: 0, name: "LMT" }),
                    (-1767226415, FixedTimespan { utc_offset: 1800, dst_offset: 0, name: "+0030" }),
                    (-1588465800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "WAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 815,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Lagos => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2035584815, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1940889600, FixedTimespan { utc_offset: 815, dst_offset: 0, name: "LMT" }),
                    (-1767226415, FixedTimespan { utc_offset: 1800, dst_offset: 0, name: "+0030" }),
                    (-1588465800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "WAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 815,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Libreville => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2035584815, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1940889600, FixedTimespan { utc_offset: 815, dst_offset: 0, name: "LMT" }),
                    (-1767226415, FixedTimespan { utc_offset: 1800, dst_offset: 0, name: "+0030" }),
                    (-1588465800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "WAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 815,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Lome => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830383032, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -968,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Luanda => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2035584815, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1940889600, FixedTimespan { utc_offset: 815, dst_offset: 0, name: "LMT" }),
                    (-1767226415, FixedTimespan { utc_offset: 1800, dst_offset: 0, name: "+0030" }),
                    (-1588465800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "WAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 815,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Lubumbashi => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2109291020, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7820,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Lusaka => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2109291020, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7820,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Malabo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2035584815, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1940889600, FixedTimespan { utc_offset: 815, dst_offset: 0, name: "LMT" }),
                    (-1767226415, FixedTimespan { utc_offset: 1800, dst_offset: 0, name: "+0030" }),
                    (-1588465800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "WAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 815,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Maputo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2109291020, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7820,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Maseru => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2458173120, FixedTimespan { utc_offset: 5400, dst_offset: 0, name: "SAST" }),
                    (-2109288600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "SAST" }),
                    (-860976000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "SAST" }),
                    (-845254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "SAST" }),
                    (-829526400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "SAST" }),
                    (-813805200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "SAST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 6720,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Mbabane => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2458173120, FixedTimespan { utc_offset: 5400, dst_offset: 0, name: "SAST" }),
                    (-2109288600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "SAST" }),
                    (-860976000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "SAST" }),
                    (-845254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "SAST" }),
                    (-829526400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "SAST" }),
                    (-813805200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "SAST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 6720,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Mogadishu => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1946168836, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1309746600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                    (-1261969200, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1041388200, FixedTimespan { utc_offset: 9900, dst_offset: 0, name: "+0245" }),
                    (-865305900, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8836,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Monrovia => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2776979812, FixedTimespan { utc_offset: -2588, dst_offset: 0, name: "MMT" }),
                    (-1604359012, FixedTimespan { utc_offset: -2670, dst_offset: 0, name: "MMT" }),
                    (63593070, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -2588,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Nairobi => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1946168836, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1309746600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                    (-1261969200, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1041388200, FixedTimespan { utc_offset: 9900, dst_offset: 0, name: "+0245" }),
                    (-865305900, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8836,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Ndjamena => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830387612, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "WAT" }),
                    (308703600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "WAST" }),
                    (321314400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "WAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3612,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Niamey => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2035584815, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1940889600, FixedTimespan { utc_offset: 815, dst_offset: 0, name: "LMT" }),
                    (-1767226415, FixedTimespan { utc_offset: 1800, dst_offset: 0, name: "+0030" }),
                    (-1588465800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "WAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 815,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Nouakchott => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830383032, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -968,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Ouagadougou => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830383032, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -968,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__PortoNovo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2035584815, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1940889600, FixedTimespan { utc_offset: 815, dst_offset: 0, name: "LMT" }),
                    (-1767226415, FixedTimespan { utc_offset: 1800, dst_offset: 0, name: "+0030" }),
                    (-1588465800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "WAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 815,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Sao_Tome => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713912016, FixedTimespan { utc_offset: -2205, dst_offset: 0, name: "LMT" }),
                    (-1830381795, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1514768400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "WAT" }),
                    (1546304400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 1616,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Timbuktu => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830383032, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -968,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Tripoli => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577926364, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-574902000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-568087200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-512175600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-504928800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-449888400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-441856800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-347158800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (378684000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386463600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (402271200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417999600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433807200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449622000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465429600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481590000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496965600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512953200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528674400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (544230000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (560037600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575852400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591660000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (607388400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (623196000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (641775600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (844034400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (860108400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (875916000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1352505600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364515200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382659200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3164,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Tunis => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2797202444, FixedTimespan { utc_offset: 561, dst_offset: 0, name: "PMT" }),
                    (-1855958961, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-969242400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-950493600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-941940000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-891136800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-877827600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-842918400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-842223600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828230400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796269600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766634400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (231202800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (243903600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (262825200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (276044400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (581122800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591145200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606870000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622594800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (641516400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654649200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1114902000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1128038400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 2444,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Africa__Windhoek => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2458170504, FixedTimespan { utc_offset: 5400, dst_offset: 0, name: "+0130" }),
                    (-2109288600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "SAST" }),
                    (-860976000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "SAST" }),
                    (-845254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "SAST" }),
                    (637970400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (764200800, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (778640400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (796780800, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (810090000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (828835200, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (841539600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (860284800, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (873594000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (891734400, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (905043600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (923184000, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (936493200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (954633600, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (967942800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (986083200, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (999392400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1018137600, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1030842000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1049587200, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1062896400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1081036800, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1094346000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1112486400, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1125795600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1143936000, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1157245200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1175385600, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1188694800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1207440000, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1220749200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1238889600, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1252198800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1270339200, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1283648400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1301788800, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1315098000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1333238400, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1346547600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1365292800, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1377997200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1396742400, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1410051600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1428192000, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1441501200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1459641600, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1472950800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                    (1491091200, FixedTimespan { utc_offset: 7200, dst_offset: -3600, name: "WAT" }),
                    (1504400400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "CAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 4104,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Adak => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3225223727, FixedTimespan { utc_offset: -42398, dst_offset: 0, name: "LMT" }),
                    (-2188944802, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "NST" }),
                    (-880196400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "NWT" }),
                    (-769395600, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "NPT" }),
                    (-765374400, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "NST" }),
                    (-86878800, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (-21466800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (-5745600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (9982800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (25704000, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (41432400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (57758400, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (73486800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (89208000, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (104936400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (120657600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (126709200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (152107200, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (162392400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (183556800, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (199285200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (215611200, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (230734800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (247060800, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (262789200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (278510400, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (294238800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (309960000, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (325688400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (341409600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (357138000, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (372859200, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (388587600, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (404913600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (420037200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (436363200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (439034400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (452088000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (467809200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (483537600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (499258800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (514987200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (530708400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (544622400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (562158000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (576072000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (594212400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (607521600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (625662000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (638971200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (657111600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (671025600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (688561200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (702475200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (720010800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (733924800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (752065200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (765374400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (783514800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (796824000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (814964400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (828878400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (846414000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (860328000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (877863600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (891777600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (909313200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (923227200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (941367600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (954676800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (972817200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (986126400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1004266800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1018180800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1035716400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1049630400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1067166000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1081080000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1099220400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1112529600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1130670000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1143979200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1162119600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1173614400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1194174000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1205064000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1225623600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1236513600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1257073200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1268568000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1289127600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1300017600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1320577200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1331467200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1352026800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1362916800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1383476400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1394366400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1414926000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1425816000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1446375600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1457870400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1478430000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1489320000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1509879600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1520769600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1541329200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1552219200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1572778800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1583668800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1604228400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1615723200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1636282800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1647172800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1667732400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1678622400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1699182000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1710072000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1730631600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1741521600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1762081200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1772971200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1793530800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1805025600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1825585200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1836475200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1857034800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1867924800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1888484400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1899374400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1919934000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1930824000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1951383600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1962878400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1983438000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1994328000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2014887600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2025777600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2046337200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2057227200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2077786800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2088676800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2109236400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2120126400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2140686000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2152180800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2172740400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2183630400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2204190000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2215080000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2235639600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2246529600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2267089200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2277979200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2298538800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2309428800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2329988400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2341483200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2362042800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2372932800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2393492400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2404382400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2424942000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2435832000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2456391600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2467281600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2487841200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2499336000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2519895600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2530785600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2551345200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2562235200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2582794800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2593684800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2614244400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2625134400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2645694000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2656584000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2677143600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2688638400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2709198000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2720088000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2740647600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2751537600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2772097200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2782987200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2803546800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2814436800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2834996400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2846491200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2867050800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2877940800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2898500400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2909390400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2929950000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2940840000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2961399600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2972289600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2992849200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3003739200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3024298800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3035793600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3056353200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3067243200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3087802800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3098692800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3119252400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3130142400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3150702000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3161592000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3182151600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3193041600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3213601200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3225096000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3245655600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3256545600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3277105200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3287995200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3308554800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3319444800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3340004400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3350894400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3371454000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3382948800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3403508400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3414398400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3434958000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3445848000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3466407600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3477297600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3497857200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3508747200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3529306800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3540196800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3560756400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3572251200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3592810800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3603700800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3624260400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3635150400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3655710000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3666600000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3687159600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3698049600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3718609200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3730104000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3750663600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3761553600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3782113200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3793003200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3813562800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3824452800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3845012400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3855902400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3876462000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3887352000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3907911600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3919406400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3939966000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3950856000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3971415600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3982305600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (4002865200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (4013755200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (4034314800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (4045204800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (4065764400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (4076654400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (4097214000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 44002,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Anchorage => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3225223727, FixedTimespan { utc_offset: -35976, dst_offset: 0, name: "LMT" }),
                    (-2188951224, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AST" }),
                    (-880200000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "APT" }),
                    (-765378000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AST" }),
                    (-86882400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (-21470400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (-5749200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (9979200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (25700400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (41428800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (57754800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (73483200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (89204400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (104932800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (120654000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (126705600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (152103600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (162388800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (183553200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (199281600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (215607600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (230731200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (247057200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (262785600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (278506800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (294235200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (309956400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (325684800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (341406000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (357134400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (372855600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (388584000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (404910000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (420033600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (436359600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (439030800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (452084400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (467805600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (483534000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (499255200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (514983600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (530704800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (544618800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (562154400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (576068400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (594208800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (607518000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (625658400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (638967600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (657108000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (671022000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (688557600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (702471600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (720007200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (733921200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (752061600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (765370800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (783511200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (796820400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (814960800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (828874800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (846410400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (860324400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (877860000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (891774000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (909309600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (923223600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (941364000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (954673200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (972813600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (986122800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1004263200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1018177200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1035712800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1049626800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1067162400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1081076400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1099216800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1112526000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1130666400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1143975600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1162116000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1173610800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1194170400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1205060400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1225620000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1236510000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1257069600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1268564400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1289124000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1300014000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1320573600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1331463600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1352023200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1362913200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1383472800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1394362800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1414922400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1425812400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1446372000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1457866800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1478426400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1489316400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1509876000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1520766000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1541325600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1552215600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1572775200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1583665200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1604224800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1615719600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1636279200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1647169200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1667728800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1678618800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1699178400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1710068400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1730628000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1741518000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1762077600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1772967600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1793527200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1805022000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1825581600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1836471600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1857031200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1867921200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1888480800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1899370800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1919930400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1930820400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1951380000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1962874800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1983434400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1994324400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2014884000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2025774000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2046333600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2057223600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2077783200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2088673200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2109232800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2120122800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2140682400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2152177200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2172736800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2183626800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2204186400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2215076400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2235636000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2246526000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2267085600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2277975600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2298535200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2309425200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2329984800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2341479600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2362039200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2372929200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2393488800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2404378800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2424938400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2435828400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2456388000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2467278000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2487837600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2499332400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2519892000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2530782000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2551341600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2562231600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2582791200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2593681200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2614240800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2625130800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2645690400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2656580400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2677140000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2688634800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2709194400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2720084400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2740644000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2751534000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2772093600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2782983600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2803543200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2814433200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2834992800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2846487600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2867047200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2877937200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2898496800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2909386800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2929946400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2940836400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2961396000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2972286000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2992845600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3003735600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3024295200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3035790000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3056349600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3067239600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3087799200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3098689200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3119248800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3130138800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3150698400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3161588400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3182148000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3193038000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3213597600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3225092400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3245652000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3256542000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3277101600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3287991600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3308551200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3319441200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3340000800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3350890800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3371450400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3382945200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3403504800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3414394800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3434954400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3445844400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3466404000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3477294000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3497853600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3508743600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3529303200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3540193200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3560752800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3572247600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3592807200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3603697200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3624256800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3635146800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3655706400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3666596400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3687156000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3698046000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3718605600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3730100400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3750660000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3761550000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3782109600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3792999600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3813559200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3824449200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3845008800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3855898800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3876458400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3887348400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3907908000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3919402800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3939962400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3950852400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3971412000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3982302000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4002861600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4013751600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4034311200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4045201200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4065760800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4076650800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4097210400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 50424,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Anguilla => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Antigua => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Araguaina => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767214032, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1206957600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1191362400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1175374800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1159826400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-633819600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-622069200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-602283600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-591832800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-570747600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-560210400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-539125200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-531352800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-191365200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-184197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-155163600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-150069600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-128898000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-121125600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-99954000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-89589600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-68418000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-57967200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (499748400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (511236000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (530593200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (540266400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (562129200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (571197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (592974000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (602042400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (634701600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (813726000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (824004000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (844570800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (856058400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (876106800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (888717600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (908074800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (919562400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (951616800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (970974000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (982461600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1003028400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1013911200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1036292400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1045360800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1350788400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1361066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -11568,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Argentina__Buenos_Aires => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372097972, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (667965600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (687927600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1224385200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1237082400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -14028,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Argentina__Catamarca => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372096212, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (667965600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687931200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1086058800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1087704000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15788,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Argentina__ComodRivadavia => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372096212, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (667965600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687931200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1086058800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1087704000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15788,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Argentina__Cordoba => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372096592, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (667965600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687931200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1224385200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1237082400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15408,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Argentina__Jujuy => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372096328, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (657086400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (669178800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (686721600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15672,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Argentina__La_Rioja => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372095956, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (667792800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (673588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (687927600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1086058800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1087704000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16044,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Argentina__Mendoza => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372095484, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (655963200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (667796400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687499200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (699418800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (719380800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1085281200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1096171200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16516,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Argentina__Rio_Gallegos => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372095388, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (667965600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (687927600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1086058800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1087704000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16612,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Argentina__Salta => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372096300, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (667965600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687931200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15700,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Argentina__San_Juan => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372095556, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (667792800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (673588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (687927600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1085972400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1090728000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16444,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Argentina__San_Luis => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372096076, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (637380000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (655963200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (667796400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (675748800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1085972400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1090728000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1200880800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1205031600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1223784000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1236481200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1255233600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15924,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Argentina__Tucuman => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372096348, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (667965600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687931200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1086058800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1087099200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1224385200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1237082400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15652,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Argentina__Ushuaia => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372095608, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (667965600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (687927600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1085886000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1087704000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16392,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Aruba => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Asuncion => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524507760, FixedTimespan { utc_offset: -13840, dst_offset: 0, name: "AMT" }),
                    (-1206389360, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (86760000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (134017200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (181368000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (194497200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (212990400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (226033200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (244526400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (257569200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (276062400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (291783600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (307598400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (323406000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (339220800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (354942000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (370756800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (386478000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (402292800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (418014000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (433828800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (449636400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (465451200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (481172400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (496987200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (512708400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (528523200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (544244400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (560059200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (575866800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (591681600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (607402800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (625032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (638938800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (654753600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (670474800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (686721600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (699418800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (718257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (733546800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (749448000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (762318000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (780984000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (793767600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (812520000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (825649200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (844574400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (856666800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (876024000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (888721200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (907473600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (920775600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (938923200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952225200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (970372800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (983674800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1002427200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1018148400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1030852800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1049598000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1062907200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1081047600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1097985600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1110682800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1129435200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1142132400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1160884800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1173582000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1192939200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1205031600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1224388800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1236481200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1255838400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1270954800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1286078400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1302404400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1317528000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1333854000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1349582400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1364094000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1381032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1395543600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1412481600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1426993200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1443931200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1459047600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1475380800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1490497200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1506830400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1521946800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1538884800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1553396400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1570334400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1584846000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1601784000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1616900400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1633233600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1648350000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1664683200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1679799600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1696132800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1711249200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1728187200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1742698800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1759636800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1774148400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1791086400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1806202800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1822536000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1837652400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1853985600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1869102000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1886040000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1900551600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1917489600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1932001200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1948939200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1964055600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1980388800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1995505200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2011838400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2026954800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2043288000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2058404400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2075342400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2089854000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2106792000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2121303600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2138241600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2153358000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2169691200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2184807600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2201140800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2216257200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2233195200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2247706800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2264644800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2279156400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2296094400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2310606000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2327544000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2342660400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2358993600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2374110000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2390443200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2405559600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2422497600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2437009200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2453947200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2468458800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2485396800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2500513200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2516846400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2531962800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2548296000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2563412400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2579745600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2594862000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2611800000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2626311600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2643249600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2657761200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2674699200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2689815600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2706148800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2721265200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2737598400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2752714800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2769652800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2784164400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2801102400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2815614000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2832552000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2847668400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2864001600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2879118000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2895451200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2910567600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2926900800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2942017200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2958955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2973466800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2990404800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3004916400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3021854400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3036970800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3053304000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3068420400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3084753600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3099870000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3116808000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3131319600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3148257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3162769200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3179707200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3194218800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3211156800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3226273200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3242606400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3257722800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3274056000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3289172400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3306110400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3320622000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3337560000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3352071600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3369009600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3384126000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3400459200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3415575600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3431908800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3447025200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3463358400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3478474800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3495412800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3509924400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3526862400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3541374000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3558312000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3573428400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3589761600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3604878000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3621211200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3636327600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3653265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3667777200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3684715200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3699226800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3716164800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3731281200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3747614400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3762730800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3779064000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3794180400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3810513600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3825630000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3842568000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3857079600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3874017600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3888529200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3905467200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3920583600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3936916800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3952033200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3968366400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3983482800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (4000420800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (4014932400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (4031870400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (4046382000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (4063320000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (4077831600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (4094769600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -13840,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Atikokan => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524502512, FixedTimespan { utc_offset: -19176, dst_offset: 0, name: "CMT" }),
                    (-1946918424, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -19088,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Atka => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3225223727, FixedTimespan { utc_offset: -42398, dst_offset: 0, name: "LMT" }),
                    (-2188944802, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "NST" }),
                    (-880196400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "NWT" }),
                    (-769395600, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "NPT" }),
                    (-765374400, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "NST" }),
                    (-86878800, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (-21466800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (-5745600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (9982800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (25704000, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (41432400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (57758400, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (73486800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (89208000, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (104936400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (120657600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (126709200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (152107200, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (162392400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (183556800, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (199285200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (215611200, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (230734800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (247060800, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (262789200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (278510400, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (294238800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (309960000, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (325688400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (341409600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (357138000, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (372859200, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (388587600, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (404913600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (420037200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (436363200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (439034400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (452088000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (467809200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (483537600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (499258800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (514987200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (530708400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (544622400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (562158000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (576072000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (594212400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (607521600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (625662000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (638971200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (657111600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (671025600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (688561200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (702475200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (720010800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (733924800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (752065200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (765374400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (783514800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (796824000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (814964400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (828878400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (846414000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (860328000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (877863600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (891777600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (909313200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (923227200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (941367600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (954676800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (972817200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (986126400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1004266800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1018180800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1035716400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1049630400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1067166000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1081080000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1099220400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1112529600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1130670000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1143979200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1162119600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1173614400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1194174000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1205064000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1225623600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1236513600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1257073200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1268568000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1289127600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1300017600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1320577200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1331467200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1352026800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1362916800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1383476400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1394366400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1414926000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1425816000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1446375600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1457870400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1478430000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1489320000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1509879600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1520769600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1541329200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1552219200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1572778800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1583668800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1604228400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1615723200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1636282800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1647172800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1667732400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1678622400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1699182000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1710072000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1730631600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1741521600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1762081200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1772971200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1793530800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1805025600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1825585200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1836475200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1857034800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1867924800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1888484400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1899374400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1919934000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1930824000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1951383600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1962878400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1983438000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1994328000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2014887600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2025777600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2046337200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2057227200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2077786800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2088676800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2109236400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2120126400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2140686000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2152180800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2172740400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2183630400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2204190000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2215080000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2235639600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2246529600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2267089200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2277979200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2298538800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2309428800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2329988400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2341483200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2362042800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2372932800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2393492400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2404382400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2424942000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2435832000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2456391600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2467281600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2487841200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2499336000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2519895600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2530785600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2551345200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2562235200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2582794800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2593684800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2614244400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2625134400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2645694000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2656584000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2677143600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2688638400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2709198000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2720088000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2740647600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2751537600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2772097200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2782987200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2803546800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2814436800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2834996400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2846491200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2867050800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2877940800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2898500400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2909390400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2929950000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2940840000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2961399600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2972289600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2992849200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3003739200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3024298800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3035793600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3056353200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3067243200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3087802800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3098692800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3119252400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3130142400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3150702000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3161592000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3182151600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3193041600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3213601200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3225096000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3245655600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3256545600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3277105200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3287995200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3308554800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3319444800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3340004400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3350894400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3371454000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3382948800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3403508400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3414398400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3434958000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3445848000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3466407600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3477297600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3497857200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3508747200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3529306800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3540196800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3560756400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3572251200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3592810800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3603700800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3624260400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3635150400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3655710000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3666600000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3687159600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3698049600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3718609200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3730104000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3750663600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3761553600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3782113200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3793003200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3813562800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3824452800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3845012400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3855902400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3876462000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3887352000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3907911600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3919406400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3939966000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3950856000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3971415600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3982305600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (4002865200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (4013755200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (4034314800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (4045204800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (4065764400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (4076654400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (4097214000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 44002,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Bahia => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767216356, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1206957600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1191362400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1175374800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1159826400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-633819600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-622069200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-602283600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-591832800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-570747600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-560210400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-539125200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-531352800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-191365200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-184197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-155163600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-150069600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-128898000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-121125600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-99954000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-89589600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-68418000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-57967200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (499748400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (511236000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (530593200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (540266400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (562129200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (571197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (592974000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (602042400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (634701600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (666756000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (687927600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (697600800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719982000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (728445600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (750826800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (761709600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (782276400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (793159200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (813726000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (824004000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (844570800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (856058400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (876106800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (888717600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (908074800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (919562400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (951616800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (970974000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (982461600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1003028400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1013911200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1036292400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1045360800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1318734000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1330221600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -9244,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Bahia_Banderas => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514714340, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1343066400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1234807200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1220292000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1207159200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1191344400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-873828000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-661539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (28800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (989139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1001836800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1175418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1193558400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1207472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225008000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1238922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1256457600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1270371600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1288508400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1301817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1319958000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1333267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1351407600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1365321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1382857200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1396771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414306800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1428220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1445756400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1459670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1477810800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1491120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509260400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1522569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1540710000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1554624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1586073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1603609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1617523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1635663600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1648972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25260,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Barbados => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1841256091, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-874263600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-862682400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-841604400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-830714400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-811882800, FixedTimespan { utc_offset: -14400, dst_offset: 1800, name: "-0330" }),
                    (-798660000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (234943200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (244616400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (261554400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (276066000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (293004000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (307515600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (325058400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (338706000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -14309,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Belem => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767213964, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1206957600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1191362400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1175374800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1159826400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-633819600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-622069200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-602283600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-591832800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-570747600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-560210400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-539125200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-531352800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-191365200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-184197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-155163600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-150069600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-128898000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-121125600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-99954000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-89589600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-68418000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-57967200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (499748400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (511236000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (530593200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (540266400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (562129200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (571197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -11636,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Belize => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1822500432, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1616954400, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1606069800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1585504800, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1574015400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1554055200, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1542565800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1522605600, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1511116200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1490551200, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1479666600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1459101600, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1448217000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1427652000, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1416162600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1396202400, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1384713000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1364752800, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1353263400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1333303200, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1321813800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1301248800, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1290364200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1269799200, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1258914600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1238349600, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1226860200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1206900000, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1195410600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1175450400, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1163961000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1143396000, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1132511400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1111946400, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1101061800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1080496800, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1069612200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1049047200, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1037557800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1017597600, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-1006108200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-986148000, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-974658600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-954093600, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-943209000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-922644000, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-911759400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-891194400, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-879705000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-868212000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-758746800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-701892000, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-690402600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-670442400, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-658953000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-638992800, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-627503400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-606938400, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-596053800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-575488800, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-564604200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-544039200, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-532549800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-512589600, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-501100200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-481140000, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-469650600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-449690400, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-438201000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-417636000, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-406751400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-386186400, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-375301800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-354736800, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-343247400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-323287200, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-311797800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-291837600, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-280348200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-259783200, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-248898600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-228333600, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-217449000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-196884000, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-185999400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-165434400, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-153945000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-133984800, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-122495400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-102535200, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-91045800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-70480800, FixedTimespan { utc_offset: -21600, dst_offset: 1800, name: "-0530" }),
                    (-59596200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (123919200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (129618000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (409039200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (413874000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -21168,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__BlancSablon => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Boa_Vista => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767211040, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1206954000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1191358800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1175371200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1159822800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-633816000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-622065600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-602280000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-591829200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-570744000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-560206800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-539121600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-531349200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-191361600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-155160000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-150066000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-128894400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-99950400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-89586000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-68414400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-57963600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (499752000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (511239600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (530596800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (540270000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (562132800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (571201200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (938923200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (951620400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (970977600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (971578800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -14560,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Bogota => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2707671824, FixedTimespan { utc_offset: -17776, dst_offset: 0, name: "BMT" }),
                    (-1739041424, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (704869200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (729057600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -17776,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Boise => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717612111, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1633269600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1615129200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1601820000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1583679600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1471788000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-84380400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-68659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-52930800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-37209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-21481200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-5760000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (9968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (25689600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (41418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (57744000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (73472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (89193600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (104922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (120643200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (129114000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (152092800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (162378000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (183542400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (199270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (215596800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (230720400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (247046400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (262774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (278496000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (294224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (309945600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (325674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (341395200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (357123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (372844800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (388573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (404899200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (420022800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (436348800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (452077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (467798400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (483526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (499248000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (514976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (530697600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (544611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (562147200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (576061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (594201600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (607510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (625651200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (638960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (657100800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (671014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (688550400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (702464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (720000000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (733914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (752054400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (765363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (783504000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (796813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (814953600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (986115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1004256000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1173603600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1194163200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1205053200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225612800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1236502800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1257062400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1268557200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1289116800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1300006800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1320566400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1331456400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1352016000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1362906000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1383465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1394355600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414915200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1425805200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1446364800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1457859600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1478419200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1489309200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509868800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1520758800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1541318400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1552208400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572768000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1583658000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1604217600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1615712400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1636272000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1647162000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667721600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1678611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1699171200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1710061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1730620800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1741510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1762070400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1772960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1793520000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1805014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1825574400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1836464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1857024000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1867914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1888473600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1899363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1919923200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1930813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1951372800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1962867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1983427200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1994317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2014876800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2025766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2046326400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2057216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2077776000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2088666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2109225600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2120115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2140675200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2152170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2172729600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2183619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2204179200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2215069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2235628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2246518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2267078400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2277968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2298528000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2309418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2329977600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2341472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2362032000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2372922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2393481600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2404371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2424931200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2435821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2456380800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2467270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2487830400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2499325200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2519884800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2530774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2551334400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2562224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2582784000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2593674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2614233600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2625123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2645683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2656573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2677132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2688627600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2709187200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2720077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2740636800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2751526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2772086400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2782976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2803536000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2814426000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2834985600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2846480400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2867040000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2877930000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2898489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2909379600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2929939200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2940829200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2961388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2972278800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2992838400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3003728400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3024288000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3035782800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3056342400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3067232400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3087792000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3098682000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3119241600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3130131600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3150691200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3161581200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3182140800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3193030800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3213590400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3225085200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3245644800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3256534800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3277094400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3287984400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3308544000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3319434000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3339993600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3350883600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3371443200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3382938000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3403497600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3414387600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3434947200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3445837200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3466396800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3477286800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3497846400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3508736400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3529296000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3540186000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3560745600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3572240400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3592800000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3603690000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3624249600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3635139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3655699200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3666589200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3687148800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3698038800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3718598400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3730093200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3750652800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3761542800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3782102400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3792992400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3813552000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3824442000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3845001600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3855891600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3876451200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3887341200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3907900800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3919395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3939955200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3950845200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3971404800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3982294800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4002854400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4013744400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4034304000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4045194000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4065753600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4076643600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4097203200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -27889,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Buenos_Aires => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372097972, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (667965600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (687927600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1224385200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1237082400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -14028,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Cambridge_Bay => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577923200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (73472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (89193600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (104922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (120643200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (136371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (152092800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (167821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (183542400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (199270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (215596800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (230720400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (247046400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (262774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (278496000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (294224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (309945600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (325674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (341395200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (357123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (372844800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (388573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (404899200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (420022800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (436348800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (452077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (467798400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (483526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (499248000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (514976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (530697600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (544611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (562147200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (576061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (594201600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (607510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (625651200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (638960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (657100800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (671014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (688550400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (702464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (720000000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (733914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (752054400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (765363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (783504000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (796813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (814953600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (973400400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (986115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1004256000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1173603600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1194163200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1205053200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225612800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1236502800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1257062400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1268557200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1289116800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1300006800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1320566400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1331456400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1352016000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1362906000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1383465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1394355600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414915200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1425805200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1446364800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1457859600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1478419200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1489309200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509868800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1520758800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1541318400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1552208400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572768000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1583658000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1604217600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1615712400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1636272000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1647162000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667721600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1678611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1699171200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1710061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1730620800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1741510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1762070400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1772960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1793520000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1805014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1825574400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1836464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1857024000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1867914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1888473600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1899363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1919923200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1930813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1951372800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1962867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1983427200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1994317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2014876800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2025766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2046326400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2057216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2077776000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2088666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2109225600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2120115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2140675200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2152170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2172729600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2183619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2204179200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2215069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2235628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2246518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2267078400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2277968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2298528000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2309418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2329977600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2341472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2362032000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2372922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2393481600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2404371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2424931200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2435821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2456380800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2467270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2487830400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2499325200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2519884800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2530774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2551334400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2562224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2582784000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2593674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2614233600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2625123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2645683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2656573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2677132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2688627600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2709187200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2720077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2740636800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2751526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2772086400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2782976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2803536000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2814426000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2834985600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2846480400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2867040000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2877930000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2898489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2909379600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2929939200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2940829200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2961388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2972278800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2992838400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3003728400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3024288000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3035782800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3056342400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3067232400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3087792000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3098682000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3119241600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3130131600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3150691200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3161581200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3182140800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3193030800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3213590400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3225085200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3245644800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3256534800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3277094400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3287984400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3308544000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3319434000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3339993600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3350883600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3371443200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3382938000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3403497600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3414387600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3434947200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3445837200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3466396800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3477286800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3497846400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3508736400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3529296000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3540186000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3560745600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3572240400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3592800000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3603690000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3624249600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3635139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3655699200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3666589200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3687148800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3698038800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3718598400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3730093200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3750652800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3761542800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3782102400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3792992400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3813552000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3824442000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3845001600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3855891600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3876451200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3887341200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3907900800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3919395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3939955200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3950845200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3971404800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3982294800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4002854400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4013744400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4034304000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4045194000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4065753600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4076643600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4097203200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::America__Campo_Grande => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767212492, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1206954000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1191358800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1175371200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1159822800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-633816000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-622065600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-602280000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-591829200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-570744000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-560206800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-539121600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-531349200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-191361600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-155160000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-150066000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-128894400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-99950400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-89586000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-68414400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-57963600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (499752000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (511239600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (530596800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (540270000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (562132800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (571201200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (592977600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (602046000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (624427200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (634705200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (656481600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (666759600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687931200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (697604400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (719985600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (728449200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (750830400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (761713200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (782280000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (793162800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (813729600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (824007600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (844574400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (856062000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (876110400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (888721200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (908078400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (919566000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (938923200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (951620400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (970977600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (982465200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1003032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1013914800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1036296000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1045364400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1066536000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1076814000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1099368000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1108868400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1129435200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1140318000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1162699200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1172372400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1192334400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1203217200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1224388800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1234666800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1255838400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1266721200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1287288000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1298170800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1318737600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1330225200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1350792000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1361070000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1382241600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1392519600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1413691200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1424574000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1445140800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1456023600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1476590400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1487473200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1508040000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1518922800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1541304000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1550372400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -13108,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Cancun => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514722376, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (377935200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (902037600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909298800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941353200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (989136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1001833200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035702000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067151600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099206000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130655600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1175414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1193554800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1207468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225004400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1238918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1256454000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1270368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1288508400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1301817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1319958000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1333267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1351407600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1365321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1382857200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1396771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414306800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1422777600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20824,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Caracas => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524505536, FixedTimespan { utc_offset: -16060, dst_offset: 0, name: "CMT" }),
                    (-1826739140, FixedTimespan { utc_offset: -16200, dst_offset: 0, name: "-0430" }),
                    (-157750200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1197183600, FixedTimespan { utc_offset: -16200, dst_offset: 0, name: "-0430" }),
                    (1462086000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16064,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Catamarca => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372096212, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (667965600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687931200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1086058800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1087704000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15788,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Cayenne => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1846269040, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -12560,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Cayman => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524502512, FixedTimespan { utc_offset: -19176, dst_offset: 0, name: "CMT" }),
                    (-1946918424, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -19088,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Chicago => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626164, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1563724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1551632400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1538928000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1520182800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1504454400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1491757200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1473004800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1459702800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1441555200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1428253200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1410105600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1396803600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1378656000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1365354000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1347206400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1333904400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1315152000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1301850000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1283702400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1270400400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1252252800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1238950800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1220803200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1207501200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1189353600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1176051600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1157299200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1144602000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1125849600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1112547600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1094400000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1081098000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1067788800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1045414800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1031500800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1018198800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1000051200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-986749200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-967996800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-955299600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-936547200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-923245200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-905097600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-891795600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-747244800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-733942800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-715795200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-702493200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-684345600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-671043600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-652896000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-639594000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-620841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-608144400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-557942400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-544640400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-447267600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-431539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-415818000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-400089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-384368400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-352918800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-337190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-321469200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-305740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-289414800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-273686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-257965200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-242236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-226515600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-210787200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-195066000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-179337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-163616400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-147888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-131562000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-116438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-100112400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-84384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-68662800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-52934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-37213200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-21484800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-5763600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (9964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (25686000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (41414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (57740400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (73468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (89190000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (104918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (120639600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (126691200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152089200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (162374400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (183538800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (199267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (215593200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (230716800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (247042800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (262771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (278492400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (294220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (309942000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (325670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (341391600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (357120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (372841200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (388569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (404895600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (420019200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (436345200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (452073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (467794800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (483523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (499244400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (514972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (530694000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (544608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (562143600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (576057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (594198000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (607507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (625647600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (638956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (657097200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (671011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (688546800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (702460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (719996400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (733910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (752050800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (765360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (783500400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (796809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (814950000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909298800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941353200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (986112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1004252400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035702000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067151600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099206000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130655600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1205049600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1236499200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1257058800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -21036,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Chihuahua => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514714140, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1343066400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1234807200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1220292000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1207159200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1191344400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (989139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1001836800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1175418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1193558400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1207472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225008000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1238922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1256457600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1270371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1288512000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1301821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1319961600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1333270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1351411200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1365325200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1382860800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1396774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414310400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1428224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1445760000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1459674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1477814400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1491123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509264000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1522573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1540713600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1554627600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572163200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1586077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1603612800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1617526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1635667200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1648976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667116800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25460,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Ciudad_Juarez => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514714044, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1343066400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1234807200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1220292000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1207159200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1191344400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (989139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1001836800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1175418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1193558400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1207472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225008000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1238922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1256457600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1268557200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1289116800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1300006800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1320566400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1331456400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1352016000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1362906000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1383465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1394355600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414915200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1425805200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1446364800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1457859600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1478419200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1489309200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509868800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1520758800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1541318400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1552208400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572768000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1583658000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1604217600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1615712400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1636272000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1647162000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667116800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1669788000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1678611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1699171200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1710061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1730620800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1741510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1762070400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1772960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1793520000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1805014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1825574400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1836464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1857024000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1867914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1888473600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1899363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1919923200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1930813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1951372800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1962867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1983427200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1994317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2014876800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2025766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2046326400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2057216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2077776000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2088666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2109225600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2120115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2140675200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2152170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2172729600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2183619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2204179200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2215069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2235628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2246518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2267078400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2277968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2298528000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2309418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2329977600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2341472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2362032000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2372922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2393481600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2404371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2424931200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2435821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2456380800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2467270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2487830400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2499325200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2519884800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2530774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2551334400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2562224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2582784000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2593674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2614233600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2625123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2645683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2656573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2677132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2688627600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2709187200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2720077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2740636800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2751526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2772086400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2782976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2803536000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2814426000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2834985600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2846480400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2867040000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2877930000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2898489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2909379600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2929939200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2940829200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2961388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2972278800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2992838400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3003728400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3024288000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3035782800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3056342400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3067232400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3087792000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3098682000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3119241600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3130131600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3150691200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3161581200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3182140800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3193030800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3213590400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3225085200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3245644800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3256534800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3277094400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3287984400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3308544000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3319434000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3339993600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3350883600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3371443200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3382938000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3403497600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3414387600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3434947200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3445837200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3466396800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3477286800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3497846400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3508736400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3529296000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3540186000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3560745600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3572240400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3592800000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3603690000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3624249600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3635139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3655699200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3666589200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3687148800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3698038800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3718598400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3730093200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3750652800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3761542800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3782102400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3792992400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3813552000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3824442000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3845001600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3855891600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3876451200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3887341200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3907900800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3919395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3939955200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3950845200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3971404800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3982294800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4002854400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4013744400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4034304000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4045194000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4065753600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4076643600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4097203200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25556,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Coral_Harbour => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524502512, FixedTimespan { utc_offset: -19176, dst_offset: 0, name: "CMT" }),
                    (-1946918424, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -19088,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Cordoba => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372096592, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (667965600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687931200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1224385200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1237082400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15408,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Costa_Rica => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524501427, FixedTimespan { utc_offset: -20173, dst_offset: 0, name: "SJMT" }),
                    (-1545071027, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (288770400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (297234000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (320220000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (328683600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (664264800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (678344400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (695714400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (700635600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20173,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Creston => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717616702, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1633273200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1601823600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1583683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-820519140, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-812653140, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-796845540, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-84380400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-68659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -26898,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Cuiaba => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767212140, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1206954000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1191358800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1175371200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1159822800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-633816000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-622065600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-602280000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-591829200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-570744000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-560206800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-539121600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-531349200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-191361600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-155160000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-150066000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-128894400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-99950400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-89586000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-68414400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-57963600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (499752000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (511239600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (530596800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (540270000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (562132800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (571201200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (592977600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (602046000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (624427200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (634705200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (656481600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (666759600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687931200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (697604400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (719985600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (728449200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (750830400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (761713200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (782280000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (793162800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (813729600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (824007600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (844574400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (856062000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (876110400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (888721200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (908078400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (919566000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (938923200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (951620400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (970977600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (982465200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1003032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1013914800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1036296000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1045364400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1099368000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1108868400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1129435200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1140318000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1162699200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1172372400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1192334400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1203217200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1224388800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1234666800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1255838400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1266721200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1287288000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1298170800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1318737600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1330225200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1350792000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1361070000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1382241600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1392519600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1413691200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1424574000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1445140800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1456023600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1476590400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1487473200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1508040000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1518922800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1541304000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1550372400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -13460,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Curacao => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Danmarkshavn => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1686091520, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (323845200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (338950800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (354675600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (370400400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (386125200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (401850000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (417574800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (433299600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (449024400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (465354000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (481078800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (496803600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (512528400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (528253200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (543978000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (559702800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (575427600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (591152400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (606877200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (622602000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (638326800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (654656400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (670381200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (686106000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (701830800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (717555600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (733280400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (749005200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (764730000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (780454800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (796179600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (811904400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (820465200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -4480,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Dawson => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2188996940, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-1632056400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (-1615125600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-1596978000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (-1583164800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-880203600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YWT" }),
                    (-769395600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YPT" }),
                    (-765381600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-147884400, FixedTimespan { utc_offset: -32400, dst_offset: 7200, name: "YDDT" }),
                    (-131554800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (120646800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (341398800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (452080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (467802000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (483530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (499251600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (514980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (530701200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (544615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (562150800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (576064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (594205200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (607514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (625654800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (638964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (657104400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (671018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (688554000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (702468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (720003600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (733917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (752058000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (765367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (783507600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (796816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (814957200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (828871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (846406800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (860320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (877856400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (891770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (909306000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (923220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (941360400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (954669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (972810000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (986119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1004259600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1018173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1035709200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1049623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1067158800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1081072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1099213200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1112522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1130662800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1143972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1162112400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1173607200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1194166800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1205056800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1225616400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1236506400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1257066000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1268560800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1289120400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1300010400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1320570000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1331460000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1352019600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1362909600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1383469200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1394359200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1414918800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1425808800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1446368400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1457863200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1478422800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1489312800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1509872400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1520762400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1541322000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1552212000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1572771600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1583661600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1604214000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -33460,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Dawson_Creek => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713881544, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1632060000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1615129200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-880207200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PWT" }),
                    (-769395600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PPT" }),
                    (-765385200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-715788000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-702486000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-684338400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-671036400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-652888800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-639586800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-620834400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-608137200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-589384800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-576082800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-557935200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-544633200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-526485600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-513183600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-495036000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-481734000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-463586400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-450284400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-431532000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-418230000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-400082400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-386780400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-368632800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-355330800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-337183200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-323881200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-305733600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-292431600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-273679200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-260982000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-242229600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-226508400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-210780000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-195058800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-179330400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-163609200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-147880800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-131554800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-116431200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-100105200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-84376800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-68655600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-52927200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-37206000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-21477600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-5756400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (9972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (25693200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (41421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (57747600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (73476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (84013200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -28856,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Denver => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717618404, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1633273200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1601823600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1583683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1570374000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1551628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1538924400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1534089600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-147884400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-131558400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-116434800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-100108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-84380400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-68659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-52930800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-37209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-21481200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-5760000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (9968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (25689600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (41418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (57744000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (73472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (89193600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (104922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (120643200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (126694800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (152092800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (162378000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (183542400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (199270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (215596800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (230720400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (247046400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (262774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (278496000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (294224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (309945600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (325674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (341395200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (357123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (372844800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (388573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (404899200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (420022800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (436348800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (452077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (467798400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (483526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (499248000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (514976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (530697600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (544611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (562147200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (576061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (594201600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (607510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (625651200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (638960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (657100800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (671014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (688550400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (702464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (720000000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (733914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (752054400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (765363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (783504000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (796813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (814953600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (986115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1004256000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1173603600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1194163200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1205053200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225612800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1236502800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1257062400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1268557200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1289116800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1300006800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1320566400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1331456400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1352016000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1362906000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1383465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1394355600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414915200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1425805200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1446364800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1457859600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1478419200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1489309200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509868800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1520758800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1541318400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1552208400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572768000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1583658000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1604217600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1615712400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1636272000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1647162000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667721600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1678611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1699171200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1710061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1730620800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1741510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1762070400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1772960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1793520000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1805014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1825574400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1836464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1857024000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1867914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1888473600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1899363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1919923200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1930813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1951372800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1962867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1983427200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1994317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2014876800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2025766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2046326400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2057216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2077776000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2088666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2109225600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2120115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2140675200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2152170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2172729600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2183619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2204179200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2215069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2235628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2246518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2267078400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2277968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2298528000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2309418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2329977600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2341472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2362032000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2372922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2393481600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2404371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2424931200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2435821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2456380800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2467270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2487830400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2499325200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2519884800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2530774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2551334400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2562224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2582784000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2593674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2614233600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2625123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2645683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2656573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2677132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2688627600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2709187200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2720077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2740636800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2751526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2772086400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2782976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2803536000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2814426000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2834985600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2846480400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2867040000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2877930000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2898489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2909379600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2929939200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2940829200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2961388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2972278800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2992838400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3003728400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3024288000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3035782800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3056342400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3067232400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3087792000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3098682000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3119241600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3130131600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3150691200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3161581200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3182140800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3193030800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3213590400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3225085200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3245644800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3256534800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3277094400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3287984400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3308544000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3319434000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3339993600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3350883600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3371443200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3382938000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3403497600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3414387600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3434947200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3445837200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3466396800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3477286800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3497846400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3508736400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3529296000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3540186000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3560745600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3572240400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3592800000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3603690000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3624249600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3635139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3655699200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3666589200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3687148800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3698038800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3718598400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3730093200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3750652800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3761542800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3782102400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3792992400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3813552000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3824442000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3845001600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3855891600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3876451200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3887341200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3907900800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3919395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3939955200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3950845200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3971404800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3982294800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4002854400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4013744400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4034304000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4045194000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4065753600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4076643600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4097203200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25196,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Detroit => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2051202469, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1724083200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-880218000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EWT" }),
                    (-769395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EPT" }),
                    (-765396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-684349200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-671047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-80506740, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-68666400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-52938000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-37216800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (126687600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (152085600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (167814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (954658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (972799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -19931,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Dominica => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Edmonton => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1998663968, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1632063600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1600614000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1596816000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1567954800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1551628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1536505200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1523203200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1504450800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1491753600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1473001200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1459699200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-715791600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-702489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (73472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (89193600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (104922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (120643200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (136371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (152092800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (167821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (183542400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (199270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (215596800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (230720400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (247046400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (262774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (278496000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (294224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (309945600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (325674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (341395200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (357123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (372844800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (388573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (404899200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (420022800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (436348800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (452077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (467798400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (483526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (499248000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (514976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (530697600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (544611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (562147200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (576061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (594201600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (607510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (625651200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (638960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (657100800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (671014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (688550400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (702464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (720000000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (733914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (752054400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (765363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (783504000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (796813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (814953600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (986115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1004256000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1173603600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1194163200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1205053200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225612800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1236502800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1257062400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1268557200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1289116800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1300006800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1320566400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1331456400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1352016000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1362906000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1383465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1394355600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414915200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1425805200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1446364800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1457859600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1478419200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1489309200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509868800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1520758800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1541318400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1552208400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572768000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1583658000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1604217600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1615712400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1636272000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1647162000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667721600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1678611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1699171200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1710061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1730620800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1741510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1762070400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1772960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1793520000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1805014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1825574400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1836464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1857024000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1867914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1888473600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1899363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1919923200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1930813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1951372800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1962867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1983427200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1994317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2014876800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2025766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2046326400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2057216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2077776000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2088666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2109225600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2120115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2140675200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2152170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2172729600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2183619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2204179200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2215069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2235628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2246518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2267078400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2277968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2298528000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2309418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2329977600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2341472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2362032000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2372922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2393481600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2404371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2424931200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2435821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2456380800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2467270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2487830400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2499325200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2519884800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2530774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2551334400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2562224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2582784000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2593674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2614233600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2625123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2645683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2656573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2677132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2688627600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2709187200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2720077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2740636800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2751526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2772086400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2782976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2803536000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2814426000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2834985600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2846480400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2867040000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2877930000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2898489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2909379600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2929939200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2940829200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2961388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2972278800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2992838400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3003728400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3024288000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3035782800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3056342400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3067232400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3087792000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3098682000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3119241600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3130131600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3150691200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3161581200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3182140800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3193030800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3213590400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3225085200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3245644800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3256534800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3277094400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3287984400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3308544000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3319434000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3339993600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3350883600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3371443200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3382938000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3403497600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3414387600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3434947200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3445837200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3466396800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3477286800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3497846400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3508736400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3529296000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3540186000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3560745600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3572240400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3592800000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3603690000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3624249600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3635139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3655699200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3666589200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3687148800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3698038800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3718598400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3730093200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3750652800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3761542800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3782102400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3792992400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3813552000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3824442000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3845001600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3855891600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3876451200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3887341200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3907900800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3919395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3939955200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3950845200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3971404800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3982294800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4002854400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4013744400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4034304000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4045194000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4065753600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4076643600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4097203200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -27232,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Eirunepe => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767208832, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1206950400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1191355200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1175367600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1159819200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-633812400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-622062000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-602276400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-591825600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-570740400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-560203200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-539118000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-531345600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-191358000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-184190400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-155156400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-150062400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-128890800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-121118400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-99946800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-89582400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-68410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-57960000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (499755600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (511243200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (530600400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (540273600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (562136400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (571204800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (750834000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (761716800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (1214283600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1384056000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16768,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__El_Salvador => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1546279392, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (547020000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (559717200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (578469600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (591166800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -21408,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Ensenada => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514711516, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1451667600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1343062800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1234803600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1222963200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1207242000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-873820800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PWT" }),
                    (-769370400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PPT" }),
                    (-761677200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-686073600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-661539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-495039600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-481734000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-463590000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-450284400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-431535600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-418230000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-400086000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-386780400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-368636400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-355330800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-337186800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-323881200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-305737200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-292431600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (199274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (215600400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (230724000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (247050000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (262778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (278499600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (294228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (309949200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (341398800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (452080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (467802000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (483530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (499251600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (514980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (530701200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (544615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (562150800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (576064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (594205200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (607514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (625654800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (638964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (657104400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (671018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (688554000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (702468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (720003600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (733917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (752058000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (765367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (783507600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (796816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (814957200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (828871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (846406800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (860320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (877856400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (891770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (909306000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (923220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (941360400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (954669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (972810000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (986119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1004259600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1018173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1035709200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1049623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1067158800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1081072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1099213200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1112522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1130662800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1143972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1162112400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1175421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1193562000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1207476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1225011600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1238925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1256461200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1268560800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1289120400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1300010400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1320570000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1331460000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1352019600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1362909600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1383469200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1394359200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1414918800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1425808800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1446368400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1457863200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1478422800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1489312800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1509872400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1520762400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1541322000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1552212000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1572771600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1583661600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1604221200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1615716000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1636275600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1647165600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1667725200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1678615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1699174800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1710064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1730624400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1741514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1762074000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1772964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1793523600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1805018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1825578000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1836468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1857027600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1867917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1888477200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1899367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1919926800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1930816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1951376400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1962871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1983430800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1994320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2014880400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2025770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2046330000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2057220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2077779600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2088669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2109229200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2120119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2140678800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2152173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2172733200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2183623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2204182800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2215072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2235632400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2246522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2267082000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2277972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2298531600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2309421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2329981200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2341476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2362035600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2372925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2393485200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2404375200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2424934800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2435824800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2456384400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2467274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2487834000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2499328800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2519888400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2530778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2551338000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2562228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2582787600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2593677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2614237200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2625127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2645686800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2656576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2677136400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2688631200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2709190800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2720080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2740640400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2751530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2772090000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2782980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2803539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2814429600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2834989200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2846484000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2867043600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2877933600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2898493200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2909383200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2929942800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2940832800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2961392400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2972282400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2992842000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3003732000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3024291600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3035786400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3056346000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3067236000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3087795600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3098685600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3119245200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3130135200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3150694800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3161584800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3182144400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3193034400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3213594000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3225088800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3245648400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3256538400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3277098000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3287988000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3308547600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3319437600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3339997200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3350887200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3371446800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3382941600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3403501200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3414391200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3434950800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3445840800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3466400400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3477290400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3497850000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3508740000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3529299600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3540189600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3560749200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3572244000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3592803600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3603693600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3624253200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3635143200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3655702800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3666592800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3687152400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3698042400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3718602000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3730096800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3750656400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3761546400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3782106000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3792996000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3813555600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3824445600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3845005200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3855895200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3876454800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3887344800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3907904400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3919399200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3939958800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3950848800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3971408400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3982298400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4002858000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4013748000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4034307600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4045197600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4065757200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4076647200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4097206800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -28084,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Fort_Nelson => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713880953, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1632060000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1615129200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-880207200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PWT" }),
                    (-769395600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PPT" }),
                    (-765385200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-715788000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-702486000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-684338400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-671036400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-652888800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-639586800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-620834400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-608137200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-589384800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-576082800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-557935200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-544633200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-526485600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-513183600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-495036000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-481734000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-463586400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-450284400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-431532000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-418230000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-400082400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-386780400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-368632800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-355330800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-337183200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-323881200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-305733600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-292431600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-273679200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-260982000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-242229600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-226508400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-210780000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-195058800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-179330400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-163609200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-147880800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-131554800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-116431200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-100105200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-84376800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-68655600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-52927200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-37206000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-21477600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-5756400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (9972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (25693200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (41421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (57747600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (73476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (89197200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (104925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (120646800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (136375200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (152096400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (167824800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (183546000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (199274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (215600400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (230724000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (247050000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (262778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (278499600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (294228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (309949200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (341398800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (452080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (467802000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (483530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (499251600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (514980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (530701200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (544615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (562150800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (576064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (594205200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (607514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (625654800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (638964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (657104400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (671018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (688554000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (702468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (720003600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (733917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (752058000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (765367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (783507600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (796816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (814957200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (828871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (846406800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (860320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (877856400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (891770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (909306000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (923220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (941360400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (954669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (972810000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (986119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1004259600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1018173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1035709200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1049623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1067158800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1081072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1099213200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1112522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1130662800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1143972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1162112400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1173607200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1194166800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1205056800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1225616400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1236506400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1257066000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1268560800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1289120400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1300010400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1320570000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1331460000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1352019600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1362909600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1383469200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1394359200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1414918800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1425808800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -29447,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Fort_Wayne => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626522, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-900259200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-891795600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-747244800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-733942800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-715795200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-702493200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-684345600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-671043600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-652896000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-639594000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-620841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-608144400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-557942400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-544640400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-386787600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20678,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Fortaleza => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767216360, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1206957600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1191362400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1175374800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1159826400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-633819600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-622069200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-602283600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-591832800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-570747600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-560210400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-539125200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-531352800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-191365200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-184197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-155163600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-150069600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-128898000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-121125600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-99954000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-89589600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-68418000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-57967200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (499748400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (511236000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (530593200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (540266400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (562129200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (571197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (592974000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (602042400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (634701600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (951616800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (970974000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (972180000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1003028400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1013911200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -9240,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Glace_Bay => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2131646412, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1632074400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1615143600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-880221600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-526500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-513198000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (73461600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (89182800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (104911200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (120632400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (136360800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (152082000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (167810400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (183531600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (199260000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (215586000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (230709600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (247035600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (262764000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (278485200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (294213600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (309934800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (325663200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (341384400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (357112800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (372834000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (388562400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (404888400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (420012000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (436338000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (452066400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (467787600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (483516000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (499237200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (514965600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (530686800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (544600800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (562136400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (576050400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (594190800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (607500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (625640400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (638949600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (657090000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (671004000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (688539600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (702453600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (719989200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (733903200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (752043600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (765352800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (783493200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (796802400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (814942800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (828856800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (846392400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (860306400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (877842000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (891756000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (909291600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (923205600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (941346000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (954655200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (972795600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (986104800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1004245200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1018159200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1035694800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1049608800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1067144400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1081058400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1099198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1112508000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1130648400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1143957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1162098000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1173592800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1194152400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1205042400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1225602000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1236492000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1257051600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1268546400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1289106000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1299996000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1320555600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1331445600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1352005200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1362895200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1383454800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1394344800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1414904400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1425794400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1446354000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1457848800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1478408400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1489298400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1509858000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1520748000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1541307600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1552197600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1572757200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1583647200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1604206800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1615701600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1636261200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1647151200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1667710800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1678600800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1699160400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1710050400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1730610000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1741500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1762059600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1772949600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1793509200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1805004000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1825563600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1836453600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1857013200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1867903200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1888462800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1899352800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1919912400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1930802400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1951362000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1962856800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1983416400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1994306400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2014866000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2025756000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2046315600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2057205600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2077765200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2088655200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2109214800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2120104800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2140664400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2152159200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2172718800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2183608800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2204168400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2215058400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2235618000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2246508000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2267067600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2277957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2298517200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2309407200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2329966800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2341461600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2362021200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2372911200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2393470800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2404360800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2424920400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2435810400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2456370000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2467260000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2487819600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2499314400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2519874000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2530764000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2551323600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2562213600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2582773200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2593663200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2614222800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2625112800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2645672400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2656562400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2677122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2688616800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2709176400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2720066400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2740626000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2751516000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2772075600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2782965600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2803525200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2814415200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2834974800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2846469600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2867029200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2877919200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2898478800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2909368800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2929928400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2940818400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2961378000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2972268000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2992827600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3003717600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3024277200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3035772000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3056331600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3067221600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3087781200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3098671200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3119230800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3130120800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3150680400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3161570400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3182130000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3193020000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3213579600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3225074400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3245634000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3256524000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3277083600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3287973600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3308533200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3319423200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3339982800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3350872800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3371432400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3382927200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3403486800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3414376800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3434936400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3445826400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3466386000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3477276000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3497835600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3508725600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3529285200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3540175200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3560734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3572229600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3592789200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3603679200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3624238800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3635128800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3655688400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3666578400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3687138000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3698028000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3718587600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3730082400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3750642000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3761532000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3782091600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3792981600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3813541200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3824431200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3844990800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3855880800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3876440400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3887330400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3907890000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3919384800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3939944400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3950834400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3971394000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3982284000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4002843600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4013733600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4034293200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4045183200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4065742800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4076632800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4097192400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -14388,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Godthab => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1686083584, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (323845200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (338950800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (354675600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (370400400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (386125200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (401850000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (417574800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (433299600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (449024400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (465354000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (481078800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (496803600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (512528400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (528253200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (543978000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (559702800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (575427600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (591152400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (606877200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (622602000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (638326800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (654656400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (670381200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (686106000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (701830800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (717555600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (733280400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (749005200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (764730000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (780454800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (796179600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (811904400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (828234000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (846378000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (859683600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (877827600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (891133200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (909277200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (922582800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (941331600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (954032400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (972781200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (985482000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1004230800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1017536400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1035680400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1048986000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1067130000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1080435600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1099184400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1111885200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1130634000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1143334800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1162083600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1174784400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1193533200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1206838800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1224982800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1238288400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1256432400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1269738000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1288486800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1301187600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1319936400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1332637200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1351386000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1364691600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1382835600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1396141200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1414285200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1427590800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1445734800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1459040400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1477789200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1490490000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1509238800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1521939600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1540688400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1553994000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1572138000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1585443600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1603587600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1616893200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1635642000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1648342800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1667091600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1679792400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1679799600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1711846800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1729990800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1743296400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1761440400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1774746000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1792890000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1806195600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1824944400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1837645200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1856394000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1869094800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1887843600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1901149200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1919293200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1932598800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1950742800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1964048400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1982797200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1995498000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2014246800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2026947600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2045696400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2058397200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2077146000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2090451600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2108595600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2121901200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2140045200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2153350800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2172099600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2184800400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2203549200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2216250000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2234998800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2248304400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2266448400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2279754000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2297898000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2311203600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2329347600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2342653200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2361402000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2374102800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2392851600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2405552400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2424301200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2437606800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2455750800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2469056400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2487200400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2500506000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2519254800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2531955600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2550704400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2563405200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2582154000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2595459600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2613603600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2626909200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2645053200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2658358800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2676502800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2689808400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2708557200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2721258000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2740006800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2752707600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2771456400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2784762000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2802906000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2816211600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2834355600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2847661200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2866410000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2879110800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2897859600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2910560400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2929309200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2942010000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2960758800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2974064400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2992208400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3005514000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3023658000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3036963600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3055712400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3068413200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3087162000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3099862800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3118611600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3131917200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3150061200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3163366800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3181510800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3194816400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3212960400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3226266000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3245014800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3257715600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3276464400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3289165200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3307914000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3321219600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3339363600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3352669200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3370813200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3384118800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3402867600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3415568400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3434317200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3447018000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3465766800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3479072400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3497216400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3510522000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3528666000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3541971600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3560115600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3573421200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3592170000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3604870800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3623619600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3636320400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3655069200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3668374800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3686518800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3699824400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3717968400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3731274000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3750022800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3762723600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3781472400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3794173200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3812922000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3825622800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3844371600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3857677200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3875821200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3889126800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3907270800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3920576400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3939325200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3952026000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3970774800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3983475600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (4002224400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (4015530000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (4033674000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (4046979600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (4065123600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (4078429200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (4096573200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -12416,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Goose_Bay => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713895900, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1632076148, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1615145348, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1096921748, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-1061670600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-1048973400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-1030221000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-1017523800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-998771400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-986074200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-966717000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-954624600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-935267400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-922570200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-903817800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-891120600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-872368200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NWT" }),
                    (-769395600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NPT" }),
                    (-765401400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-746044200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-733347000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-714594600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-701897400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-683145000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-670447800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-651695400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-638998200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-619641000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-606943800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-589401000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-576099000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-557951400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-544649400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-526501800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-513199800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-495052200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-481750200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-463602600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-450300600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-431548200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-418246200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-400098600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-386796600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-368649000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-355347000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-337199400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-323897400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-305749800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-289423800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-273695400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-257974200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-242245800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-226524600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-210796200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-195075000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-179346600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-163625400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-147897000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-131571000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-119903400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-116445600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-100119600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-84391200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-68670000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-52941600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-37220400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-21492000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-5770800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (9957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (25678800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (41407200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (57733200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (73461600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (89182800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (104911200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (120632400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (136360800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (152082000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (167810400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (183531600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (199260000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (215586000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (230709600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (247035600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (262764000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (278485200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (294213600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (309934800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (325663200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (341384400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (357112800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (372834000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (388562400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (404888400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (420012000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (436338000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (452066400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (467787600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (483516000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (499237200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (514965600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (530686800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (544593660, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (562129260, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (576043260, FixedTimespan { utc_offset: -14400, dst_offset: 7200, name: "ADDT" }),
                    (594180060, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (607492860, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (625633260, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (638942460, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (657082860, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (670996860, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (688532460, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (702446460, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (719982060, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (733896060, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (752036460, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (765345660, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (783486060, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (796795260, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (814935660, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (828849660, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (846385260, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (860299260, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (877834860, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (891748860, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (909284460, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (923198460, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (941338860, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (954648060, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (972788460, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (986097660, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1004238060, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1018152060, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1035687660, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1049601660, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1067137260, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1081051260, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1099191660, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1112500860, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1130641260, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1143950460, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1162090860, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1173585660, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1194145260, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1205035260, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1225594860, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1236484860, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1257044460, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1268539260, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1289098860, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1299988860, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1320555600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1331445600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1352005200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1362895200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1383454800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1394344800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1414904400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1425794400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1446354000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1457848800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1478408400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1489298400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1509858000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1520748000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1541307600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1552197600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1572757200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1583647200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1604206800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1615701600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1636261200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1647151200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1667710800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1678600800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1699160400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1710050400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1730610000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1741500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1762059600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1772949600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1793509200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1805004000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1825563600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1836453600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1857013200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1867903200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1888462800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1899352800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1919912400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1930802400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1951362000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1962856800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1983416400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1994306400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2014866000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2025756000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2046315600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2057205600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2077765200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2088655200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2109214800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2120104800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2140664400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2152159200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2172718800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2183608800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2204168400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2215058400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2235618000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2246508000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2267067600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2277957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2298517200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2309407200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2329966800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2341461600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2362021200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2372911200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2393470800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2404360800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2424920400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2435810400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2456370000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2467260000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2487819600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2499314400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2519874000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2530764000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2551323600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2562213600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2582773200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2593663200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2614222800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2625112800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2645672400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2656562400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2677122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2688616800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2709176400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2720066400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2740626000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2751516000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2772075600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2782965600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2803525200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2814415200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2834974800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2846469600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2867029200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2877919200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2898478800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2909368800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2929928400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2940818400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2961378000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2972268000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2992827600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3003717600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3024277200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3035772000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3056331600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3067221600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3087781200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3098671200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3119230800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3130120800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3150680400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3161570400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3182130000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3193020000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3213579600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3225074400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3245634000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3256524000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3277083600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3287973600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3308533200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3319423200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3339982800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3350872800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3371432400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3382927200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3403486800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3414376800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3434936400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3445826400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3466386000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3477276000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3497835600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3508725600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3529285200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3540175200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3560734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3572229600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3592789200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3603679200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3624238800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3635128800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3655688400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3666578400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3687138000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3698028000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3718587600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3730082400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3750642000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3761532000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3782091600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3792981600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3813541200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3824431200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3844990800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3855880800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3876440400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3887330400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3907890000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3919384800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3939944400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3950834400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3971394000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3982284000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4002843600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4013733600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4034293200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4045183200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4065742800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4076632800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4097192400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -14500,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Grand_Turk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524504528, FixedTimespan { utc_offset: -18430, dst_offset: 0, name: "KMT" }),
                    (-1827687170, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (954658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (972799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -17072,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Grenada => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Guadeloupe => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Guatemala => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1617040676, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (123055200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (130914000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (422344800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (433054800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (669708000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (684219600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1146376800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1159678800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -21724,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Guayaquil => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524502440, FixedTimespan { utc_offset: -18840, dst_offset: 0, name: "QMT" }),
                    (-1230749160, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (722926800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (728884800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -19160,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Guyana => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1843589241, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1730577600, FixedTimespan { utc_offset: -13500, dst_offset: 0, name: "-0345" }),
                    (176096700, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (701841600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -13959,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Halifax => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2131645536, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1696276800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1680469200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1632074400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1615143600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1566763200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1557090000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1535486400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1524949200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1504468800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1493413200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1472414400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1461963600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1440964800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1429390800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1409515200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1396731600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1376856000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1366491600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1346616000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1333832400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1313956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1303678800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1282507200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1272661200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1251057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1240088400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1219608000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1207429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1188763200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1175979600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1157313600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1143925200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1124049600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1113771600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1091390400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1081026000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1059854400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1050786000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1030910400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1018126800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-999460800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-986677200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-965592000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-955227600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-935956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-923173200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-904507200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-891723600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-880221600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-747252000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-733950000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-715802400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-702500400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-684352800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-671050800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-652903200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-639601200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-589399200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-576097200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-557949600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-544647600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-526500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-513198000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-495050400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-481748400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-431546400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-418244400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-400096800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-386794800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-368647200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-355345200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-337197600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-323895600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-242244000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-226522800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-210794400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-195073200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-179344800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-163623600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-147895200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-131569200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-116445600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-100119600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-84391200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-68670000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-52941600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-37220400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-21492000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-5770800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (9957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (25678800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (41407200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (57733200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (73461600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (89182800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (104911200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (120632400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (136360800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (152082000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (167810400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (183531600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (199260000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (215586000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (230709600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (247035600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (262764000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (278485200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (294213600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (309934800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (325663200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (341384400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (357112800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (372834000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (388562400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (404888400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (420012000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (436338000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (452066400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (467787600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (483516000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (499237200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (514965600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (530686800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (544600800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (562136400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (576050400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (594190800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (607500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (625640400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (638949600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (657090000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (671004000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (688539600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (702453600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (719989200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (733903200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (752043600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (765352800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (783493200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (796802400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (814942800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (828856800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (846392400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (860306400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (877842000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (891756000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (909291600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (923205600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (941346000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (954655200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (972795600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (986104800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1004245200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1018159200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1035694800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1049608800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1067144400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1081058400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1099198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1112508000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1130648400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1143957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1162098000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1173592800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1194152400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1205042400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1225602000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1236492000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1257051600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1268546400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1289106000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1299996000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1320555600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1331445600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1352005200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1362895200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1383454800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1394344800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1414904400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1425794400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1446354000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1457848800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1478408400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1489298400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1509858000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1520748000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1541307600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1552197600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1572757200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1583647200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1604206800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1615701600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1636261200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1647151200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1667710800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1678600800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1699160400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1710050400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1730610000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1741500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1762059600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1772949600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1793509200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1805004000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1825563600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1836453600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1857013200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1867903200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1888462800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1899352800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1919912400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1930802400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1951362000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1962856800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1983416400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1994306400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2014866000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2025756000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2046315600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2057205600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2077765200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2088655200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2109214800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2120104800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2140664400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2152159200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2172718800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2183608800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2204168400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2215058400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2235618000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2246508000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2267067600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2277957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2298517200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2309407200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2329966800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2341461600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2362021200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2372911200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2393470800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2404360800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2424920400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2435810400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2456370000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2467260000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2487819600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2499314400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2519874000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2530764000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2551323600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2562213600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2582773200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2593663200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2614222800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2625112800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2645672400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2656562400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2677122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2688616800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2709176400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2720066400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2740626000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2751516000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2772075600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2782965600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2803525200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2814415200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2834974800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2846469600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2867029200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2877919200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2898478800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2909368800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2929928400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2940818400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2961378000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2972268000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2992827600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3003717600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3024277200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3035772000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3056331600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3067221600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3087781200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3098671200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3119230800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3130120800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3150680400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3161570400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3182130000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3193020000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3213579600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3225074400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3245634000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3256524000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3277083600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3287973600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3308533200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3319423200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3339982800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3350872800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3371432400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3382927200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3403486800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3414376800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3434936400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3445826400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3466386000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3477276000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3497835600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3508725600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3529285200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3540175200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3560734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3572229600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3592789200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3603679200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3624238800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3635128800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3655688400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3666578400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3687138000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3698028000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3718587600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3730082400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3750642000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3761532000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3782091600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3792981600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3813541200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3824431200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3844990800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3855880800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3876440400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3887330400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3907890000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3919384800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3939944400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3950834400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3971394000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3982284000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4002843600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4013733600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4034293200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4045183200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4065742800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4076632800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4097192400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15264,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Havana => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524501832, FixedTimespan { utc_offset: -19776, dst_offset: 0, name: "HMT" }),
                    (-1402813824, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-1311534000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-1300996800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-933534000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-925675200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-902084400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-893620800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-870030000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-862171200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-775681200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-767822400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-744231600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-736372800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-144702000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-134251200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-113425200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-102542400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-86295600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-72907200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-54154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-41457600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-21495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-5774400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (9954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (25675200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (41403600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (57729600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (73458000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (87364800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (104907600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (118900800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (136357200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (150436800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (167806800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (183528000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (199256400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (215582400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (230706000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (247032000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (263365200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (276667200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (290581200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (308721600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (322030800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (340171200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (358318800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (371620800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (389768400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (403070400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (421218000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (434520000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (452667600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (466574400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (484117200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (498024000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (511333200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (529473600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (542782800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (560923200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (574837200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (592372800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (606286800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (623822400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (638946000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (655876800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (671000400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (687330000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (702450000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (718779600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (733899600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (750229200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (765349200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (781678800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (796798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (813128400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (828853200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (844578000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (860302800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (876632400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (891147600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (909291600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (922597200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (941346000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (954651600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (972795600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (986101200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1004245200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1018155600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1035694800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1049605200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1067144400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1080450000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1162098000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1173589200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1193547600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1205643600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1224997200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1236488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1256446800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1268542800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1288501200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1300597200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1321160400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1333256400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1352005200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1362891600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1383454800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1394341200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1414904400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1425790800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1446354000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1457845200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1478408400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1489294800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1509858000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1520744400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1541307600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1552194000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1572757200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1583643600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1604206800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1615698000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1636261200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1647147600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1667710800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1678597200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1699160400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1710046800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1730610000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1741496400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1762059600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1772946000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1793509200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1805000400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1825563600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1836450000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1857013200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1867899600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1888462800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1899349200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1919912400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1930798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1951362000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1962853200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1983416400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1994302800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2014866000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2025752400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2046315600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2057202000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2077765200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2088651600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2109214800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2120101200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2140664400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2152155600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2172718800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2183605200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2204168400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2215054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2235618000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2246504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2267067600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2277954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2298517200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2309403600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2329966800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2341458000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2362021200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2372907600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2393470800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2404357200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2424920400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2435806800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2456370000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2467256400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2487819600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2499310800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2519874000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2530760400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2551323600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2562210000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2582773200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2593659600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2614222800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2625109200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2645672400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2656558800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2677122000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2688613200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2709176400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2720062800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2740626000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2751512400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2772075600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2782962000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2803525200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2814411600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2834974800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2846466000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2867029200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2877915600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2898478800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2909365200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2929928400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2940814800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2961378000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2972264400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2992827600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3003714000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3024277200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3035768400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3056331600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3067218000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3087781200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3098667600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3119230800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3130117200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3150680400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3161566800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3182130000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3193016400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3213579600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3225070800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3245634000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3256520400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3277083600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3287970000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3308533200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3319419600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3339982800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3350869200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3371432400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3382923600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3403486800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3414373200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3434936400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3445822800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3466386000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3477272400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3497835600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3508722000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3529285200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3540171600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3560734800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3572226000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3592789200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3603675600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3624238800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3635125200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3655688400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3666574800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3687138000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3698024400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3718587600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3730078800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3750642000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3761528400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3782091600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3792978000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3813541200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3824427600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3844990800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3855877200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3876440400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3887326800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3907890000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3919381200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3939944400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3950830800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3971394000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3982280400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (4002843600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (4013730000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (4034293200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (4045179600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (4065742800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (4076629200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (4097192400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -19768,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Hermosillo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514712968, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1343066400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1234807200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1220292000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1207159200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1191344400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-873828000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-661539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (28800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -26632,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Indiana__Indianapolis => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626522, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-900259200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-891795600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-747244800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-733942800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-715795200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-702493200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-684345600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-671043600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-652896000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-639594000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-620841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-608144400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-557942400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-544640400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-386787600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20678,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Indiana__Knox => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626410, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-715795200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-702493200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-684345600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-671043600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-652896000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-639594000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-620841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-608144400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-557942400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-544640400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-447267600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-431539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-415818000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-400089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-386787600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-355338000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-337190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-321469200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-305740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-289414800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-273686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-257965200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-242236800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-195066000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-84384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-68662800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-52934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-37213200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-21484800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-5763600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (9964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (25686000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (41414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (57740400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (73468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (89190000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (104918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (120639600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (126691200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152089200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (162374400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (183538800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (199267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (215593200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (230716800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (247042800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (262771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (278492400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (294220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (309942000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (325670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (341391600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (357120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (372841200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (388569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (404895600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (420019200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (436345200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (452073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (467794800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (483523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (499244400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (514972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (530694000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (544608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (562143600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (576057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (594198000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (607507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (625647600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (638956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (657097200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (671011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (688546800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1205049600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1236499200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1257058800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20790,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Indiana__Marengo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626477, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-450291600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-431539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-418237200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-400089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-386787600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-355338000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-337190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-323888400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-305740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-292438800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-273686400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (41410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (57736800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (73465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (89186400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (126687600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152089200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (162370800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20723,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Indiana__Petersburg => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626253, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-462996000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-450291600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-431539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-418237200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-400089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-386787600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-355338000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-337190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-323888400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-305740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-292438800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-273686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-257965200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-242236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-226515600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-210787200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-195066000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-179337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-163616400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-147888000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-100112400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-84384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-68662800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-52934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-37213200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-21484800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-5763600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (9964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (25686000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (41414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (57740400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (73468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (89190000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (104918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (120639600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (126691200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152089200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (162374400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (183538800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (199267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (215593200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (230716800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (247042800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20947,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Indiana__Tell_City => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626377, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-462996000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-450291600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-431539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-418237200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-400089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-386787600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-355338000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-337190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-323888400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-305740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-292438800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-273686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-257965200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-242236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-226515600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-210787200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-195066000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-179337600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-68662800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-52934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-37213200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-21484800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1205049600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1236499200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1257058800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20823,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Indiana__Vevay => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626784, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (41410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (57736800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (73465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (89186400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20416,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Indiana__Vincennes => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626193, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-747244800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-733942800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-462996000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-450291600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-431539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-418237200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-400089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-386787600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-355338000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-337190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-323888400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-305740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-289414800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-273686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-260989200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-242236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-226515600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-210787200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-195066000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-179337600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -21007,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Indiana__Winamac => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626415, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-747244800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-733942800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-715795200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-702493200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-684345600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-671043600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-652896000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-639594000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-620841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-608144400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-557942400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-544640400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-447267600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-431539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-415818000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-400089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-386787600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-355338000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-337190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-323888400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-305740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-292438800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-273686400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20785,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Indianapolis => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626522, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-900259200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-891795600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-747244800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-733942800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-715795200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-702493200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-684345600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-671043600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-652896000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-639594000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-620841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-608144400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-557942400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-544640400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-386787600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20678,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Inuvik => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-536457600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (73476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (89197200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (104925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (120646800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (136375200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (152096400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (167824800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (183546000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (199274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (215600400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (230724000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (247050000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (262778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (278499600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (294228000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (309945600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (325674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (341395200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (357123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (372844800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (388573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (404899200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (420022800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (436348800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (452077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (467798400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (483526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (499248000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (514976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (530697600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (544611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (562147200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (576061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (594201600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (607510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (625651200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (638960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (657100800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (671014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (688550400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (702464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (720000000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (733914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (752054400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (765363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (783504000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (796813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (814953600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (986115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1004256000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1173603600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1194163200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1205053200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225612800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1236502800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1257062400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1268557200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1289116800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1300006800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1320566400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1331456400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1352016000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1362906000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1383465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1394355600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414915200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1425805200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1446364800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1457859600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1478419200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1489309200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509868800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1520758800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1541318400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1552208400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572768000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1583658000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1604217600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1615712400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1636272000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1647162000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667721600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1678611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1699171200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1710061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1730620800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1741510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1762070400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1772960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1793520000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1805014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1825574400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1836464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1857024000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1867914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1888473600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1899363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1919923200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1930813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1951372800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1962867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1983427200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1994317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2014876800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2025766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2046326400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2057216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2077776000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2088666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2109225600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2120115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2140675200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2152170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2172729600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2183619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2204179200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2215069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2235628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2246518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2267078400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2277968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2298528000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2309418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2329977600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2341472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2362032000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2372922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2393481600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2404371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2424931200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2435821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2456380800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2467270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2487830400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2499325200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2519884800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2530774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2551334400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2562224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2582784000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2593674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2614233600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2625123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2645683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2656573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2677132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2688627600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2709187200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2720077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2740636800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2751526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2772086400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2782976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2803536000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2814426000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2834985600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2846480400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2867040000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2877930000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2898489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2909379600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2929939200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2940829200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2961388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2972278800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2992838400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3003728400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3024288000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3035782800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3056342400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3067232400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3087792000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3098682000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3119241600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3130131600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3150691200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3161581200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3182140800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3193030800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3213590400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3225085200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3245644800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3256534800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3277094400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3287984400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3308544000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3319434000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3339993600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3350883600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3371443200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3382938000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3403497600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3414387600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3434947200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3445837200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3466396800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3477286800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3497846400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3508736400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3529296000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3540186000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3560745600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3572240400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3592800000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3603690000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3624249600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3635139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3655699200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3666589200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3687148800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3698038800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3718598400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3730093200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3750652800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3761542800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3782102400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3792992400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3813552000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3824442000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3845001600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3855891600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3876451200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3887341200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3907900800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3919395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3939955200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3950845200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3971404800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3982294800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4002854400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4013744400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4034304000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4045194000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4065753600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4076643600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4097203200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::America__Iqaluit => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-865296000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EWT" }),
                    (-769395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EPT" }),
                    (-765396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (73465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (89186400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (136364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (152085600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (167814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::America__Jamaica => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524503170, FixedTimespan { utc_offset: -18430, dst_offset: 0, name: "KMT" }),
                    (-1827687170, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (126687600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (152085600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (162370800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -18430,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Jujuy => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372096328, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (657086400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (669178800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (686721600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15672,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Juneau => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3225223727, FixedTimespan { utc_offset: -32261, dst_offset: 0, name: "LMT" }),
                    (-2188954939, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-880207200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PWT" }),
                    (-769395600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PPT" }),
                    (-765385200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-21477600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-5756400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (9972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (25693200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (41421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (57747600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (73476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (89197200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (104925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (120646800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (126698400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (152096400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (162381600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (183546000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (199274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (215600400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (230724000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (247050000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (262778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (278499600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (294228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (309949200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (341402400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (439030800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (452084400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (467805600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (483534000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (499255200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (514983600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (530704800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (544618800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (562154400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (576068400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (594208800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (607518000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (625658400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (638967600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (657108000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (671022000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (688557600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (702471600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (720007200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (733921200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (752061600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (765370800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (783511200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (796820400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (814960800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (828874800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (846410400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (860324400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (877860000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (891774000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (909309600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (923223600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (941364000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (954673200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (972813600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (986122800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1004263200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1018177200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1035712800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1049626800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1067162400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1081076400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1099216800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1112526000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1130666400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1143975600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1162116000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1173610800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1194170400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1205060400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1225620000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1236510000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1257069600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1268564400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1289124000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1300014000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1320573600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1331463600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1352023200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1362913200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1383472800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1394362800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1414922400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1425812400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1446372000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1457866800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1478426400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1489316400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1509876000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1520766000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1541325600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1552215600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1572775200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1583665200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1604224800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1615719600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1636279200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1647169200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1667728800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1678618800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1699178400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1710068400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1730628000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1741518000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1762077600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1772967600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1793527200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1805022000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1825581600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1836471600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1857031200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1867921200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1888480800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1899370800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1919930400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1930820400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1951380000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1962874800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1983434400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1994324400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2014884000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2025774000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2046333600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2057223600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2077783200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2088673200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2109232800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2120122800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2140682400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2152177200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2172736800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2183626800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2204186400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2215076400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2235636000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2246526000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2267085600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2277975600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2298535200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2309425200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2329984800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2341479600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2362039200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2372929200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2393488800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2404378800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2424938400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2435828400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2456388000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2467278000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2487837600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2499332400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2519892000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2530782000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2551341600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2562231600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2582791200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2593681200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2614240800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2625130800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2645690400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2656580400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2677140000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2688634800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2709194400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2720084400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2740644000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2751534000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2772093600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2782983600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2803543200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2814433200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2834992800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2846487600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2867047200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2877937200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2898496800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2909386800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2929946400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2940836400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2961396000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2972286000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2992845600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3003735600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3024295200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3035790000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3056349600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3067239600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3087799200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3098689200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3119248800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3130138800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3150698400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3161588400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3182148000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3193038000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3213597600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3225092400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3245652000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3256542000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3277101600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3287991600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3308551200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3319441200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3340000800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3350890800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3371450400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3382945200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3403504800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3414394800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3434954400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3445844400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3466404000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3477294000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3497853600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3508743600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3529303200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3540193200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3560752800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3572247600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3592807200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3603697200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3624256800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3635146800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3655706400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3666596400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3687156000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3698046000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3718605600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3730100400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3750660000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3761550000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3782109600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3792999600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3813559200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3824449200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3845008800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3855898800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3876458400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3887348400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3907908000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3919402800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3939962400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3950852400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3971412000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3982302000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4002861600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4013751600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4034311200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4045201200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4065760800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4076650800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4097210400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 54139,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Kentucky__Louisville => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626618, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1535904000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1525280400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-905097600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-891795600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-747251940, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-744224400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-620841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-608144400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-557942400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-544640400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-450291600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-431539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-415818000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-400089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-384368400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-352918800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-337190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-321469200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-305740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-289414800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-273686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-266432400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-52938000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-37216800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (41410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (57736800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (73465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (89186400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (126687600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152089200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (162370800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (954658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (972799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20582,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Kentucky__Monticello => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626836, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-52934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-37213200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-21484800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-5763600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (9964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (25686000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (41414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (57740400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (73468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (89190000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (104918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (120639600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (126691200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152089200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (162374400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (183538800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (199267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (215593200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (230716800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (247042800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (262771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (278492400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (294220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (309942000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (325670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (341391600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (357120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (372841200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (388569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (404895600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (420019200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (436345200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (452073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (467794800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (483523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (499244400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (514972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (530694000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (544608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (562143600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (576057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (594198000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (607507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (625647600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (638956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (657097200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (671011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (688546800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (702460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (719996400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (733910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (752050800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (765360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (783500400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (796809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (814950000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909298800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941353200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20364,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Knox_IN => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626410, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-715795200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-702493200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-684345600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-671043600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-652896000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-639594000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-620841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-608144400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-557942400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-544640400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-447267600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-431539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-415818000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-400089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-386787600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-355338000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-337190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-321469200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-305740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-289414800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-273686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-257965200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-242236800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-195066000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-84384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-68662800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-52934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-37213200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-21484800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-5763600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (9964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (25686000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (41414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (57740400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (73468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (89190000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (104918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (120639600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (126691200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152089200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (162374400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (183538800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (199267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (215593200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (230716800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (247042800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (262771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (278492400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (294220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (309942000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (325670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (341391600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (357120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (372841200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (388569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (404895600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (420019200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (436345200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (452073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (467794800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (483523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (499244400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (514972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (530694000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (544608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (562143600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (576057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (594198000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (607507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (625647600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (638956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (657097200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (671011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (688546800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1205049600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1236499200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1257058800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20790,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Kralendijk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__La_Paz => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524505244, FixedTimespan { utc_offset: -16356, dst_offset: 0, name: "CMT" }),
                    (-1205954844, FixedTimespan { utc_offset: -16356, dst_offset: 3600, name: "BST" }),
                    (-1192307244, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16356,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Lima => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524503108, FixedTimespan { utc_offset: -18516, dst_offset: 0, name: "LMT" }),
                    (-1938538284, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1009825200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1002052800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-986756400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-971035200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-955306800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-939585600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (504939600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (512712000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (536475600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (544248000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (631170000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (638942400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (757400400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (765172800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -18492,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Los_Angeles => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717611622, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1633269600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1615129200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1601820000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1583679600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-880207200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PWT" }),
                    (-769395600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PPT" }),
                    (-765385200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-687967140, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-662655600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-620838000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-608137200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-589388400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-576082800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-557938800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-544633200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-526489200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-513183600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-495039600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-481734000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-463590000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-450284400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-431535600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-418230000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-400086000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-386780400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-368636400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-355330800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-337186800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-323881200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-305737200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-292431600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-273682800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-260982000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-242233200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-226508400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-210783600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-195058800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-179334000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-163609200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-147884400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-131554800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-116434800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-100105200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-84376800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-68655600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-52927200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-37206000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-21477600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-5756400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (9972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (25693200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (41421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (57747600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (73476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (89197200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (104925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (120646800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (126698400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (152096400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (162381600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (183546000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (199274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (215600400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (230724000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (247050000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (262778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (278499600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (294228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (309949200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (341398800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (452080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (467802000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (483530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (499251600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (514980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (530701200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (544615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (562150800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (576064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (594205200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (607514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (625654800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (638964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (657104400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (671018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (688554000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (702468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (720003600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (733917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (752058000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (765367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (783507600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (796816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (814957200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (828871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (846406800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (860320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (877856400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (891770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (909306000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (923220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (941360400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (954669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (972810000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (986119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1004259600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1018173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1035709200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1049623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1067158800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1081072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1099213200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1112522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1130662800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1143972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1162112400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1173607200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1194166800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1205056800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1225616400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1236506400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1257066000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1268560800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1289120400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1300010400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1320570000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1331460000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1352019600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1362909600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1383469200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1394359200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1414918800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1425808800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1446368400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1457863200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1478422800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1489312800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1509872400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1520762400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1541322000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1552212000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1572771600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1583661600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1604221200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1615716000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1636275600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1647165600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1667725200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1678615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1699174800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1710064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1730624400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1741514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1762074000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1772964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1793523600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1805018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1825578000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1836468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1857027600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1867917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1888477200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1899367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1919926800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1930816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1951376400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1962871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1983430800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1994320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2014880400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2025770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2046330000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2057220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2077779600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2088669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2109229200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2120119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2140678800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2152173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2172733200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2183623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2204182800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2215072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2235632400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2246522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2267082000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2277972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2298531600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2309421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2329981200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2341476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2362035600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2372925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2393485200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2404375200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2424934800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2435824800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2456384400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2467274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2487834000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2499328800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2519888400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2530778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2551338000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2562228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2582787600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2593677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2614237200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2625127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2645686800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2656576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2677136400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2688631200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2709190800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2720080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2740640400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2751530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2772090000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2782980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2803539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2814429600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2834989200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2846484000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2867043600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2877933600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2898493200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2909383200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2929942800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2940832800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2961392400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2972282400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2992842000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3003732000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3024291600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3035786400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3056346000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3067236000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3087795600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3098685600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3119245200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3130135200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3150694800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3161584800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3182144400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3193034400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3213594000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3225088800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3245648400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3256538400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3277098000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3287988000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3308547600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3319437600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3339997200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3350887200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3371446800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3382941600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3403501200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3414391200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3434950800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3445840800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3466400400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3477290400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3497850000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3508740000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3529299600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3540189600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3560749200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3572244000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3592803600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3603693600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3624253200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3635143200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3655702800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3666592800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3687152400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3698042400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3718602000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3730096800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3750656400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3761546400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3782106000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3792996000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3813555600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3824445600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3845005200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3855895200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3876454800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3887344800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3907904400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3919399200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3939958800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3950848800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3971408400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3982298400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4002858000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4013748000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4034307600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4045197600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4065757200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4076647200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4097206800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -28378,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Louisville => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626618, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1535904000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1525280400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-905097600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-891795600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-747251940, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-744224400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-620841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-608144400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-557942400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-544640400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-450291600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-431539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-415818000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-400089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-384368400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-352918800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-337190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-321469200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-305740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-289414800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-273686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-266432400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-52938000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-37216800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (41410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (57736800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (73465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (89186400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (126687600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152089200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (162370800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (954658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (972799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20582,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Lower_Princes => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Maceio => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767217028, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1206957600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1191362400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1175374800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1159826400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-633819600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-622069200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-602283600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-591832800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-570747600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-560210400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-539125200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-531352800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-191365200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-184197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-155163600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-150069600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-128898000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-121125600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-99954000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-89589600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-68418000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-57967200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (499748400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (511236000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (530593200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (540266400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (562129200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (571197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (592974000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (602042400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (634701600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (813726000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (824004000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (951616800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (970974000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (972180000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1003028400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1013911200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -8572,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Managua => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524500892, FixedTimespan { utc_offset: -20712, dst_offset: 0, name: "MMT" }),
                    (-1121105688, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (105084000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (161758800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (290584800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (299134800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (322034400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (330584400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (694260000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (717310800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (725868000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (852094800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1113112800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1128229200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1146384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1159682400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20708,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Manaus => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767211196, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1206954000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1191358800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1175371200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1159822800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-633816000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-622065600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-602280000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-591829200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-570744000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-560206800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-539121600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-531349200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-191361600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-155160000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-150066000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-128894400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-99950400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-89586000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-68414400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-57963600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (499752000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (511239600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (530596800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (540270000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (562132800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (571201200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (750830400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (761713200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -14404,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Marigot => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Martinique => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524506940, FixedTimespan { utc_offset: -14660, dst_offset: 0, name: "FFMT" }),
                    (-1851537340, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (323841600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (338958000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -14660,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Matamoros => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514719800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (576057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (594198000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909298800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941353200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (989136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1001833200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035702000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067151600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099206000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130655600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1175414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1193554800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1207468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225004400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1238918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1256454000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -23400,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Mazatlan => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514714060, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1343066400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1234807200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1220292000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1207159200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1191344400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-873828000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-661539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (28800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (989139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1001836800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1175418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1193558400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1207472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225008000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1238922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1256457600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1270371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1288512000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1301821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1319961600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1333270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1351411200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1365325200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1382860800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1396774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414310400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1428224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1445760000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1459674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1477814400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1491123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509264000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1522573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1540713600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1554627600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572163200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1586077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1603612800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1617526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1635667200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1648976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667116800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25540,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Mendoza => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372095484, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (655963200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (667796400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687499200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (699418800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (719380800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1085281200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1096171200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16516,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Menominee => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2659759773, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-747244800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-733942800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-116438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-100112400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-21484800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (120639600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (126691200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152089200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (162374400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (183538800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (199267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (215593200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (230716800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (247042800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (262771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (278492400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (294220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (309942000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (325670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (341391600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (357120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (372841200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (388569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (404895600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (420019200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (436345200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (452073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (467794800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (483523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (499244400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (514972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (530694000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (544608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (562143600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (576057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (594198000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (607507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (625647600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (638956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (657097200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (671011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (688546800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (702460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (719996400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (733910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (752050800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (765360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (783500400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (796809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (814950000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909298800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941353200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (986112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1004252400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035702000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067151600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099206000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130655600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1205049600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1236499200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1257058800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -21027,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Merida => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514721692, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (377935200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (407653200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909298800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941353200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (989136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1001833200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035702000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067151600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099206000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130655600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1175414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1193554800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1207468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225004400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1238918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1256454000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1270368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1288508400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1301817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1319958000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1333267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1351407600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1365321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1382857200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1396771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414306800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1428220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1445756400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1459670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1477810800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1491120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509260400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1522569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1540710000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1554624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1586073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1603609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1617523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1635663600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1648972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -21508,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Metlakatla => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3225223727, FixedTimespan { utc_offset: -31578, dst_offset: 0, name: "LMT" }),
                    (-2188955622, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-880207200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PWT" }),
                    (-769395600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PPT" }),
                    (-765385200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-21477600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-5756400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (9972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (25693200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (41421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (57747600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (73476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (89197200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (104925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (120646800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (126698400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (152096400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (162381600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (183546000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (199274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (215600400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (230724000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (247050000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (262778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (278499600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (294228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (309949200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (341398800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1446372000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1457866800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1478426400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1489316400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1509876000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1520766000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1541325600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1547978400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1552215600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1572775200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1583665200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1604224800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1615719600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1636279200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1647169200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1667728800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1678618800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1699178400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1710068400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1730628000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1741518000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1762077600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1772967600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1793527200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1805022000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1825581600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1836471600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1857031200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1867921200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1888480800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1899370800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1919930400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1930820400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1951380000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1962874800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1983434400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1994324400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2014884000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2025774000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2046333600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2057223600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2077783200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2088673200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2109232800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2120122800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2140682400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2152177200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2172736800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2183626800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2204186400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2215076400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2235636000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2246526000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2267085600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2277975600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2298535200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2309425200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2329984800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2341479600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2362039200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2372929200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2393488800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2404378800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2424938400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2435828400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2456388000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2467278000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2487837600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2499332400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2519892000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2530782000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2551341600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2562231600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2582791200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2593681200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2614240800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2625130800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2645690400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2656580400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2677140000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2688634800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2709194400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2720084400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2740644000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2751534000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2772093600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2782983600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2803543200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2814433200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2834992800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2846487600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2867047200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2877937200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2898496800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2909386800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2929946400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2940836400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2961396000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2972286000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2992845600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3003735600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3024295200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3035790000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3056349600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3067239600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3087799200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3098689200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3119248800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3130138800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3150698400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3161588400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3182148000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3193038000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3213597600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3225092400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3245652000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3256542000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3277101600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3287991600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3308551200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3319441200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3340000800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3350890800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3371450400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3382945200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3403504800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3414394800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3434954400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3445844400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3466404000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3477294000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3497853600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3508743600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3529303200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3540193200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3560752800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3572247600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3592807200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3603697200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3624256800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3635146800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3655706400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3666596400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3687156000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3698046000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3718605600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3730100400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3750660000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3761550000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3782109600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3792999600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3813559200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3824449200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3845008800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3855898800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3876458400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3887348400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3907908000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3919402800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3939962400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3950852400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3971412000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3982302000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4002861600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4013751600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4034311200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4045201200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4065760800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4076650800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4097210400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 54822,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Mexico_City => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514715804, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1343066400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1234807200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1220292000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1207159200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1191344400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-975261600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-963169200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-917114400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-907354800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-821901600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-810068400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-627501600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-612990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909298800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941353200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (989136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1001833200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035702000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067151600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099206000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130655600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1175414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1193554800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1207468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225004400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1238918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1256454000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1270368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1288508400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1301817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1319958000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1333267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1351407600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1365321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1382857200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1396771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414306800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1428220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1445756400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1459670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1477810800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1491120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509260400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1522569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1540710000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1554624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1586073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1603609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1617523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1635663600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1648972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -23796,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Miquelon => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1847650520, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (326001600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (544597200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (562132800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (576046800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (594187200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (607496400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (625636800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (638946000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (657086400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (671000400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (688536000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (702450000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (719985600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (733899600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (752040000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (765349200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (783489600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (796798800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (814939200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (828853200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (846388800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (860302800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (877838400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (891752400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (909288000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (923202000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (941342400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (954651600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (972792000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (986101200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1004241600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1018155600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1035691200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1049605200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1067140800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1081054800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1099195200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1112504400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1130644800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1143954000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1162094400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1173589200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1194148800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1205038800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1225598400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1236488400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1257048000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1268542800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1289102400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1299992400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1320552000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1331442000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1352001600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1362891600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1383451200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1394341200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1414900800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1425790800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1446350400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1457845200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1478404800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1489294800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1509854400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1520744400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1541304000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1552194000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1572753600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1583643600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1604203200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1615698000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1636257600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1647147600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1667707200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1678597200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1699156800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1710046800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1730606400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1741496400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1762056000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1772946000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1793505600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1805000400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1825560000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1836450000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1857009600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1867899600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1888459200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1899349200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1919908800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1930798800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1951358400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1962853200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1983412800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1994302800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2014862400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2025752400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2046312000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2057202000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2077761600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2088651600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2109211200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2120101200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2140660800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2152155600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2172715200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2183605200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2204164800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2215054800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2235614400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2246504400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2267064000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2277954000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2298513600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2309403600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2329963200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2341458000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2362017600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2372907600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2393467200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2404357200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2424916800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2435806800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2456366400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2467256400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2487816000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2499310800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2519870400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2530760400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2551320000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2562210000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2582769600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2593659600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2614219200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2625109200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2645668800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2656558800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2677118400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2688613200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2709172800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2720062800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2740622400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2751512400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2772072000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2782962000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2803521600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2814411600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2834971200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2846466000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2867025600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2877915600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2898475200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2909365200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2929924800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2940814800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2961374400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (2972264400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (2992824000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3003714000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3024273600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3035768400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3056328000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3067218000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3087777600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3098667600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3119227200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3130117200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3150676800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3161566800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3182126400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3193016400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3213576000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3225070800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3245630400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3256520400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3277080000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3287970000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3308529600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3319419600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3339979200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3350869200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3371428800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3382923600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3403483200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3414373200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3434932800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3445822800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3466382400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3477272400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3497832000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3508722000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3529281600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3540171600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3560731200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3572226000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3592785600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3603675600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3624235200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3635125200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3655684800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3666574800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3687134400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3698024400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3718584000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3730078800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3750638400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3761528400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3782088000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3792978000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3813537600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3824427600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3844987200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3855877200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3876436800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3887326800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3907886400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3919381200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3939940800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3950830800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (3971390400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (3982280400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (4002840000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (4013730000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (4034289600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (4045179600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (4065739200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (4076629200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (4097188800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -13480,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Moncton => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2715882052, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-2131642800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1632074400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1615143600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1153681200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1145822400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1122231600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1114372800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1090782000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1082923200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1059332400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1051473600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1027882800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1020024000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-996433200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-988574400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-965674800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-955396800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-934743600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-923947200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-904503600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-891892800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-880221600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-747252000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-733950000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-715802400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-702500400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-684352800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-671050800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-652903200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-639601200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-620848800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-608151600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-589399200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-576097200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-557949600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-544647600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-526500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-513198000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-495050400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-481748400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-463600800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-450298800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-431546400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-418244400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-400096800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-384375600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-368647200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-352926000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-337197600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-321476400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-305748000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-289422000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-273693600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-257972400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-242244000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-226522800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-210794400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-195073200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-179344800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-163623600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-147895200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-131569200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-116445600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-100119600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-84391200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-68670000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-52941600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-37220400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-21492000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-5770800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (9957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (25678800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (41407200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (57733200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (73461600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (89182800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (136360800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (152082000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (167810400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (183531600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (199260000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (215586000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (230709600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (247035600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (262764000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (278485200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (294213600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (309934800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (325663200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (341384400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (357112800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (372834000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (388562400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (404888400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (420012000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (436338000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (452066400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (467787600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (483516000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (499237200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (514965600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (530686800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (544600800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (562136400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (576050400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (594190800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (607500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (625640400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (638949600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (657090000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (671004000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (688539600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (702453600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (719989200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (733896060, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (752036460, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (765345660, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (783486060, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (796795260, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (814935660, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (828849660, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (846385260, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (860299260, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (877834860, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (891748860, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (909284460, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (923198460, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (941338860, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (954648060, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (972788460, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (986097660, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1004238060, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1018152060, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1035687660, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1049601660, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1067137260, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1081051260, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1099191660, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1112500860, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1130641260, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1143950460, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1162090860, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1173592800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1194152400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1205042400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1225602000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1236492000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1257051600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1268546400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1289106000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1299996000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1320555600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1331445600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1352005200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1362895200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1383454800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1394344800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1414904400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1425794400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1446354000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1457848800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1478408400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1489298400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1509858000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1520748000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1541307600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1552197600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1572757200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1583647200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1604206800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1615701600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1636261200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1647151200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1667710800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1678600800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1699160400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1710050400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1730610000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1741500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1762059600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1772949600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1793509200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1805004000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1825563600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1836453600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1857013200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1867903200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1888462800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1899352800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1919912400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1930802400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1951362000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1962856800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1983416400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1994306400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2014866000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2025756000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2046315600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2057205600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2077765200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2088655200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2109214800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2120104800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2140664400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2152159200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2172718800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2183608800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2204168400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2215058400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2235618000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2246508000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2267067600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2277957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2298517200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2309407200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2329966800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2341461600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2362021200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2372911200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2393470800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2404360800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2424920400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2435810400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2456370000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2467260000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2487819600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2499314400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2519874000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2530764000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2551323600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2562213600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2582773200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2593663200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2614222800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2625112800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2645672400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2656562400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2677122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2688616800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2709176400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2720066400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2740626000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2751516000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2772075600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2782965600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2803525200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2814415200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2834974800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2846469600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2867029200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2877919200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2898478800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2909368800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2929928400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2940818400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2961378000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2972268000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2992827600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3003717600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3024277200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3035772000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3056331600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3067221600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3087781200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3098671200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3119230800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3130120800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3150680400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3161570400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3182130000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3193020000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3213579600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3225074400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3245634000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3256524000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3277083600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3287973600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3308533200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3319423200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3339982800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3350872800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3371432400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3382927200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3403486800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3414376800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3434936400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3445826400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3466386000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3477276000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3497835600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3508725600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3529285200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3540175200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3560734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3572229600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3592789200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3603679200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3624238800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3635128800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3655688400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3666578400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3687138000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3698028000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3718587600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3730082400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3750642000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3761532000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3782091600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3792981600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3813541200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3824431200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3844990800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3855880800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3876440400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3887330400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3907890000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3919384800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3939944400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3950834400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3971394000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3982284000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4002843600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4013733600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4034293200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4045183200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4065742800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4076632800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4097192400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15548,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Monterrey => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514719124, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (576057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (594198000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909298800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941353200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (989136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1001833200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035702000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067151600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099206000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130655600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1175414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1193554800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1207468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225004400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1238918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1256454000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1270368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1288508400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1301817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1319958000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1333267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1351407600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1365321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1382857200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1396771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414306800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1428220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1445756400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1459670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1477810800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1491120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509260400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1522569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1540710000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1554624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1586073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1603609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1617523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1635663600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1648972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -24076,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Montevideo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1942690509, FixedTimespan { utc_offset: -13491, dst_offset: 0, name: "MMT" }),
                    (-1567455309, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1459627200, FixedTimespan { utc_offset: -12600, dst_offset: 1800, name: "-03" }),
                    (-1443819600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "-0330" }),
                    (-1428006600, FixedTimespan { utc_offset: -12600, dst_offset: 1800, name: "-03" }),
                    (-1412283600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "-0330" }),
                    (-1396470600, FixedTimespan { utc_offset: -12600, dst_offset: 1800, name: "-03" }),
                    (-1380747600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "-0330" }),
                    (-1141590600, FixedTimespan { utc_offset: -12600, dst_offset: 1800, name: "-03" }),
                    (-1128286800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "-0330" }),
                    (-1110141000, FixedTimespan { utc_offset: -12600, dst_offset: 1800, name: "-03" }),
                    (-1096837200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "-0330" }),
                    (-1078691400, FixedTimespan { utc_offset: -12600, dst_offset: 1800, name: "-03" }),
                    (-1065387600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "-0330" }),
                    (-1047241800, FixedTimespan { utc_offset: -12600, dst_offset: 1800, name: "-03" }),
                    (-1033938000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "-0330" }),
                    (-1015187400, FixedTimespan { utc_offset: -12600, dst_offset: 1800, name: "-03" }),
                    (-1002488400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "-0330" }),
                    (-983737800, FixedTimespan { utc_offset: -12600, dst_offset: 1800, name: "-03" }),
                    (-971038800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "-0330" }),
                    (-954707400, FixedTimespan { utc_offset: -12600, dst_offset: 1800, name: "-03" }),
                    (-938984400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "-0330" }),
                    (-920838600, FixedTimespan { utc_offset: -12600, dst_offset: 1800, name: "-03" }),
                    (-907534800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "-0330" }),
                    (-896819400, FixedTimespan { utc_offset: -12600, dst_offset: 1800, name: "-03" }),
                    (-853621200, FixedTimespan { utc_offset: -10800, dst_offset: 1800, name: "-0230" }),
                    (-845847000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-334789200, FixedTimespan { utc_offset: -10800, dst_offset: 1800, name: "-0230" }),
                    (-319671000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-314226000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-309996000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-149720400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-134604000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-50446800, FixedTimespan { utc_offset: -10800, dst_offset: 1800, name: "-0230" }),
                    (-34205400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (9860400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (14176800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (72846000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (80100000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (127278000, FixedTimespan { utc_offset: -10800, dst_offset: 5400, name: "-0130" }),
                    (132111000, FixedTimespan { utc_offset: -10800, dst_offset: 1800, name: "-0230" }),
                    (147234600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (156913200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (165376800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (219812400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (226461600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (250052400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (257911200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (282711600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (289360800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (294202800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (322020000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (566449200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (573012000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (597812400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (625633200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (635911200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (667965600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (688532400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (730864800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1095562800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1111896000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1128834000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1142136000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1159678800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1173585600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1191733200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205035200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1223182800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1236484800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1254632400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1268539200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1286082000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1299988800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1317531600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1331438400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1349586000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1362888000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1381035600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1394337600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1412485200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1425787200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -13491,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Montreal => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2366736148, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1632070800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1615140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1601753400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1583697600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1567357200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1554667200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1534698000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1524074400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1503248400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1492365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1471798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1460916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1440954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1428861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1409504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1397412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1378054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1365962400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1346605200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1333908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1315155600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1301853600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1283706000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1270404000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1252256400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1238954400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1220806800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1207504800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1188752400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1176055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1157302800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1144000800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1125853200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1112551200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1094403600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1081101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1062954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1049652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1031504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1018202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1000054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-986752800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-968000400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-955303200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-936550800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-880221600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EWT" }),
                    (-769395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EPT" }),
                    (-765396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-747248400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-733946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-715798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-702496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-684349200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-671047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-652899600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-634154400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-620845200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-602704800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-589395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-576093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-557946000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-544644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-526496400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-513194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-495046800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-481744800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-463597200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-450295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-431542800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-418240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-400093200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-384372000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-368643600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-352922400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-337194000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-321472800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-305744400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-289418400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-273690000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-257968800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-242240400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-226519200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-210790800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-195069600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-179341200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-163620000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-147891600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-131565600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-116442000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-100116000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-84387600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-68666400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-52938000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-37216800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (41410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (57736800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (73465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (89186400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (136364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (152085600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (167814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (954658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (972799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -19052,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Montserrat => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Nassau => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2366736148, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1632070800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1615140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1601753400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1583697600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1567357200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1554667200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1534698000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1524074400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1503248400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1492365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1471798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1460916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1440954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1428861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1409504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1397412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1378054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1365962400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1346605200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1333908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1315155600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1301853600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1283706000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1270404000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1252256400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1238954400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1220806800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1207504800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1188752400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1176055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1157302800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1144000800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1125853200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1112551200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1094403600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1081101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1062954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1049652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1031504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1018202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1000054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-986752800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-968000400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-955303200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-936550800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-880221600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EWT" }),
                    (-769395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EPT" }),
                    (-765396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-747248400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-733946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-715798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-702496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-684349200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-671047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-652899600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-634154400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-620845200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-602704800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-589395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-576093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-557946000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-544644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-526496400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-513194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-495046800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-481744800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-463597200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-450295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-431542800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-418240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-400093200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-384372000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-368643600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-352922400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-337194000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-321472800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-305744400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-289418400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-273690000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-257968800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-242240400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-226519200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-210790800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-195069600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-179341200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-163620000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-147891600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-131565600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-116442000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-100116000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-84387600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-68666400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-52938000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-37216800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (41410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (57736800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (73465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (89186400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (136364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (152085600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (167814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (954658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (972799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -19052,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__New_York => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717633038, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1633280400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1615140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1601830800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1583690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1570381200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1551636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1536512400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1523210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1504458000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1491760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1473008400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1459706400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1441558800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1428256800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1410109200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1396807200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1378659600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1365357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1347210000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1333908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1315155600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1301853600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1283706000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1270404000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1252256400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1238954400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1220806800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1207504800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1189357200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1176055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1157302800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1144605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1125853200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1112551200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1094403600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1081101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1062954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1049652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1031504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1018202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1000054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-986752800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-968000400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-955303200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-936550800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-923248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-905101200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-891799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-880218000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EWT" }),
                    (-769395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EPT" }),
                    (-765396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-747248400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-733946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-715798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-702496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-684349200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-671047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-652899600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-639597600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-620845200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-608148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-589395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-576093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-557946000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-544644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-526496400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-513194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-495046800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-481744800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-463597200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-447271200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-431542800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-415821600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-400093200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-384372000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-368643600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-352922400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-337194000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-321472800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-305744400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-289418400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-273690000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-257968800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-242240400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-226519200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-210790800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-195069600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-179341200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-163620000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-147891600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-131565600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-116442000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-100116000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-84387600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-68666400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-52938000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-37216800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (41410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (57736800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (73465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (89186400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (126687600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (152085600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (162370800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (954658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (972799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -17762,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Nipigon => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2366736148, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1632070800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1615140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1601753400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1583697600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1567357200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1554667200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1534698000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1524074400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1503248400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1492365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1471798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1460916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1440954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1428861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1409504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1397412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1378054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1365962400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1346605200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1333908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1315155600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1301853600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1283706000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1270404000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1252256400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1238954400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1220806800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1207504800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1188752400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1176055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1157302800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1144000800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1125853200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1112551200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1094403600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1081101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1062954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1049652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1031504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1018202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1000054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-986752800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-968000400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-955303200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-936550800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-880221600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EWT" }),
                    (-769395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EPT" }),
                    (-765396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-747248400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-733946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-715798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-702496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-684349200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-671047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-652899600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-634154400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-620845200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-602704800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-589395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-576093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-557946000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-544644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-526496400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-513194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-495046800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-481744800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-463597200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-450295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-431542800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-418240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-400093200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-384372000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-368643600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-352922400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-337194000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-321472800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-305744400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-289418400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-273690000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-257968800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-242240400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-226519200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-210790800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-195069600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-179341200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-163620000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-147891600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-131565600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-116442000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-100116000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-84387600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-68666400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-52938000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-37216800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (41410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (57736800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (73465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (89186400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (136364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (152085600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (167814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (954658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (972799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -19052,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Nome => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3225223727, FixedTimespan { utc_offset: -39698, dst_offset: 0, name: "LMT" }),
                    (-2188947502, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "NST" }),
                    (-880196400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "NWT" }),
                    (-769395600, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "NPT" }),
                    (-765374400, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "NST" }),
                    (-86878800, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (-21466800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (-5745600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (9982800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (25704000, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (41432400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (57758400, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (73486800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (89208000, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (104936400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (120657600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (126709200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (152107200, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (162392400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (183556800, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (199285200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (215611200, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (230734800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (247060800, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (262789200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (278510400, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (294238800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (309960000, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (325688400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (341409600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (357138000, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (372859200, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (388587600, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (404913600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (420037200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (436363200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (439030800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (452084400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (467805600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (483534000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (499255200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (514983600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (530704800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (544618800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (562154400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (576068400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (594208800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (607518000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (625658400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (638967600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (657108000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (671022000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (688557600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (702471600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (720007200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (733921200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (752061600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (765370800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (783511200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (796820400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (814960800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (828874800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (846410400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (860324400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (877860000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (891774000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (909309600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (923223600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (941364000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (954673200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (972813600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (986122800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1004263200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1018177200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1035712800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1049626800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1067162400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1081076400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1099216800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1112526000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1130666400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1143975600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1162116000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1173610800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1194170400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1205060400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1225620000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1236510000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1257069600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1268564400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1289124000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1300014000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1320573600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1331463600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1352023200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1362913200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1383472800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1394362800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1414922400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1425812400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1446372000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1457866800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1478426400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1489316400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1509876000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1520766000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1541325600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1552215600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1572775200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1583665200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1604224800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1615719600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1636279200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1647169200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1667728800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1678618800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1699178400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1710068400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1730628000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1741518000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1762077600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1772967600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1793527200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1805022000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1825581600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1836471600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1857031200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1867921200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1888480800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1899370800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1919930400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1930820400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1951380000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1962874800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1983434400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1994324400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2014884000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2025774000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2046333600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2057223600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2077783200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2088673200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2109232800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2120122800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2140682400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2152177200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2172736800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2183626800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2204186400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2215076400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2235636000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2246526000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2267085600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2277975600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2298535200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2309425200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2329984800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2341479600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2362039200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2372929200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2393488800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2404378800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2424938400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2435828400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2456388000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2467278000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2487837600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2499332400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2519892000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2530782000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2551341600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2562231600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2582791200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2593681200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2614240800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2625130800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2645690400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2656580400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2677140000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2688634800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2709194400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2720084400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2740644000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2751534000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2772093600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2782983600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2803543200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2814433200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2834992800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2846487600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2867047200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2877937200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2898496800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2909386800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2929946400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2940836400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2961396000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2972286000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2992845600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3003735600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3024295200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3035790000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3056349600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3067239600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3087799200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3098689200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3119248800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3130138800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3150698400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3161588400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3182148000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3193038000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3213597600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3225092400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3245652000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3256542000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3277101600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3287991600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3308551200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3319441200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3340000800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3350890800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3371450400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3382945200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3403504800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3414394800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3434954400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3445844400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3466404000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3477294000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3497853600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3508743600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3529303200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3540193200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3560752800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3572247600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3592807200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3603697200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3624256800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3635146800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3655706400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3666596400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3687156000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3698046000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3718605600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3730100400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3750660000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3761550000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3782109600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3792999600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3813559200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3824449200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3845008800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3855898800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3876458400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3887348400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3907908000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3919402800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3939962400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3950852400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3971412000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3982302000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4002861600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4013751600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4034311200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4045201200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4065760800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4076650800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4097210400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 46702,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Noronha => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767217820, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1206961200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1191366000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1175378400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1159830000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-633823200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-622072800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-602287200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-591836400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-570751200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-560214000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-539128800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-531356400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-191368800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-184201200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-155167200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-150073200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-128901600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-121129200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-99957600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-89593200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-68421600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-57970800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (499744800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (511232400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (530589600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (540262800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (562125600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (571194000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (592970400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (602038800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (624420000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (634698000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (938916000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (951613200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (970970400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (971571600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1003024800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1013907600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -7780,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__North_Dakota__Beulah => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717619173, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1633273200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1601823600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1583683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-84380400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-68659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-52930800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-37209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-21481200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-5760000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (9968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (25689600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (41418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (57744000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (73472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (89193600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (104922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (120643200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (126694800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (152092800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (162378000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (183542400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (199270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (215596800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (230720400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (247046400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (262774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (278496000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (294224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (309945600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (325674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (341395200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (357123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (372844800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (388573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (404899200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (420022800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (436348800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (452077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (467798400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (483526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (499248000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (514976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (530697600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (544611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (562147200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (576061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (594201600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (607510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (625651200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (638960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (657100800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (671014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (688550400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (702464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (720000000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (733914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (752054400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (765363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (783504000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (796813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (814953600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (986115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1004256000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1173603600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1194163200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1205053200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225612800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1236502800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1257062400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1268557200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1289116800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -24427,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__North_Dakota__Center => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717619288, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1633273200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1601823600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1583683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-84380400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-68659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-52930800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-37209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-21481200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-5760000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (9968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (25689600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (41418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (57744000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (73472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (89193600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (104922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (120643200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (126694800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (152092800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (162378000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (183542400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (199270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (215596800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (230720400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (247046400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (262774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (278496000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (294224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (309945600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (325674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (341395200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (357123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (372844800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (388573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (404899200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (420022800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (436348800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (452077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (467798400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (483526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (499248000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (514976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (530697600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (544611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (562147200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (576061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (594201600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (607510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (625651200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (638960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (657100800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (671014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (688550400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (702464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (720000000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (733910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (752050800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (765360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (783500400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (796809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (814950000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909298800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941353200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (986112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1004252400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035702000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067151600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099206000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130655600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1205049600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1236499200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1257058800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -24312,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__North_Dakota__New_Salem => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717619261, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1633273200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1601823600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1583683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-84380400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-68659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-52930800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-37209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-21481200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-5760000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (9968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (25689600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (41418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (57744000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (73472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (89193600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (104922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (120643200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (126694800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (152092800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (162378000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (183542400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (199270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (215596800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (230720400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (247046400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (262774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (278496000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (294224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (309945600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (325674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (341395200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (357123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (372844800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (388573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (404899200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (420022800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (436348800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (452077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (467798400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (483526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (499248000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (514976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (530697600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (544611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (562147200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (576061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (594201600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (607510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (625651200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (638960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (657100800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (671014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (688550400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (702464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (720000000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (733914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (752054400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (765363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (783504000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (796813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (814953600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (986115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1004256000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099206000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130655600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1205049600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1236499200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1257058800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -24339,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Nuuk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1686083584, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (323845200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (338950800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (354675600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (370400400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (386125200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (401850000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (417574800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (433299600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (449024400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (465354000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (481078800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (496803600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (512528400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (528253200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (543978000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (559702800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (575427600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (591152400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (606877200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (622602000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (638326800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (654656400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (670381200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (686106000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (701830800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (717555600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (733280400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (749005200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (764730000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (780454800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (796179600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (811904400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (828234000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (846378000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (859683600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (877827600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (891133200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (909277200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (922582800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (941331600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (954032400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (972781200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (985482000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1004230800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1017536400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1035680400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1048986000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1067130000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1080435600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1099184400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1111885200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1130634000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1143334800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1162083600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1174784400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1193533200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1206838800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1224982800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1238288400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1256432400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1269738000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1288486800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1301187600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1319936400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1332637200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1351386000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1364691600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1382835600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1396141200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1414285200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1427590800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1445734800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1459040400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1477789200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1490490000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1509238800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1521939600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1540688400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1553994000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1572138000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1585443600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1603587600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1616893200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1635642000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1648342800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1667091600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1679792400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1679799600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1711846800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1729990800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1743296400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1761440400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1774746000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1792890000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1806195600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1824944400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1837645200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1856394000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1869094800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1887843600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1901149200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1919293200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1932598800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1950742800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1964048400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1982797200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1995498000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2014246800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2026947600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2045696400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2058397200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2077146000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2090451600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2108595600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2121901200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2140045200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2153350800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2172099600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2184800400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2203549200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2216250000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2234998800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2248304400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2266448400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2279754000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2297898000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2311203600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2329347600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2342653200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2361402000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2374102800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2392851600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2405552400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2424301200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2437606800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2455750800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2469056400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2487200400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2500506000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2519254800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2531955600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2550704400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2563405200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2582154000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2595459600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2613603600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2626909200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2645053200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2658358800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2676502800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2689808400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2708557200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2721258000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2740006800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2752707600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2771456400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2784762000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2802906000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2816211600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2834355600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2847661200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2866410000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2879110800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2897859600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2910560400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2929309200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2942010000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2960758800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2974064400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2992208400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3005514000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3023658000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3036963600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3055712400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3068413200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3087162000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3099862800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3118611600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3131917200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3150061200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3163366800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3181510800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3194816400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3212960400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3226266000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3245014800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3257715600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3276464400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3289165200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3307914000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3321219600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3339363600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3352669200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3370813200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3384118800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3402867600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3415568400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3434317200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3447018000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3465766800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3479072400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3497216400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3510522000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3528666000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3541971600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3560115600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3573421200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3592170000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3604870800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3623619600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3636320400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3655069200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3668374800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3686518800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3699824400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3717968400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3731274000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3750022800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3762723600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3781472400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3794173200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3812922000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3825622800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3844371600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3857677200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3875821200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3889126800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3907270800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3920576400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3939325200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3952026000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3970774800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3983475600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (4002224400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (4015530000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (4033674000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (4046979600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (4065123600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (4078429200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (4096573200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -12416,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Ojinaga => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514714540, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1343066400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1234807200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1220292000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1207159200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1191344400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (989139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1001836800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1175418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1193558400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1207472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225008000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1238922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1256457600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1268557200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1289116800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1300006800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1320566400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1331456400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1352016000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1362906000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1383465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1394355600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414915200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1425805200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1446364800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1457859600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1478419200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1489309200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509868800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1520758800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1541318400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1552208400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572768000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1583658000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1604217600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1615712400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1636272000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1647162000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667116800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25060,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Panama => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524502512, FixedTimespan { utc_offset: -19176, dst_offset: 0, name: "CMT" }),
                    (-1946918424, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -19088,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Pangnirtung => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-865296000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EWT" }),
                    (-769395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EPT" }),
                    (-765396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (73465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (89186400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (136364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (152085600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (167814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::America__Paramaribo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1861906760, FixedTimespan { utc_offset: -13252, dst_offset: 0, name: "PMT" }),
                    (-1104524348, FixedTimespan { utc_offset: -13236, dst_offset: 0, name: "PMT" }),
                    (-765317964, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "-0330" }),
                    (465449400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -13240,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Phoenix => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717616702, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1633273200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1601823600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1583683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-820519140, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-812653140, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-796845540, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-84380400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-68659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -26898,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__PortauPrince => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524504240, FixedTimespan { utc_offset: -17340, dst_offset: 0, name: "PPMT" }),
                    (-1670483460, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (421218000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436334400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452062800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467784000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483512400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499233600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514962000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530683200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (546411600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562132800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576050400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607500000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638949600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671004000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702453600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733903200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765352800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796802400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828856800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860306400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130644800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162094400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -17360,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Port_of_Spain => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Porto_Acre => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767209328, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1206950400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1191355200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1175367600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1159819200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-633812400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-622062000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-602276400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-591825600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-570740400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-560203200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-539118000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-531345600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-191358000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-184190400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-155156400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-150062400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-128890800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-121118400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-99946800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-89582400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-68410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-57960000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (499755600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (511243200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (530600400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (540273600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (562136400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (571204800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (1214283600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1384056000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16272,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Porto_Velho => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767210264, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1206954000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1191358800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1175371200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1159822800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-633816000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-622065600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-602280000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-591829200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-570744000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-560206800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-539121600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-531349200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-191361600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-155160000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-150066000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-128894400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-99950400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-89586000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-68414400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-57963600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (499752000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (511239600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (530596800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (540270000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (562132800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (571201200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15336,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Puerto_Rico => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Punta_Arenas => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524504580, FixedTimespan { utc_offset: -16965, dst_offset: 0, name: "SMT" }),
                    (-1892661435, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1688410800, FixedTimespan { utc_offset: -16965, dst_offset: 0, name: "SMT" }),
                    (-1619205435, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1593806400, FixedTimespan { utc_offset: -16965, dst_offset: 0, name: "SMT" }),
                    (-1335986235, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1317585600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1304362800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1286049600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1272826800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1254513600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1241290800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1222977600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1209754800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1191355200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1178132400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-870552000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-865278000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-736632000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-718056000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-713649600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-36619200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23922000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-3355200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (7527600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (24465600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (37767600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (55915200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (69217200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (87969600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (100666800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (118209600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (132116400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (150868800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (163566000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (182318400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (195620400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (213768000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (227070000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (245217600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (258519600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (277272000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (289969200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (308721600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (321418800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (340171200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (353473200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (371620800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (384922800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (403070400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (416372400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (434520000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (447822000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (466574400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (479271600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (498024000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (510721200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (529473600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (545194800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (560923200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (574225200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (592372800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (605674800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (624427200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (637124400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (653457600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (668574000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687326400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (700628400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (718776000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (732078000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (750225600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (763527600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (781675200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (794977200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (813729600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (826426800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (845179200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (859690800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (876628800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (889930800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (906868800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (923194800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (939528000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952830000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (971582400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (984279600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1003032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1015729200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1034481600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1047178800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1065931200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1079233200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1097380800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1110682800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1128830400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1142132400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1160884800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1173582000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1192334400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1206846000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1223784000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1237086000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1255233600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1270350000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1286683200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1304823600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1313899200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1335668400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1346558400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1367118000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1378612800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1398567600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1410062400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1463281200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1471147200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1480820400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -17020,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Rainy_River => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2602258284, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1694368800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1681671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1632067200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1029686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1018198800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-746035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-732733200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-715795200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-702493200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-684345600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-671043600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-652896000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-639594000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-620755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-607626000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-557942400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-544640400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-450291600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-431539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-418237200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-400089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-386787600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-355338000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-337190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-321469200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-305740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-292438800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-210787200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-198090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-116438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-100108800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-84384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-68659200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-52934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-37209600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-21484800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-5760000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (9964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (25689600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (41414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (57744000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (73468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (89193600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (104918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (120643200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (136368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152092800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (167817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (183542400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (199267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (215596800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (230716800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (247046400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (262771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (278496000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (294220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (309945600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (325670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (341395200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (357120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (372844800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (388569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (404899200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (420019200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (436348800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (452073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (467798400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (483523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (499248000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (514972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (530697600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (544608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (562147200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (576057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (594201600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (607507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (625651200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (638956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (657100800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (671011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (688550400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (702460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (720000000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (733910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (752054400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (765360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (783504000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (796809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (814953600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846403200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877852800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909302400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941356800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972806400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (986112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1004256000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035705600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067155200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099209600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130659200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1205049600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1236499200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1257058800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -23316,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Rankin_Inlet => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-410227200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (73468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (89190000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (104918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (120639600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (136368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152089200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (167817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (183538800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (199267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (215593200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (230716800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (247042800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (262771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (278492400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (294220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (309942000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (325670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (341391600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (357120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (372841200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (388569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (404895600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (420019200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (436345200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (452073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (467794800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (483523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (499244400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (514972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (530694000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (544608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (562143600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (576057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (594198000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (607507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (625647600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (638956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (657097200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (671011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (688546800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (702460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (719996400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (733910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (752050800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (765360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (783500400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (796809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (814950000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909298800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941353200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1004252400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035702000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067151600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099206000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130655600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1205049600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1236499200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1257058800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::America__Recife => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767217224, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1206957600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1191362400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1175374800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1159826400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-633819600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-622069200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-602283600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-591832800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-570747600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-560210400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-539125200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-531352800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-191365200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-184197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-155163600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-150069600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-128898000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-121125600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-99954000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-89589600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-68418000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-57967200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (499748400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (511236000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (530593200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (540266400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (562129200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (571197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (592974000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (602042400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (634701600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (951616800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (970974000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (971575200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1003028400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1013911200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -8376,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Regina => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2030202084, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1632063600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1251651600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1238349600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1220202000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1206900000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1188752400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1175450400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1156698000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1144000800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1125248400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1111946400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1032714000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1016992800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1001264400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-986148000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-969814800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-954093600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-937760400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-922039200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-906310800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-890589600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-748450800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-732729600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-715791600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-702489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-684342000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-671040000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-652892400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-639590400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-620838000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-608140800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-589388400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-576086400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-557938800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-544636800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-526489200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-513187200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-495039600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-481737600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-463590000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-450288000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-431535600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-418233600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-400086000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-386784000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-337186800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-321465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-305737200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25116,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Resolute => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-704937600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (73468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (89190000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (104918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (120639600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (136368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152089200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (167817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (183538800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (199267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (215593200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (230716800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (247042800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (262771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (278492400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (294220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (309942000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (325670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (341391600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (357120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (372841200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (388569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (404895600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (420019200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (436345200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (452073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (467794800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (483523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (499244400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (514972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (530694000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (544608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (562143600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (576057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (594198000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (607507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (625647600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (638956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (657097200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (671011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (688546800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (702460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (719996400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (733910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (752050800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (765360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (783500400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (796809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (814950000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909298800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941353200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1004252400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035702000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067151600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099206000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130655600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1205049600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1236499200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1257058800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::America__Rio_Branco => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767209328, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1206950400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1191355200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1175367600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1159819200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-633812400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-622062000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-602276400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-591825600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-570740400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-560203200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-539118000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-531345600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-191358000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-184190400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-155156400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-150062400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-128890800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-121118400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-99946800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-89582400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-68410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-57960000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (499755600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (511243200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (530600400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (540273600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (562136400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (571204800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (1214283600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1384056000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16272,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Rosario => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2372096592, FixedTimespan { utc_offset: -15408, dst_offset: 0, name: "CMT" }),
                    (-1567453392, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1233432000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1222981200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1205956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1194037200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1172865600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1162501200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1141329600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1130965200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1109793600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1099429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1078257600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1067806800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1046635200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1036270800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1015099200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1004734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-983563200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-973198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-952027200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-941576400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-931032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-900882000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-890337600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-833749200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-827265600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-752274000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-733780800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-197326800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-190843200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-164491200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (596948400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (605066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (636516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (667965600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687931200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (699415200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719377200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (731469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952052400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1198983600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1205632800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1224385200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1237082400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15408,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Santa_Isabel => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514711516, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1451667600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1343062800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1234803600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1222963200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1207242000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-873820800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PWT" }),
                    (-769370400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PPT" }),
                    (-761677200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-686073600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-661539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-495039600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-481734000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-463590000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-450284400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-431535600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-418230000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-400086000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-386780400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-368636400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-355330800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-337186800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-323881200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-305737200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-292431600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (199274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (215600400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (230724000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (247050000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (262778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (278499600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (294228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (309949200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (341398800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (452080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (467802000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (483530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (499251600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (514980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (530701200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (544615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (562150800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (576064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (594205200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (607514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (625654800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (638964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (657104400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (671018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (688554000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (702468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (720003600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (733917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (752058000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (765367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (783507600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (796816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (814957200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (828871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (846406800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (860320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (877856400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (891770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (909306000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (923220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (941360400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (954669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (972810000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (986119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1004259600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1018173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1035709200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1049623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1067158800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1081072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1099213200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1112522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1130662800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1143972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1162112400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1175421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1193562000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1207476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1225011600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1238925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1256461200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1268560800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1289120400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1300010400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1320570000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1331460000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1352019600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1362909600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1383469200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1394359200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1414918800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1425808800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1446368400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1457863200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1478422800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1489312800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1509872400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1520762400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1541322000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1552212000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1572771600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1583661600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1604221200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1615716000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1636275600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1647165600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1667725200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1678615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1699174800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1710064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1730624400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1741514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1762074000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1772964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1793523600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1805018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1825578000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1836468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1857027600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1867917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1888477200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1899367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1919926800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1930816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1951376400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1962871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1983430800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1994320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2014880400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2025770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2046330000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2057220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2077779600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2088669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2109229200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2120119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2140678800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2152173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2172733200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2183623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2204182800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2215072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2235632400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2246522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2267082000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2277972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2298531600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2309421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2329981200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2341476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2362035600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2372925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2393485200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2404375200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2424934800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2435824800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2456384400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2467274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2487834000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2499328800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2519888400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2530778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2551338000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2562228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2582787600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2593677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2614237200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2625127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2645686800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2656576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2677136400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2688631200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2709190800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2720080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2740640400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2751530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2772090000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2782980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2803539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2814429600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2834989200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2846484000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2867043600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2877933600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2898493200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2909383200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2929942800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2940832800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2961392400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2972282400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2992842000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3003732000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3024291600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3035786400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3056346000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3067236000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3087795600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3098685600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3119245200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3130135200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3150694800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3161584800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3182144400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3193034400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3213594000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3225088800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3245648400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3256538400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3277098000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3287988000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3308547600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3319437600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3339997200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3350887200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3371446800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3382941600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3403501200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3414391200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3434950800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3445840800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3466400400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3477290400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3497850000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3508740000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3529299600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3540189600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3560749200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3572244000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3592803600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3603693600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3624253200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3635143200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3655702800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3666592800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3687152400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3698042400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3718602000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3730096800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3750656400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3761546400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3782106000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3792996000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3813555600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3824445600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3845005200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3855895200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3876454800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3887344800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3907904400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3919399200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3939958800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3950848800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3971408400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3982298400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4002858000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4013748000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4034307600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4045197600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4065757200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4076647200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4097206800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -28084,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Santarem => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767212472, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1206954000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1191358800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1175371200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1159822800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-633816000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-622065600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-602280000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-591829200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-570744000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-560206800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-539121600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-531349200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-191361600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-155160000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-150066000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-128894400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-99950400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-89586000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-68414400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-57963600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (499752000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (511239600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (530596800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (540270000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (562132800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (571201200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1214280000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -13128,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Santiago => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524504635, FixedTimespan { utc_offset: -16965, dst_offset: 0, name: "SMT" }),
                    (-1892661435, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1688410800, FixedTimespan { utc_offset: -16965, dst_offset: 0, name: "SMT" }),
                    (-1619205435, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1593806400, FixedTimespan { utc_offset: -16965, dst_offset: 0, name: "SMT" }),
                    (-1335986235, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1317585600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1304362800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1286049600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1272826800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1254513600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1241290800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1222977600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1209754800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1191355200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1178132400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-870552000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-865278000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-740520000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-736635600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-718056000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-713649600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-36619200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23922000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-3355200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (7527600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (24465600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (37767600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (55915200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (69217200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (87969600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (100666800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (118209600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (132116400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (150868800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (163566000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (182318400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (195620400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (213768000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (227070000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (245217600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (258519600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (277272000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (289969200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (308721600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (321418800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (340171200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (353473200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (371620800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (384922800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (403070400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (416372400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (434520000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (447822000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (466574400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (479271600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (498024000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (510721200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (529473600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (545194800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (560923200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (574225200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (592372800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (605674800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (624427200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (637124400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (653457600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (668574000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687326400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (700628400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (718776000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (732078000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (750225600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (763527600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (781675200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (794977200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (813729600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (826426800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (845179200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (859690800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (876628800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (889930800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (906868800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (923194800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (939528000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952830000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (971582400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (984279600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1003032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1015729200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1034481600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1047178800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1065931200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1079233200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1097380800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1110682800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1128830400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1142132400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1160884800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1173582000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1192334400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1206846000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1223784000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1237086000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1255233600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1270350000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1286683200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1304823600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1313899200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1335668400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1346558400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1367118000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1378612800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1398567600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1410062400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1463281200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1471147200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1494730800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1502596800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1526180400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1534046400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1554606000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1567915200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1586055600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1599364800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1617505200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1630814400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1648954800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1662868800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1680404400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1693713600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1712458800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1725768000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1743908400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1757217600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1775358000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1788667200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1806807600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1820116800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1838257200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1851566400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1870311600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1883016000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1901761200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1915070400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1933210800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1946520000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1964660400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1977969600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1996110000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2009419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2027559600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2040868800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2059614000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2072318400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2091063600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2104372800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2122513200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2135822400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2153962800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2167272000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2185412400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2198721600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2217466800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2230171200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2248916400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2262225600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2280366000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2293675200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2311815600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2325124800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2343265200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2356574400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2374714800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2388024000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2406769200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2419473600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2438218800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2451528000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2469668400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2482977600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2501118000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2514427200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2532567600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2545876800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2564017200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2577326400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2596071600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2609380800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2627521200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2640830400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2658970800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2672280000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2690420400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2703729600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2721870000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2735179200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2753924400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2766628800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2785374000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2798683200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2816823600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2830132800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2848273200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2861582400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2879722800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2893032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2911172400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2924481600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2943226800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2955931200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2974676400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2987985600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3006126000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3019435200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3037575600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3050884800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3069025200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3082334400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3101079600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3113784000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3132529200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3145838400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3163978800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3177288000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3195428400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3208737600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3226878000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3240187200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3258327600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3271636800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3290382000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3303086400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3321831600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3335140800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3353281200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3366590400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3384730800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3398040000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3416180400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3429489600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3447630000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3460939200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3479684400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3492993600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3511134000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3524443200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3542583600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3555892800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3574033200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3587342400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3605482800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3618792000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3637537200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3650241600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3668986800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3682296000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3700436400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3713745600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3731886000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3745195200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3763335600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3776644800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3794785200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3808094400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3826839600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3839544000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3858289200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3871598400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3889738800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3903048000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3921188400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3934497600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3952638000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3965947200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3984692400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3997396800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (4016142000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (4029451200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (4047591600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (4060900800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (4079041200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (4092350400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16965,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Santo_Domingo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524504824, FixedTimespan { utc_offset: -16800, dst_offset: 0, name: "SDMT" }),
                    (-1159773600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-100119600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-89668800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-5770800, FixedTimespan { utc_offset: -18000, dst_offset: 1800, name: "-0430" }),
                    (4422600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (25678800, FixedTimespan { utc_offset: -18000, dst_offset: 1800, name: "-0430" }),
                    (33193800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (57733200, FixedTimespan { utc_offset: -18000, dst_offset: 1800, name: "-0430" }),
                    (64816200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (89182800, FixedTimespan { utc_offset: -18000, dst_offset: 1800, name: "-0430" }),
                    (96438600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (120632400, FixedTimespan { utc_offset: -18000, dst_offset: 1800, name: "-0430" }),
                    (127974600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (152082000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (972799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (975823200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16776,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Sao_Paulo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767214412, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1206957600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1191362400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1175374800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1159826400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-633819600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-622069200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-602283600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-591832800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-570747600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-560210400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-539125200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-531352800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-195426000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-184197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-155163600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-150069600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-128898000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-121125600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-99954000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-89589600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-68418000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-57967200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (499748400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (511236000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (530593200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (540266400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (562129200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (571197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (592974000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (602042400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (634701600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (666756000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (687927600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (697600800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719982000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (728445600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (750826800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (761709600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (782276400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (793159200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (813726000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (824004000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (844570800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (856058400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (876106800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (888717600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (908074800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (919562400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (951616800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (970974000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (982461600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1003028400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1013911200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1036292400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1045360800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1066532400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1076810400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1099364400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1108864800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1129431600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1140314400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1162695600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1172368800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1192330800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1203213600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1224385200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1234663200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1255834800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1266717600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1287284400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1298167200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1318734000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1330221600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1350788400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1361066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1382238000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1392516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1413687600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1424570400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1445137200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1456020000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1476586800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1487469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1508036400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1518919200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1541300400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1550368800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -11188,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Scoresbysund => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1686090728, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (323841600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (338961600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (354679200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (370400400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (386125200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (401850000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (417574800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (433299600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (449024400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (465354000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (481078800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (496803600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (512528400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (528253200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (543978000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (559702800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (575427600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (591152400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (606877200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (622602000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (638326800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (654656400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (670381200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (686106000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (701830800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (717555600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (733280400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (749005200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (764730000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (780454800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (796179600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (811904400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (828234000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (846378000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (859683600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (877827600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (891133200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (909277200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (922582800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (941331600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (954032400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (972781200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (985482000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1004230800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1017536400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1035680400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1048986000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1067130000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1080435600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1099184400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1111885200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1130634000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1143334800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1162083600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1174784400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1193533200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1206838800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1224982800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1238288400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1256432400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1269738000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1288486800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1301187600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1319936400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1332637200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1351386000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1364691600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1382835600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1396141200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1414285200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1427590800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1445734800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1459040400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1477789200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1490490000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1509238800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1521939600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1540688400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1553994000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1572138000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1585443600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1603587600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1616893200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1635642000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1648342800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1667091600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1679792400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1698541200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1711846800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1729990800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1743296400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1761440400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1774746000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1792890000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1806195600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1824944400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1837645200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1856394000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1869094800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1887843600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1901149200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1919293200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1932598800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1950742800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1964048400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1982797200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1995498000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2014246800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2026947600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2045696400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2058397200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2077146000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2090451600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2108595600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2121901200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2140045200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2153350800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2172099600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2184800400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2203549200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2216250000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2234998800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2248304400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2266448400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2279754000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2297898000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2311203600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2329347600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2342653200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2361402000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2374102800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2392851600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2405552400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2424301200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2437606800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2455750800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2469056400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2487200400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2500506000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2519254800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2531955600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2550704400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2563405200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2582154000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2595459600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2613603600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2626909200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2645053200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2658358800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2676502800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2689808400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2708557200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2721258000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2740006800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2752707600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2771456400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2784762000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2802906000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2816211600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2834355600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2847661200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2866410000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2879110800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2897859600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2910560400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2929309200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2942010000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2960758800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (2974064400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (2992208400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3005514000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3023658000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3036963600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3055712400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3068413200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3087162000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3099862800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3118611600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3131917200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3150061200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3163366800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3181510800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3194816400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3212960400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3226266000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3245014800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3257715600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3276464400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3289165200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3307914000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3321219600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3339363600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3352669200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3370813200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3384118800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3402867600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3415568400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3434317200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3447018000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3465766800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3479072400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3497216400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3510522000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3528666000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3541971600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3560115600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3573421200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3592170000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3604870800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3623619600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3636320400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3655069200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3668374800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3686518800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3699824400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3717968400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3731274000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3750022800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3762723600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3781472400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3794173200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3812922000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3825622800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3844371600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3857677200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3875821200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3889126800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3907270800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3920576400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3939325200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3952026000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (3970774800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (3983475600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (4002224400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (4015530000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (4033674000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (4046979600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (4065123600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (4078429200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (4096573200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -5272,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Shiprock => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717618404, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1633273200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1601823600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1583683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1570374000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1551628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1538924400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1534089600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-147884400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-131558400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-116434800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-100108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-84380400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-68659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-52930800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-37209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-21481200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-5760000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (9968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (25689600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (41418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (57744000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (73472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (89193600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (104922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (120643200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (126694800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (152092800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (162378000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (183542400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (199270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (215596800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (230720400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (247046400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (262774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (278496000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (294224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (309945600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (325674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (341395200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (357123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (372844800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (388573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (404899200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (420022800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (436348800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (452077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (467798400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (483526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (499248000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (514976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (530697600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (544611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (562147200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (576061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (594201600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (607510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (625651200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (638960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (657100800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (671014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (688550400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (702464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (720000000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (733914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (752054400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (765363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (783504000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (796813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (814953600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (986115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1004256000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1173603600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1194163200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1205053200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225612800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1236502800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1257062400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1268557200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1289116800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1300006800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1320566400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1331456400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1352016000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1362906000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1383465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1394355600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414915200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1425805200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1446364800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1457859600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1478419200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1489309200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509868800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1520758800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1541318400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1552208400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572768000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1583658000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1604217600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1615712400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1636272000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1647162000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667721600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1678611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1699171200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1710061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1730620800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1741510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1762070400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1772960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1793520000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1805014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1825574400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1836464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1857024000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1867914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1888473600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1899363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1919923200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1930813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1951372800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1962867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1983427200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1994317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2014876800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2025766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2046326400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2057216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2077776000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2088666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2109225600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2120115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2140675200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2152170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2172729600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2183619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2204179200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2215069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2235628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2246518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2267078400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2277968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2298528000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2309418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2329977600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2341472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2362032000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2372922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2393481600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2404371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2424931200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2435821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2456380800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2467270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2487830400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2499325200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2519884800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2530774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2551334400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2562224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2582784000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2593674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2614233600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2625123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2645683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2656573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2677132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2688627600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2709187200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2720077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2740636800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2751526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2772086400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2782976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2803536000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2814426000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2834985600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2846480400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2867040000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2877930000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2898489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2909379600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2929939200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2940829200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2961388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2972278800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2992838400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3003728400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3024288000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3035782800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3056342400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3067232400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3087792000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3098682000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3119241600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3130131600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3150691200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3161581200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3182140800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3193030800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3213590400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3225085200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3245644800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3256534800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3277094400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3287984400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3308544000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3319434000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3339993600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3350883600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3371443200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3382938000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3403497600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3414387600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3434947200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3445837200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3466396800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3477286800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3497846400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3508736400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3529296000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3540186000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3560745600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3572240400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3592800000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3603690000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3624249600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3635139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3655699200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3666589200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3687148800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3698038800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3718598400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3730093200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3750652800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3761542800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3782102400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3792992400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3813552000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3824442000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3845001600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3855891600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3876451200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3887341200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3907900800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3919395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3939955200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3950845200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3971404800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3982294800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4002854400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4013744400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4034304000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4045194000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4065753600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4076643600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4097203200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25196,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Sitka => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3225223727, FixedTimespan { utc_offset: -32473, dst_offset: 0, name: "LMT" }),
                    (-2188954727, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-880207200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PWT" }),
                    (-769395600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PPT" }),
                    (-765385200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-21477600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-5756400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (9972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (25693200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (41421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (57747600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (73476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (89197200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (104925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (120646800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (126698400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (152096400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (162381600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (183546000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (199274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (215600400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (230724000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (247050000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (262778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (278499600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (294228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (309949200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (341398800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (439030800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (452084400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (467805600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (483534000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (499255200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (514983600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (530704800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (544618800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (562154400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (576068400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (594208800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (607518000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (625658400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (638967600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (657108000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (671022000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (688557600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (702471600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (720007200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (733921200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (752061600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (765370800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (783511200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (796820400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (814960800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (828874800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (846410400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (860324400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (877860000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (891774000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (909309600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (923223600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (941364000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (954673200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (972813600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (986122800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1004263200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1018177200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1035712800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1049626800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1067162400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1081076400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1099216800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1112526000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1130666400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1143975600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1162116000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1173610800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1194170400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1205060400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1225620000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1236510000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1257069600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1268564400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1289124000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1300014000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1320573600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1331463600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1352023200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1362913200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1383472800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1394362800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1414922400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1425812400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1446372000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1457866800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1478426400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1489316400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1509876000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1520766000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1541325600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1552215600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1572775200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1583665200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1604224800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1615719600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1636279200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1647169200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1667728800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1678618800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1699178400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1710068400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1730628000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1741518000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1762077600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1772967600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1793527200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1805022000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1825581600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1836471600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1857031200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1867921200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1888480800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1899370800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1919930400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1930820400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1951380000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1962874800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1983434400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1994324400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2014884000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2025774000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2046333600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2057223600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2077783200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2088673200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2109232800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2120122800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2140682400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2152177200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2172736800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2183626800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2204186400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2215076400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2235636000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2246526000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2267085600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2277975600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2298535200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2309425200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2329984800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2341479600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2362039200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2372929200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2393488800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2404378800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2424938400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2435828400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2456388000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2467278000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2487837600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2499332400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2519892000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2530782000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2551341600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2562231600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2582791200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2593681200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2614240800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2625130800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2645690400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2656580400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2677140000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2688634800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2709194400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2720084400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2740644000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2751534000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2772093600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2782983600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2803543200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2814433200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2834992800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2846487600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2867047200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2877937200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2898496800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2909386800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2929946400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2940836400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2961396000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2972286000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2992845600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3003735600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3024295200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3035790000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3056349600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3067239600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3087799200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3098689200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3119248800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3130138800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3150698400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3161588400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3182148000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3193038000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3213597600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3225092400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3245652000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3256542000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3277101600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3287991600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3308551200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3319441200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3340000800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3350890800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3371450400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3382945200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3403504800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3414394800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3434954400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3445844400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3466404000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3477294000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3497853600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3508743600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3529303200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3540193200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3560752800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3572247600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3592807200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3603697200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3624256800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3635146800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3655706400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3666596400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3687156000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3698046000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3718605600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3730100400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3750660000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3761550000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3782109600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3792999600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3813559200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3824449200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3845008800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3855898800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3876458400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3887348400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3907908000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3919402800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3939962400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3950852400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3971412000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3982302000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4002861600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4013751600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4034311200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4045201200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4065760800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4076650800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4097210400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 53927,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__St_Barthelemy => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__St_Johns => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713897748, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1664130548, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1650137348, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1632076148, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1615145348, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1598650148, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1590100148, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1567286948, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1551565748, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1535837348, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1520116148, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1503782948, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1488666548, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1472333348, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1457216948, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1440883748, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1425767348, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1409434148, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1394317748, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1377984548, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1362263348, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1346534948, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1330813748, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1314480548, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1299364148, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1283030948, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1267914548, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1251581348, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1236464948, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1220131748, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1205015348, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1188682148, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1172960948, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1156627748, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1141511348, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1125178148, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1110061748, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1096921748, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-1093728600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-1078612200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-1061670600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-1048973400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-1030221000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-1017523800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-998771400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-986074200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-966717000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-954624600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-935267400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-922570200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-903817800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-891120600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-872368200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NWT" }),
                    (-769395600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NPT" }),
                    (-765401400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-746044200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-733347000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-714594600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-701897400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-683145000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-670447800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-651695400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-638998200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-619641000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-606943800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-589401000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-576099000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-557951400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-544649400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-526501800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-513199800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-495052200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-481750200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-463602600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-450300600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-431548200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-418246200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-400098600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-386796600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-368649000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-355347000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-337199400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-323897400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-305749800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-289423800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-273695400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-257974200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-242245800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-226524600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-210796200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-195075000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-179346600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-163625400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-147897000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-131571000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-116447400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-100121400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-84393000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-68671800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-52943400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-37222200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-21493800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-5772600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (9955800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (25677000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (41405400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (57731400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (73459800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (89181000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (104909400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (120630600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (136359000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (152080200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (167808600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (183529800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (199258200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (215584200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (230707800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (247033800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (262762200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (278483400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (294211800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (309933000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (325661400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (341382600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (357111000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (372832200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (388560600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (404886600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (420010200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (436336200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (452064600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (467785800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (483514200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (499235400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (514963800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (530685000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (544591860, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (562127460, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (576041460, FixedTimespan { utc_offset: -12600, dst_offset: 7200, name: "NDDT" }),
                    (594178260, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (607491060, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (625631460, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (638940660, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (657081060, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (670995060, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (688530660, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (702444660, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (719980260, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (733894260, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (752034660, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (765343860, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (783484260, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (796793460, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (814933860, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (828847860, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (846383460, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (860297460, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (877833060, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (891747060, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (909282660, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (923196660, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (941337060, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (954646260, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (972786660, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (986095860, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1004236260, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1018150260, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1035685860, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1049599860, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1067135460, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1081049460, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1099189860, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1112499060, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1130639460, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1143948660, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1162089060, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1173583860, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1194143460, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1205033460, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1225593060, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1236483060, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1257042660, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1268537460, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1289097060, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1299987060, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1320553800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1331443800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1352003400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1362893400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1383453000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1394343000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1414902600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1425792600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1446352200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1457847000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1478406600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1489296600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1509856200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1520746200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1541305800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1552195800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1572755400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1583645400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1604205000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1615699800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1636259400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1647149400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1667709000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1678599000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1699158600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1710048600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1730608200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1741498200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1762057800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1772947800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1793507400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1805002200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1825561800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1836451800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1857011400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1867901400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1888461000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1899351000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1919910600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1930800600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1951360200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1962855000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1983414600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1994304600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2014864200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2025754200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2046313800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2057203800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2077763400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2088653400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2109213000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2120103000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2140662600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2152157400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2172717000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2183607000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2204166600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2215056600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2235616200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2246506200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2267065800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2277955800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2298515400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2309405400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2329965000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2341459800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2362019400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2372909400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2393469000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2404359000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2424918600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2435808600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2456368200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2467258200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2487817800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2499312600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2519872200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2530762200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2551321800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2562211800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2582771400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2593661400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2614221000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2625111000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2645670600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2656560600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2677120200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2688615000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2709174600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2720064600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2740624200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2751514200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2772073800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2782963800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2803523400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2814413400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2834973000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2846467800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2867027400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2877917400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2898477000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2909367000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2929926600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2940816600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2961376200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2972266200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2992825800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3003715800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3024275400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3035770200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3056329800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3067219800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3087779400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3098669400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3119229000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3130119000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3150678600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3161568600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3182128200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3193018200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3213577800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3225072600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3245632200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3256522200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3277081800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3287971800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3308531400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3319421400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3339981000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3350871000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3371430600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3382925400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3403485000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3414375000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3434934600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3445824600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3466384200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3477274200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3497833800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3508723800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3529283400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3540173400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3560733000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3572227800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3592787400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3603677400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3624237000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3635127000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3655686600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3666576600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3687136200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3698026200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3718585800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3730080600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3750640200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3761530200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3782089800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3792979800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3813539400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3824429400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3844989000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3855879000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3876438600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3887328600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3907888200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3919383000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3939942600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3950832600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3971392200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3982282200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (4002841800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (4013731800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (4034291400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (4045181400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (4065741000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (4076631000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (4097190600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -12652,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__St_Kitts => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__St_Lucia => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__St_Thomas => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__St_Vincent => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Swift_Current => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2030201320, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1632063600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-747241200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-732729600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-715791600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-702489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-684342000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-671040000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-652892400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-639590400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-400086000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-384364800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-337186800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-321465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-305737200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-292435200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-273682800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-260985600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (73472400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25880,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Tegucigalpa => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1538503868, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (547020000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (559717200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (578469600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (591166800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1146981600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1154926800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20932,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Thule => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1686079492, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (670399200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (686120400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (701848800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (717570000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (733903200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (752043600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (765352800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (783493200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (796802400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (814942800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (828856800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (846392400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (860306400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (877842000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (891756000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (909291600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (923205600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (941346000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (954655200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (972795600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (986104800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1004245200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1018159200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1035694800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1049608800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1067144400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1081058400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1099198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1112508000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1130648400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1143957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1162098000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1173592800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1194152400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1205042400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1225602000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1236492000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1257051600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1268546400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1289106000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1299996000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1320555600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1331445600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1352005200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1362895200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1383454800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1394344800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1414904400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1425794400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1446354000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1457848800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1478408400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1489298400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1509858000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1520748000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1541307600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1552197600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1572757200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1583647200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1604206800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1615701600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1636261200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1647151200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1667710800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1678600800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1699160400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1710050400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1730610000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1741500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1762059600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1772949600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1793509200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1805004000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1825563600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1836453600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1857013200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1867903200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1888462800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1899352800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1919912400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1930802400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1951362000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1962856800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1983416400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1994306400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2014866000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2025756000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2046315600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2057205600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2077765200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2088655200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2109214800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2120104800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2140664400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2152159200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2172718800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2183608800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2204168400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2215058400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2235618000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2246508000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2267067600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2277957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2298517200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2309407200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2329966800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2341461600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2362021200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2372911200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2393470800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2404360800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2424920400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2435810400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2456370000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2467260000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2487819600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2499314400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2519874000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2530764000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2551323600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2562213600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2582773200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2593663200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2614222800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2625112800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2645672400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2656562400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2677122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2688616800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2709176400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2720066400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2740626000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2751516000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2772075600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2782965600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2803525200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2814415200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2834974800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2846469600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2867029200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2877919200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2898478800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2909368800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2929928400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2940818400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2961378000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2972268000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2992827600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3003717600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3024277200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3035772000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3056331600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3067221600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3087781200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3098671200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3119230800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3130120800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3150680400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3161570400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3182130000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3193020000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3213579600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3225074400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3245634000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3256524000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3277083600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3287973600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3308533200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3319423200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3339982800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3350872800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3371432400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3382927200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3403486800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3414376800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3434936400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3445826400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3466386000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3477276000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3497835600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3508725600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3529285200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3540175200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3560734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3572229600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3592789200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3603679200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3624238800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3635128800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3655688400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3666578400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3687138000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3698028000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3718587600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3730082400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3750642000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3761532000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3782091600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3792981600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3813541200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3824431200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3844990800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3855880800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3876440400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3887330400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3907890000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3919384800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3939944400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3950834400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3971394000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3982284000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4002843600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4013733600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4034293200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4045183200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4065742800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4076632800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4097192400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16508,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Thunder_Bay => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2366736148, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1632070800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1615140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1601753400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1583697600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1567357200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1554667200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1534698000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1524074400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1503248400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1492365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1471798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1460916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1440954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1428861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1409504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1397412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1378054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1365962400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1346605200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1333908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1315155600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1301853600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1283706000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1270404000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1252256400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1238954400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1220806800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1207504800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1188752400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1176055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1157302800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1144000800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1125853200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1112551200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1094403600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1081101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1062954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1049652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1031504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1018202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1000054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-986752800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-968000400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-955303200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-936550800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-880221600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EWT" }),
                    (-769395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EPT" }),
                    (-765396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-747248400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-733946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-715798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-702496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-684349200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-671047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-652899600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-634154400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-620845200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-602704800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-589395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-576093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-557946000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-544644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-526496400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-513194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-495046800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-481744800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-463597200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-450295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-431542800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-418240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-400093200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-384372000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-368643600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-352922400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-337194000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-321472800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-305744400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-289418400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-273690000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-257968800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-242240400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-226519200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-210790800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-195069600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-179341200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-163620000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-147891600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-131565600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-116442000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-100116000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-84387600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-68666400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-52938000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-37216800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (41410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (57736800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (73465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (89186400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (136364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (152085600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (167814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (954658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (972799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -19052,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Tijuana => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514711516, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1451667600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1343062800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1234803600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1222963200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1207242000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-873820800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PWT" }),
                    (-769370400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PPT" }),
                    (-761677200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-686073600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-661539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-495039600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-481734000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-463590000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-450284400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-431535600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-418230000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-400086000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-386780400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-368636400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-355330800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-337186800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-323881200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-305737200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-292431600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (199274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (215600400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (230724000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (247050000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (262778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (278499600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (294228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (309949200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (341398800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (452080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (467802000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (483530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (499251600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (514980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (530701200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (544615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (562150800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (576064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (594205200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (607514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (625654800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (638964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (657104400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (671018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (688554000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (702468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (720003600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (733917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (752058000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (765367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (783507600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (796816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (814957200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (828871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (846406800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (860320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (877856400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (891770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (909306000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (923220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (941360400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (954669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (972810000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (986119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1004259600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1018173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1035709200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1049623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1067158800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1081072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1099213200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1112522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1130662800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1143972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1162112400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1175421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1193562000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1207476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1225011600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1238925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1256461200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1268560800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1289120400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1300010400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1320570000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1331460000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1352019600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1362909600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1383469200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1394359200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1414918800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1425808800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1446368400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1457863200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1478422800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1489312800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1509872400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1520762400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1541322000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1552212000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1572771600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1583661600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1604221200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1615716000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1636275600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1647165600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1667725200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1678615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1699174800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1710064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1730624400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1741514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1762074000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1772964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1793523600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1805018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1825578000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1836468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1857027600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1867917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1888477200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1899367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1919926800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1930816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1951376400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1962871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1983430800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1994320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2014880400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2025770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2046330000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2057220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2077779600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2088669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2109229200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2120119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2140678800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2152173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2172733200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2183623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2204182800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2215072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2235632400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2246522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2267082000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2277972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2298531600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2309421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2329981200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2341476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2362035600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2372925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2393485200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2404375200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2424934800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2435824800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2456384400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2467274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2487834000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2499328800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2519888400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2530778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2551338000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2562228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2582787600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2593677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2614237200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2625127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2645686800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2656576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2677136400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2688631200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2709190800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2720080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2740640400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2751530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2772090000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2782980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2803539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2814429600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2834989200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2846484000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2867043600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2877933600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2898493200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2909383200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2929942800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2940832800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2961392400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2972282400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2992842000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3003732000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3024291600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3035786400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3056346000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3067236000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3087795600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3098685600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3119245200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3130135200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3150694800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3161584800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3182144400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3193034400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3213594000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3225088800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3245648400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3256538400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3277098000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3287988000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3308547600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3319437600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3339997200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3350887200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3371446800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3382941600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3403501200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3414391200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3434950800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3445840800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3466400400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3477290400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3497850000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3508740000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3529299600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3540189600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3560749200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3572244000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3592803600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3603693600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3624253200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3635143200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3655702800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3666592800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3687152400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3698042400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3718602000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3730096800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3750656400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3761546400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3782106000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3792996000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3813555600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3824445600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3845005200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3855895200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3876454800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3887344800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3907904400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3919399200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3939958800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3950848800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3971408400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3982298400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4002858000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4013748000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4034307600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4045197600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4065757200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4076647200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4097206800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -28084,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Toronto => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2366736148, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1632070800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1615140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1601753400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1583697600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1567357200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1554667200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1534698000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1524074400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1503248400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1492365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1471798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1460916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1440954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1428861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1409504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1397412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1378054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1365962400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1346605200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1333908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1315155600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1301853600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1283706000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1270404000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1252256400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1238954400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1220806800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1207504800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1188752400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1176055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1157302800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1144000800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1125853200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1112551200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1094403600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1081101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1062954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1049652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1031504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1018202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1000054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-986752800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-968000400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-955303200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-936550800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-880221600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EWT" }),
                    (-769395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EPT" }),
                    (-765396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-747248400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-733946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-715798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-702496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-684349200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-671047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-652899600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-634154400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-620845200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-602704800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-589395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-576093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-557946000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-544644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-526496400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-513194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-495046800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-481744800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-463597200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-450295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-431542800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-418240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-400093200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-384372000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-368643600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-352922400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-337194000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-321472800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-305744400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-289418400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-273690000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-257968800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-242240400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-226519200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-210790800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-195069600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-179341200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-163620000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-147891600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-131565600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-116442000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-100116000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-84387600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-68666400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-52938000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-37216800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (41410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (57736800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (73465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (89186400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (136364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (152085600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (167814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (954658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (972799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -19052,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Tortola => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Vancouver => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713880852, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1632060000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1615129200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-880207200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PWT" }),
                    (-769395600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PPT" }),
                    (-765385200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-747237600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-733935600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-715788000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-702486000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-684338400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-671036400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-652888800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-639586800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-620834400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-608137200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-589384800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-576082800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-557935200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-544633200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-526485600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-513183600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-495036000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-481734000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-463586400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-450284400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-431532000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-418230000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-400082400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-386780400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-368632800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-355330800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-337183200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-323881200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-305733600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-292431600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-273679200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-260982000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-242229600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-226508400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-210780000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-195058800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-179330400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-163609200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-147880800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-131554800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-116431200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-100105200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-84376800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-68655600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-52927200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-37206000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-21477600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-5756400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (9972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (25693200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (41421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (57747600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (73476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (89197200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (104925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (120646800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (136375200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (152096400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (167824800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (183546000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (199274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (215600400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (230724000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (247050000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (262778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (278499600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (294228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (309949200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (341398800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (452080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (467802000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (483530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (499251600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (514980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (530701200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (544615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (562150800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (576064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (594205200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (607514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (625654800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (638964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (657104400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (671018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (688554000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (702468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (720003600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (733917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (752058000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (765367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (783507600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (796816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (814957200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (828871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (846406800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (860320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (877856400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (891770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (909306000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (923220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (941360400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (954669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (972810000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (986119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1004259600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1018173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1035709200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1049623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1067158800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1081072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1099213200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1112522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1130662800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1143972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1162112400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1173607200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1194166800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1205056800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1225616400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1236506400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1257066000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1268560800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1289120400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1300010400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1320570000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1331460000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1352019600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1362909600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1383469200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1394359200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1414918800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1425808800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1446368400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1457863200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1478422800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1489312800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1509872400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1520762400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1541322000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1552212000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1572771600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1583661600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1604221200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1615716000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1636275600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1647165600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1667725200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1678615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1699174800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1710064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1730624400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1741514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1762074000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1772964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1793523600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1805018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1825578000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1836468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1857027600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1867917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1888477200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1899367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1919926800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1930816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1951376400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1962871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1983430800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1994320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2014880400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2025770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2046330000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2057220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2077779600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2088669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2109229200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2120119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2140678800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2152173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2172733200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2183623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2204182800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2215072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2235632400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2246522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2267082000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2277972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2298531600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2309421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2329981200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2341476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2362035600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2372925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2393485200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2404375200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2424934800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2435824800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2456384400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2467274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2487834000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2499328800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2519888400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2530778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2551338000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2562228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2582787600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2593677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2614237200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2625127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2645686800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2656576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2677136400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2688631200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2709190800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2720080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2740640400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2751530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2772090000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2782980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2803539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2814429600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2834989200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2846484000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2867043600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2877933600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2898493200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2909383200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2929942800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2940832800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2961392400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2972282400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2992842000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3003732000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3024291600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3035786400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3056346000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3067236000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3087795600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3098685600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3119245200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3130135200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3150694800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3161584800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3182144400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3193034400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3213594000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3225088800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3245648400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3256538400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3277098000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3287988000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3308547600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3319437600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3339997200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3350887200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3371446800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3382941600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3403501200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3414391200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3434950800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3445840800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3466400400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3477290400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3497850000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3508740000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3529299600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3540189600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3560749200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3572244000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3592803600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3603693600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3624253200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3635143200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3655702800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3666592800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3687152400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3698042400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3718602000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3730096800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3750656400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3761546400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3782106000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3792996000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3813555600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3824445600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3845005200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3855895200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3876454800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3887344800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3907904400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3919399200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3939958800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3950848800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3971408400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3982298400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4002858000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4013748000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4034307600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4045197600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4065757200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4076647200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4097206800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -29548,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Virgin => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2233035335, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-873057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Whitehorse => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2188997988, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-1632056400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (-1615125600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-1596978000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (-1583164800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-880203600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YWT" }),
                    (-769395600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YPT" }),
                    (-765381600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-147884400, FixedTimespan { utc_offset: -32400, dst_offset: 7200, name: "YDDT" }),
                    (-131554800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-121273200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (341398800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (452080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (467802000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (483530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (499251600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (514980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (530701200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (544615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (562150800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (576064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (594205200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (607514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (625654800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (638964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (657104400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (671018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (688554000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (702468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (720003600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (733917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (752058000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (765367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (783507600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (796816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (814957200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (828871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (846406800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (860320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (877856400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (891770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (909306000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (923220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (941360400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (954669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (972810000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (986119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1004259600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1018173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1035709200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1049623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1067158800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1081072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1099213200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1112522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1130662800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1143972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1162112400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1173607200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1194166800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1205056800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1225616400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1236506400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1257066000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1268560800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1289120400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1300010400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1320570000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1331460000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1352019600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1362909600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1383469200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1394359200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1414918800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1425808800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1446368400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1457863200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1478422800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1489312800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1509872400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1520762400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1541322000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1552212000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1572771600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1583661600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1604214000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -32412,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Winnipeg => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2602258284, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1694368800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1681671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1632067200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1029686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1018198800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-746035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-732733200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-715795200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-702493200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-684345600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-671043600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-652896000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-639594000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-620755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-607626000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-557942400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-544640400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-450291600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-431539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-418237200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-400089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-386787600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-355338000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-337190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-321469200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-305740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-292438800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-210787200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-198090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-116438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-100108800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-84384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-68659200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-52934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-37209600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-21484800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-5760000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (9964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (25689600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (41414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (57744000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (73468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (89193600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (104918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (120643200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (136368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152092800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (167817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (183542400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (199267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (215596800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (230716800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (247046400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (262771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (278496000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (294220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (309945600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (325670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (341395200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (357120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (372844800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (388569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (404899200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (420019200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (436348800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (452073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (467798400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (483523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (499248000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (514972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (530697600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (544608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (562147200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (576057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (594201600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (607507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (625651200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (638956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (657100800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (671011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (688550400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (702460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (720000000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (733910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (752054400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (765360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (783504000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (796809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (814953600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846403200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877852800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909302400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941356800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972806400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (986112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1004256000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035705600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067155200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099209600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130659200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1205049600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1236499200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1257058800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -23316,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Yakutat => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3225223727, FixedTimespan { utc_offset: -33535, dst_offset: 0, name: "LMT" }),
                    (-2188953665, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-880203600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YWT" }),
                    (-769395600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YPT" }),
                    (-765381600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-21474000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (-5752800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (9975600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (25696800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (41425200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (57751200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (73479600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (89200800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (104929200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (120650400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (126702000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (152100000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (162385200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (183549600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (199278000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (215604000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (230727600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (247053600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (262782000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (278503200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (294231600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (309952800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (325681200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (341402400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (357130800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (372852000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (388580400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (404906400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (420030000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (436356000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (439030800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (452084400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (467805600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (483534000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (499255200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (514983600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (530704800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (544618800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (562154400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (576068400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (594208800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (607518000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (625658400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (638967600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (657108000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (671022000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (688557600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (702471600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (720007200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (733921200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (752061600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (765370800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (783511200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (796820400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (814960800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (828874800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (846410400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (860324400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (877860000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (891774000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (909309600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (923223600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (941364000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (954673200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (972813600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (986122800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1004263200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1018177200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1035712800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1049626800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1067162400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1081076400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1099216800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1112526000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1130666400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1143975600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1162116000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1173610800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1194170400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1205060400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1225620000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1236510000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1257069600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1268564400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1289124000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1300014000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1320573600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1331463600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1352023200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1362913200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1383472800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1394362800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1414922400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1425812400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1446372000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1457866800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1478426400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1489316400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1509876000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1520766000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1541325600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1552215600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1572775200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1583665200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1604224800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1615719600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1636279200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1647169200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1667728800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1678618800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1699178400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1710068400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1730628000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1741518000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1762077600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1772967600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1793527200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1805022000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1825581600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1836471600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1857031200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1867921200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1888480800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1899370800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1919930400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1930820400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1951380000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1962874800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1983434400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1994324400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2014884000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2025774000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2046333600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2057223600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2077783200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2088673200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2109232800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2120122800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2140682400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2152177200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2172736800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2183626800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2204186400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2215076400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2235636000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2246526000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2267085600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2277975600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2298535200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2309425200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2329984800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2341479600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2362039200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2372929200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2393488800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2404378800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2424938400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2435828400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2456388000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2467278000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2487837600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2499332400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2519892000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2530782000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2551341600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2562231600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2582791200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2593681200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2614240800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2625130800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2645690400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2656580400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2677140000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2688634800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2709194400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2720084400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2740644000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2751534000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2772093600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2782983600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2803543200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2814433200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2834992800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2846487600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2867047200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2877937200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2898496800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2909386800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2929946400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2940836400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2961396000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2972286000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2992845600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3003735600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3024295200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3035790000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3056349600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3067239600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3087799200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3098689200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3119248800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3130138800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3150698400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3161588400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3182148000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3193038000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3213597600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3225092400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3245652000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3256542000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3277101600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3287991600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3308551200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3319441200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3340000800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3350890800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3371450400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3382945200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3403504800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3414394800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3434954400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3445844400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3466404000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3477294000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3497853600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3508743600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3529303200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3540193200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3560752800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3572247600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3592807200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3603697200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3624256800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3635146800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3655706400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3666596400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3687156000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3698046000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3718605600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3730100400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3750660000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3761550000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3782109600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3792999600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3813559200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3824449200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3845008800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3855898800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3876458400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3887348400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3907908000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3919402800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3939962400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3950852400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3971412000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3982302000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4002861600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4013751600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4034311200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4045201200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4065760800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4076650800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4097210400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 52865,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::America__Yellowknife => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1998663968, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1632063600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1600614000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1596816000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1567954800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1551628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1536505200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1523203200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1504450800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1491753600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1473001200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1459699200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-715791600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-702489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (73472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (89193600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (104922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (120643200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (136371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (152092800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (167821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (183542400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (199270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (215596800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (230720400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (247046400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (262774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (278496000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (294224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (309945600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (325674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (341395200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (357123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (372844800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (388573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (404899200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (420022800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (436348800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (452077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (467798400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (483526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (499248000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (514976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (530697600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (544611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (562147200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (576061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (594201600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (607510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (625651200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (638960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (657100800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (671014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (688550400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (702464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (720000000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (733914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (752054400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (765363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (783504000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (796813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (814953600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (986115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1004256000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1173603600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1194163200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1205053200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225612800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1236502800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1257062400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1268557200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1289116800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1300006800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1320566400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1331456400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1352016000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1362906000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1383465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1394355600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414915200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1425805200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1446364800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1457859600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1478419200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1489309200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509868800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1520758800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1541318400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1552208400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572768000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1583658000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1604217600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1615712400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1636272000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1647162000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667721600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1678611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1699171200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1710061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1730620800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1741510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1762070400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1772960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1793520000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1805014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1825574400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1836464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1857024000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1867914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1888473600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1899363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1919923200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1930813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1951372800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1962867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1983427200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1994317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2014876800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2025766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2046326400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2057216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2077776000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2088666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2109225600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2120115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2140675200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2152170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2172729600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2183619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2204179200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2215069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2235628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2246518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2267078400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2277968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2298528000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2309418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2329977600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2341472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2362032000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2372922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2393481600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2404371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2424931200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2435821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2456380800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2467270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2487830400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2499325200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2519884800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2530774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2551334400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2562224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2582784000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2593674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2614233600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2625123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2645683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2656573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2677132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2688627600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2709187200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2720077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2740636800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2751526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2772086400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2782976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2803536000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2814426000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2834985600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2846480400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2867040000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2877930000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2898489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2909379600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2929939200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2940829200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2961388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2972278800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2992838400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3003728400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3024288000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3035782800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3056342400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3067232400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3087792000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3098682000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3119241600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3130131600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3150691200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3161581200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3182140800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3193030800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3213590400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3225085200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3245644800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3256534800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3277094400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3287984400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3308544000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3319434000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3339993600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3350883600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3371443200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3382938000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3403497600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3414387600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3434947200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3445837200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3466396800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3477286800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3497846400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3508736400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3529296000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3540186000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3560745600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3572240400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3592800000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3603690000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3624249600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3635139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3655699200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3666589200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3687148800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3698038800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3718598400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3730093200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3750652800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3761542800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3782102400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3792992400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3813552000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3824442000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3845001600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3855891600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3876451200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3887341200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3907900800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3919395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3939955200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3950845200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3971404800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3982294800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4002854400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4013744400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4034304000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4045194000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4065753600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4076643600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4097203200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -27232,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Antarctica__Casey => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-31536000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1255802400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1267714800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1319738400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1329804000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1477065600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1520701200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1538856000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1552752000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1570129200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1583596800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1601740860, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1615640400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1633190460, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1647090000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1664640060, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1678291200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::Antarctica__Davis => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-409190400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (-163062000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "-00" }),
                    (-28857600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1255806000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1268233200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1319742000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1329836400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::Antarctica__DumontDUrville => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840176120, FixedTimespan { utc_offset: 35312, dst_offset: 0, name: "PMMT" }),
                    (-2366790512, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 35320,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Antarctica__Macquarie => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2214259200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1680508800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1665388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1601719200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "-00" }),
                    (-687052800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-71136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-55411200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-37267200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-25776000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-5817600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (5673600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (25632000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (37728000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (57686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (67968000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (89136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (100022400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (120585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (131472000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (152035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (162921600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (183484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (194976000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (215539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (226425600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (246988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (257875200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (278438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (289324800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (309888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (320774400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (341337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (352224000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (372787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (386092800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (404841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (417542400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (436291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (447177600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (467740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (478627200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (499190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (510076800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (530035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (542736000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (562089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (574790400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (594144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (606240000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (625593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (637689600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (657043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (670348800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (686678400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (701798400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (718128000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (733248000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (749577600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (764697600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (781027200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (796147200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (812476800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (828201600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (844531200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (859651200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (875980800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (891100800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (907430400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (922550400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (938880000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (954000000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (967305600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (985449600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1002384000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1017504000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1033833600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1048953600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1065283200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1080403200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1096732800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1111852800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1128182400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1143907200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1159632000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1174752000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1191686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1207411200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1223136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1238860800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1254585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1301760000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1317484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1333209600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1349539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1365264000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1380988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1396713600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1412438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1428163200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1443888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1459612800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1475337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1491062400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1506787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1522512000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1538841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1554566400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1570291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1586016000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1601740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1617465600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1633190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1648915200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1664640000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1680364800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1696089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1712419200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1728144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1743868800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1759593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1775318400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1791043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1806768000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1822492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1838217600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1853942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1869667200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1885996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1901721600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1917446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1933171200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1948896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1964620800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1980345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1996070400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2011795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2027520000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2043244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2058969600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2075299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2091024000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2106748800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2122473600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2138198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2153923200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2169648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2185372800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2201097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2216822400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2233152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2248876800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2264601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2280326400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2296051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2311776000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2327500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2343225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2358950400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2374675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2390400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2406124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2422454400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2438179200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2453904000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2469628800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2485353600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2501078400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2516803200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2532528000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2548252800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2563977600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2579702400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2596032000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2611756800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2627481600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2643206400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2658931200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2674656000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2690380800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2706105600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2721830400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2737555200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2753280000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2769609600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2785334400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2801059200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2816784000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2832508800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2848233600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2863958400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2879683200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2895408000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2911132800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2926857600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2942582400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2958912000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2974636800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2990361600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3006086400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3021811200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3037536000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3053260800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3068985600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3084710400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3100435200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3116764800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3132489600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3148214400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3163939200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3179664000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3195388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3211113600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3226838400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3242563200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3258288000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3274012800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3289737600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3306067200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3321792000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3337516800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3353241600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3368966400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3384691200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3400416000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3416140800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3431865600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3447590400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3463315200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3479644800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3495369600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3511094400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3526819200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3542544000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3558268800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3573993600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3589718400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3605443200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3621168000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3636892800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3653222400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3668947200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3684672000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3700396800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3716121600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3731846400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3747571200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3763296000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3779020800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3794745600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3810470400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3826195200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3842524800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3858249600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3873974400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3889699200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3905424000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3921148800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3936873600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3952598400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3968323200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3984048000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4000377600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4016102400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4031827200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4047552000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4063276800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4079001600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4094726400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::Antarctica__Mawson => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-501206400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1255809600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::Antarctica__McMurdo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3192435544, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1330335000, FixedTimespan { utc_offset: 41400, dst_offset: 3600, name: "NZST" }),
                    (-1320057000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1300699800, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1287396000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1269250200, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1255946400, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1237800600, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1224496800, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1206351000, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1192442400, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1174901400, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1160992800, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1143451800, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1125914400, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1112607000, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1094464800, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1081157400, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1063015200, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1049707800, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1031565600, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1018258200, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1000116000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-986808600, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-968061600, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-955359000, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-936612000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-923304600, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-757425600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (152632800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (162309600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (183477600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (194968800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (215532000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (226418400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (246981600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (257868000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (278431200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (289317600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (309880800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (320767200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (341330400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (352216800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (372780000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (384271200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (404834400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (415720800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (436284000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (447170400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (467733600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (478620000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (499183200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (510069600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (530632800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (541519200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (562082400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (573573600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (594136800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (605023200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (623772000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (637682400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (655221600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (669132000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (686671200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (700581600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (718120800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (732636000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (749570400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (764085600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (781020000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (795535200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (812469600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (826984800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (844524000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (858434400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (875973600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (889884000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (907423200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (921938400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (938872800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (953388000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (970322400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (984837600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1002376800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1016287200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1033826400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1047736800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1065276000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1079791200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1096725600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1111240800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1128175200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1142690400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1159624800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1174140000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1191074400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1207404000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1222524000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1238853600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1253973600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1270303200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1285423200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1301752800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1316872800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1333202400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1348927200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1365256800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1380376800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1396706400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1411826400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1428156000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1443276000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1459605600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1474725600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1491055200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1506175200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1522504800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1538229600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1554559200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1569679200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1586008800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1601128800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1617458400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1632578400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1648908000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1664028000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1680357600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1695477600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1712412000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1727532000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1743861600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1758981600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1775311200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1790431200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1806760800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1821880800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1838210400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1853330400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1869660000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1885384800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1901714400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1916834400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1933164000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1948284000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1964613600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1979733600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1996063200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2011183200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2027512800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2042632800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2058962400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2074687200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2091016800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2106136800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2122466400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2137586400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2153916000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2169036000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2185365600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2200485600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2216815200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2232540000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2248869600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2263989600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2280319200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2295439200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2311768800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2326888800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2343218400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2358338400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2374668000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2389788000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2406117600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2421842400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2438172000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2453292000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2469621600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2484741600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2501071200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2516191200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2532520800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2547640800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2563970400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2579090400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2596024800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2611144800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2627474400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2642594400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2658924000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2674044000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2690373600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2705493600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2721823200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2736943200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2753272800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2768997600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2785327200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2800447200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2816776800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2831896800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2848226400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2863346400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2879676000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2894796000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2911125600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2926245600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2942575200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2958300000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2974629600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2989749600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3006079200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3021199200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3037528800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3052648800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3068978400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3084098400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3100428000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3116152800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3132482400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3147602400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3163932000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3179052000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3195381600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3210501600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3226831200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3241951200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3258280800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3273400800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3289730400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3305455200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3321784800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3336904800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3353234400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3368354400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3384684000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3399804000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3416133600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3431253600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3447583200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3462703200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3479637600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3494757600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3511087200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3526207200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3542536800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3557656800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3573986400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3589106400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3605436000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3620556000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3636885600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3652610400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3668940000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3684060000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3700389600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3715509600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3731839200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3746959200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3763288800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3778408800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3794738400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3809858400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3826188000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3841912800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3858242400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3873362400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3889692000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3904812000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3921141600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3936261600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3952591200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3967711200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3984040800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3999765600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (4016095200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (4031215200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (4047544800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (4062664800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (4078994400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (4094114400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 41944,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Antarctica__Palmer => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-157766400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-152658000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-132955200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-101419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-86821200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-71092800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-54766800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-39038400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23317200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-7588800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (128142000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (136605600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (389070000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (403070400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (416372400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (434520000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (447822000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (466574400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (479271600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (498024000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (510721200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (529473600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (545194800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (560923200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (574225200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (592372800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (605674800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (624427200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (637124400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (653457600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (668574000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687326400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (700628400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (718776000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (732078000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (750225600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (763527600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (781675200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (794977200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (813729600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (826426800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (845179200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (859690800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (876628800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (889930800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (906868800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (923194800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (939528000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952830000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (971582400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (984279600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1003032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1015729200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1034481600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1047178800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1065931200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1079233200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1097380800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1110682800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1128830400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1142132400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1160884800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1173582000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1192334400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1206846000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1223784000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1237086000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1255233600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1270350000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1286683200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1304823600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1313899200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1335668400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1346558400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1367118000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1378612800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1398567600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1410062400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1463281200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1471147200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1480820400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::Antarctica__Rothera => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (218246400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::Antarctica__South_Pole => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3192435544, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1330335000, FixedTimespan { utc_offset: 41400, dst_offset: 3600, name: "NZST" }),
                    (-1320057000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1300699800, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1287396000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1269250200, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1255946400, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1237800600, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1224496800, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1206351000, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1192442400, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1174901400, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1160992800, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1143451800, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1125914400, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1112607000, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1094464800, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1081157400, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1063015200, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1049707800, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1031565600, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1018258200, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1000116000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-986808600, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-968061600, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-955359000, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-936612000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-923304600, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-757425600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (152632800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (162309600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (183477600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (194968800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (215532000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (226418400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (246981600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (257868000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (278431200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (289317600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (309880800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (320767200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (341330400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (352216800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (372780000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (384271200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (404834400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (415720800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (436284000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (447170400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (467733600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (478620000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (499183200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (510069600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (530632800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (541519200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (562082400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (573573600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (594136800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (605023200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (623772000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (637682400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (655221600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (669132000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (686671200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (700581600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (718120800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (732636000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (749570400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (764085600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (781020000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (795535200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (812469600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (826984800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (844524000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (858434400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (875973600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (889884000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (907423200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (921938400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (938872800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (953388000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (970322400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (984837600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1002376800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1016287200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1033826400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1047736800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1065276000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1079791200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1096725600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1111240800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1128175200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1142690400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1159624800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1174140000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1191074400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1207404000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1222524000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1238853600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1253973600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1270303200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1285423200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1301752800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1316872800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1333202400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1348927200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1365256800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1380376800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1396706400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1411826400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1428156000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1443276000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1459605600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1474725600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1491055200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1506175200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1522504800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1538229600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1554559200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1569679200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1586008800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1601128800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1617458400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1632578400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1648908000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1664028000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1680357600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1695477600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1712412000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1727532000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1743861600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1758981600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1775311200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1790431200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1806760800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1821880800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1838210400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1853330400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1869660000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1885384800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1901714400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1916834400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1933164000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1948284000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1964613600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1979733600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1996063200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2011183200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2027512800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2042632800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2058962400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2074687200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2091016800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2106136800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2122466400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2137586400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2153916000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2169036000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2185365600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2200485600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2216815200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2232540000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2248869600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2263989600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2280319200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2295439200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2311768800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2326888800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2343218400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2358338400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2374668000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2389788000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2406117600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2421842400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2438172000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2453292000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2469621600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2484741600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2501071200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2516191200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2532520800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2547640800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2563970400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2579090400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2596024800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2611144800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2627474400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2642594400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2658924000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2674044000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2690373600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2705493600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2721823200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2736943200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2753272800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2768997600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2785327200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2800447200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2816776800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2831896800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2848226400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2863346400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2879676000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2894796000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2911125600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2926245600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2942575200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2958300000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2974629600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2989749600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3006079200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3021199200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3037528800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3052648800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3068978400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3084098400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3100428000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3116152800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3132482400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3147602400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3163932000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3179052000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3195381600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3210501600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3226831200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3241951200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3258280800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3273400800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3289730400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3305455200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3321784800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3336904800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3353234400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3368354400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3384684000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3399804000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3416133600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3431253600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3447583200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3462703200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3479637600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3494757600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3511087200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3526207200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3542536800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3557656800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3573986400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3589106400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3605436000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3620556000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3636885600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3652610400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3668940000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3684060000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3700389600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3715509600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3731839200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3746959200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3763288800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3778408800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3794738400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3809858400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3826188000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3841912800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3858242400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3873362400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3889692000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3904812000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3921141600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3936261600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3952591200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3967711200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3984040800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3999765600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (4016095200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (4031215200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (4047544800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (4062664800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (4078994400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (4094114400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 41944,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Antarctica__Syowa => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-719636812, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 11212,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Antarctica__Troll => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (1108166400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1111885200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1130634000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1143334800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1162083600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1174784400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1193533200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1206838800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1224982800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1238288400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1256432400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1269738000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1288486800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1301187600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1319936400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1332637200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1351386000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1364691600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1382835600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1396141200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1414285200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1427590800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1445734800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1459040400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1477789200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1490490000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1509238800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1521939600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1540688400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1553994000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1572138000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1585443600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1603587600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1616893200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1635642000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1648342800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1667091600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1679792400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1698541200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1711846800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1729990800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1743296400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1761440400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1774746000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1792890000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1806195600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1824944400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1837645200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1856394000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1869094800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1887843600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1901149200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1919293200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1932598800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1950742800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1964048400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (1982797200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (1995498000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2014246800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2026947600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2045696400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2058397200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2077146000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2090451600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2108595600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2121901200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2140045200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2153350800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2172099600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2184800400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2203549200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2216250000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2234998800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2248304400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2266448400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2279754000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2297898000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2311203600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2329347600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2342653200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2361402000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2374102800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2392851600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2405552400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2424301200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2437606800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2455750800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2469056400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2487200400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2500506000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2519254800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2531955600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2550704400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2563405200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2582154000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2595459600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2613603600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2626909200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2645053200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2658358800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2676502800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2689808400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2708557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2721258000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2740006800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2752707600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2771456400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2784762000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2802906000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2816211600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2834355600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2847661200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2866410000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2879110800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2897859600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2910560400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2929309200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2942010000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2960758800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (2974064400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (2992208400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3005514000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3023658000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3036963600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3055712400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3068413200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3087162000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3099862800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3118611600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3131917200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3150061200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3163366800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3181510800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3194816400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3212960400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3226266000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3245014800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3257715600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3276464400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3289165200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3307914000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3321219600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3339363600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3352669200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3370813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3384118800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3402867600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3415568400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3434317200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3447018000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3465766800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3479072400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3497216400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3510522000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3528666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3541971600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3560115600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3573421200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3592170000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3604870800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3623619600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3636320400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3655069200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3668374800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3686518800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3699824400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3717968400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3731274000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3750022800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3762723600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3781472400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3794173200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3812922000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3825622800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3844371600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3857677200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3875821200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3889126800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3907270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3920576400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3939325200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3952026000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (3970774800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (3983475600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (4002224400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (4015530000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (4033674000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (4046979600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (4065123600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                    (4078429200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "+02" }),
                    (4096573200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "+00" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::Antarctica__Vostok => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-380073600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (760035600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "-00" }),
                    (783648000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1702839600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::Arctic__Longyearbyen => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2422054408, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693706400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-938905200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-776563200, FixedTimespan { utc_offset: 3600, dst_offset: 7200, name: "CEMT" }),
                    (-765936000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-761180400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-748479600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733273200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-717631200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-714610800, FixedTimespan { utc_offset: 3600, dst_offset: 7200, name: "CEMT" }),
                    (-710380800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701910000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-684975600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-654130800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-639010800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3208,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Aden => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-719636812, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 11212,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Almaty => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441170468, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (-1247547600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (354909600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (370717200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (386445600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (402253200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (417981600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (433789200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (449604000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (465336000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (481060800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (496785600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (512510400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (528235200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (543960000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (559684800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (575409600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (591134400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (606859200, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (622584000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (638308800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (654638400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (670363200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (686091600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (695768400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (701812800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (717537600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (733262400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (748987200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (764712000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (780436800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (796161600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (811886400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (828216000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (846360000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (859665600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (877809600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (891115200, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (909259200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (922564800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (941313600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (954014400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (972763200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (985464000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1004212800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1017518400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1035662400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1048968000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1067112000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1080417600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1099162800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1709229600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 18468,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Amman => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1230776624, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (108165600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (118270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (136591200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (149806800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (168127200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (181342800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (199749600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (215643600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (231285600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (244501200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (262735200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (275950800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (481154400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (496962000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (512949600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (528670800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (544399200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (560120400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (575848800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (592174800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (610581600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (623624400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (641167200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (655074000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (671839200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (685918800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (702856800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717973200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733701600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749422800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (765151200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (779662800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (797205600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811116000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828655200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (843170400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (860104800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (874620000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891554400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (906069600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (930780000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (938124000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954367200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (970178400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985816800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1001628000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017352800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1033077600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048802400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1066946400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080252000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1097791200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1112306400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1128031200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143756000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1161900000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1175205600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193349600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206655200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1225404000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238104800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256853600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269554400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288303200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301608800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319752800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1333058400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1387486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1395957600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414706400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427407200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1446156000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459461600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477605600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490911200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509055200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1522360800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540504800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553810400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1571954400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585260000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1604008800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616709600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635458400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1645740000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1666904400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8624,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Anadyr => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441194596, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (-1247572800, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                    (354884400, FixedTimespan { utc_offset: 46800, dst_offset: 3600, name: "+14" }),
                    (370692000, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                    (386420400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (402231600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (417960000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (433767600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (449582400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (465314400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (481039200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (496764000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (512488800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (528213600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (543938400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (559663200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (575388000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (591112800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (606837600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (622562400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (638287200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (654616800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (670341600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (686070000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (695746800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (701791200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (717516000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (733240800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (748965600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (764690400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (780415200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (796140000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (811864800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (828194400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (846338400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (859644000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (877788000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (891093600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (909237600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (922543200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (941292000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (953992800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (972741600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (985442400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1004191200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1017496800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1035640800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1048946400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1067090400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1080396000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1099144800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1111845600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1130594400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1143295200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1162044000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1174744800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1193493600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1206799200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1224943200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1238248800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1256392800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1269698400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1288450800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1301151600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 42596,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Aqtau => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441164064, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (-1247544000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (370724400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (386445600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (402256800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (417985200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (433792800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (449607600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (465339600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (481064400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (496789200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (512514000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (528238800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (543963600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (559688400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (575413200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (591138000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (606862800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (622587600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (638312400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (654642000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (670366800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (686095200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (695772000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (701816400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (717541200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (733266000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (748990800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (764715600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (780436800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (780444000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (796168800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (811893600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (828223200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (846367200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (859672800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (877816800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (891122400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (909266400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (922572000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (941320800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (954021600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (972770400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (985471200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1004220000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1017525600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1035669600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1048975200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1067119200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1080424800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1099170000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 12064,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Aqtobe => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441165720, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (-1247544000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (354913200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (370720800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (386445600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (402256800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (417985200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (433792800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (449607600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (465339600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (481064400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (496789200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (512514000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (528238800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (543963600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (559688400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (575413200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (591138000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (606862800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (622587600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (638312400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (654642000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (670366800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (686095200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (695772000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (701816400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (717541200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (733266000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (748990800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (764715600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (780440400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (796165200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (811890000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (828219600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (846363600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (859669200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (877813200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (891118800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (909262800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (922568400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (941317200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (954018000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (972766800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (985467600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1004216400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1017522000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1035666000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1048971600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1067115600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1080421200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1099166400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 13720,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Ashgabat => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441166012, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (-1247544000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (354913200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (370720800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (386449200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (402256800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (417985200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (433792800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (449607600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (465339600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (481064400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (496789200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (512514000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (528238800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (543963600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (559688400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (575413200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (591138000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (606862800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (622587600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (638312400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (654642000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (670366800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (686095200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (695772000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 14012,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Ashkhabad => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441166012, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (-1247544000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (354913200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (370720800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (386449200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (402256800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (417985200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (433792800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (449607600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (465339600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (481064400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (496789200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (512514000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (528238800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (543963600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (559688400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (575413200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (591138000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (606862800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (622587600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (638312400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (654642000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (670366800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (686095200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (695772000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 14012,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Atyrau => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441164464, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (-1247540400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (370724400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (386445600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (402256800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (417985200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (433792800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (449607600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (465339600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (481064400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (496789200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (512514000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (528238800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (543963600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (559688400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (575413200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (591138000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (606862800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (622587600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (638312400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (654642000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (670366800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (686095200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (695772000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (701816400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (717541200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (733266000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (748990800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (764715600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (780440400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (796165200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (811890000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (828219600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (846363600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (859669200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (877813200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (891118800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (909262800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (922568400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (941320800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (954021600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (972770400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (985471200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1004220000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1017525600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1035669600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1048975200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1067119200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1080424800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1099170000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 12464,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Baghdad => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524532260, FixedTimespan { utc_offset: 10656, dst_offset: 0, name: "BMT" }),
                    (-1641005856, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (389048400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (402264000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (417906000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (433800000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (449614800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (465422400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (481150800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (496792800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (512517600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (528242400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (543967200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (559692000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (575416800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (591141600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (606866400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (622591200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (638316000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (654645600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (670464000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (686275200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (702086400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (717897600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (733622400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (749433600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (765158400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (780969600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (796694400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (812505600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (828316800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (844128000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (859852800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (875664000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (891388800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (907200000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (922924800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (938736000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (954547200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (970358400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (986083200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1001894400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1017619200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1033430400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1049155200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1064966400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1080777600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1096588800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1112313600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1128124800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1143849600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1159660800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1175385600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1191196800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 10660,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Bahrain => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577935568, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (76190400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 12368,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Baku => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441163964, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (-405140400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (354916800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (370724400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (386452800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (402260400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (417988800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (433796400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (449611200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (465343200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (481068000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (496792800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (512517600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (528242400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (543967200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (559692000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (575416800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (591141600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (606866400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (622591200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (638316000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (654645600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (670370400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (686098800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (701823600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (717544800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (828234000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (846378000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (859680000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (877824000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (891129600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (909273600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (922579200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (941328000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (954028800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (972777600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (985478400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1004227200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1017532800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1035676800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1048982400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1067126400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1080432000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1099180800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1111881600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1130630400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1143331200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1162080000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1174780800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1193529600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1206835200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1224979200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1238284800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1256428800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1269734400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1288483200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1301184000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1319932800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1332633600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1351382400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1364688000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1382832000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1396137600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1414281600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1427587200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1445731200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 11964,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Bangkok => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840164924, FixedTimespan { utc_offset: 24124, dst_offset: 0, name: "BMT" }),
                    (-1570084924, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 24124,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Barnaul => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1579844100, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (-1247551200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (354906000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (370713600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (386442000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (402249600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (417978000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (433785600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (449600400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (465332400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (481057200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (496782000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (512506800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (528231600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (543956400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (559681200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (575406000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (591130800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (606855600, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (622580400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (638305200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (654634800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (670359600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (686088000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (695764800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (701809200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (717534000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (733258800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (748983600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (764708400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (780433200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (796158000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (801590400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (811886400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (828216000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (846360000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (859665600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (877809600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (891115200, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (909259200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (922564800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (941313600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (954014400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (972763200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (985464000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1004212800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1017518400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1035662400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1048968000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1067112000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1080417600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1099166400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1111867200, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1130616000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1143316800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1162065600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1174766400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1193515200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1206820800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1224964800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1238270400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1256414400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1269720000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1288468800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1301169600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1414263600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1459022400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 20100,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Beirut => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840149320, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1570413600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1552186800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1538359200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1522551600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1507514400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1490583600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1473645600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1460948400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-399866400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-386650800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-368330400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-355114800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-336794400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-323578800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-305172000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-291956400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-273636000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-260420400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (78012000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (86734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (105055200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (118270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (136591200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (149806800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (168127200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (181342800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (199749600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (212965200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (231285600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (244501200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (262735200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (275950800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (452210400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (466722000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (483746400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (498258000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (515282400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (529794000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (546818400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (561330000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (581119200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (592952400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (610754400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (624488400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (641512800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (656024400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (673048800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (687560400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (704671200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (718146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733269600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (748990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764719200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796168800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828223200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (843944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859672800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (875394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891122400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (906843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922572000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954021600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985471200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017525600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048975200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080424800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111874400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143324000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174773600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206828000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238277600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256418000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269727200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301176800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332626400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364680800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396130400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427580000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445720400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459029600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490479200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521928800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553983200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585432800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616882400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635627600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648332000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667077200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679781600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698526800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711836000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729976400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743285600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761426000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774735200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792875600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806184800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824930000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837634400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856379600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869084000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887829200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901138400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919278800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932588000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950728400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964037600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982782800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995487200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014232400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026936800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045682000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058386400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077131600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090440800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108581200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121890400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140030800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153340000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172085200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184789600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203534800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216239200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234984400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248293600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266434000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279743200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297883600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311192800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329333200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342642400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361387600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374092000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392837200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405541600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424286800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437596000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455736400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469045600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487186000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500495200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519240400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531944800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550690000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563394400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582139600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595448800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613589200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626898400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645038800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658348000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676488400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689797600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708542800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721247200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2739992400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752696800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771442000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784751200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802891600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816200800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834341200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847650400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866395600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879100000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897845200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910549600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929294800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2941999200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960744400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974053600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992194000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005503200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023643600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036952800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055698000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068402400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087147600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099852000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118597200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131906400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150046800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163356000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181496400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194805600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212946000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226255200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245000400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257704800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276450000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289154400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307899600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321208800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339349200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352658400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370798800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384108000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402853200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415557600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434302800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447007200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465752400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479061600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497202000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510511200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528651600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541960800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560101200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573410400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592155600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604860000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623605200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636309600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655054800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668364000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686504400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699813600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717954000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731263200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750008400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762712800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781458000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794162400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812907600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825612000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844357200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857666400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875806800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889116000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907256400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920565600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939310800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952015200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970760400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983464800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002210000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015519200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033659600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046968800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065109200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078418400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096558800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8520,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Bishkek => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441169904, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (-1247547600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (354909600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (370717200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (386445600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (402253200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (417981600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (433789200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (449604000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (465336000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (481060800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (496785600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (512510400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (528235200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (543960000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (559684800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (575409600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (591134400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (606859200, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (622584000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (638308800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (654638400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (670363200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (683582400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (703018800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (717530400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (734468400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (748980000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (765918000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (780429600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (797367600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (811879200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (828817200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (843933600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (859671000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (877811400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (891120600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (909261000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (922570200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (941315400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (954019800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (972765000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (985469400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1004214600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1017523800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1035664200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1048973400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1067113800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1080423000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1099168200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1111872600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1123783200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 17904,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Brunei => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1383463280, FixedTimespan { utc_offset: 27000, dst_offset: 0, name: "+0730" }),
                    (-1167636600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-1082448000, FixedTimespan { utc_offset: 28800, dst_offset: 1200, name: "+0820" }),
                    (-1074586800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-1050825600, FixedTimespan { utc_offset: 28800, dst_offset: 1200, name: "+0820" }),
                    (-1042964400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-1019289600, FixedTimespan { utc_offset: 28800, dst_offset: 1200, name: "+0820" }),
                    (-1011428400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-987753600, FixedTimespan { utc_offset: 28800, dst_offset: 1200, name: "+0820" }),
                    (-979892400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-956217600, FixedTimespan { utc_offset: 28800, dst_offset: 1200, name: "+0820" }),
                    (-948356400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-924595200, FixedTimespan { utc_offset: 28800, dst_offset: 1200, name: "+0820" }),
                    (-916734000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-893059200, FixedTimespan { utc_offset: 28800, dst_offset: 1200, name: "+0820" }),
                    (-885198000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-879667200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-767005200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 26480,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Calcutta => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3645237208, FixedTimespan { utc_offset: 21200, dst_offset: 0, name: "HMT" }),
                    (-3155694800, FixedTimespan { utc_offset: 19270, dst_offset: 0, name: "MMT" }),
                    (-2019705670, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "IST" }),
                    (-891581400, FixedTimespan { utc_offset: 19800, dst_offset: 3600, name: "+0630" }),
                    (-872058600, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "IST" }),
                    (-862637400, FixedTimespan { utc_offset: 19800, dst_offset: 3600, name: "+0630" }),
                    (-764145000, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "IST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 21208,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Chita => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1579419232, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-1247558400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (354898800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (370706400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (386434800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (402242400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (417970800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (433778400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (449593200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (465325200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (481050000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (496774800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (512499600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (528224400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (543949200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (559674000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (575398800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (591123600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (606848400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (622573200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (638298000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (654627600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (670352400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (686080800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (695757600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (701802000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (717526800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (733251600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (748976400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (764701200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (780426000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (796150800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (811875600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (828205200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (846349200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (859654800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (877798800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (891104400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (909248400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (922554000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (941302800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (954003600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (972752400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (985453200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1004202000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1017507600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1035651600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1048957200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1067101200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1080406800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1099155600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1111856400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1130605200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1143306000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1162054800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1174755600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1193504400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1206810000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1224954000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1238259600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1256403600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1269709200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1288458000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1301158800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1414252800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1459015200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 27232,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Choibalsan => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2032933080, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (252435600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (417974400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (433778400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (449593200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (465314400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (481042800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (496764000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (512492400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (528213600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (543942000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (559663200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (575391600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (591112800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (606841200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (622562400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (638290800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (654616800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (670345200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (686066400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (701794800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (717516000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (733244400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (748965600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (764694000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (780415200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (796143600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (811864800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (828198000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (843919200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (859647600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (875368800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (891097200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (906818400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (988390800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1001692800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1017421200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1033142400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1048870800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1064592000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1080320400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1096041600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1111770000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1127491200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1143219600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1159545600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1206889200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1427479200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1443193200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1458928800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1474642800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 27480,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Chongqing => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177481943, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-1600675200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-1585904400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-933667200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-922093200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-908870400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-888829200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-881049600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-767869200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-745833600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-733827600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-716889600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-699613200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-683884800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-670669200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-652348800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-650019600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (515527200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (527014800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (545162400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (558464400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (577216800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (589914000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (608666400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (621968400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (640116000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (653418000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (671565600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (684867600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 29143,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Chungking => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177481943, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-1600675200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-1585904400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-933667200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-922093200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-908870400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-888829200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-881049600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-767869200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-745833600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-733827600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-716889600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-699613200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-683884800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-670669200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-652348800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-650019600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (515527200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (527014800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (545162400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (558464400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (577216800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (589914000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (608666400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (621968400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (640116000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (653418000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (671565600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (684867600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 29143,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Colombo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840159964, FixedTimespan { utc_offset: 19172, dst_offset: 0, name: "MMT" }),
                    (-2019705572, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "+0530" }),
                    (-883287000, FixedTimespan { utc_offset: 19800, dst_offset: 1800, name: "+06" }),
                    (-862639200, FixedTimespan { utc_offset: 19800, dst_offset: 3600, name: "+0630" }),
                    (-764051400, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "+0530" }),
                    (832962600, FixedTimespan { utc_offset: 23400, dst_offset: 0, name: "+0630" }),
                    (846266400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1145039400, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "+0530" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 19164,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Dacca => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524543300, FixedTimespan { utc_offset: 21200, dst_offset: 0, name: "HMT" }),
                    (-891582800, FixedTimespan { utc_offset: 23400, dst_offset: 0, name: "+0630" }),
                    (-872058600, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "+0530" }),
                    (-862637400, FixedTimespan { utc_offset: 23400, dst_offset: 0, name: "+0630" }),
                    (-576138600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1245430800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1262278800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 21700,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Damascus => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577931912, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1568592000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1554080400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1537142400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1522630800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1505692800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1491181200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1474243200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1459126800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-242265600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-228877200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-210556800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-197427600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-178934400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-165718800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-147398400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-134269200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-116467200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-102646800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-84326400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-71110800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-52704000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-39488400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-21168000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-7952400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (10368000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (23583600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (41904000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (55119600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (73526400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (86742000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (105062400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (118278000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (136598400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (149814000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (168134400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (181350000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (199756800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (212972400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (231292800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (241916400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (262828800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (273452400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (418694400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (433810800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (450316800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (465433200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (508896000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (529196400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (541555200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (562633200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (574387200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (594255600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (607305600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (623199600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638928000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654649200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670456800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686264400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (702684000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717886800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733096800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (748904400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (765151200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780958800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796687200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (812494800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828309600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (844117200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859759200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (875653200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891208800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (907189200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922917600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (938725200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954540000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (970347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (986076000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1001883600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017612000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1033419600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1049148000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1064955600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080770400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1096578000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1112306400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1128114000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143842400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1158872400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1175205600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193950800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1207260000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1225486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238104800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256850000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1270159200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288299600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301608800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319749200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1333058400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351198800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364508000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382648400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1395957600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414702800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427407200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1446152400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1458856800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477602000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490911200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509051600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1522360800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540501200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553810400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1571950800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585260000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1604005200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616709600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635454800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648159200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1666904400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8712,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Dhaka => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524543300, FixedTimespan { utc_offset: 21200, dst_offset: 0, name: "HMT" }),
                    (-891582800, FixedTimespan { utc_offset: 23400, dst_offset: 0, name: "+0630" }),
                    (-872058600, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "+0530" }),
                    (-862637400, FixedTimespan { utc_offset: 23400, dst_offset: 0, name: "+0630" }),
                    (-576138600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1245430800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1262278800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 21700,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Dili => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830414140, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-879152400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (199897200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (969120000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 30140,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Dubai => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577936472, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 13272,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Dushanbe => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441168512, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (-1247547600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (354909600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (370717200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (386445600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (402253200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (417981600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (433789200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (449604000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (465336000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (481060800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (496785600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (512510400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (528235200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (543960000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (559684800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (575409600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (591134400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (606859200, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (622584000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (638308800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (654638400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (670363200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (684360000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 16512,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Famagusta => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1518920148, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (166572000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (182293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (200959200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (213829200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (228866400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (243982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (260316000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (276123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (291765600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (307486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (323820000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (338936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (354664800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (370386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (386114400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (401835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (417564000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (433285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (449013600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (465339600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (481068000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (496789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (512517600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (528238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (543967200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (559688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (575416800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (591138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (606866400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638316000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670370400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701820000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733269600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (748990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764719200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796168800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828223200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (843944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859672800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (875394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891122400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954032400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985482000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1473282000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1509228000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8148,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Gaza => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2185409872, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-933638400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-923097600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-919036800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-857347200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-844300800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-825811200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-812678400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-794188800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-779846400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-762652800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-748310400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-731116800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-399088800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-386650800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-368330400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-355114800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-336790800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-323654400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-305168400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-292032000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-273632400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-260496000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-242096400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-228960000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-210560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-197424000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-178938000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-165801600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-147402000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-134265600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-115866000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-102643200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-84330000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-81313200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (142380000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (150843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (167176800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (178664400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (334101600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (337730400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (452642400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (462319200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (482277600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (494370000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (516751200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (526424400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (545436000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (558478800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (576626400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (589323600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (609890400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (620773200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (638316000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (651618000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (669765600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (683672400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (701820000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (715726800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (733701600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (747176400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (765151200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (778021200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (796600800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (810075600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (820447200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828655200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (843170400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (860104800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (874620000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891554400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (906069600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (924213600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (939934800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (956268000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (971989200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (987717600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1003438800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1019167200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1034888400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1050616800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1066338000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1082066400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1096581600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1113516000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1128380400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143842400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1158872400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1175378400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1189638000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206655200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1219957200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238104800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1252015200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269640860, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1281474000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301608860, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1312146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1333058400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1348178400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364508000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1380229200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1395957600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414098000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427493600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445551200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1458946800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477692000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490396400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509141600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521846000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540591200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553810400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572037200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585346400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603490400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616796000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635458400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648332000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1666998000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1682726400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698447600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1713571200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729897200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1744416000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761346800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774656000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792796400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806105600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824850800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837555200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856300400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869004800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887750000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901059200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919199600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932508800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950649200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1963958400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982703600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995408000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014153200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026857600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045602800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058307200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077052400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090361600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2107897200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121811200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2138742000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153260800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2168982000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184710400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2199826800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216160000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2230066800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2234304000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234905200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248214400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2260911600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2264544000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266354800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279664000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2291756400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2295388800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297804400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311113600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2321996400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2326233600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329254000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342563200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2352841200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2356473600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361308400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374012800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2383686000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2387318400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392758000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405462400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2413926000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2418163200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424207600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437516800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2444770800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2448403200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455657200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2468966400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2475010800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2479248000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487106800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500416000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2505855600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2509488000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519161200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531865600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2536700400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2540332800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550610800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563315200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2566940400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2571177600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582060400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595369600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2597785200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2601417600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613510000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626819200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2628025200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2632262400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2644959600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658268800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2658870000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2663107200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676409200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2693347200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708463600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2724192000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2739913200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2754432000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771362800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2785276800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802812400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816121600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834262000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847571200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866316400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879020800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897766000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910470400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929215600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2941920000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960665200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2973974400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992114800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005424000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023564400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036873600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055618800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068323200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087068400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099772800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3117913200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131827200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3148758000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163276800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3179602800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194726400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3209842800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226176000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3240687600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3244320000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3244921200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257625600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3271532400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3275164800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276370800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289075200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3301772400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3306009600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307820400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321129600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3332617200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3336249600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339270000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352579200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3362857200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3367094400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370719600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384028800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3393702000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3397939200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402774000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415478400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3424546800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3428179200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434223600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3446928000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3454786800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3459024000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465673200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3478982400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3485631600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3489264000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497122800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510432000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3516476400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3520108800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528572400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541881600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3546716400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3550953600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560022000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573331200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3577561200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3581193600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592076400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604780800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3607801200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3612038400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623526000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636230400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3638646000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3642883200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3654975600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668284800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3669490800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3673123200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686425200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699734400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717874800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731184000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3749929200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762633600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781378800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794083200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812828400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825532800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844278000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857587200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875727600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889036800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907177200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920486400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939231600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3951936000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970681200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983385600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002130800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015440000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033580400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046889600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065030000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078339200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096479600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8272,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Harbin => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177481943, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-1600675200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-1585904400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-933667200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-922093200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-908870400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-888829200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-881049600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-767869200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-745833600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-733827600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-716889600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-699613200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-683884800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-670669200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-652348800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-650019600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (515527200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (527014800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (545162400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (558464400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (577216800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (589914000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (608666400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (621968400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (640116000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (653418000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (671565600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (684867600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 29143,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Hebron => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2185410023, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-933638400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-923097600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-919036800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-857347200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-844300800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-825811200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-812678400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-794188800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-779846400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-762652800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-748310400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-731116800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-399088800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-386650800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-368330400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-355114800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-336790800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-323654400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-305168400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-292032000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-273632400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-260496000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-242096400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-228960000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-210560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-197424000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-178938000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-165801600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-147402000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-134265600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-115866000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-102643200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-84330000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-81313200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (142380000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (150843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (167176800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (178664400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (334101600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (337730400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (452642400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (462319200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (482277600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (494370000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (516751200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (526424400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (545436000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (558478800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (576626400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (589323600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (609890400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (620773200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (638316000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (651618000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (669765600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (683672400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (701820000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (715726800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (733701600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (747176400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (765151200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (778021200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (796600800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (810075600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (820447200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828655200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (843170400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (860104800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (874620000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891554400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (906069600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (924213600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (939934800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (956268000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (971989200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (987717600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1003438800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1019167200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1034888400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1050616800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1066338000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1082066400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1096581600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1113516000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1128380400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143842400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1158872400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1175378400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1189638000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206655200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1220216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238104800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1252015200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269554400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1281474000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301608860, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1312146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1314655200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1317330000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1333058400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1348178400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364508000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1380229200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1395957600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414098000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427493600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445551200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1458946800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477692000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490396400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509141600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521846000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540591200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553810400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572037200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585346400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603490400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616796000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635458400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648332000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1666998000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1682726400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698447600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1713571200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729897200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1744416000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761346800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774656000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792796400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806105600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824850800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837555200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856300400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869004800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887750000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901059200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919199600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932508800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950649200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1963958400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982703600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995408000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014153200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026857600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045602800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058307200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077052400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090361600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2107897200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121811200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2138742000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153260800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2168982000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184710400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2199826800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216160000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2230066800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2234304000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234905200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248214400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2260911600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2264544000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266354800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279664000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2291756400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2295388800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297804400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311113600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2321996400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2326233600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329254000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342563200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2352841200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2356473600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361308400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374012800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2383686000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2387318400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392758000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405462400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2413926000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2418163200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424207600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437516800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2444770800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2448403200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455657200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2468966400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2475010800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2479248000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487106800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500416000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2505855600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2509488000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519161200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531865600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2536700400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2540332800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550610800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563315200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2566940400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2571177600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582060400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595369600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2597785200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2601417600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613510000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626819200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2628025200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2632262400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2644959600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658268800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2658870000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2663107200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676409200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2693347200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708463600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2724192000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2739913200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2754432000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771362800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2785276800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802812400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816121600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834262000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847571200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866316400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879020800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897766000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910470400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929215600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2941920000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960665200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2973974400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992114800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005424000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023564400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036873600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055618800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068323200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087068400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099772800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3117913200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131827200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3148758000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163276800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3179602800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194726400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3209842800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226176000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3240687600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3244320000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3244921200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257625600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3271532400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3275164800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276370800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289075200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3301772400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3306009600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307820400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321129600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3332617200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3336249600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339270000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352579200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3362857200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3367094400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370719600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384028800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3393702000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3397939200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402774000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415478400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3424546800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3428179200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434223600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3446928000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3454786800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3459024000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465673200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3478982400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3485631600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3489264000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497122800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510432000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3516476400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3520108800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528572400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541881600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3546716400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3550953600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560022000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573331200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3577561200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3581193600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592076400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604780800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3607801200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3612038400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623526000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636230400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3638646000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3642883200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3654975600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668284800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3669490800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3673123200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686425200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699734400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717874800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731184000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3749929200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762633600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781378800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794083200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812828400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825532800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844278000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857587200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875727600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889036800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907177200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920486400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939231600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3951936000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970681200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983385600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002130800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015440000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033580400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046889600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065030000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078339200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096479600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8423,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Ho_Chi_Minh => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2004073590, FixedTimespan { utc_offset: 25590, dst_offset: 0, name: "PLMT" }),
                    (-1851577590, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (-852105600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-782643600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-767869200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (-718095600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-457772400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (-315648000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (171820800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 25590,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Hong_Kong => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2056718202, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-900910800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-891579600, FixedTimespan { utc_offset: 28800, dst_offset: 1800, name: "HKWT" }),
                    (-884248200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-761209200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-747907200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-728541000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-717049800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-697091400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-683785800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-668061000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-654755400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-636611400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-623305800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-605161800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-591856200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-573712200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-559801800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-541657800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-528352200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-510211800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-498112200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-478762200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-466662600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-446707800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-435213000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-415258200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-403158600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-383808600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-371709000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-352359000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-340259400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-320909400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-308809800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-288855000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-277360200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-257405400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-245910600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-225955800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-213856200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-194506200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-182406600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-163056600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-148537800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-132816600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-117088200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-101367000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-85638600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-69312600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-53584200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-37863000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-22134600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-6413400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (9315000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (25036200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (40764600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (56485800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (72214200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (88540200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (104268600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (119989800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (126041400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (151439400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (167167800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (182889000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (198617400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (214338600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (295385400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (309292200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 27402,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Hovd => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2032927596, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (252439200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (417978000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (433785600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (449600400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (465321600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (481050000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (496771200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (512499600, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (528220800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (543949200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (559670400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (575398800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (591120000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (606848400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (622569600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (638298000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (654624000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (670352400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (686073600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (701802000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (717523200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (733251600, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (748972800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (764701200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (780422400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (796150800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (811872000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (828205200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (843926400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (859654800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (875376000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (891104400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (906825600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (988398000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1001700000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1017428400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1033149600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1048878000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1064599200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1080327600, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1096048800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1111777200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1127498400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1143226800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1159552800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1427482800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1443196800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1458932400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1474646400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 21996,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Irkutsk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840165825, FixedTimespan { utc_offset: 25025, dst_offset: 0, name: "IMT" }),
                    (-1575874625, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (-1247554800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (354902400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (370710000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (386438400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (402246000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (417974400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (433782000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (449596800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (465328800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (481053600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (496778400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (512503200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (528228000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (543952800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (559677600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (575402400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (591127200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (606852000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (622576800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (638301600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (654631200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (670356000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (686084400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (695761200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (701805600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (717530400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (733255200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (748980000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (764704800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (780429600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (796154400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (811879200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (828208800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (846352800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (859658400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (877802400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (891108000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (909252000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (922557600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (941306400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (954007200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (972756000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (985456800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1004205600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1017511200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1035655200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1048960800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1067104800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1080410400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1099159200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1111860000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1130608800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1143309600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1162058400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1174759200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1193508000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1206813600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1224957600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1238263200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1256407200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1269712800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1288461600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1301162400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1414256400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 25025,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Istanbul => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840147752, FixedTimespan { utc_offset: 7016, dst_offset: 0, name: "IMT" }),
                    (-1869875816, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1693706400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1680490800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1570413600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1552186800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1538359200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1522551600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1507514400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1490583600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1440208800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1428030000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1409709600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1396494000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-931053600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-922676400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-917834400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-892436400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-875844000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-764737200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-744343200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-733806000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-716436000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-701924400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-684986400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-670474800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-654141600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-639025200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-622087200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-606970800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-590032800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-575521200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-235620000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-194842800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-177732000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-165726000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (107910000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (121215600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (133920000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (152665200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (164678400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (184114800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (196214400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (215564400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (228873600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (245804400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (260323200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (267915600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (428454000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (433893600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (468111600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (482799600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (496710000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (512521200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (528246000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (543970800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (559695600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (575420400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (591145200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (606870000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622594800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638319600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654649200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670374000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686098800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701823600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717548400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733273200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (748998000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764118000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780447600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796172400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811897200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828226800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846370800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859676400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877820400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891126000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909270000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922575600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941324400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954025200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972774000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985474800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004223600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017529200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035673200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048978800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067122800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080428400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099177200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111878000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130626800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143327600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162076400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301266800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396220400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1446933600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1473195600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 6952,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Jakarta => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3231299232, FixedTimespan { utc_offset: 25632, dst_offset: 0, name: "BMT" }),
                    (-1451744832, FixedTimespan { utc_offset: 26400, dst_offset: 0, name: "+0720" }),
                    (-1172906400, FixedTimespan { utc_offset: 27000, dst_offset: 0, name: "+0730" }),
                    (-876641400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-766054800, FixedTimespan { utc_offset: 27000, dst_offset: 0, name: "+0730" }),
                    (-683883000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-620812800, FixedTimespan { utc_offset: 27000, dst_offset: 0, name: "+0730" }),
                    (-189415800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "WIB" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 25632,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Jayapura => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1172913768, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-799491600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "+0930" }),
                    (-189423000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "WIT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 33768,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Jerusalem => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840149254, FixedTimespan { utc_offset: 8440, dst_offset: 0, name: "JMT" }),
                    (-1641003640, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-933638400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-923097600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-919036800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-857347200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-844300800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-825811200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-812678400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-794188800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-779846400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-762652800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-748310400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-731116800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-681955200, FixedTimespan { utc_offset: 7200, dst_offset: 7200, name: "IDDT" }),
                    (-673228800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-667958400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-652320000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-636422400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-622080000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-608947200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-591840000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-572486400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-558576000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-542851200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-527731200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-514425600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-490838400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-482976000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-459388800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-451526400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-428544000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-418262400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-400118400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-387417600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (142380000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (150843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (167176800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (178664400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (334101600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (337730400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (452642400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (462319200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (482277600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (494370000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (516751200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (526424400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (545436000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (558478800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (576626400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (589323600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (609890400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (620773200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (638316000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (651618000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (669765600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (683672400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (701820000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (715726800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (733701600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (747176400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (765151200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (778021200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (796600800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (810075600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (826840800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (842821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (858895200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (874184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (890344800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (905029200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (923011200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (936313200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (955670400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (970783200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (986770800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1001282400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1017356400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1033941600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1048806000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1065132000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1081292400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1095804000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1112313600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1128812400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1143763200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1159657200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1175212800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1189897200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1206662400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1223161200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1238112000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1254006000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1269561600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1284246000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1301616000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1317510000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1333065600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1348354800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1364515200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1382828400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1395964800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1414278000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1427414400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1445727600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1458864000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1477782000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1490313600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1509231600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1521763200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1540681200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1553817600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1572130800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1585267200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1603580400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1616716800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1635634800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1648166400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1667084400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1679616000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1698534000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1711670400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1729983600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1743120000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1761433200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1774569600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1792882800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1806019200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1824937200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1837468800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1856386800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1868918400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1887836400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1900972800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1919286000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1932422400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1950735600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1963872000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1982790000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1995321600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2014239600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2026771200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2045689200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2058220800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2077138800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2090275200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2108588400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2121724800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2140038000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2153174400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2172092400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2184624000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2203542000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2216073600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2234991600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2248128000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2266441200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2279577600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2297890800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2311027200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2329340400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2342476800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2361394800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2373926400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2392844400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2405376000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2424294000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2437430400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2455743600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2468880000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2487193200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2500329600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2519247600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2531779200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2550697200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2563228800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2582146800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2595283200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2613596400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2626732800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2645046000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2658182400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2676495600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2689632000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2708550000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2721081600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2739999600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2752531200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2771449200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2784585600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2802898800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2816035200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2834348400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2847484800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2866402800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2878934400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2897852400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2910384000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2929302000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2941833600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2960751600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2973888000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2992201200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3005337600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3023650800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3036787200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3055705200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3068236800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3087154800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3099686400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3118604400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3131740800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3150054000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3163190400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3181503600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3194640000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3212953200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3226089600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3245007600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3257539200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3276457200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3288988800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3307906800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3321043200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3339356400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3352492800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3370806000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3383942400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3402860400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3415392000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3434310000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3446841600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3465759600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3478896000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3497209200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3510345600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3528658800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3541795200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3560108400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3573244800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3592162800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3604694400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3623612400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3636144000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3655062000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3668198400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3686511600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3699648000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3717961200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3731097600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3750015600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3762547200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3781465200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3793996800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3812914800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3825446400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3844364400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3857500800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3875814000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3888950400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3907263600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3920400000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3939318000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3951849600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3970767600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3983299200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (4002217200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (4015353600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (4033666800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (4046803200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (4065116400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (4078252800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (4096566000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8454,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Kabul => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524538208, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (-788932800, FixedTimespan { utc_offset: 16200, dst_offset: 0, name: "+0430" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 16608,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Kamchatka => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1487759676, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (-1247569200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (354888000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (370695600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (386424000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (402231600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (417960000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (433767600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (449582400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (465314400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (481039200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (496764000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (512488800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (528213600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (543938400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (559663200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (575388000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (591112800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (606837600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (622562400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (638287200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (654616800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (670341600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (686070000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (695746800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (701791200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (717516000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (733240800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (748965600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (764690400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (780415200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (796140000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (811864800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (828194400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (846338400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (859644000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (877788000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (891093600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (909237600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (922543200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (941292000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (953992800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (972741600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (985442400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1004191200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1017496800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1035640800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1048946400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1067090400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1080396000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1099144800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1111845600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1130594400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1143295200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1162044000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1174744800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1193493600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1206799200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1224943200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1238248800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1256392800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1269698400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1288450800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1301151600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 38076,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Karachi => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1988166492, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "+0530" }),
                    (-862637400, FixedTimespan { utc_offset: 19800, dst_offset: 3600, name: "+0630" }),
                    (-764145000, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "+0530" }),
                    (-576135000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (38775600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "PKT" }),
                    (1018119600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "PKST" }),
                    (1033840800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "PKT" }),
                    (1212260400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "PKST" }),
                    (1225476000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "PKT" }),
                    (1239735600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "PKST" }),
                    (1257012000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "PKT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 16092,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Kashgar => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1325483420, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 21020,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Kathmandu => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577943676, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "+0530" }),
                    (504901800, FixedTimespan { utc_offset: 20700, dst_offset: 0, name: "+0545" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 20476,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Katmandu => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577943676, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "+0530" }),
                    (504901800, FixedTimespan { utc_offset: 20700, dst_offset: 0, name: "+0545" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 20476,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Khandyga => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1579424533, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-1247558400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (354898800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (370706400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (386434800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (402242400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (417970800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (433778400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (449593200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (465325200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (481050000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (496774800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (512499600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (528224400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (543949200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (559674000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (575398800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (591123600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (606848400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (622573200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (638298000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (654627600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (670352400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (686080800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (695757600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (701802000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (717526800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (733251600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (748976400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (764701200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (780426000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (796150800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (811875600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (828205200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (846349200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (859654800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (877798800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (891104400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (909248400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (922554000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (941302800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (954003600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (972752400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (985453200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1004202000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1017507600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1035651600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1048957200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1067101200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1072882800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1080403200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1099152000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1111852800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1130601600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1143302400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1162051200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1174752000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1193500800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1206806400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1224950400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1238256000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1256400000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1269705600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1288454400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1301155200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1315832400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1414252800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 32533,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Kolkata => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3645237208, FixedTimespan { utc_offset: 21200, dst_offset: 0, name: "HMT" }),
                    (-3155694800, FixedTimespan { utc_offset: 19270, dst_offset: 0, name: "MMT" }),
                    (-2019705670, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "IST" }),
                    (-891581400, FixedTimespan { utc_offset: 19800, dst_offset: 3600, name: "+0630" }),
                    (-872058600, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "IST" }),
                    (-862637400, FixedTimespan { utc_offset: 19800, dst_offset: 3600, name: "+0630" }),
                    (-764145000, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "IST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 21208,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Krasnoyarsk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577513486, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (-1247551200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (354906000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (370713600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (386442000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (402249600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (417978000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (433785600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (449600400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (465332400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (481057200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (496782000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (512506800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (528231600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (543956400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (559681200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (575406000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (591130800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (606855600, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (622580400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (638305200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (654634800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (670359600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (686088000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (695764800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (701809200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (717534000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (733258800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (748983600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (764708400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (780433200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (796158000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (811882800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (828212400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (846356400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (859662000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (877806000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (891111600, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (909255600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (922561200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (941310000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (954010800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (972759600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (985460400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1004209200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1017514800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1035658800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1048964400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1067108400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1080414000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1099162800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1111863600, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1130612400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1143313200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1162062000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1174762800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1193511600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1206817200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1224961200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1238266800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1256410800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1269716400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1288465200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1301166000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1414260000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 22286,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Kuala_Lumpur => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177477725, FixedTimespan { utc_offset: 24925, dst_offset: 0, name: "SMT" }),
                    (-2038200925, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (-1167634800, FixedTimespan { utc_offset: 25200, dst_offset: 1200, name: "+0720" }),
                    (-1073028000, FixedTimespan { utc_offset: 26400, dst_offset: 0, name: "+0720" }),
                    (-894180000, FixedTimespan { utc_offset: 27000, dst_offset: 0, name: "+0730" }),
                    (-879665400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-767005200, FixedTimespan { utc_offset: 27000, dst_offset: 0, name: "+0730" }),
                    (378635400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 24925,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Kuching => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1383463280, FixedTimespan { utc_offset: 27000, dst_offset: 0, name: "+0730" }),
                    (-1167636600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-1082448000, FixedTimespan { utc_offset: 28800, dst_offset: 1200, name: "+0820" }),
                    (-1074586800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-1050825600, FixedTimespan { utc_offset: 28800, dst_offset: 1200, name: "+0820" }),
                    (-1042964400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-1019289600, FixedTimespan { utc_offset: 28800, dst_offset: 1200, name: "+0820" }),
                    (-1011428400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-987753600, FixedTimespan { utc_offset: 28800, dst_offset: 1200, name: "+0820" }),
                    (-979892400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-956217600, FixedTimespan { utc_offset: 28800, dst_offset: 1200, name: "+0820" }),
                    (-948356400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-924595200, FixedTimespan { utc_offset: 28800, dst_offset: 1200, name: "+0820" }),
                    (-916734000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-893059200, FixedTimespan { utc_offset: 28800, dst_offset: 1200, name: "+0820" }),
                    (-885198000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-879667200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-767005200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 26480,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Kuwait => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-719636812, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 11212,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Macao => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2056692850, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-884509200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-873280800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (-855918000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-841744800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (-828529200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-765363600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-747046800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-733827600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-716461200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-697021200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-683715600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-667990800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-654771600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-636627600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-623322000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-605178000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-591872400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-573642000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-559818000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-541674000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-528368400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-510224400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-498128400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-478774800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-466678800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-446720400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-435229200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-415258200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-403158600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-383808600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-371709000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-352359000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-340259400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-320909400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-308809800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-288855000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-277360200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-257405400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-245910600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-225955800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-213856200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-194506200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-182406600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-163056600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-148537800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-132820200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-117088200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-101370600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-85638600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-69312600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-53584200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-37863000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-22134600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-6413400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (9315000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (25036200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (40764600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (56485800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (72214200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (88540200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (104268600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (119989800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (126041400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (151439400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (167167800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (182889000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (198617400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (214338600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (295385400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (309292200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 27250,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Macau => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2056692850, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-884509200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-873280800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (-855918000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-841744800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (-828529200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-765363600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-747046800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-733827600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-716461200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-697021200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-683715600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-667990800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-654771600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-636627600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-623322000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-605178000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-591872400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-573642000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-559818000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-541674000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-528368400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-510224400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-498128400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-478774800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-466678800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-446720400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-435229200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-415258200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-403158600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-383808600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-371709000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-352359000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-340259400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-320909400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-308809800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-288855000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-277360200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-257405400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-245910600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-225955800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-213856200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-194506200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-182406600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-163056600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-148537800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-132820200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-117088200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-101370600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-85638600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-69312600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-53584200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-37863000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-22134600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-6413400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (9315000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (25036200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (40764600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (56485800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (72214200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (88540200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (104268600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (119989800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (126041400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (151439400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (167167800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (182889000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (198617400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (214338600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (295385400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (309292200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 27250,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Magadan => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441188192, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (-1247565600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (354891600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (370699200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (386427600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (402235200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (417963600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (433771200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (449586000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (465318000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (481042800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (496767600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (512492400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (528217200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (543942000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (559666800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (575391600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (591116400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (606841200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (622566000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (638290800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (654620400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (670345200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (686073600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (695750400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (701794800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (717519600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (733244400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (748969200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (764694000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (780418800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (796143600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (811868400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (828198000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (846342000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (859647600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (877791600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (891097200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (909241200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (922546800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (941295600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (953996400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (972745200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (985446000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1004194800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1017500400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1035644400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1048950000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1067094000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1080399600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1099148400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1111849200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1130598000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1143298800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1162047600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1174748400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1193497200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1206802800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1224946800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1238252400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1256396400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1269702000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1288450800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1301151600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1414245600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1461427200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 36192,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Makassar => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577951856, FixedTimespan { utc_offset: 28656, dst_offset: 0, name: "MMT" }),
                    (-1172908656, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-880272000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-766054800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "WITA" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 28656,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Manila => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3944621040, FixedTimespan { utc_offset: 29040, dst_offset: 0, name: "LMT" }),
                    (-2229321840, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "PST" }),
                    (-1046678400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "PDT" }),
                    (-1038733200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "PST" }),
                    (-873273600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-794221200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "PST" }),
                    (-496224000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "PDT" }),
                    (-489315600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "PST" }),
                    (259344000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "PDT" }),
                    (275151600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "PST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -57360,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Muscat => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577936472, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 13272,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Nicosia => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1518920008, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (166572000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (182293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (200959200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (213829200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (228866400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (243982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (260316000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (276123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (291765600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (307486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (323820000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (338936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (354664800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (370386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (386114400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (401835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (417564000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (433285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (449013600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (465339600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (481068000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (496789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (512517600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (528238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (543967200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (559688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (575416800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (591138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (606866400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638316000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670370400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701820000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733269600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (748990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764719200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796168800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828223200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (843944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859672800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (875394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891122400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954032400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985482000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490490000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8008,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Novokuznetsk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441259328, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (-1247551200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (354906000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (370713600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (386442000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (402249600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (417978000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (433785600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (449600400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (465332400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (481057200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (496782000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (512506800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (528231600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (543956400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (559681200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (575406000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (591130800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (606855600, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (622580400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (638305200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (654634800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (670359600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (686088000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (695764800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (701809200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (717534000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (733258800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (748983600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (764708400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (780433200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (796158000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (811882800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (828212400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (846356400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (859662000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (877806000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (891111600, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (909255600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (922561200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (941310000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (954010800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (972759600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (985460400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1004209200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1017514800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1035658800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1048964400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1067108400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1080414000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1099162800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1111863600, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1130612400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1143313200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1162062000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1174762800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1193511600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1206817200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1224961200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1238266800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1256410800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1269716400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1288468800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1301169600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 20928,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Novosibirsk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1579476700, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (-1247551200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (354906000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (370713600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (386442000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (402249600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (417978000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (433785600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (449600400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (465332400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (481057200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (496782000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (512506800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (528231600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (543956400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (559681200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (575406000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (591130800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (606855600, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (622580400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (638305200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (654634800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (670359600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (686088000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (695764800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (701809200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (717534000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (733258800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (738086400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (748987200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (764712000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (780436800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (796161600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (811886400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (828216000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (846360000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (859665600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (877809600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (891115200, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (909259200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (922564800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (941313600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (954014400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (972763200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (985464000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1004212800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1017518400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1035662400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1048968000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1067112000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1080417600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1099166400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1111867200, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1130616000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1143316800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1162065600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1174766400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1193515200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1206820800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1224964800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1238270400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1256414400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1269720000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1288468800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1301169600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1414263600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1469304000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 19900,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Omsk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1582088010, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (-1247547600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (354909600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (370717200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (386445600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (402253200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (417981600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (433789200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (449604000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (465336000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (481060800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (496785600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (512510400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (528235200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (543960000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (559684800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (575409600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (591134400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (606859200, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (622584000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (638308800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (654638400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (670363200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (686091600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (695768400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (701812800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (717537600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (733262400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (748987200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (764712000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (780436800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (796161600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (811886400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (828216000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (846360000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (859665600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (877809600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (891115200, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (909259200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (922564800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (941313600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (954014400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (972763200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (985464000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1004212800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1017518400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1035662400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1048968000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1067112000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1080417600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1099166400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1111867200, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1130616000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1143316800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1162065600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1174766400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1193515200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1206820800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1224964800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1238270400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1256414400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1269720000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1288468800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1301169600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1414263600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 17610,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Oral => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441164324, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (-1247540400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (354913200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (370720800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (386445600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (402256800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (417985200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (433792800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (449607600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (465339600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (481064400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (496789200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (512514000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (528238800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (543963600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (559688400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (575413200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (591138000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (606862800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (622591200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (638316000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (654645600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (670370400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (686095200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (695772000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (701816400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (717544800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (733269600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (748994400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (764719200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (780444000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (796168800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (811893600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (828223200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (846367200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (859672800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (877816800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (891122400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (909266400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (922572000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (941320800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (954021600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (972770400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (985471200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1004220000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1017525600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1035669600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1048975200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1067119200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1080424800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1099170000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 12324,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Phnom_Penh => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840164924, FixedTimespan { utc_offset: 24124, dst_offset: 0, name: "BMT" }),
                    (-1570084924, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 24124,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Pontianak => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1946186240, FixedTimespan { utc_offset: 26240, dst_offset: 0, name: "PMT" }),
                    (-1172906240, FixedTimespan { utc_offset: 27000, dst_offset: 0, name: "+0730" }),
                    (-881220600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-766054800, FixedTimespan { utc_offset: 27000, dst_offset: 0, name: "+0730" }),
                    (-683883000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-620812800, FixedTimespan { utc_offset: 27000, dst_offset: 0, name: "+0730" }),
                    (-189415800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "WITA" }),
                    (567964800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "WIB" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 26240,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Pyongyang => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1948782180, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-1830414600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-768646800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                    (1439564400, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (1525446000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 30180,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Qatar => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577935568, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (76190400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 12368,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Qostanay => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441167268, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (-1247544000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (354913200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (370720800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (386445600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (402256800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (417985200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (433792800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (449607600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (465339600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (481064400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (496789200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (512514000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (528238800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (543963600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (559688400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (575413200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (591138000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (606862800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (622587600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (638312400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (654642000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (670366800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (686095200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (695772000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (701816400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (717541200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (733266000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (748990800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (764715600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (780440400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (796165200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (811890000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (828219600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (846363600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (859669200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (877813200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (891118800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (909262800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (922568400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (941317200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (954018000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (972766800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (985467600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1004216400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1017522000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1035666000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1048971600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1067115600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1080421200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1099166400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1709229600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 15268,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Qyzylorda => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441167712, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (-1247544000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (354913200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (370720800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (386445600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (402256800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (417985200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (433792800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (449607600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (465339600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (481064400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (496789200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (512514000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (528238800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (543963600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (559688400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (575413200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (591138000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (606862800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (622587600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (638312400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (654642000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (670366800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (686091600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (695768400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (701812800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (717541200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (733266000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (748990800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (764715600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (780440400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (796165200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (811890000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (828219600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (846363600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (859669200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (877813200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (891118800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (909262800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (922568400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (941317200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (954018000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (972766800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (985467600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1004216400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1017522000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1035666000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1048971600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1067115600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1080421200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1099166400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1545328800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 15712,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Rangoon => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840163887, FixedTimespan { utc_offset: 23087, dst_offset: 0, name: "RMT" }),
                    (-1577946287, FixedTimespan { utc_offset: 23400, dst_offset: 0, name: "+0630" }),
                    (-873268200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-778410000, FixedTimespan { utc_offset: 23400, dst_offset: 0, name: "+0630" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 23087,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Riyadh => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-719636812, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 11212,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Saigon => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2004073590, FixedTimespan { utc_offset: 25590, dst_offset: 0, name: "PLMT" }),
                    (-1851577590, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (-852105600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-782643600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-767869200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (-718095600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-457772400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (-315648000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (171820800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 25590,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Sakhalin => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2031039048, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-768560400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (354891600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (370699200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (386427600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (402235200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (417963600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (433771200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (449586000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (465318000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (481042800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (496767600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (512492400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (528217200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (543942000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (559666800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (575391600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (591116400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (606841200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (622566000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (638290800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (654620400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (670345200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (686073600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (695750400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (701794800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (717519600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (733244400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (748969200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (764694000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (780418800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (796143600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (811868400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (828198000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (846342000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (859647600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (877795200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (891100800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (909244800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (922550400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (941299200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (954000000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (972748800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (985449600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1004198400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1017504000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1035648000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1048953600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1067097600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1080403200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1099152000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1111852800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1130601600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1143302400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1162051200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1174752000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1193500800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1206806400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1224950400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1238256000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1256400000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1269705600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1288454400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1301155200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1414249200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1459008000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 34248,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Samarkand => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441168073, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (-1247544000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (354913200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (370720800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (386445600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (402256800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (417985200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (433792800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (449607600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (465339600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (481064400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (496789200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (512514000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (528238800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (543963600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (559688400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (575413200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (591138000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (606862800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (622587600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (638312400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (654642000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (670366800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (686091600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 16073,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Seoul => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1948782472, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-1830414600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-767350800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                    (-681210000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "KDT" }),
                    (-672228000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                    (-654771600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "KDT" }),
                    (-640864800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                    (-623408400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "KDT" }),
                    (-609415200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                    (-588848400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "KDT" }),
                    (-577965600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                    (-498128400, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-462702600, FixedTimespan { utc_offset: 30600, dst_offset: 3600, name: "KDT" }),
                    (-451733400, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-429784200, FixedTimespan { utc_offset: 30600, dst_offset: 3600, name: "KDT" }),
                    (-418296600, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-399544200, FixedTimespan { utc_offset: 30600, dst_offset: 3600, name: "KDT" }),
                    (-387451800, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-368094600, FixedTimespan { utc_offset: 30600, dst_offset: 3600, name: "KDT" }),
                    (-356002200, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-336645000, FixedTimespan { utc_offset: 30600, dst_offset: 3600, name: "KDT" }),
                    (-324552600, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-305195400, FixedTimespan { utc_offset: 30600, dst_offset: 3600, name: "KDT" }),
                    (-293103000, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-264933000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                    (547578000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "KDT" }),
                    (560883600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                    (579027600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "KDT" }),
                    (592333200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 30472,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Shanghai => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177481943, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-1600675200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-1585904400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-933667200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-922093200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-908870400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-888829200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-881049600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-767869200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-745833600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-733827600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-716889600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-699613200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-683884800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-670669200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-652348800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-650019600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (515527200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (527014800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (545162400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (558464400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (577216800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (589914000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (608666400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (621968400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (640116000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (653418000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (671565600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (684867600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 29143,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Singapore => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177477725, FixedTimespan { utc_offset: 24925, dst_offset: 0, name: "SMT" }),
                    (-2038200925, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (-1167634800, FixedTimespan { utc_offset: 25200, dst_offset: 1200, name: "+0720" }),
                    (-1073028000, FixedTimespan { utc_offset: 26400, dst_offset: 0, name: "+0720" }),
                    (-894180000, FixedTimespan { utc_offset: 27000, dst_offset: 0, name: "+0730" }),
                    (-879665400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-767005200, FixedTimespan { utc_offset: 27000, dst_offset: 0, name: "+0730" }),
                    (378635400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 24925,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Srednekolymsk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441188892, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (-1247565600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (354891600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (370699200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (386427600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (402235200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (417963600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (433771200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (449586000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (465318000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (481042800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (496767600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (512492400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (528217200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (543942000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (559666800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (575391600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (591116400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (606841200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (622566000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (638290800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (654620400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (670345200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (686073600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (695750400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (701794800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (717519600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (733244400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (748969200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (764694000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (780418800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (796143600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (811868400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (828198000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (846342000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (859647600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (877791600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (891097200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (909241200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (922546800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (941295600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (953996400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (972745200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (985446000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1004194800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1017500400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1035644400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1048950000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1067094000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1080399600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1099148400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1111849200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1130598000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1143298800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1162047600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1174748400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1193497200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1206802800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1224946800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1238252400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1256396400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1269702000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1288450800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1301151600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1414245600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 36892,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Taipei => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2335248360, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-1017820800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-766224000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-745833600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-733827600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-716889600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-699613200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-683884800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-670669200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-652348800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-639133200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-620812800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-607597200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-589276800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-576061200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-562924800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-541760400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-528710400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-510224400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-497174400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-478688400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-465638400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-449830800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-434016000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-418208400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-402480000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-386672400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-370944000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-355136400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-339408000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-323600400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-302515200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-291978000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-270979200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-260442000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (133977600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (149785200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (165513600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (181321200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (299606400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (307551600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 29160,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Tashkent => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441168631, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (-1247547600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (354909600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (370717200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (386445600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (402253200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (417981600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (433789200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (449604000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (465336000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (481060800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (496785600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (512510400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (528235200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (543960000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (559684800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (575409600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (591134400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (606859200, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (622584000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (638308800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (654638400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (670363200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (686091600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 16631,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Tbilisi => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840151551, FixedTimespan { utc_offset: 10751, dst_offset: 0, name: "TBMT" }),
                    (-1441162751, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (-405140400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (354916800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (370724400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (386452800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (402260400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (417988800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (433796400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (449611200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (465343200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (481068000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (496792800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (512517600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (528242400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (543967200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (559692000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (575416800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (591141600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (606866400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (622591200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (638316000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (654645600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (670370400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (686098800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (701816400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (717537600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (733266000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (748987200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (764715600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (780436800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (796161600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (811882800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (828216000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (859662000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (877806000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (891115200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (909255600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (922564800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (941310000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (954014400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (972759600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (985464000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1004209200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1017518400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1035658800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1048968000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1067108400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1080417600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1088276400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1099177200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1111878000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 10751,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Tehran => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1704165944, FixedTimespan { utc_offset: 12344, dst_offset: 0, name: "TMT" }),
                    (-1090466744, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (227820600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (246223800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (259617600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (271108800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (283982400, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (296598600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (306531000, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (322432200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (338499000, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (673216200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (685481400, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (701209800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (717103800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (732745800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (748639800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (764281800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (780175800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (795817800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (811711800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (827353800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (843247800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (858976200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (874870200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (890512200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (906406200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (922048200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (937942200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (953584200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (969478200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (985206600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1001100600, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1016742600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1032636600, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1048278600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1064172600, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1079814600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1095708600, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1111437000, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1127331000, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1206045000, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1221939000, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1237667400, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1253561400, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1269203400, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1285097400, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1300739400, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1316633400, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1332275400, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1348169400, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1363897800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1379791800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1395433800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1411327800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1426969800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1442863800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1458505800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1474399800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1490128200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1506022200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1521664200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1537558200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1553200200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1569094200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1584736200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1600630200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1616358600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1632252600, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1647894600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1663788600, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 12344,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Tel_Aviv => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840149254, FixedTimespan { utc_offset: 8440, dst_offset: 0, name: "JMT" }),
                    (-1641003640, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-933638400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-923097600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-919036800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-857347200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-844300800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-825811200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-812678400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-794188800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-779846400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-762652800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-748310400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-731116800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-681955200, FixedTimespan { utc_offset: 7200, dst_offset: 7200, name: "IDDT" }),
                    (-673228800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-667958400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-652320000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-636422400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-622080000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-608947200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-591840000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-572486400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-558576000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-542851200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-527731200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-514425600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-490838400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-482976000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-459388800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-451526400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-428544000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-418262400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-400118400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-387417600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (142380000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (150843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (167176800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (178664400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (334101600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (337730400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (452642400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (462319200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (482277600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (494370000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (516751200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (526424400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (545436000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (558478800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (576626400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (589323600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (609890400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (620773200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (638316000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (651618000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (669765600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (683672400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (701820000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (715726800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (733701600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (747176400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (765151200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (778021200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (796600800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (810075600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (826840800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (842821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (858895200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (874184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (890344800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (905029200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (923011200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (936313200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (955670400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (970783200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (986770800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1001282400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1017356400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1033941600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1048806000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1065132000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1081292400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1095804000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1112313600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1128812400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1143763200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1159657200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1175212800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1189897200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1206662400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1223161200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1238112000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1254006000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1269561600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1284246000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1301616000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1317510000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1333065600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1348354800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1364515200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1382828400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1395964800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1414278000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1427414400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1445727600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1458864000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1477782000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1490313600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1509231600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1521763200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1540681200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1553817600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1572130800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1585267200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1603580400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1616716800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1635634800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1648166400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1667084400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1679616000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1698534000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1711670400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1729983600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1743120000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1761433200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1774569600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1792882800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1806019200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1824937200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1837468800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1856386800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1868918400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1887836400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1900972800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1919286000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1932422400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1950735600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1963872000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1982790000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1995321600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2014239600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2026771200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2045689200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2058220800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2077138800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2090275200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2108588400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2121724800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2140038000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2153174400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2172092400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2184624000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2203542000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2216073600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2234991600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2248128000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2266441200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2279577600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2297890800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2311027200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2329340400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2342476800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2361394800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2373926400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2392844400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2405376000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2424294000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2437430400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2455743600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2468880000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2487193200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2500329600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2519247600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2531779200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2550697200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2563228800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2582146800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2595283200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2613596400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2626732800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2645046000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2658182400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2676495600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2689632000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2708550000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2721081600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2739999600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2752531200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2771449200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2784585600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2802898800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2816035200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2834348400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2847484800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2866402800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2878934400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2897852400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2910384000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2929302000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2941833600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2960751600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2973888000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2992201200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3005337600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3023650800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3036787200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3055705200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3068236800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3087154800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3099686400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3118604400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3131740800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3150054000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3163190400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3181503600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3194640000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3212953200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3226089600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3245007600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3257539200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3276457200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3288988800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3307906800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3321043200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3339356400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3352492800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3370806000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3383942400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3402860400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3415392000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3434310000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3446841600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3465759600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3478896000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3497209200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3510345600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3528658800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3541795200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3560108400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3573244800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3592162800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3604694400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3623612400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3636144000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3655062000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3668198400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3686511600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3699648000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3717961200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3731097600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3750015600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3762547200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3781465200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3793996800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3812914800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3825446400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3844364400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3857500800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3875814000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3888950400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3907263600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3920400000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3939318000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3951849600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3970767600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3983299200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (4002217200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (4015353600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (4033666800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (4046803200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (4065116400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (4078252800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (4096566000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8454,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Thimbu => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-706341516, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "+0530" }),
                    (560025000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 21516,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Thimphu => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-706341516, FixedTimespan { utc_offset: 19800, dst_offset: 0, name: "+0530" }),
                    (560025000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 21516,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Tokyo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2587745939, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-683802000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "JDT" }),
                    (-672310800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-654771600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "JDT" }),
                    (-640861200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-620298000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "JDT" }),
                    (-609411600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-588848400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "JDT" }),
                    (-577962000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 33539,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Tomsk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1578807591, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (-1247551200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (354906000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (370713600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (386442000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (402249600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (417978000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (433785600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (449600400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (465332400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (481057200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (496782000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (512506800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (528231600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (543956400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (559681200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (575406000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (591130800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (606855600, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (622580400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (638305200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (654634800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (670359600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (686088000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (695764800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (701809200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (717534000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (733258800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (748983600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (764708400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (780433200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (796158000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (811882800, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (828212400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (846356400, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (859662000, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (877806000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (891111600, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (909255600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (922561200, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (941310000, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (954010800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (972759600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (985460400, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1004209200, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1017514800, FixedTimespan { utc_offset: 25200, dst_offset: 3600, name: "+08" }),
                    (1020193200, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1035662400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1048968000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1067112000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1080417600, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1099166400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1111867200, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1130616000, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1143316800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1162065600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1174766400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1193515200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1206820800, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1224964800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1238270400, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1256414400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1269720000, FixedTimespan { utc_offset: 21600, dst_offset: 3600, name: "+07" }),
                    (1288468800, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1301169600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (1414263600, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1464465600, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 20391,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Ujung_Pandang => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577951856, FixedTimespan { utc_offset: 28656, dst_offset: 0, name: "MMT" }),
                    (-1172908656, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-880272000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-766054800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "WITA" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 28656,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Ulaanbaatar => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2032931252, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (252435600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (417974400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (433782000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (449596800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (465318000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (481046400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (496767600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (512496000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (528217200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (543945600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (559666800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (575395200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (591116400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (606844800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (622566000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (638294400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (654620400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (670348800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (686070000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (701798400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (717519600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (733248000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (748969200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (764697600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (780418800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (796147200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (811868400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (828201600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (843922800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (859651200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (875372400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (891100800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (906822000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (988394400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1001696400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1017424800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1033146000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1048874400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1064595600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1080324000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1096045200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1111773600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1127494800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1143223200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1159549200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1427479200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1443193200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1458928800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1474642800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 25652,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Ulan_Bator => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2032931252, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (252435600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (417974400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (433782000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (449596800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (465318000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (481046400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (496767600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (512496000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (528217200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (543945600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (559666800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (575395200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (591116400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (606844800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (622566000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (638294400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (654620400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (670348800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (686070000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (701798400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (717519600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (733248000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (748969200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (764697600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (780418800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (796147200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (811868400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (828201600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (843922800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (859651200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (875372400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (891100800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (906822000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (988394400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1001696400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1017424800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1033146000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1048874400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1064595600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1080324000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1096045200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1111773600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1127494800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1143223200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1159549200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1427479200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1443193200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (1458928800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (1474642800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 25652,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Urumqi => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1325483420, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 21020,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__UstNera => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1579426374, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-1247558400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (354898800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (370699200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (386427600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (402235200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (417963600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (433771200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (449586000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (465318000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (481042800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (496767600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (512492400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (528217200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (543942000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (559666800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (575391600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (591116400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (606841200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (622566000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (638290800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (654620400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (670345200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (686073600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (695750400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (701794800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (717519600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (733244400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (748969200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (764694000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (780418800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (796143600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (811868400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (828198000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (846342000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (859647600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (877791600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (891097200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (909241200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (922546800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (941295600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (953996400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (972745200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (985446000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1004194800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1017500400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1035644400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1048950000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1067094000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1080399600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1099148400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1111849200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1130598000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1143298800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1162047600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1174748400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1193497200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1206802800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1224946800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1238252400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1256396400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1269702000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1288450800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1301151600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1315828800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1414249200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 34374,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Vientiane => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840164924, FixedTimespan { utc_offset: 24124, dst_offset: 0, name: "BMT" }),
                    (-1570084924, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 24124,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Vladivostok => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1487321251, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-1247562000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (354895200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (370702800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (386431200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (402238800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (417967200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (433774800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (449589600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (465321600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (481046400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (496771200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (512496000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (528220800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (543945600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (559670400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (575395200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (591120000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (606844800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (622569600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (638294400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (654624000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (670348800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (686077200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (695754000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (701798400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (717523200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (733248000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (748972800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (764697600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (780422400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (796147200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (811872000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (828201600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (846345600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (859651200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (877795200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (891100800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (909244800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (922550400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (941299200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (954000000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (972748800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (985449600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1004198400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1017504000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1035648000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1048953600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1067097600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1080403200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1099152000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1111852800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1130601600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1143302400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1162051200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1174752000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1193500800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1206806400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1224950400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1238256000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1256400000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1269705600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "+11" }),
                    (1288454400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1301155200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1414249200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 31651,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Yakutsk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1579423138, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (-1247558400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (354898800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (370706400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (386434800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (402242400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (417970800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (433778400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (449593200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (465325200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (481050000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (496774800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (512499600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (528224400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (543949200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (559674000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (575398800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (591123600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (606848400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (622573200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (638298000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (654627600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (670352400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "+09" }),
                    (686080800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                    (695757600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (701802000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (717526800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (733251600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (748976400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (764701200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (780426000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (796150800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (811875600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (828205200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (846349200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (859654800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (877798800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (891104400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (909248400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (922554000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (941302800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (954003600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (972752400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (985453200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1004202000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1017507600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1035651600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1048957200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1067101200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1080406800, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1099155600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1111856400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1130605200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1143306000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1162054800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1174755600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1193504400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1206810000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1224954000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1238259600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1256403600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1269709200, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "+10" }),
                    (1288458000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (1301158800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1414252800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 31138,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Yangon => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840163887, FixedTimespan { utc_offset: 23087, dst_offset: 0, name: "RMT" }),
                    (-1577946287, FixedTimespan { utc_offset: 23400, dst_offset: 0, name: "+0630" }),
                    (-873268200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-778410000, FixedTimespan { utc_offset: 23400, dst_offset: 0, name: "+0630" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 23087,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Yekaterinburg => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1688270553, FixedTimespan { utc_offset: 13505, dst_offset: 0, name: "PMT" }),
                    (-1592610305, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (-1247544000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (354913200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (370720800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (386449200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (402256800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (417985200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (433792800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (449607600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (465339600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (481064400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (496789200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (512514000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (528238800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (543963600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (559688400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (575413200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (591138000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (606862800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (622587600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (638312400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (654642000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (670366800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (686095200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (695772000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (701816400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (717541200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (733266000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (748990800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (764715600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (780440400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (796165200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (811890000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (828219600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (846363600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (859669200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (877813200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (891118800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (909262800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (922568400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (941317200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (954018000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (972766800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (985467600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1004216400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1017522000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1035666000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1048971600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1067115600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1080421200, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1099170000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1111870800, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1130619600, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1143320400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1162069200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1174770000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1193518800, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1206824400, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1224968400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1238274000, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1256418000, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1269723600, FixedTimespan { utc_offset: 18000, dst_offset: 3600, name: "+06" }),
                    (1288472400, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (1301173200, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                    (1414267200, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 14553,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Asia__Yerevan => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441162680, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (-405140400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (354916800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (370724400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (386452800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (402260400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (417988800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (433796400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (449611200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (465343200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (481068000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (496792800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (512517600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (528242400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (543967200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (559692000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (575416800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (591141600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (606866400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (622591200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (638316000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (654645600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (670370400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (686098800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (701823600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (717548400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (733273200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (748998000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (764722800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (780447600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (796172400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (811893600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (859672800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (877816800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (891122400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (909266400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (922572000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (941320800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (954021600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (972770400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (985471200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1004220000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1017525600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1035669600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1048975200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1067119200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1080424800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1099173600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1111874400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1130623200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1143324000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1162072800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1174773600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1193522400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1206828000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1224972000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1238277600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1256421600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1269727200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1288476000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1301176800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1319925600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 10680,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Atlantic__Azores => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713904240, FixedTimespan { utc_offset: -6872, dst_offset: 0, name: "HMT" }),
                    (-1830369928, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1689548400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1677794400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1667430000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1647730800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1635807600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1616194800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1604358000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1584658800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1572735600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1553036400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1541199600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1521500400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1442444400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1426806000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1379286000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1364770800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1348441200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1333321200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1316386800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1301266800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1284332400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1269817200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1221433200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1206918000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1191193200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1175468400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1127689200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1111964400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1096844400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1080514800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1063580400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1049065200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1033340400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1017615600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1002495600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-986166000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-969231600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-950482800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-942015600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-922662000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-906937200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-891126000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-877302000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-873680400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "+00" }),
                    (-873676800, FixedTimespan { utc_offset: -7200, dst_offset: 7200, name: "+00" }),
                    (-864007200, FixedTimespan { utc_offset: -7200, dst_offset: 7200, name: "-01" }),
                    (-864000000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-857948400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-845852400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-842835600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "+00" }),
                    (-842832000, FixedTimespan { utc_offset: -7200, dst_offset: 7200, name: "+00" }),
                    (-831348000, FixedTimespan { utc_offset: -7200, dst_offset: 7200, name: "-01" }),
                    (-831340800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-825894000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-814402800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-810781200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "+00" }),
                    (-810777600, FixedTimespan { utc_offset: -7200, dst_offset: 7200, name: "+00" }),
                    (-799898400, FixedTimespan { utc_offset: -7200, dst_offset: 7200, name: "-01" }),
                    (-799891200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-794444400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-782953200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-779331600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "+00" }),
                    (-779328000, FixedTimespan { utc_offset: -7200, dst_offset: 7200, name: "+00" }),
                    (-768448800, FixedTimespan { utc_offset: -7200, dst_offset: 7200, name: "-01" }),
                    (-768441600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-762994800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-749084400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-733359600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-717624000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-701899200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-686174400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-670449600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-654724800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-639000000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-623275200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-607550400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-591825600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-575496000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-559771200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-544046400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-528321600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-512596800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-496872000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-481147200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-465422400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-449697600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-433972800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-417643200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-401918400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-386193600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-370468800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-354744000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-339019200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-323294400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-307569600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-291844800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-276120000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-260395200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-244670400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-228340800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-212616000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-196891200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-181166400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-165441600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-149716800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-133992000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-118267200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (228272400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (243997200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (260326800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (276051600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (291776400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (307504800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (323226000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (338954400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (354679200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (370404000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (386128800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (401853600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (417582000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (433303200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (449028000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (465357600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (481082400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (496807200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (512532000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (528256800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (543981600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (559706400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (575431200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (591156000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (606880800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (622605600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (638330400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (654660000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (670384800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (686109600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (701834400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (717555600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (733280400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (749005200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (764730000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (780454800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (796179600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (811904400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (828234000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (846378000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (859683600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (877827600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (891133200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (909277200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (922582800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (941331600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (954032400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (972781200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (985482000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1004230800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1017536400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1035680400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1048986000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1067130000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1080435600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1099184400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1111885200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1130634000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1143334800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1162083600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1174784400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1193533200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1206838800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1224982800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1238288400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1256432400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1269738000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1288486800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1301187600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1319936400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1332637200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1351386000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1364691600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1382835600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1396141200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1414285200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1427590800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1445734800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1459040400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1477789200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1490490000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1509238800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1521939600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1540688400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1553994000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1572138000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1585443600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1603587600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1616893200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1635642000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1648342800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1667091600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1679792400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1698541200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1711846800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1729990800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1743296400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1761440400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1774746000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1792890000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1806195600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1824944400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1837645200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1856394000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1869094800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1887843600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1901149200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1919293200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1932598800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1950742800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1964048400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (1982797200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (1995498000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2014246800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2026947600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2045696400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2058397200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2077146000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2090451600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2108595600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2121901200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2140045200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2153350800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2172099600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2184800400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2203549200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2216250000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2234998800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2248304400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2266448400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2279754000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2297898000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2311203600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2329347600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2342653200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2361402000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2374102800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2392851600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2405552400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2424301200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2437606800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2455750800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2469056400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2487200400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2500506000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2519254800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2531955600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2550704400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2563405200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2582154000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2595459600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2613603600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2626909200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2645053200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2658358800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2676502800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2689808400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2708557200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2721258000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2740006800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2752707600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2771456400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2784762000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2802906000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2816211600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2834355600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2847661200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2866410000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2879110800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2897859600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2910560400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2929309200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2942010000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2960758800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (2974064400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (2992208400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3005514000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3023658000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3036963600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3055712400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3068413200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3087162000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3099862800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3118611600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3131917200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3150061200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3163366800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3181510800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3194816400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3212960400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3226266000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3245014800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3257715600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3276464400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3289165200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3307914000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3321219600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3339363600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3352669200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3370813200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3384118800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3402867600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3415568400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3434317200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3447018000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3465766800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3479072400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3497216400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3510522000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3528666000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3541971600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3560115600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3573421200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3592170000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3604870800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3623619600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3636320400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3655069200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3668374800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3686518800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3699824400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3717968400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3731274000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3750022800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3762723600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3781472400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3794173200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3812922000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3825622800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3844371600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3857677200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3875821200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3889126800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3907270800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3920576400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3939325200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3952026000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (3970774800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (3983475600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (4002224400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (4015530000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (4033674000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (4046979600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (4065123600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (4078429200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (4096573200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -6160,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Atlantic__Bermuda => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524506042, FixedTimespan { utc_offset: -15558, dst_offset: 0, name: "BMT" }),
                    (-1664307642, FixedTimespan { utc_offset: -15558, dst_offset: 3600, name: "BST" }),
                    (-1648932042, FixedTimespan { utc_offset: -15558, dst_offset: 0, name: "BMT" }),
                    (-1632080442, FixedTimespan { utc_offset: -15558, dst_offset: 3600, name: "BST" }),
                    (-1618692042, FixedTimespan { utc_offset: -15558, dst_offset: 0, name: "BMT" }),
                    (-1262281242, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-882727200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-858538800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-845229600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-825879600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-814384800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-793825200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-782935200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-762375600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-713988000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-703710000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-681933600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-672865200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-650484000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-641415600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-618429600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-609966000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-586980000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-578516400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-555530400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-546462000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-429127200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-415825200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (136360800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (152082000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (167810400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (183531600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (199260000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (215586000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (230709600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (247035600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (262764000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (278485200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (294213600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (309934800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (325663200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (341384400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (357112800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (372834000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (388562400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (404888400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (420012000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (436338000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (452066400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (467787600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (483516000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (499237200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (514965600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (530686800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (544600800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (562136400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (576050400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (594190800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (607500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (625640400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (638949600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (657090000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (671004000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (688539600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (702453600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (719989200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (733903200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (752043600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (765352800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (783493200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (796802400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (814942800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (828856800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (846392400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (860306400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (877842000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (891756000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (909291600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (923205600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (941346000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (954655200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (972795600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (986104800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1004245200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1018159200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1035694800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1049608800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1067144400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1081058400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1099198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1112508000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1130648400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1143957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1162098000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1173592800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1194152400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1205042400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1225602000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1236492000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1257051600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1268546400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1289106000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1299996000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1320555600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1331445600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1352005200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1362895200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1383454800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1394344800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1414904400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1425794400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1446354000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1457848800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1478408400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1489298400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1509858000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1520748000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1541307600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1552197600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1572757200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1583647200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1604206800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1615701600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1636261200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1647151200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1667710800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1678600800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1699160400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1710050400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1730610000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1741500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1762059600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1772949600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1793509200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1805004000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1825563600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1836453600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1857013200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1867903200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1888462800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1899352800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1919912400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1930802400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1951362000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1962856800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1983416400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1994306400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2014866000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2025756000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2046315600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2057205600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2077765200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2088655200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2109214800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2120104800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2140664400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2152159200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2172718800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2183608800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2204168400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2215058400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2235618000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2246508000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2267067600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2277957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2298517200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2309407200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2329966800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2341461600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2362021200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2372911200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2393470800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2404360800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2424920400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2435810400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2456370000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2467260000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2487819600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2499314400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2519874000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2530764000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2551323600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2562213600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2582773200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2593663200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2614222800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2625112800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2645672400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2656562400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2677122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2688616800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2709176400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2720066400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2740626000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2751516000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2772075600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2782965600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2803525200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2814415200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2834974800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2846469600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2867029200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2877919200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2898478800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2909368800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2929928400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2940818400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2961378000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2972268000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2992827600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3003717600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3024277200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3035772000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3056331600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3067221600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3087781200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3098671200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3119230800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3130120800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3150680400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3161570400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3182130000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3193020000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3213579600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3225074400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3245634000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3256524000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3277083600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3287973600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3308533200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3319423200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3339982800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3350872800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3371432400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3382927200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3403486800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3414376800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3434936400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3445826400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3466386000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3477276000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3497835600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3508725600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3529285200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3540175200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3560734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3572229600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3592789200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3603679200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3624238800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3635128800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3655688400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3666578400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3687138000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3698028000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3718587600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3730082400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3750642000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3761532000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3782091600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3792981600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3813541200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3824431200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3844990800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3855880800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3876440400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3887330400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3907890000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3919384800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3939944400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3950834400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3971394000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3982284000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4002843600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4013733600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4034293200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4045183200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4065742800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4076632800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4097192400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15558,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Atlantic__Canary => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1509663504, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-733874400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (323827200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (338950800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (354675600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (370400400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (386125200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (401850000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (417574800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (433299600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (449024400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (465354000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (481078800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (496803600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (512528400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (528253200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (543978000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (559702800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (575427600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (591152400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (606877200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (622602000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (638326800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (654656400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (670381200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (686106000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (701830800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (717555600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (733280400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (749005200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (764730000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (780454800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (796179600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (811904400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (828234000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (846378000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (859683600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (877827600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (891133200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (909277200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (922582800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (941331600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (954032400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (972781200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (985482000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1004230800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1017536400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1035680400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1048986000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1067130000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1080435600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1099184400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1111885200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1130634000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1143334800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1162083600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1174784400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1193533200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1206838800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1224982800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1238288400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1256432400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1269738000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1288486800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1301187600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1319936400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1332637200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1351386000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1364691600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1382835600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1396141200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1414285200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1427590800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1445734800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1459040400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1477789200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1490490000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1509238800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1521939600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1540688400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1553994000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1572138000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1585443600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1603587600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1616893200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1635642000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1648342800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1667091600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1679792400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1698541200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1711846800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1729990800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1743296400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1761440400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1774746000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1792890000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1806195600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1824944400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1837645200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1856394000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1869094800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1887843600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1901149200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1919293200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1932598800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1950742800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1964048400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1982797200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1995498000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2014246800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2026947600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2045696400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2058397200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2077146000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2090451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2108595600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2121901200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2140045200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2153350800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2172099600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2184800400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2203549200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2216250000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2234998800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2248304400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2266448400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2279754000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2297898000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2311203600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2329347600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2342653200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2361402000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2374102800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2392851600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2405552400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2424301200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2437606800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2455750800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2469056400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2487200400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2500506000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2519254800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2531955600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2550704400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2563405200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2582154000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2595459600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2613603600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2626909200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2645053200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2658358800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2676502800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2689808400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2708557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2721258000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2740006800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2752707600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2771456400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2784762000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2802906000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2816211600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2834355600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2847661200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2866410000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2879110800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2897859600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2910560400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2929309200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2942010000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2960758800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2974064400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2992208400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3005514000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3023658000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3036963600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3055712400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3068413200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3087162000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3099862800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3118611600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3131917200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3150061200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3163366800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3181510800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3194816400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3212960400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3226266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3245014800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3257715600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3276464400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3289165200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3307914000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3321219600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3339363600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3352669200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3370813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3384118800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3402867600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3415568400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3434317200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3447018000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3465766800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3479072400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3497216400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3510522000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3528666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3541971600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3560115600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3573421200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3592170000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3604870800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3623619600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3636320400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3655069200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3668374800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3686518800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3699824400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3717968400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3731274000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3750022800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3762723600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3781472400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3794173200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3812922000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3825622800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3844371600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3857677200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3875821200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3889126800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3907270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3920576400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3939325200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3952026000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3970774800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3983475600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4002224400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4015530000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4033674000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4046979600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4065123600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4078429200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4096573200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -3696,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Atlantic__Cape_Verde => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830371156, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-862610400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-764118000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (186120000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -5644,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Atlantic__Faeroe => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1955748776, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (354675600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (370400400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (386125200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (401850000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (417574800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (433299600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (449024400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (465354000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (481078800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (496803600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (512528400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (528253200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (543978000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (559702800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (575427600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (591152400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (606877200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (622602000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (638326800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (654656400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (670381200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (686106000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (701830800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (717555600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (733280400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (749005200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (764730000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (780454800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (796179600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (811904400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (828234000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (846378000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (859683600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (877827600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (891133200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (909277200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (922582800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (941331600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (954032400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (972781200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (985482000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1004230800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1017536400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1035680400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1048986000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1067130000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1080435600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1099184400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1111885200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1130634000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1143334800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1162083600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1174784400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1193533200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1206838800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1224982800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1238288400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1256432400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1269738000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1288486800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1301187600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1319936400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1332637200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1351386000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1364691600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1382835600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1396141200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1414285200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1427590800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1445734800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1459040400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1477789200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1490490000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1509238800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1521939600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1540688400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1553994000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1572138000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1585443600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1603587600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1616893200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1635642000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1648342800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1667091600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1679792400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1698541200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1711846800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1729990800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1743296400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1761440400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1774746000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1792890000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1806195600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1824944400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1837645200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1856394000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1869094800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1887843600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1901149200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1919293200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1932598800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1950742800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1964048400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1982797200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1995498000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2014246800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2026947600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2045696400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2058397200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2077146000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2090451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2108595600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2121901200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2140045200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2153350800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2172099600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2184800400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2203549200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2216250000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2234998800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2248304400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2266448400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2279754000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2297898000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2311203600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2329347600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2342653200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2361402000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2374102800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2392851600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2405552400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2424301200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2437606800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2455750800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2469056400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2487200400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2500506000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2519254800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2531955600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2550704400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2563405200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2582154000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2595459600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2613603600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2626909200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2645053200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2658358800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2676502800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2689808400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2708557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2721258000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2740006800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2752707600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2771456400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2784762000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2802906000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2816211600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2834355600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2847661200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2866410000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2879110800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2897859600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2910560400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2929309200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2942010000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2960758800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2974064400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2992208400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3005514000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3023658000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3036963600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3055712400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3068413200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3087162000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3099862800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3118611600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3131917200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3150061200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3163366800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3181510800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3194816400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3212960400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3226266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3245014800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3257715600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3276464400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3289165200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3307914000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3321219600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3339363600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3352669200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3370813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3384118800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3402867600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3415568400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3434317200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3447018000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3465766800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3479072400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3497216400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3510522000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3528666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3541971600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3560115600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3573421200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3592170000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3604870800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3623619600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3636320400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3655069200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3668374800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3686518800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3699824400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3717968400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3731274000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3750022800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3762723600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3781472400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3794173200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3812922000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3825622800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3844371600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3857677200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3875821200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3889126800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3907270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3920576400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3939325200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3952026000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3970774800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3983475600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4002224400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4015530000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4033674000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4046979600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4065123600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4078429200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4096573200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -1624,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Atlantic__Faroe => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1955748776, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (354675600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (370400400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (386125200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (401850000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (417574800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (433299600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (449024400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (465354000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (481078800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (496803600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (512528400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (528253200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (543978000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (559702800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (575427600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (591152400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (606877200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (622602000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (638326800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (654656400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (670381200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (686106000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (701830800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (717555600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (733280400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (749005200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (764730000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (780454800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (796179600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (811904400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (828234000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (846378000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (859683600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (877827600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (891133200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (909277200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (922582800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (941331600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (954032400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (972781200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (985482000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1004230800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1017536400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1035680400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1048986000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1067130000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1080435600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1099184400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1111885200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1130634000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1143334800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1162083600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1174784400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1193533200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1206838800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1224982800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1238288400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1256432400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1269738000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1288486800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1301187600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1319936400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1332637200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1351386000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1364691600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1382835600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1396141200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1414285200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1427590800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1445734800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1459040400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1477789200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1490490000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1509238800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1521939600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1540688400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1553994000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1572138000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1585443600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1603587600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1616893200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1635642000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1648342800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1667091600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1679792400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1698541200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1711846800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1729990800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1743296400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1761440400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1774746000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1792890000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1806195600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1824944400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1837645200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1856394000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1869094800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1887843600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1901149200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1919293200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1932598800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1950742800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1964048400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1982797200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1995498000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2014246800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2026947600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2045696400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2058397200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2077146000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2090451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2108595600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2121901200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2140045200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2153350800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2172099600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2184800400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2203549200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2216250000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2234998800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2248304400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2266448400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2279754000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2297898000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2311203600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2329347600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2342653200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2361402000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2374102800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2392851600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2405552400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2424301200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2437606800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2455750800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2469056400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2487200400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2500506000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2519254800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2531955600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2550704400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2563405200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2582154000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2595459600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2613603600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2626909200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2645053200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2658358800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2676502800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2689808400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2708557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2721258000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2740006800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2752707600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2771456400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2784762000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2802906000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2816211600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2834355600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2847661200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2866410000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2879110800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2897859600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2910560400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2929309200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2942010000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2960758800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2974064400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2992208400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3005514000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3023658000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3036963600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3055712400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3068413200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3087162000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3099862800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3118611600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3131917200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3150061200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3163366800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3181510800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3194816400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3212960400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3226266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3245014800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3257715600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3276464400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3289165200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3307914000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3321219600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3339363600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3352669200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3370813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3384118800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3402867600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3415568400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3434317200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3447018000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3465766800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3479072400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3497216400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3510522000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3528666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3541971600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3560115600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3573421200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3592170000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3604870800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3623619600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3636320400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3655069200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3668374800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3686518800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3699824400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3717968400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3731274000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3750022800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3762723600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3781472400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3794173200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3812922000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3825622800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3844371600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3857677200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3875821200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3889126800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3907270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3920576400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3939325200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3952026000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3970774800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3983475600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4002224400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4015530000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4033674000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4046979600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4065123600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4078429200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4096573200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -1624,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Atlantic__Jan_Mayen => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2422054408, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693706400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-938905200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-776563200, FixedTimespan { utc_offset: 3600, dst_offset: 7200, name: "CEMT" }),
                    (-765936000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-761180400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-748479600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733273200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-717631200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-714610800, FixedTimespan { utc_offset: 3600, dst_offset: 7200, name: "CEMT" }),
                    (-710380800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701910000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-684975600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-654130800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-639010800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3208,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Atlantic__Madeira => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713906344, FixedTimespan { utc_offset: -4056, dst_offset: 0, name: "FMT" }),
                    (-1830376344, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1689552000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1677798000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1667433600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1647734400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1635811200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1616198400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1604361600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1584662400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1572739200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1553040000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1541203200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1521504000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1442448000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1426809600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1379289600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1364774400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1348444800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1333324800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1316390400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1301270400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1284336000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1269820800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1221436800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1206921600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1191196800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1175472000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1127692800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1111968000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1096848000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1080518400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1063584000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1049068800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1033344000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-1017619200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-1002499200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-986169600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-969235200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-950486400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-942019200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-922665600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-906940800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-891129600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-877305600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-873684000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+01" }),
                    (-873680400, FixedTimespan { utc_offset: -3600, dst_offset: 7200, name: "+01" }),
                    (-864010800, FixedTimespan { utc_offset: -3600, dst_offset: 7200, name: "+00" }),
                    (-864003600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-857952000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-845856000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-842839200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+01" }),
                    (-842835600, FixedTimespan { utc_offset: -3600, dst_offset: 7200, name: "+01" }),
                    (-831351600, FixedTimespan { utc_offset: -3600, dst_offset: 7200, name: "+00" }),
                    (-831344400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-825897600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-814406400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-810784800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+01" }),
                    (-810781200, FixedTimespan { utc_offset: -3600, dst_offset: 7200, name: "+01" }),
                    (-799902000, FixedTimespan { utc_offset: -3600, dst_offset: 7200, name: "+00" }),
                    (-799894800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-794448000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-782956800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-779335200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+01" }),
                    (-779331600, FixedTimespan { utc_offset: -3600, dst_offset: 7200, name: "+01" }),
                    (-768452400, FixedTimespan { utc_offset: -3600, dst_offset: 7200, name: "+00" }),
                    (-768445200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-762998400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-749088000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-733363200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-717627600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-701902800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-686178000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-670453200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-654728400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-639003600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-623278800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-607554000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-591829200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-575499600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-559774800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-544050000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-528325200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-512600400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-496875600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-481150800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-465426000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-449701200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-433976400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-417646800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-401922000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-386197200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-370472400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-354747600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-339022800, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-323298000, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-307573200, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-291848400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-276123600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-260398800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-244674000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-228344400, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-212619600, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-196894800, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-181170000, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-165445200, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-149720400, FixedTimespan { utc_offset: -3600, dst_offset: 3600, name: "+00" }),
                    (-133995600, FixedTimespan { utc_offset: -3600, dst_offset: 0, name: "-01" }),
                    (-118270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (228268800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (243993600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (260323200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (276048000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (291772800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (307501200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (323222400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (338950800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (354675600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (370400400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (386125200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (401850000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (417578400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (433299600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (449024400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (465354000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (481078800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (496803600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (512528400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (528253200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (543978000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (559702800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (575427600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (591152400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (606877200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (622602000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (638326800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (654656400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (670381200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (686106000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (701830800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (717555600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (733280400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (749005200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (764730000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (780454800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (796179600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (811904400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (828234000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (846378000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (859683600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (877827600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (891133200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (909277200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (922582800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (941331600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (954032400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (972781200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (985482000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1004230800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1017536400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1035680400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1048986000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1067130000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1080435600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1099184400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1111885200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1130634000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1143334800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1162083600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1174784400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1193533200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1206838800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1224982800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1238288400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1256432400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1269738000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1288486800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1301187600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1319936400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1332637200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1351386000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1364691600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1382835600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1396141200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1414285200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1427590800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1445734800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1459040400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1477789200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1490490000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1509238800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1521939600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1540688400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1553994000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1572138000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1585443600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1603587600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1616893200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1635642000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1648342800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1667091600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1679792400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1698541200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1711846800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1729990800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1743296400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1761440400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1774746000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1792890000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1806195600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1824944400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1837645200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1856394000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1869094800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1887843600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1901149200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1919293200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1932598800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1950742800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1964048400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1982797200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1995498000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2014246800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2026947600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2045696400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2058397200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2077146000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2090451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2108595600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2121901200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2140045200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2153350800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2172099600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2184800400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2203549200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2216250000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2234998800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2248304400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2266448400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2279754000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2297898000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2311203600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2329347600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2342653200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2361402000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2374102800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2392851600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2405552400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2424301200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2437606800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2455750800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2469056400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2487200400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2500506000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2519254800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2531955600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2550704400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2563405200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2582154000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2595459600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2613603600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2626909200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2645053200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2658358800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2676502800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2689808400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2708557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2721258000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2740006800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2752707600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2771456400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2784762000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2802906000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2816211600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2834355600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2847661200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2866410000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2879110800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2897859600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2910560400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2929309200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2942010000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2960758800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2974064400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2992208400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3005514000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3023658000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3036963600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3055712400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3068413200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3087162000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3099862800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3118611600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3131917200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3150061200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3163366800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3181510800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3194816400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3212960400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3226266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3245014800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3257715600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3276464400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3289165200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3307914000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3321219600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3339363600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3352669200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3370813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3384118800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3402867600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3415568400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3434317200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3447018000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3465766800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3479072400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3497216400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3510522000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3528666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3541971600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3560115600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3573421200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3592170000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3604870800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3623619600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3636320400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3655069200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3668374800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3686518800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3699824400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3717968400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3731274000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3750022800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3762723600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3781472400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3794173200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3812922000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3825622800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3844371600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3857677200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3875821200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3889126800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3907270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3920576400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3939325200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3952026000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3970774800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3983475600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4002224400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4015530000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4033674000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4046979600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4065123600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4078429200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4096573200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -4056,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Atlantic__Reykjavik => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830383032, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -968,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Atlantic__South_Georgia => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524512832, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -8768,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Atlantic__St_Helena => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830383032, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -968,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Atlantic__Stanley => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524507716, FixedTimespan { utc_offset: -13884, dst_offset: 0, name: "SMT" }),
                    (-1824235716, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1018209600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1003093200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-986760000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-971643600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-954705600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-939589200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-923256000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-908139600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-891806400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-876690000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-860356800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-852066000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (420609600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (433306800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (452052000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (464151600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (483501600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (495601200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (514350000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (527054400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (545799600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (558504000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (577249200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (589953600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (608698800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (621403200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (640753200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (652852800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (672202800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (684907200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (703652400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (716356800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (735102000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (747806400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (766551600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (779256000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (798001200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (810705600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (830055600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (842760000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (861505200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (874209600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (892954800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (905659200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (924404400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (937108800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (955854000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (968558400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (987310800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (999410400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1019365200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1030860000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1050814800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1062914400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1082264400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1094364000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1113714000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1125813600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1145163600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1157263200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1176613200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1188712800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1208667600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1220767200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1240117200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1252216800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1271566800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1283666400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -13884,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__ACT => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2364113092, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1672560000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1665388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-883641600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-876124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-860400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-844675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-828345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-813225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (57686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (67968000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (89136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (100022400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (120585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (131472000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (152035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (162921600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (183484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (194976000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (215539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (226425600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (246988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (257875200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (278438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (289324800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (309888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (320774400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (341337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (352224000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (372787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (386697600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (404841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (415728000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (436291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (447177600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (467740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (478627200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (499190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (511286400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (530035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (542736000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (562089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (574790400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (594144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (606240000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (625593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (636480000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (657043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (667929600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (688492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (699379200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (719942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (731433600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (751996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (762883200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (783446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (794332800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (814896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (828201600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (846345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (859651200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (877795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (891100800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (909244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (922550400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (941299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (954000000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (967305600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (985449600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1004198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1017504000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1035648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1048953600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1067097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1080403200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1099152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1111852800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1130601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1143907200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1162051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1174752000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1193500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1207411200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1223136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1238860800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1254585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1270310400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1286035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1301760000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1317484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1333209600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1349539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1365264000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1380988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1396713600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1412438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1428163200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1443888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1459612800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1475337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1491062400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1506787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1522512000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1538841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1554566400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1570291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1586016000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1601740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1617465600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1633190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1648915200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1664640000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1680364800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1696089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1712419200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1728144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1743868800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1759593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1775318400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1791043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1806768000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1822492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1838217600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1853942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1869667200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1885996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1901721600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1917446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1933171200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1948896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1964620800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1980345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1996070400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2011795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2027520000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2043244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2058969600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2075299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2091024000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2106748800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2122473600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2138198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2153923200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2169648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2185372800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2201097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2216822400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2233152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2248876800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2264601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2280326400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2296051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2311776000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2327500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2343225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2358950400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2374675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2390400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2406124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2422454400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2438179200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2453904000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2469628800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2485353600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2501078400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2516803200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2532528000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2548252800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2563977600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2579702400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2596032000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2611756800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2627481600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2643206400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2658931200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2674656000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2690380800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2706105600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2721830400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2737555200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2753280000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2769609600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2785334400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2801059200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2816784000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2832508800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2848233600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2863958400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2879683200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2895408000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2911132800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2926857600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2942582400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2958912000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2974636800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2990361600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3006086400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3021811200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3037536000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3053260800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3068985600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3084710400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3100435200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3116764800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3132489600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3148214400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3163939200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3179664000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3195388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3211113600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3226838400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3242563200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3258288000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3274012800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3289737600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3306067200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3321792000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3337516800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3353241600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3368966400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3384691200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3400416000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3416140800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3431865600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3447590400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3463315200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3479644800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3495369600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3511094400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3526819200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3542544000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3558268800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3573993600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3589718400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3605443200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3621168000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3636892800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3653222400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3668947200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3684672000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3700396800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3716121600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3731846400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3747571200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3763296000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3779020800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3794745600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3810470400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3826195200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3842524800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3858249600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3873974400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3889699200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3905424000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3921148800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3936873600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3952598400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3968323200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3984048000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4000377600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4016102400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4031827200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4047552000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4063276800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4079001600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4094726400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 36292,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Adelaide => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2364110060, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "ACST" }),
                    (-2230189200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-1672558200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-1665387000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-883639800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-876123000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-860398200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-844673400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-828343800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-813223800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (57688200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (67969800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (89137800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (100024200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (120587400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (131473800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (152037000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (162923400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (183486600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (194977800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (215541000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (226427400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (246990600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (257877000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (278440200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (289326600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (309889800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (320776200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (341339400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (352225800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (372789000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (384280200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (404843400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (415729800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (436293000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (447179400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (467742600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (478629000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (499192200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (511288200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (530037000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (542737800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (562091400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (574792200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (594145800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (606241800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (625595400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (637691400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (657045000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (667931400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (688494600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (701195400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (719944200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (731435400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (751998600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (764094600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (783448200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (796149000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (814897800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (828203400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (846347400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (859653000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (877797000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (891102600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (909246600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (922552200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (941301000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (954001800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (972750600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (985451400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1004200200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1017505800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1035649800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1048955400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1067099400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1080405000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1099153800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1111854600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1130603400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1143909000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1162053000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1174753800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1193502600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1207413000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1223137800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1238862600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1254587400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1270312200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1286037000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1301761800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1317486600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1333211400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1349541000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1365265800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1380990600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1396715400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1412440200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1428165000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1443889800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1459614600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1475339400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1491064200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1506789000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1522513800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1538843400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1554568200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1570293000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1586017800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1601742600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1617467400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1633192200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1648917000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1664641800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1680366600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1696091400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1712421000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1728145800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1743870600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1759595400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1775320200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1791045000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1806769800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1822494600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1838219400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1853944200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1869669000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1885998600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1901723400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1917448200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1933173000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1948897800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1964622600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1980347400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1996072200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2011797000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2027521800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2043246600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2058971400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2075301000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2091025800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2106750600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2122475400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2138200200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2153925000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2169649800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2185374600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2201099400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2216824200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2233153800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2248878600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2264603400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2280328200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2296053000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2311777800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2327502600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2343227400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2358952200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2374677000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2390401800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2406126600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2422456200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2438181000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2453905800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2469630600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2485355400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2501080200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2516805000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2532529800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2548254600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2563979400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2579704200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2596033800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2611758600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2627483400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2643208200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2658933000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2674657800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2690382600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2706107400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2721832200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2737557000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2753281800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2769611400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2785336200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2801061000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2816785800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2832510600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2848235400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2863960200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2879685000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2895409800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2911134600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2926859400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2942584200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2958913800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2974638600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2990363400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3006088200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3021813000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3037537800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3053262600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3068987400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3084712200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3100437000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3116766600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3132491400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3148216200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3163941000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3179665800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3195390600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3211115400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3226840200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3242565000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3258289800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3274014600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3289739400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3306069000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3321793800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3337518600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3353243400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3368968200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3384693000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3400417800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3416142600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3431867400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3447592200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3463317000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3479646600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3495371400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3511096200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3526821000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3542545800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3558270600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3573995400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3589720200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3605445000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3621169800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3636894600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3653224200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3668949000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3684673800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3700398600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3716123400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3731848200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3747573000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3763297800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3779022600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3794747400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3810472200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3826197000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3842526600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3858251400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3873976200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3889701000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3905425800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3921150600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3936875400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3952600200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3968325000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3984049800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4000379400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (4016104200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4031829000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (4047553800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4063278600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (4079003400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4094728200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 33260,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Brisbane => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2366791928, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1672560000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1665388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-883641600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-876124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-860400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-844675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-828345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-813225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (57686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (67968000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (625593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (636480000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (657043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (667929600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (688492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (699379200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 36728,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Broken_Hill => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2364110748, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-2314951200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "ACST" }),
                    (-2230189200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-1672558200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-1665387000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-883639800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-876123000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-860398200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-844673400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-828343800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-813223800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (57688200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (67969800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (89137800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (100024200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (120587400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (131473800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (152037000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (162923400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (183486600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (194977800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (215541000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (226427400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (246990600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (257877000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (278440200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (289326600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (309889800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (320776200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (341339400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (352225800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (372789000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (386699400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (404843400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (415729800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (436293000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (447179400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (467742600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (478629000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (499192200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (511288200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (530037000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (542737800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (562091400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (574792200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (594145800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (606241800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (625595400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (636481800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (657045000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (667931400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (688494600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (699381000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (719944200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (731435400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (751998600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (762885000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (783448200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (794334600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (814897800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (828203400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (846347400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (859653000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (877797000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (891102600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (909246600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (922552200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (941301000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (954001800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (972750600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (985451400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1004200200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1017505800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1035649800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1048955400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1067099400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1080405000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1099153800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1111854600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1130603400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1143909000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1162053000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1174753800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1193502600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1207413000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1223137800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1238862600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1254587400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1270312200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1286037000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1301761800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1317486600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1333211400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1349541000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1365265800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1380990600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1396715400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1412440200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1428165000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1443889800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1459614600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1475339400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1491064200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1506789000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1522513800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1538843400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1554568200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1570293000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1586017800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1601742600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1617467400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1633192200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1648917000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1664641800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1680366600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1696091400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1712421000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1728145800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1743870600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1759595400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1775320200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1791045000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1806769800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1822494600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1838219400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1853944200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1869669000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1885998600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1901723400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1917448200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1933173000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1948897800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1964622600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1980347400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1996072200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2011797000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2027521800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2043246600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2058971400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2075301000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2091025800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2106750600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2122475400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2138200200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2153925000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2169649800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2185374600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2201099400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2216824200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2233153800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2248878600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2264603400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2280328200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2296053000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2311777800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2327502600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2343227400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2358952200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2374677000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2390401800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2406126600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2422456200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2438181000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2453905800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2469630600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2485355400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2501080200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2516805000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2532529800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2548254600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2563979400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2579704200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2596033800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2611758600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2627483400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2643208200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2658933000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2674657800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2690382600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2706107400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2721832200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2737557000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2753281800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2769611400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2785336200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2801061000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2816785800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2832510600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2848235400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2863960200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2879685000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2895409800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2911134600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2926859400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2942584200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2958913800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2974638600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2990363400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3006088200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3021813000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3037537800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3053262600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3068987400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3084712200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3100437000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3116766600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3132491400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3148216200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3163941000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3179665800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3195390600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3211115400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3226840200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3242565000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3258289800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3274014600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3289739400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3306069000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3321793800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3337518600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3353243400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3368968200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3384693000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3400417800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3416142600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3431867400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3447592200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3463317000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3479646600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3495371400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3511096200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3526821000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3542545800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3558270600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3573995400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3589720200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3605445000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3621169800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3636894600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3653224200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3668949000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3684673800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3700398600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3716123400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3731848200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3747573000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3763297800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3779022600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3794747400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3810472200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3826197000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3842526600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3858251400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3873976200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3889701000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3905425800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3921150600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3936875400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3952600200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3968325000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3984049800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4000379400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (4016104200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4031829000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (4047553800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4063278600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (4079003400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4094728200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 33948,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Canberra => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2364113092, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1672560000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1665388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-883641600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-876124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-860400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-844675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-828345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-813225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (57686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (67968000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (89136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (100022400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (120585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (131472000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (152035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (162921600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (183484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (194976000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (215539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (226425600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (246988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (257875200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (278438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (289324800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (309888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (320774400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (341337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (352224000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (372787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (386697600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (404841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (415728000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (436291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (447177600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (467740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (478627200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (499190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (511286400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (530035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (542736000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (562089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (574790400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (594144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (606240000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (625593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (636480000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (657043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (667929600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (688492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (699379200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (719942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (731433600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (751996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (762883200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (783446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (794332800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (814896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (828201600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (846345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (859651200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (877795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (891100800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (909244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (922550400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (941299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (954000000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (967305600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (985449600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1004198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1017504000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1035648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1048953600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1067097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1080403200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1099152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1111852800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1130601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1143907200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1162051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1174752000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1193500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1207411200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1223136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1238860800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1254585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1270310400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1286035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1301760000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1317484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1333209600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1349539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1365264000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1380988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1396713600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1412438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1428163200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1443888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1459612800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1475337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1491062400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1506787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1522512000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1538841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1554566400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1570291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1586016000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1601740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1617465600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1633190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1648915200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1664640000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1680364800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1696089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1712419200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1728144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1743868800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1759593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1775318400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1791043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1806768000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1822492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1838217600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1853942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1869667200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1885996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1901721600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1917446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1933171200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1948896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1964620800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1980345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1996070400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2011795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2027520000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2043244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2058969600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2075299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2091024000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2106748800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2122473600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2138198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2153923200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2169648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2185372800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2201097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2216822400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2233152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2248876800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2264601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2280326400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2296051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2311776000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2327500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2343225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2358950400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2374675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2390400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2406124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2422454400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2438179200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2453904000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2469628800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2485353600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2501078400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2516803200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2532528000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2548252800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2563977600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2579702400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2596032000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2611756800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2627481600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2643206400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2658931200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2674656000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2690380800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2706105600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2721830400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2737555200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2753280000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2769609600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2785334400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2801059200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2816784000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2832508800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2848233600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2863958400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2879683200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2895408000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2911132800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2926857600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2942582400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2958912000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2974636800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2990361600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3006086400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3021811200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3037536000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3053260800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3068985600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3084710400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3100435200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3116764800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3132489600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3148214400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3163939200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3179664000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3195388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3211113600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3226838400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3242563200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3258288000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3274012800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3289737600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3306067200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3321792000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3337516800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3353241600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3368966400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3384691200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3400416000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3416140800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3431865600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3447590400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3463315200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3479644800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3495369600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3511094400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3526819200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3542544000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3558268800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3573993600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3589718400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3605443200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3621168000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3636892800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3653222400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3668947200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3684672000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3700396800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3716121600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3731846400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3747571200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3763296000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3779020800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3794745600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3810470400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3826195200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3842524800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3858249600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3873974400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3889699200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3905424000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3921148800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3936873600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3952598400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3968323200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3984048000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4000377600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4016102400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4031827200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4047552000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4063276800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4079001600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4094726400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 36292,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Currie => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2345795356, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1680508800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1665388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1646640000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1635753600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1615190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1604304000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-883641600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-876124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-860400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-844675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-828345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-813225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-71136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-55411200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-37267200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-25776000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-5817600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (5673600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (25632000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (37728000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (57686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (67968000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (89136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (100022400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (120585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (131472000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (152035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (162921600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (183484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (194976000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (215539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (226425600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (246988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (257875200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (278438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (289324800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (309888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (320774400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (341337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (352224000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (372787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (386092800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (404841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (417542400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (436291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (447177600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (467740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (478627200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (499190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (510076800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (530035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (542736000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (562089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (574790400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (594144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (606240000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (625593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (637689600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (657043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (670348800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (686678400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (701798400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (718128000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (733248000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (749577600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (764697600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (781027200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (796147200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (812476800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (828201600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (844531200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (859651200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (875980800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (891100800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (907430400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (922550400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (938880000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (954000000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (967305600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (985449600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1002384000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1017504000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1033833600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1048953600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1065283200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1080403200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1096732800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1111852800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1128182400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1143907200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1159632000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1174752000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1191686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1207411200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1223136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1238860800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1254585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1270310400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1286035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1301760000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1317484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1333209600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1349539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1365264000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1380988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1396713600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1412438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1428163200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1443888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1459612800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1475337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1491062400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1506787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1522512000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1538841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1554566400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1570291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1586016000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1601740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1617465600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1633190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1648915200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1664640000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1680364800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1696089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1712419200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1728144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1743868800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1759593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1775318400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1791043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1806768000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1822492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1838217600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1853942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1869667200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1885996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1901721600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1917446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1933171200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1948896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1964620800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1980345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1996070400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2011795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2027520000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2043244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2058969600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2075299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2091024000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2106748800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2122473600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2138198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2153923200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2169648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2185372800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2201097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2216822400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2233152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2248876800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2264601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2280326400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2296051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2311776000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2327500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2343225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2358950400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2374675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2390400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2406124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2422454400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2438179200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2453904000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2469628800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2485353600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2501078400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2516803200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2532528000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2548252800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2563977600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2579702400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2596032000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2611756800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2627481600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2643206400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2658931200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2674656000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2690380800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2706105600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2721830400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2737555200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2753280000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2769609600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2785334400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2801059200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2816784000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2832508800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2848233600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2863958400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2879683200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2895408000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2911132800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2926857600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2942582400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2958912000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2974636800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2990361600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3006086400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3021811200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3037536000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3053260800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3068985600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3084710400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3100435200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3116764800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3132489600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3148214400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3163939200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3179664000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3195388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3211113600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3226838400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3242563200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3258288000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3274012800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3289737600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3306067200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3321792000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3337516800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3353241600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3368966400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3384691200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3400416000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3416140800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3431865600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3447590400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3463315200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3479644800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3495369600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3511094400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3526819200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3542544000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3558268800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3573993600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3589718400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3605443200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3621168000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3636892800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3653222400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3668947200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3684672000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3700396800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3716121600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3731846400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3747571200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3763296000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3779020800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3794745600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3810470400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3826195200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3842524800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3858249600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3873974400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3889699200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3905424000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3921148800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3936873600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3952598400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3968323200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3984048000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4000377600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4016102400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4031827200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4047552000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4063276800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4079001600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4094726400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 35356,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Darwin => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2364108200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "ACST" }),
                    (-2230189200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-1672558200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-1665387000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-883639800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-876123000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-860398200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-844673400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-828343800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-813223800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 31400,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Eucla => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2337928528, FixedTimespan { utc_offset: 31500, dst_offset: 0, name: "+0845" }),
                    (-1672555500, FixedTimespan { utc_offset: 31500, dst_offset: 3600, name: "+0945" }),
                    (-1665384300, FixedTimespan { utc_offset: 31500, dst_offset: 0, name: "+0845" }),
                    (-883637100, FixedTimespan { utc_offset: 31500, dst_offset: 3600, name: "+0945" }),
                    (-876120300, FixedTimespan { utc_offset: 31500, dst_offset: 0, name: "+0845" }),
                    (-860395500, FixedTimespan { utc_offset: 31500, dst_offset: 3600, name: "+0945" }),
                    (-844670700, FixedTimespan { utc_offset: 31500, dst_offset: 0, name: "+0845" }),
                    (152039700, FixedTimespan { utc_offset: 31500, dst_offset: 3600, name: "+0945" }),
                    (162926100, FixedTimespan { utc_offset: 31500, dst_offset: 0, name: "+0845" }),
                    (436295700, FixedTimespan { utc_offset: 31500, dst_offset: 3600, name: "+0945" }),
                    (447182100, FixedTimespan { utc_offset: 31500, dst_offset: 0, name: "+0845" }),
                    (690311700, FixedTimespan { utc_offset: 31500, dst_offset: 3600, name: "+0945" }),
                    (699383700, FixedTimespan { utc_offset: 31500, dst_offset: 0, name: "+0845" }),
                    (1165079700, FixedTimespan { utc_offset: 31500, dst_offset: 3600, name: "+0945" }),
                    (1174756500, FixedTimespan { utc_offset: 31500, dst_offset: 0, name: "+0845" }),
                    (1193505300, FixedTimespan { utc_offset: 31500, dst_offset: 3600, name: "+0945" }),
                    (1206810900, FixedTimespan { utc_offset: 31500, dst_offset: 0, name: "+0845" }),
                    (1224954900, FixedTimespan { utc_offset: 31500, dst_offset: 3600, name: "+0945" }),
                    (1238260500, FixedTimespan { utc_offset: 31500, dst_offset: 0, name: "+0845" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 30928,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Hobart => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2345795356, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1680508800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1665388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1646640000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1635753600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1615190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1604304000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-883641600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-876124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-860400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-844675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-828345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-813225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-71136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-55411200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-37267200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-25776000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-5817600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (5673600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (25632000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (37728000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (57686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (67968000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (89136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (100022400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (120585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (131472000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (152035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (162921600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (183484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (194976000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (215539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (226425600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (246988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (257875200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (278438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (289324800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (309888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (320774400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (341337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (352224000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (372787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (386092800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (404841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (417542400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (436291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (447177600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (467740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (478627200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (499190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (510076800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (530035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (542736000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (562089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (574790400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (594144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (606240000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (625593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (637689600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (657043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (670348800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (686678400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (701798400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (718128000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (733248000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (749577600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (764697600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (781027200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (796147200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (812476800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (828201600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (844531200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (859651200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (875980800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (891100800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (907430400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (922550400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (938880000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (954000000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (967305600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (985449600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1002384000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1017504000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1033833600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1048953600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1065283200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1080403200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1096732800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1111852800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1128182400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1143907200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1159632000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1174752000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1191686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1207411200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1223136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1238860800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1254585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1270310400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1286035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1301760000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1317484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1333209600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1349539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1365264000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1380988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1396713600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1412438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1428163200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1443888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1459612800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1475337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1491062400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1506787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1522512000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1538841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1554566400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1570291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1586016000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1601740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1617465600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1633190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1648915200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1664640000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1680364800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1696089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1712419200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1728144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1743868800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1759593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1775318400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1791043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1806768000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1822492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1838217600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1853942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1869667200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1885996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1901721600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1917446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1933171200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1948896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1964620800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1980345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1996070400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2011795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2027520000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2043244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2058969600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2075299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2091024000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2106748800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2122473600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2138198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2153923200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2169648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2185372800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2201097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2216822400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2233152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2248876800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2264601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2280326400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2296051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2311776000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2327500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2343225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2358950400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2374675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2390400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2406124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2422454400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2438179200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2453904000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2469628800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2485353600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2501078400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2516803200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2532528000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2548252800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2563977600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2579702400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2596032000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2611756800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2627481600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2643206400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2658931200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2674656000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2690380800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2706105600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2721830400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2737555200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2753280000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2769609600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2785334400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2801059200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2816784000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2832508800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2848233600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2863958400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2879683200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2895408000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2911132800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2926857600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2942582400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2958912000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2974636800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2990361600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3006086400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3021811200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3037536000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3053260800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3068985600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3084710400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3100435200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3116764800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3132489600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3148214400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3163939200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3179664000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3195388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3211113600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3226838400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3242563200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3258288000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3274012800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3289737600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3306067200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3321792000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3337516800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3353241600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3368966400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3384691200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3400416000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3416140800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3431865600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3447590400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3463315200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3479644800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3495369600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3511094400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3526819200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3542544000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3558268800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3573993600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3589718400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3605443200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3621168000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3636892800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3653222400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3668947200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3684672000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3700396800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3716121600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3731846400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3747571200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3763296000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3779020800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3794745600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3810470400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3826195200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3842524800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3858249600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3873974400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3889699200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3905424000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3921148800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3936873600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3952598400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3968323200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3984048000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4000377600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4016102400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4031827200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4047552000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4063276800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4079001600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4094726400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 35356,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__LHI => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2364114980, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (352216800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (372785400, FixedTimespan { utc_offset: 37800, dst_offset: 3600, name: "+1130" }),
                    (384273000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (404839800, FixedTimespan { utc_offset: 37800, dst_offset: 3600, name: "+1130" }),
                    (415722600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (436289400, FixedTimespan { utc_offset: 37800, dst_offset: 3600, name: "+1130" }),
                    (447172200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (467739000, FixedTimespan { utc_offset: 37800, dst_offset: 3600, name: "+1130" }),
                    (478621800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (499188600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (511282800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (530033400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (542732400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (562087800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (574786800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (594142200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (606236400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (625591800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (636476400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (657041400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (667926000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (688491000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (699375600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (719940600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (731430000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (751995000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (762879600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (783444600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (794329200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (814894200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (828198000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (846343800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (859647600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (877793400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (891097200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (909243000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (922546800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (941297400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (953996400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (967303800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (985446000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1004196600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1017500400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1035646200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1048950000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1067095800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1080399600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1099150200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1111849200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1130599800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1143903600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1162049400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1174748400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1193499000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1207407600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1223134200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1238857200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1254583800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1270306800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1286033400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1301756400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1317483000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1333206000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1349537400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1365260400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1380987000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1396710000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1412436600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1428159600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1443886200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1459609200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1475335800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1491058800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1506785400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1522508400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1538839800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1554562800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1570289400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1586012400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1601739000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1617462000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1633188600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1648911600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1664638200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1680361200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1696087800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1712415600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1728142200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1743865200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1759591800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1775314800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1791041400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1806764400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1822491000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1838214000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1853940600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1869663600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1885995000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1901718000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1917444600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1933167600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1948894200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1964617200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1980343800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1996066800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2011793400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2027516400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2043243000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2058966000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2075297400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2091020400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2106747000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2122470000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2138196600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2153919600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2169646200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2185369200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2201095800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2216818800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2233150200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2248873200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2264599800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2280322800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2296049400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2311772400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2327499000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2343222000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2358948600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2374671600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2390398200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2406121200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2422452600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2438175600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2453902200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2469625200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2485351800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2501074800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2516801400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2532524400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2548251000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2563974000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2579700600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2596028400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2611755000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2627478000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2643204600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2658927600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2674654200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2690377200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2706103800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2721826800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2737553400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2753276400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2769607800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2785330800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2801057400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2816780400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2832507000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2848230000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2863956600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2879679600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2895406200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2911129200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2926855800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2942578800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2958910200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2974633200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2990359800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3006082800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3021809400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3037532400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3053259000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3068982000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3084708600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3100431600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3116763000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3132486000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3148212600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3163935600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3179662200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3195385200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3211111800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3226834800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3242561400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3258284400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3274011000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3289734000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3306065400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3321788400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3337515000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3353238000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3368964600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3384687600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3400414200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3416137200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3431863800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3447586800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3463313400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3479641200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3495367800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3511090800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3526817400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3542540400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3558267000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3573990000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3589716600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3605439600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3621166200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3636889200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3653220600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3668943600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3684670200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3700393200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3716119800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3731842800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3747569400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3763292400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3779019000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3794742000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3810468600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3826191600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3842523000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3858246000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3873972600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3889695600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3905422200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3921145200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3936871800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3952594800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3968321400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3984044400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (4000375800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (4016098800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (4031825400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (4047548400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (4063275000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (4078998000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (4094724600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 38180,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Lindeman => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2366790956, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1672560000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1665388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-883641600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-876124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-860400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-844675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-828345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-813225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (57686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (67968000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (625593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (636480000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (657043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (667929600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (688492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (699379200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (719942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (731433600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (751996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (762883200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 35756,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Lord_Howe => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2364114980, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (352216800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (372785400, FixedTimespan { utc_offset: 37800, dst_offset: 3600, name: "+1130" }),
                    (384273000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (404839800, FixedTimespan { utc_offset: 37800, dst_offset: 3600, name: "+1130" }),
                    (415722600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (436289400, FixedTimespan { utc_offset: 37800, dst_offset: 3600, name: "+1130" }),
                    (447172200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (467739000, FixedTimespan { utc_offset: 37800, dst_offset: 3600, name: "+1130" }),
                    (478621800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (499188600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (511282800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (530033400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (542732400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (562087800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (574786800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (594142200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (606236400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (625591800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (636476400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (657041400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (667926000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (688491000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (699375600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (719940600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (731430000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (751995000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (762879600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (783444600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (794329200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (814894200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (828198000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (846343800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (859647600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (877793400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (891097200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (909243000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (922546800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (941297400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (953996400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (967303800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (985446000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1004196600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1017500400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1035646200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1048950000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1067095800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1080399600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1099150200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1111849200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1130599800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1143903600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1162049400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1174748400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1193499000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1207407600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1223134200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1238857200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1254583800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1270306800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1286033400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1301756400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1317483000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1333206000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1349537400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1365260400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1380987000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1396710000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1412436600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1428159600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1443886200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1459609200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1475335800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1491058800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1506785400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1522508400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1538839800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1554562800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1570289400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1586012400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1601739000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1617462000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1633188600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1648911600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1664638200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1680361200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1696087800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1712415600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1728142200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1743865200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1759591800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1775314800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1791041400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1806764400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1822491000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1838214000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1853940600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1869663600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1885995000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1901718000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1917444600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1933167600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1948894200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1964617200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (1980343800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (1996066800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2011793400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2027516400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2043243000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2058966000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2075297400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2091020400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2106747000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2122470000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2138196600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2153919600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2169646200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2185369200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2201095800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2216818800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2233150200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2248873200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2264599800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2280322800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2296049400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2311772400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2327499000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2343222000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2358948600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2374671600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2390398200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2406121200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2422452600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2438175600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2453902200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2469625200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2485351800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2501074800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2516801400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2532524400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2548251000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2563974000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2579700600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2596028400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2611755000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2627478000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2643204600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2658927600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2674654200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2690377200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2706103800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2721826800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2737553400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2753276400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2769607800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2785330800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2801057400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2816780400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2832507000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2848230000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2863956600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2879679600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2895406200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2911129200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2926855800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2942578800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2958910200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (2974633200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (2990359800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3006082800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3021809400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3037532400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3053259000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3068982000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3084708600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3100431600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3116763000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3132486000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3148212600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3163935600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3179662200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3195385200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3211111800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3226834800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3242561400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3258284400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3274011000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3289734000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3306065400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3321788400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3337515000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3353238000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3368964600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3384687600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3400414200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3416137200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3431863800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3447586800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3463313400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3479641200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3495367800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3511090800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3526817400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3542540400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3558267000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3573990000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3589716600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3605439600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3621166200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3636889200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3653220600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3668943600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3684670200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3700393200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3716119800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3731842800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3747569400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3763292400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3779019000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3794742000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3810468600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3826191600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3842523000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3858246000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3873972600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3889695600, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3905422200, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3921145200, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3936871800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3952594800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (3968321400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (3984044400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (4000375800, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (4016098800, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (4031825400, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (4047548400, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (4063275000, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                    (4078998000, FixedTimespan { utc_offset: 37800, dst_offset: 0, name: "+1030" }),
                    (4094724600, FixedTimespan { utc_offset: 37800, dst_offset: 1800, name: "+11" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 38180,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Melbourne => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2364111592, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1672560000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1665388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-883641600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-876124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-860400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-844675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-828345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-813225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (57686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (67968000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (89136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (100022400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (120585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (131472000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (152035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (162921600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (183484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (194976000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (215539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (226425600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (246988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (257875200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (278438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (289324800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (309888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (320774400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (341337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (352224000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (372787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (384278400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (404841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (415728000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (436291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (447177600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (467740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (478627200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (499190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (511286400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (530035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (542736000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (561484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (574790400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (594144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (606240000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (625593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (637689600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (657043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (667929600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (688492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (699379200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (719942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (731433600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (751996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (762883200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (783446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (796147200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (814896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (828201600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (846345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (859651200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (877795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (891100800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (909244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (922550400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (941299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (954000000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (967305600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (985449600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1004198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1017504000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1035648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1048953600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1067097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1080403200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1099152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1111852800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1130601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1143907200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1162051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1174752000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1193500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1207411200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1223136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1238860800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1254585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1270310400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1286035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1301760000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1317484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1333209600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1349539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1365264000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1380988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1396713600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1412438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1428163200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1443888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1459612800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1475337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1491062400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1506787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1522512000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1538841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1554566400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1570291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1586016000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1601740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1617465600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1633190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1648915200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1664640000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1680364800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1696089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1712419200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1728144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1743868800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1759593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1775318400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1791043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1806768000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1822492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1838217600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1853942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1869667200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1885996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1901721600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1917446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1933171200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1948896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1964620800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1980345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1996070400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2011795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2027520000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2043244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2058969600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2075299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2091024000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2106748800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2122473600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2138198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2153923200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2169648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2185372800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2201097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2216822400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2233152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2248876800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2264601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2280326400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2296051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2311776000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2327500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2343225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2358950400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2374675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2390400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2406124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2422454400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2438179200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2453904000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2469628800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2485353600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2501078400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2516803200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2532528000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2548252800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2563977600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2579702400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2596032000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2611756800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2627481600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2643206400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2658931200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2674656000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2690380800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2706105600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2721830400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2737555200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2753280000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2769609600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2785334400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2801059200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2816784000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2832508800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2848233600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2863958400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2879683200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2895408000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2911132800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2926857600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2942582400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2958912000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2974636800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2990361600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3006086400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3021811200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3037536000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3053260800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3068985600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3084710400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3100435200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3116764800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3132489600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3148214400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3163939200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3179664000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3195388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3211113600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3226838400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3242563200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3258288000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3274012800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3289737600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3306067200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3321792000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3337516800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3353241600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3368966400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3384691200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3400416000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3416140800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3431865600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3447590400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3463315200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3479644800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3495369600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3511094400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3526819200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3542544000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3558268800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3573993600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3589718400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3605443200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3621168000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3636892800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3653222400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3668947200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3684672000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3700396800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3716121600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3731846400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3747571200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3763296000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3779020800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3794745600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3810470400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3826195200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3842524800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3858249600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3873974400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3889699200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3905424000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3921148800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3936873600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3952598400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3968323200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3984048000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4000377600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4016102400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4031827200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4047552000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4063276800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4079001600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4094726400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 34792,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__NSW => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2364113092, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1672560000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1665388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-883641600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-876124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-860400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-844675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-828345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-813225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (57686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (67968000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (89136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (100022400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (120585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (131472000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (152035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (162921600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (183484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (194976000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (215539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (226425600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (246988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (257875200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (278438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (289324800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (309888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (320774400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (341337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (352224000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (372787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (386697600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (404841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (415728000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (436291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (447177600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (467740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (478627200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (499190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (511286400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (530035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (542736000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (562089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (574790400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (594144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (606240000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (625593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (636480000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (657043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (667929600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (688492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (699379200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (719942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (731433600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (751996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (762883200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (783446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (794332800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (814896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (828201600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (846345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (859651200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (877795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (891100800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (909244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (922550400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (941299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (954000000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (967305600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (985449600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1004198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1017504000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1035648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1048953600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1067097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1080403200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1099152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1111852800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1130601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1143907200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1162051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1174752000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1193500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1207411200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1223136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1238860800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1254585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1270310400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1286035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1301760000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1317484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1333209600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1349539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1365264000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1380988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1396713600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1412438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1428163200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1443888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1459612800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1475337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1491062400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1506787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1522512000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1538841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1554566400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1570291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1586016000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1601740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1617465600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1633190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1648915200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1664640000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1680364800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1696089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1712419200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1728144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1743868800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1759593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1775318400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1791043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1806768000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1822492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1838217600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1853942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1869667200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1885996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1901721600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1917446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1933171200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1948896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1964620800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1980345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1996070400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2011795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2027520000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2043244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2058969600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2075299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2091024000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2106748800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2122473600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2138198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2153923200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2169648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2185372800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2201097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2216822400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2233152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2248876800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2264601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2280326400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2296051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2311776000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2327500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2343225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2358950400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2374675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2390400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2406124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2422454400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2438179200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2453904000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2469628800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2485353600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2501078400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2516803200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2532528000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2548252800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2563977600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2579702400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2596032000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2611756800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2627481600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2643206400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2658931200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2674656000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2690380800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2706105600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2721830400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2737555200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2753280000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2769609600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2785334400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2801059200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2816784000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2832508800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2848233600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2863958400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2879683200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2895408000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2911132800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2926857600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2942582400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2958912000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2974636800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2990361600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3006086400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3021811200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3037536000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3053260800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3068985600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3084710400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3100435200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3116764800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3132489600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3148214400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3163939200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3179664000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3195388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3211113600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3226838400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3242563200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3258288000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3274012800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3289737600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3306067200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3321792000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3337516800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3353241600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3368966400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3384691200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3400416000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3416140800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3431865600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3447590400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3463315200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3479644800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3495369600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3511094400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3526819200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3542544000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3558268800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3573993600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3589718400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3605443200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3621168000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3636892800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3653222400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3668947200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3684672000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3700396800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3716121600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3731846400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3747571200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3763296000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3779020800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3794745600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3810470400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3826195200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3842524800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3858249600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3873974400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3889699200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3905424000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3921148800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3936873600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3952598400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3968323200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3984048000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4000377600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4016102400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4031827200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4047552000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4063276800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4079001600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4094726400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 36292,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__North => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2364108200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "ACST" }),
                    (-2230189200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-1672558200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-1665387000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-883639800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-876123000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-860398200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-844673400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-828343800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-813223800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 31400,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Perth => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2337925404, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (-1672552800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (-1665381600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (-883634400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (-876117600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (-860392800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (-844668000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (152042400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (162928800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (436298400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (447184800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (690314400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (699386400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (1165082400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (1174759200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (1193508000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (1206813600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (1224957600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (1238263200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 27804,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Queensland => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2366791928, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1672560000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1665388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-883641600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-876124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-860400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-844675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-828345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-813225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (57686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (67968000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (625593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (636480000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (657043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (667929600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (688492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (699379200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 36728,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__South => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2364110060, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "ACST" }),
                    (-2230189200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-1672558200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-1665387000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-883639800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-876123000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-860398200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-844673400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-828343800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-813223800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (57688200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (67969800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (89137800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (100024200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (120587400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (131473800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (152037000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (162923400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (183486600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (194977800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (215541000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (226427400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (246990600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (257877000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (278440200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (289326600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (309889800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (320776200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (341339400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (352225800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (372789000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (384280200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (404843400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (415729800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (436293000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (447179400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (467742600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (478629000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (499192200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (511288200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (530037000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (542737800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (562091400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (574792200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (594145800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (606241800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (625595400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (637691400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (657045000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (667931400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (688494600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (701195400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (719944200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (731435400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (751998600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (764094600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (783448200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (796149000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (814897800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (828203400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (846347400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (859653000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (877797000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (891102600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (909246600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (922552200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (941301000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (954001800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (972750600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (985451400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1004200200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1017505800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1035649800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1048955400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1067099400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1080405000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1099153800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1111854600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1130603400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1143909000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1162053000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1174753800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1193502600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1207413000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1223137800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1238862600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1254587400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1270312200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1286037000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1301761800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1317486600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1333211400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1349541000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1365265800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1380990600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1396715400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1412440200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1428165000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1443889800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1459614600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1475339400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1491064200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1506789000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1522513800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1538843400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1554568200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1570293000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1586017800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1601742600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1617467400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1633192200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1648917000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1664641800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1680366600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1696091400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1712421000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1728145800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1743870600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1759595400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1775320200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1791045000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1806769800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1822494600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1838219400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1853944200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1869669000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1885998600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1901723400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1917448200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1933173000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1948897800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1964622600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1980347400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1996072200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2011797000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2027521800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2043246600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2058971400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2075301000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2091025800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2106750600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2122475400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2138200200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2153925000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2169649800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2185374600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2201099400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2216824200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2233153800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2248878600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2264603400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2280328200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2296053000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2311777800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2327502600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2343227400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2358952200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2374677000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2390401800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2406126600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2422456200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2438181000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2453905800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2469630600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2485355400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2501080200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2516805000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2532529800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2548254600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2563979400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2579704200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2596033800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2611758600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2627483400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2643208200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2658933000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2674657800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2690382600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2706107400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2721832200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2737557000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2753281800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2769611400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2785336200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2801061000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2816785800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2832510600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2848235400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2863960200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2879685000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2895409800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2911134600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2926859400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2942584200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2958913800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2974638600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2990363400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3006088200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3021813000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3037537800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3053262600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3068987400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3084712200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3100437000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3116766600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3132491400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3148216200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3163941000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3179665800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3195390600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3211115400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3226840200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3242565000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3258289800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3274014600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3289739400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3306069000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3321793800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3337518600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3353243400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3368968200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3384693000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3400417800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3416142600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3431867400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3447592200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3463317000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3479646600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3495371400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3511096200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3526821000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3542545800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3558270600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3573995400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3589720200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3605445000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3621169800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3636894600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3653224200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3668949000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3684673800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3700398600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3716123400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3731848200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3747573000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3763297800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3779022600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3794747400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3810472200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3826197000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3842526600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3858251400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3873976200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3889701000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3905425800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3921150600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3936875400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3952600200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3968325000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3984049800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4000379400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (4016104200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4031829000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (4047553800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4063278600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (4079003400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4094728200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 33260,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Sydney => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2364113092, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1672560000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1665388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-883641600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-876124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-860400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-844675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-828345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-813225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (57686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (67968000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (89136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (100022400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (120585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (131472000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (152035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (162921600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (183484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (194976000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (215539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (226425600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (246988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (257875200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (278438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (289324800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (309888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (320774400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (341337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (352224000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (372787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (386697600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (404841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (415728000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (436291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (447177600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (467740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (478627200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (499190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (511286400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (530035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (542736000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (562089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (574790400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (594144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (606240000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (625593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (636480000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (657043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (667929600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (688492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (699379200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (719942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (731433600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (751996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (762883200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (783446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (794332800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (814896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (828201600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (846345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (859651200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (877795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (891100800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (909244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (922550400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (941299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (954000000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (967305600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (985449600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1004198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1017504000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1035648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1048953600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1067097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1080403200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1099152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1111852800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1130601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1143907200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1162051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1174752000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1193500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1207411200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1223136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1238860800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1254585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1270310400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1286035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1301760000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1317484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1333209600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1349539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1365264000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1380988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1396713600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1412438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1428163200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1443888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1459612800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1475337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1491062400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1506787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1522512000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1538841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1554566400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1570291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1586016000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1601740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1617465600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1633190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1648915200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1664640000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1680364800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1696089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1712419200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1728144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1743868800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1759593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1775318400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1791043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1806768000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1822492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1838217600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1853942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1869667200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1885996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1901721600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1917446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1933171200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1948896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1964620800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1980345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1996070400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2011795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2027520000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2043244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2058969600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2075299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2091024000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2106748800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2122473600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2138198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2153923200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2169648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2185372800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2201097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2216822400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2233152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2248876800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2264601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2280326400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2296051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2311776000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2327500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2343225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2358950400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2374675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2390400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2406124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2422454400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2438179200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2453904000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2469628800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2485353600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2501078400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2516803200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2532528000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2548252800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2563977600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2579702400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2596032000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2611756800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2627481600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2643206400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2658931200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2674656000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2690380800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2706105600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2721830400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2737555200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2753280000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2769609600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2785334400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2801059200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2816784000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2832508800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2848233600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2863958400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2879683200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2895408000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2911132800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2926857600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2942582400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2958912000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2974636800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2990361600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3006086400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3021811200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3037536000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3053260800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3068985600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3084710400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3100435200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3116764800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3132489600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3148214400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3163939200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3179664000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3195388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3211113600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3226838400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3242563200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3258288000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3274012800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3289737600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3306067200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3321792000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3337516800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3353241600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3368966400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3384691200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3400416000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3416140800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3431865600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3447590400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3463315200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3479644800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3495369600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3511094400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3526819200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3542544000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3558268800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3573993600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3589718400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3605443200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3621168000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3636892800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3653222400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3668947200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3684672000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3700396800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3716121600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3731846400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3747571200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3763296000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3779020800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3794745600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3810470400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3826195200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3842524800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3858249600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3873974400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3889699200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3905424000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3921148800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3936873600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3952598400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3968323200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3984048000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4000377600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4016102400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4031827200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4047552000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4063276800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4079001600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4094726400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 36292,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Tasmania => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2345795356, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1680508800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1665388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1646640000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1635753600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1615190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1604304000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-883641600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-876124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-860400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-844675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-828345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-813225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-71136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-55411200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-37267200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-25776000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-5817600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (5673600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (25632000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (37728000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (57686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (67968000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (89136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (100022400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (120585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (131472000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (152035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (162921600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (183484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (194976000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (215539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (226425600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (246988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (257875200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (278438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (289324800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (309888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (320774400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (341337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (352224000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (372787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (386092800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (404841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (417542400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (436291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (447177600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (467740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (478627200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (499190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (510076800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (530035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (542736000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (562089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (574790400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (594144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (606240000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (625593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (637689600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (657043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (670348800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (686678400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (701798400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (718128000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (733248000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (749577600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (764697600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (781027200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (796147200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (812476800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (828201600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (844531200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (859651200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (875980800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (891100800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (907430400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (922550400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (938880000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (954000000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (967305600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (985449600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1002384000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1017504000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1033833600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1048953600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1065283200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1080403200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1096732800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1111852800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1128182400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1143907200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1159632000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1174752000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1191686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1207411200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1223136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1238860800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1254585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1270310400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1286035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1301760000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1317484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1333209600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1349539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1365264000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1380988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1396713600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1412438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1428163200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1443888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1459612800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1475337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1491062400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1506787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1522512000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1538841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1554566400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1570291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1586016000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1601740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1617465600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1633190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1648915200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1664640000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1680364800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1696089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1712419200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1728144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1743868800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1759593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1775318400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1791043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1806768000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1822492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1838217600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1853942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1869667200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1885996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1901721600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1917446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1933171200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1948896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1964620800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1980345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1996070400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2011795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2027520000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2043244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2058969600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2075299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2091024000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2106748800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2122473600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2138198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2153923200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2169648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2185372800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2201097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2216822400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2233152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2248876800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2264601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2280326400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2296051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2311776000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2327500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2343225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2358950400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2374675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2390400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2406124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2422454400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2438179200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2453904000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2469628800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2485353600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2501078400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2516803200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2532528000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2548252800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2563977600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2579702400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2596032000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2611756800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2627481600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2643206400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2658931200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2674656000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2690380800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2706105600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2721830400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2737555200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2753280000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2769609600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2785334400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2801059200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2816784000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2832508800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2848233600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2863958400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2879683200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2895408000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2911132800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2926857600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2942582400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2958912000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2974636800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2990361600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3006086400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3021811200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3037536000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3053260800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3068985600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3084710400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3100435200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3116764800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3132489600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3148214400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3163939200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3179664000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3195388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3211113600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3226838400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3242563200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3258288000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3274012800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3289737600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3306067200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3321792000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3337516800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3353241600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3368966400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3384691200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3400416000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3416140800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3431865600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3447590400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3463315200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3479644800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3495369600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3511094400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3526819200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3542544000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3558268800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3573993600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3589718400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3605443200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3621168000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3636892800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3653222400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3668947200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3684672000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3700396800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3716121600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3731846400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3747571200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3763296000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3779020800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3794745600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3810470400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3826195200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3842524800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3858249600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3873974400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3889699200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3905424000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3921148800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3936873600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3952598400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3968323200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3984048000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4000377600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4016102400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4031827200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4047552000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4063276800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4079001600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4094726400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 35356,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Victoria => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2364111592, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-1672560000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-1665388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-883641600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-876124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-860400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-844675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-828345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (-813225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (57686400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (67968000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (89136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (100022400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (120585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (131472000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (152035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (162921600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (183484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (194976000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (215539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (226425600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (246988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (257875200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (278438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (289324800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (309888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (320774400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (341337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (352224000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (372787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (384278400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (404841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (415728000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (436291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (447177600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (467740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (478627200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (499190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (511286400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (530035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (542736000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (561484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (574790400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (594144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (606240000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (625593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (637689600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (657043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (667929600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (688492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (699379200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (719942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (731433600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (751996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (762883200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (783446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (796147200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (814896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (828201600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (846345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (859651200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (877795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (891100800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (909244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (922550400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (941299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (954000000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (967305600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (985449600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1004198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1017504000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1035648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1048953600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1067097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1080403200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1099152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1111852800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1130601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1143907200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1162051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1174752000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1193500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1207411200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1223136000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1238860800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1254585600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1270310400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1286035200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1301760000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1317484800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1333209600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1349539200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1365264000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1380988800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1396713600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1412438400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1428163200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1443888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1459612800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1475337600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1491062400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1506787200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1522512000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1538841600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1554566400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1570291200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1586016000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1601740800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1617465600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1633190400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1648915200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1664640000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1680364800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1696089600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1712419200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1728144000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1743868800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1759593600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1775318400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1791043200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1806768000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1822492800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1838217600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1853942400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1869667200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1885996800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1901721600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1917446400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1933171200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1948896000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1964620800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (1980345600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (1996070400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2011795200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2027520000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2043244800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2058969600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2075299200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2091024000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2106748800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2122473600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2138198400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2153923200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2169648000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2185372800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2201097600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2216822400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2233152000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2248876800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2264601600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2280326400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2296051200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2311776000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2327500800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2343225600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2358950400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2374675200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2390400000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2406124800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2422454400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2438179200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2453904000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2469628800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2485353600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2501078400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2516803200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2532528000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2548252800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2563977600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2579702400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2596032000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2611756800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2627481600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2643206400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2658931200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2674656000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2690380800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2706105600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2721830400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2737555200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2753280000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2769609600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2785334400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2801059200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2816784000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2832508800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2848233600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2863958400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2879683200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2895408000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2911132800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2926857600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2942582400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2958912000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (2974636800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (2990361600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3006086400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3021811200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3037536000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3053260800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3068985600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3084710400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3100435200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3116764800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3132489600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3148214400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3163939200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3179664000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3195388800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3211113600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3226838400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3242563200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3258288000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3274012800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3289737600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3306067200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3321792000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3337516800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3353241600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3368966400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3384691200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3400416000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3416140800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3431865600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3447590400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3463315200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3479644800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3495369600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3511094400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3526819200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3542544000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3558268800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3573993600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3589718400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3605443200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3621168000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3636892800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3653222400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3668947200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3684672000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3700396800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3716121600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3731846400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3747571200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3763296000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3779020800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3794745600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3810470400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3826195200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3842524800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3858249600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3873974400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3889699200, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3905424000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3921148800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3936873600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3952598400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (3968323200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (3984048000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4000377600, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4016102400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4031827200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4047552000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4063276800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                    (4079001600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (4094726400, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "AEDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 34792,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__West => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2337925404, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (-1672552800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (-1665381600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (-883634400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (-876117600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (-860392800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (-844668000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (152042400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (162928800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (436298400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (447184800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (690314400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (699386400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (1165082400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (1174759200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (1193508000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (1206813600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                    (1224957600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "AWDT" }),
                    (1238263200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "AWST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 27804,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Australia__Yancowinna => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2364110748, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "AEST" }),
                    (-2314951200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "ACST" }),
                    (-2230189200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-1672558200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-1665387000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-883639800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-876123000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-860398200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-844673400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (-828343800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (-813223800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (57688200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (67969800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (89137800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (100024200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (120587400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (131473800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (152037000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (162923400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (183486600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (194977800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (215541000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (226427400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (246990600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (257877000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (278440200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (289326600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (309889800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (320776200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (341339400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (352225800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (372789000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (386699400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (404843400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (415729800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (436293000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (447179400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (467742600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (478629000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (499192200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (511288200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (530037000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (542737800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (562091400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (574792200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (594145800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (606241800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (625595400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (636481800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (657045000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (667931400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (688494600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (699381000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (719944200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (731435400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (751998600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (762885000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (783448200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (794334600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (814897800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (828203400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (846347400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (859653000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (877797000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (891102600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (909246600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (922552200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (941301000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (954001800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (972750600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (985451400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1004200200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1017505800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1035649800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1048955400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1067099400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1080405000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1099153800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1111854600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1130603400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1143909000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1162053000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1174753800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1193502600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1207413000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1223137800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1238862600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1254587400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1270312200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1286037000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1301761800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1317486600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1333211400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1349541000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1365265800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1380990600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1396715400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1412440200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1428165000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1443889800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1459614600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1475339400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1491064200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1506789000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1522513800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1538843400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1554568200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1570293000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1586017800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1601742600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1617467400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1633192200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1648917000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1664641800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1680366600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1696091400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1712421000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1728145800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1743870600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1759595400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1775320200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1791045000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1806769800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1822494600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1838219400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1853944200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1869669000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1885998600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1901723400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1917448200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1933173000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1948897800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1964622600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (1980347400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (1996072200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2011797000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2027521800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2043246600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2058971400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2075301000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2091025800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2106750600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2122475400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2138200200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2153925000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2169649800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2185374600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2201099400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2216824200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2233153800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2248878600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2264603400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2280328200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2296053000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2311777800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2327502600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2343227400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2358952200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2374677000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2390401800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2406126600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2422456200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2438181000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2453905800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2469630600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2485355400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2501080200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2516805000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2532529800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2548254600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2563979400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2579704200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2596033800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2611758600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2627483400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2643208200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2658933000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2674657800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2690382600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2706107400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2721832200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2737557000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2753281800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2769611400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2785336200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2801061000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2816785800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2832510600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2848235400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2863960200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2879685000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2895409800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2911134600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2926859400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2942584200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2958913800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (2974638600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (2990363400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3006088200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3021813000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3037537800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3053262600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3068987400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3084712200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3100437000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3116766600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3132491400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3148216200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3163941000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3179665800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3195390600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3211115400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3226840200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3242565000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3258289800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3274014600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3289739400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3306069000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3321793800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3337518600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3353243400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3368968200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3384693000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3400417800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3416142600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3431867400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3447592200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3463317000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3479646600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3495371400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3511096200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3526821000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3542545800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3558270600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3573995400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3589720200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3605445000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3621169800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3636894600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3653224200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3668949000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3684673800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3700398600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3716123400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3731848200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3747573000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3763297800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3779022600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3794747400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3810472200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3826197000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3842526600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3858251400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3873976200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3889701000, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3905425800, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3921150600, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3936875400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3952600200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (3968325000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (3984049800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4000379400, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (4016104200, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4031829000, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (4047553800, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4063278600, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                    (4079003400, FixedTimespan { utc_offset: 34200, dst_offset: 0, name: "ACST" }),
                    (4094728200, FixedTimespan { utc_offset: 34200, dst_offset: 3600, name: "ACDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 33948,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Brazil__Acre => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767209328, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1206950400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1191355200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1175367600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1159819200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-633812400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-622062000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-602276400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-591825600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-570740400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-560203200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-539118000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-531345600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-191358000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-184190400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-155156400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-150062400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-128890800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-121118400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-99946800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-89582400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-68410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-57960000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (499755600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (511243200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (530600400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (540273600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (562136400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (571204800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (1214283600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1384056000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16272,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Brazil__DeNoronha => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767217820, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1206961200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1191366000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-1175378400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-1159830000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-633823200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-622072800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-602287200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-591836400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-570751200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-560214000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-539128800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-531356400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-191368800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-184201200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-155167200, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-150073200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-128901600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-121129200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-99957600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-89593200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (-68421600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (-57970800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (499744800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (511232400, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (530589600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (540262800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (562125600, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (571194000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (592970400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (602038800, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (624420000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (634698000, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (938916000, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (951613200, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (970970400, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (971571600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                    (1003024800, FixedTimespan { utc_offset: -7200, dst_offset: 3600, name: "-01" }),
                    (1013907600, FixedTimespan { utc_offset: -7200, dst_offset: 0, name: "-02" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -7780,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Brazil__East => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767214412, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1206957600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1191362400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-1175374800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-1159826400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-633819600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-622069200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-602283600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-591832800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-570747600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-560210400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-539125200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-531352800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-195426000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-184197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-155163600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-150069600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-128898000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-121125600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-99954000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-89589600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (-68418000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (-57967200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (499748400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (511236000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (530593200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (540266400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (562129200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (571197600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (592974000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (602042400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (624423600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (634701600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (656478000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (666756000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (687927600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (697600800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (719982000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (728445600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (750826800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (761709600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (782276400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (793159200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (813726000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (824004000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (844570800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (856058400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (876106800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (888717600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (908074800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (919562400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (938919600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (951616800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (970974000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (982461600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1003028400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1013911200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1036292400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1045360800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1066532400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1076810400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1099364400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1108864800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1129431600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1140314400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1162695600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1172368800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1192330800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1203213600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1224385200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1234663200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1255834800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1266717600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1287284400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1298167200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1318734000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1330221600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1350788400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1361066400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1382238000, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1392516000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1413687600, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1424570400, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1445137200, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1456020000, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1476586800, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1487469600, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1508036400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1518919200, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                    (1541300400, FixedTimespan { utc_offset: -10800, dst_offset: 3600, name: "-02" }),
                    (1550368800, FixedTimespan { utc_offset: -10800, dst_offset: 0, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -11188,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Brazil__West => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767211196, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1206954000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1191358800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1175371200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-1159822800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-633816000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-622065600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-602280000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-591829200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-570744000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-560206800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-539121600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-531349200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-191361600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-184194000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-155160000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-150066000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-128894400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-121122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-99950400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-89586000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-68414400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-57963600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (499752000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (511239600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (530596800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (540270000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (562132800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (571201200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (750830400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (761713200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -14404,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::CET => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1693706400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-938905200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766623600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (228877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (243997200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (260326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (276051600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (291776400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (307501200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3600,
                        dst_offset: 0,
                        name: "CET",
                    },
                    rest: REST
                }
            },

            Tz::CST6CDT => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-84384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-68662800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-52934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-37213200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-21484800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-5763600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (9964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (25686000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (41414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (57740400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (73468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (89190000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (104918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (120639600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (126691200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152089200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (162374400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (183538800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (199267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (215593200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (230716800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (247042800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (262771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (278492400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (294220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (309942000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (325670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (341391600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (357120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (372841200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (388569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (404895600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (420019200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (436345200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (452073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (467794800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (483523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (499244400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (514972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (530694000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (544608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (562143600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (576057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (594198000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (607507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (625647600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (638956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (657097200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (671011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (688546800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (702460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (719996400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (733910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (752050800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (765360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (783500400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (796809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (814950000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909298800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941353200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (986112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1004252400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035702000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067151600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099206000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130655600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1205049600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1236499200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1257058800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -21600,
                        dst_offset: 0,
                        name: "CST",
                    },
                    rest: REST
                }
            },

            Tz::Canada__Atlantic => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2131645536, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1696276800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1680469200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1632074400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1615143600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1566763200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1557090000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1535486400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1524949200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1504468800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1493413200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1472414400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1461963600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1440964800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1429390800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1409515200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1396731600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1376856000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1366491600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1346616000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1333832400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1313956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1303678800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1282507200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1272661200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1251057600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1240088400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1219608000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1207429200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1188763200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1175979600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1157313600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1143925200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1124049600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1113771600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1091390400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1081026000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1059854400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1050786000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-1030910400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-1018126800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-999460800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-986677200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-965592000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-955227600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-935956800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-923173200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-904507200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-891723600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-880221600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "APT" }),
                    (-765399600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-747252000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-733950000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-715802400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-702500400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-684352800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-671050800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-652903200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-639601200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-589399200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-576097200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-557949600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-544647600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-526500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-513198000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-495050400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-481748400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-431546400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-418244400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-400096800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-386794800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-368647200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-355345200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-337197600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-323895600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-242244000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-226522800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-210794400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-195073200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-179344800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-163623600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-147895200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-131569200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-116445600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-100119600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-84391200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-68670000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-52941600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-37220400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (-21492000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (-5770800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (9957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (25678800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (41407200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (57733200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (73461600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (89182800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (104911200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (120632400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (136360800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (152082000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (167810400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (183531600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (199260000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (215586000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (230709600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (247035600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (262764000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (278485200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (294213600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (309934800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (325663200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (341384400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (357112800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (372834000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (388562400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (404888400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (420012000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (436338000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (452066400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (467787600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (483516000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (499237200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (514965600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (530686800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (544600800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (562136400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (576050400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (594190800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (607500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (625640400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (638949600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (657090000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (671004000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (688539600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (702453600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (719989200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (733903200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (752043600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (765352800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (783493200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (796802400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (814942800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (828856800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (846392400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (860306400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (877842000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (891756000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (909291600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (923205600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (941346000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (954655200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (972795600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (986104800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1004245200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1018159200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1035694800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1049608800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1067144400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1081058400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1099198800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1112508000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1130648400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1143957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1162098000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1173592800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1194152400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1205042400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1225602000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1236492000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1257051600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1268546400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1289106000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1299996000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1320555600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1331445600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1352005200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1362895200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1383454800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1394344800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1414904400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1425794400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1446354000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1457848800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1478408400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1489298400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1509858000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1520748000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1541307600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1552197600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1572757200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1583647200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1604206800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1615701600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1636261200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1647151200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1667710800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1678600800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1699160400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1710050400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1730610000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1741500000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1762059600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1772949600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1793509200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1805004000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1825563600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1836453600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1857013200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1867903200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1888462800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1899352800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1919912400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1930802400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1951362000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1962856800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (1983416400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (1994306400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2014866000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2025756000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2046315600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2057205600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2077765200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2088655200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2109214800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2120104800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2140664400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2152159200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2172718800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2183608800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2204168400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2215058400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2235618000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2246508000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2267067600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2277957600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2298517200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2309407200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2329966800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2341461600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2362021200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2372911200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2393470800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2404360800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2424920400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2435810400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2456370000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2467260000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2487819600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2499314400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2519874000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2530764000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2551323600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2562213600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2582773200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2593663200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2614222800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2625112800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2645672400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2656562400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2677122000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2688616800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2709176400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2720066400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2740626000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2751516000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2772075600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2782965600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2803525200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2814415200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2834974800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2846469600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2867029200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2877919200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2898478800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2909368800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2929928400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2940818400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2961378000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (2972268000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (2992827600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3003717600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3024277200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3035772000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3056331600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3067221600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3087781200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3098671200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3119230800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3130120800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3150680400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3161570400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3182130000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3193020000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3213579600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3225074400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3245634000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3256524000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3277083600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3287973600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3308533200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3319423200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3339982800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3350872800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3371432400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3382927200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3403486800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3414376800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3434936400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3445826400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3466386000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3477276000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3497835600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3508725600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3529285200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3540175200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3560734800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3572229600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3592789200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3603679200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3624238800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3635128800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3655688400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3666578400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3687138000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3698028000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3718587600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3730082400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3750642000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3761532000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3782091600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3792981600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3813541200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3824431200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3844990800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3855880800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3876440400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3887330400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3907890000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3919384800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3939944400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3950834400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (3971394000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (3982284000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4002843600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4013733600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4034293200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4045183200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4065742800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                    (4076632800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "ADT" }),
                    (4097192400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "AST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -15264,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Canada__Central => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2602258284, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1694368800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1681671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1632067200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1029686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1018198800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-746035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-732733200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-715795200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-702493200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-684345600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-671043600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-652896000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-639594000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-620755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-607626000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-557942400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-544640400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-450291600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-431539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-418237200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-400089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-386787600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-355338000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-337190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-321469200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-305740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-292438800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-210787200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-198090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-116438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-100108800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-84384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-68659200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-52934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-37209600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-21484800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-5760000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (9964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (25689600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (41414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (57744000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (73468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (89193600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (104918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (120643200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (136368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152092800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (167817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (183542400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (199267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (215596800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (230716800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (247046400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (262771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (278496000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (294220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (309945600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (325670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (341395200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (357120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (372844800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (388569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (404899200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (420019200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (436348800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (452073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (467798400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (483523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (499248000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (514972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (530697600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (544608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (562147200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (576057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (594201600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (607507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (625651200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (638956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (657100800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (671011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (688550400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (702460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (720000000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (733910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (752054400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (765360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (783504000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (796809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (814953600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846403200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877852800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909302400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941356800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972806400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (986112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1004256000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035705600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067155200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099209600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130659200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1205049600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1236499200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1257058800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -23316,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Canada__Eastern => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2366736148, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1632070800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1615140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1601753400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1583697600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1567357200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1554667200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1534698000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1524074400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1503248400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1492365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1471798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1460916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1440954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1428861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1409504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1397412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1378054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1365962400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1346605200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1333908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1315155600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1301853600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1283706000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1270404000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1252256400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1238954400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1220806800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1207504800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1188752400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1176055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1157302800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1144000800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1125853200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1112551200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1094403600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1081101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1062954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1049652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1031504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1018202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1000054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-986752800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-968000400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-955303200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-936550800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-880221600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EWT" }),
                    (-769395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EPT" }),
                    (-765396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-747248400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-733946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-715798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-702496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-684349200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-671047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-652899600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-634154400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-620845200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-602704800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-589395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-576093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-557946000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-544644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-526496400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-513194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-495046800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-481744800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-463597200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-450295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-431542800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-418240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-400093200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-384372000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-368643600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-352922400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-337194000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-321472800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-305744400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-289418400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-273690000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-257968800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-242240400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-226519200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-210790800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-195069600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-179341200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-163620000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-147891600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-131565600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-116442000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-100116000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-84387600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-68666400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-52938000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-37216800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (41410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (57736800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (73465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (89186400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (136364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (152085600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (167814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (954658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (972799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -19052,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Canada__Mountain => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1998663968, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1632063600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1600614000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1596816000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1567954800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1551628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1536505200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1523203200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1504450800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1491753600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1473001200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1459699200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-715791600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-702489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (73472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (89193600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (104922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (120643200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (136371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (152092800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (167821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (183542400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (199270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (215596800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (230720400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (247046400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (262774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (278496000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (294224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (309945600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (325674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (341395200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (357123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (372844800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (388573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (404899200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (420022800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (436348800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (452077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (467798400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (483526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (499248000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (514976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (530697600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (544611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (562147200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (576061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (594201600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (607510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (625651200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (638960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (657100800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (671014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (688550400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (702464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (720000000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (733914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (752054400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (765363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (783504000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (796813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (814953600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (986115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1004256000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1173603600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1194163200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1205053200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225612800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1236502800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1257062400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1268557200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1289116800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1300006800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1320566400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1331456400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1352016000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1362906000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1383465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1394355600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414915200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1425805200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1446364800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1457859600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1478419200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1489309200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509868800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1520758800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1541318400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1552208400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572768000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1583658000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1604217600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1615712400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1636272000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1647162000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667721600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1678611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1699171200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1710061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1730620800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1741510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1762070400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1772960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1793520000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1805014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1825574400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1836464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1857024000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1867914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1888473600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1899363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1919923200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1930813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1951372800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1962867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1983427200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1994317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2014876800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2025766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2046326400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2057216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2077776000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2088666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2109225600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2120115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2140675200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2152170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2172729600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2183619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2204179200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2215069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2235628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2246518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2267078400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2277968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2298528000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2309418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2329977600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2341472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2362032000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2372922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2393481600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2404371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2424931200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2435821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2456380800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2467270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2487830400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2499325200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2519884800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2530774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2551334400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2562224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2582784000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2593674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2614233600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2625123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2645683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2656573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2677132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2688627600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2709187200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2720077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2740636800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2751526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2772086400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2782976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2803536000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2814426000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2834985600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2846480400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2867040000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2877930000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2898489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2909379600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2929939200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2940829200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2961388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2972278800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2992838400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3003728400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3024288000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3035782800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3056342400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3067232400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3087792000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3098682000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3119241600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3130131600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3150691200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3161581200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3182140800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3193030800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3213590400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3225085200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3245644800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3256534800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3277094400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3287984400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3308544000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3319434000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3339993600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3350883600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3371443200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3382938000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3403497600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3414387600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3434947200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3445837200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3466396800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3477286800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3497846400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3508736400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3529296000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3540186000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3560745600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3572240400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3592800000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3603690000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3624249600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3635139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3655699200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3666589200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3687148800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3698038800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3718598400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3730093200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3750652800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3761542800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3782102400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3792992400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3813552000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3824442000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3845001600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3855891600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3876451200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3887341200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3907900800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3919395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3939955200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3950845200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3971404800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3982294800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4002854400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4013744400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4034304000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4045194000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4065753600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4076643600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4097203200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -27232,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Canada__Newfoundland => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713897748, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1664130548, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1650137348, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1632076148, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1615145348, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1598650148, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1590100148, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1567286948, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1551565748, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1535837348, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1520116148, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1503782948, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1488666548, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1472333348, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1457216948, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1440883748, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1425767348, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1409434148, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1394317748, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1377984548, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1362263348, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1346534948, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1330813748, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1314480548, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1299364148, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1283030948, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1267914548, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1251581348, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1236464948, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1220131748, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1205015348, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1188682148, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1172960948, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1156627748, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1141511348, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1125178148, FixedTimespan { utc_offset: -12652, dst_offset: 3600, name: "NDT" }),
                    (-1110061748, FixedTimespan { utc_offset: -12652, dst_offset: 0, name: "NST" }),
                    (-1096921748, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-1093728600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-1078612200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-1061670600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-1048973400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-1030221000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-1017523800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-998771400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-986074200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-966717000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-954624600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-935267400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-922570200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-903817800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-891120600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-872368200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NWT" }),
                    (-769395600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NPT" }),
                    (-765401400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-746044200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-733347000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-714594600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-701897400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-683145000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-670447800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-651695400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-638998200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-619641000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-606943800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-589401000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-576099000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-557951400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-544649400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-526501800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-513199800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-495052200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-481750200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-463602600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-450300600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-431548200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-418246200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-400098600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-386796600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-368649000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-355347000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-337199400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-323897400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-305749800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-289423800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-273695400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-257974200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-242245800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-226524600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-210796200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-195075000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-179346600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-163625400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-147897000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-131571000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-116447400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-100121400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-84393000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-68671800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-52943400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-37222200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (-21493800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (-5772600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (9955800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (25677000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (41405400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (57731400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (73459800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (89181000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (104909400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (120630600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (136359000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (152080200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (167808600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (183529800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (199258200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (215584200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (230707800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (247033800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (262762200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (278483400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (294211800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (309933000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (325661400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (341382600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (357111000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (372832200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (388560600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (404886600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (420010200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (436336200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (452064600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (467785800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (483514200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (499235400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (514963800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (530685000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (544591860, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (562127460, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (576041460, FixedTimespan { utc_offset: -12600, dst_offset: 7200, name: "NDDT" }),
                    (594178260, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (607491060, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (625631460, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (638940660, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (657081060, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (670995060, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (688530660, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (702444660, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (719980260, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (733894260, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (752034660, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (765343860, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (783484260, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (796793460, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (814933860, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (828847860, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (846383460, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (860297460, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (877833060, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (891747060, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (909282660, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (923196660, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (941337060, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (954646260, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (972786660, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (986095860, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1004236260, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1018150260, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1035685860, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1049599860, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1067135460, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1081049460, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1099189860, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1112499060, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1130639460, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1143948660, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1162089060, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1173583860, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1194143460, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1205033460, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1225593060, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1236483060, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1257042660, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1268537460, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1289097060, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1299987060, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1320553800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1331443800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1352003400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1362893400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1383453000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1394343000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1414902600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1425792600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1446352200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1457847000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1478406600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1489296600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1509856200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1520746200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1541305800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1552195800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1572755400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1583645400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1604205000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1615699800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1636259400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1647149400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1667709000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1678599000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1699158600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1710048600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1730608200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1741498200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1762057800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1772947800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1793507400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1805002200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1825561800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1836451800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1857011400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1867901400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1888461000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1899351000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1919910600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1930800600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1951360200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1962855000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (1983414600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (1994304600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2014864200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2025754200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2046313800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2057203800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2077763400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2088653400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2109213000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2120103000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2140662600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2152157400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2172717000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2183607000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2204166600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2215056600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2235616200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2246506200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2267065800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2277955800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2298515400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2309405400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2329965000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2341459800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2362019400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2372909400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2393469000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2404359000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2424918600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2435808600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2456368200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2467258200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2487817800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2499312600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2519872200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2530762200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2551321800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2562211800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2582771400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2593661400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2614221000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2625111000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2645670600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2656560600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2677120200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2688615000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2709174600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2720064600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2740624200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2751514200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2772073800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2782963800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2803523400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2814413400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2834973000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2846467800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2867027400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2877917400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2898477000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2909367000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2929926600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2940816600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2961376200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (2972266200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (2992825800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3003715800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3024275400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3035770200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3056329800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3067219800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3087779400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3098669400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3119229000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3130119000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3150678600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3161568600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3182128200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3193018200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3213577800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3225072600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3245632200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3256522200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3277081800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3287971800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3308531400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3319421400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3339981000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3350871000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3371430600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3382925400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3403485000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3414375000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3434934600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3445824600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3466384200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3477274200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3497833800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3508723800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3529283400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3540173400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3560733000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3572227800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3592787400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3603677400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3624237000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3635127000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3655686600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3666576600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3687136200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3698026200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3718585800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3730080600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3750640200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3761530200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3782089800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3792979800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3813539400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3824429400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3844989000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3855879000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3876438600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3887328600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3907888200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3919383000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3939942600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3950832600, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (3971392200, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (3982282200, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (4002841800, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (4013731800, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (4034291400, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (4045181400, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (4065741000, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                    (4076631000, FixedTimespan { utc_offset: -12600, dst_offset: 3600, name: "NDT" }),
                    (4097190600, FixedTimespan { utc_offset: -12600, dst_offset: 0, name: "NST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -12652,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Canada__Pacific => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713880852, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1632060000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1615129200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-880207200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PWT" }),
                    (-769395600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PPT" }),
                    (-765385200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-747237600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-733935600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-715788000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-702486000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-684338400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-671036400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-652888800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-639586800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-620834400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-608137200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-589384800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-576082800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-557935200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-544633200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-526485600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-513183600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-495036000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-481734000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-463586400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-450284400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-431532000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-418230000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-400082400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-386780400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-368632800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-355330800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-337183200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-323881200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-305733600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-292431600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-273679200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-260982000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-242229600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-226508400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-210780000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-195058800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-179330400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-163609200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-147880800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-131554800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-116431200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-100105200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-84376800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-68655600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-52927200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-37206000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-21477600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-5756400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (9972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (25693200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (41421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (57747600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (73476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (89197200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (104925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (120646800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (136375200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (152096400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (167824800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (183546000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (199274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (215600400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (230724000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (247050000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (262778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (278499600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (294228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (309949200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (341398800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (452080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (467802000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (483530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (499251600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (514980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (530701200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (544615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (562150800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (576064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (594205200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (607514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (625654800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (638964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (657104400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (671018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (688554000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (702468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (720003600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (733917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (752058000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (765367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (783507600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (796816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (814957200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (828871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (846406800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (860320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (877856400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (891770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (909306000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (923220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (941360400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (954669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (972810000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (986119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1004259600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1018173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1035709200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1049623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1067158800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1081072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1099213200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1112522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1130662800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1143972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1162112400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1173607200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1194166800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1205056800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1225616400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1236506400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1257066000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1268560800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1289120400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1300010400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1320570000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1331460000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1352019600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1362909600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1383469200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1394359200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1414918800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1425808800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1446368400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1457863200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1478422800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1489312800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1509872400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1520762400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1541322000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1552212000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1572771600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1583661600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1604221200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1615716000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1636275600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1647165600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1667725200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1678615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1699174800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1710064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1730624400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1741514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1762074000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1772964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1793523600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1805018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1825578000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1836468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1857027600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1867917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1888477200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1899367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1919926800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1930816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1951376400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1962871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1983430800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1994320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2014880400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2025770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2046330000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2057220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2077779600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2088669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2109229200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2120119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2140678800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2152173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2172733200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2183623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2204182800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2215072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2235632400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2246522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2267082000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2277972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2298531600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2309421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2329981200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2341476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2362035600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2372925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2393485200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2404375200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2424934800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2435824800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2456384400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2467274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2487834000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2499328800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2519888400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2530778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2551338000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2562228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2582787600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2593677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2614237200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2625127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2645686800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2656576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2677136400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2688631200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2709190800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2720080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2740640400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2751530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2772090000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2782980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2803539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2814429600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2834989200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2846484000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2867043600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2877933600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2898493200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2909383200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2929942800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2940832800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2961392400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2972282400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2992842000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3003732000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3024291600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3035786400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3056346000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3067236000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3087795600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3098685600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3119245200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3130135200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3150694800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3161584800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3182144400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3193034400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3213594000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3225088800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3245648400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3256538400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3277098000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3287988000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3308547600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3319437600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3339997200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3350887200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3371446800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3382941600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3403501200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3414391200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3434950800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3445840800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3466400400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3477290400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3497850000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3508740000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3529299600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3540189600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3560749200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3572244000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3592803600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3603693600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3624253200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3635143200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3655702800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3666592800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3687152400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3698042400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3718602000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3730096800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3750656400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3761546400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3782106000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3792996000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3813555600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3824445600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3845005200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3855895200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3876454800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3887344800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3907904400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3919399200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3939958800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3950848800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3971408400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3982298400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4002858000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4013748000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4034307600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4045197600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4065757200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4076647200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4097206800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -29548,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Canada__Saskatchewan => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2030202084, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1632063600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1251651600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1238349600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1220202000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1206900000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1188752400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1175450400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1156698000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1144000800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1125248400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1111946400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1032714000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1016992800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1001264400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-986148000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-969814800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-954093600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-937760400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-922039200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-906310800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-890589600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-748450800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-732729600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-715791600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-702489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-684342000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-671040000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-652892400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-639590400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-620838000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-608140800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-589388400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-576086400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-557938800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-544636800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-526489200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-513187200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-495039600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-481737600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-463590000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-450288000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-431535600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-418233600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-400086000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-386784000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-337186800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-321465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-305737200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25116,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Canada__Yukon => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2188997988, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-1632056400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (-1615125600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-1596978000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YDT" }),
                    (-1583164800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-880203600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YWT" }),
                    (-769395600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "YPT" }),
                    (-765381600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-147884400, FixedTimespan { utc_offset: -32400, dst_offset: 7200, name: "YDDT" }),
                    (-131554800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (-121273200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (341398800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (452080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (467802000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (483530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (499251600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (514980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (530701200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (544615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (562150800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (576064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (594205200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (607514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (625654800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (638964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (657104400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (671018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (688554000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (702468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (720003600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (733917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (752058000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (765367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (783507600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (796816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (814957200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (828871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (846406800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (860320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (877856400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (891770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (909306000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (923220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (941360400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (954669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (972810000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (986119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1004259600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1018173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1035709200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1049623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1067158800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1081072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1099213200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1112522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1130662800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1143972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1162112400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1173607200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1194166800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1205056800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1225616400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1236506400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1257066000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1268560800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1289120400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1300010400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1320570000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1331460000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1352019600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1362909600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1383469200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1394359200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1414918800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1425808800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1446368400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1457863200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1478422800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1489312800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1509872400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1520762400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1541322000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1552212000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1572771600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1583661600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1604214000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -32412,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Chile__Continental => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524504635, FixedTimespan { utc_offset: -16965, dst_offset: 0, name: "SMT" }),
                    (-1892661435, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1688410800, FixedTimespan { utc_offset: -16965, dst_offset: 0, name: "SMT" }),
                    (-1619205435, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-1593806400, FixedTimespan { utc_offset: -16965, dst_offset: 0, name: "SMT" }),
                    (-1335986235, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1317585600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1304362800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1286049600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1272826800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1254513600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1241290800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1222977600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1209754800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-1191355200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-1178132400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-870552000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-865278000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-740520000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-736635600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "-04" }),
                    (-718056000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (-713649600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-36619200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (-23922000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (-3355200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (7527600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (24465600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (37767600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (55915200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (69217200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (87969600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (100666800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (118209600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (132116400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (150868800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (163566000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (182318400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (195620400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (213768000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (227070000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (245217600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (258519600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (277272000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (289969200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (308721600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (321418800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (340171200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (353473200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (371620800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (384922800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (403070400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (416372400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (434520000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (447822000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (466574400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (479271600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (498024000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (510721200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (529473600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (545194800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (560923200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (574225200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (592372800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (605674800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (624427200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (637124400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (653457600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (668574000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (687326400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (700628400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (718776000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (732078000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (750225600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (763527600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (781675200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (794977200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (813729600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (826426800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (845179200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (859690800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (876628800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (889930800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (906868800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (923194800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (939528000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (952830000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (971582400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (984279600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1003032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1015729200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1034481600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1047178800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1065931200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1079233200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1097380800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1110682800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1128830400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1142132400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1160884800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1173582000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1192334400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1206846000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1223784000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1237086000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1255233600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1270350000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1286683200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1304823600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1313899200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1335668400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1346558400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1367118000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1378612800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1398567600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1410062400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1463281200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1471147200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1494730800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1502596800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1526180400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1534046400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1554606000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1567915200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1586055600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1599364800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1617505200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1630814400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1648954800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1662868800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1680404400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1693713600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1712458800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1725768000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1743908400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1757217600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1775358000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1788667200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1806807600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1820116800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1838257200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1851566400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1870311600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1883016000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1901761200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1915070400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1933210800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1946520000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1964660400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (1977969600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (1996110000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2009419200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2027559600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2040868800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2059614000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2072318400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2091063600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2104372800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2122513200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2135822400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2153962800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2167272000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2185412400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2198721600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2217466800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2230171200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2248916400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2262225600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2280366000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2293675200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2311815600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2325124800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2343265200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2356574400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2374714800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2388024000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2406769200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2419473600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2438218800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2451528000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2469668400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2482977600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2501118000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2514427200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2532567600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2545876800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2564017200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2577326400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2596071600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2609380800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2627521200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2640830400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2658970800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2672280000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2690420400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2703729600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2721870000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2735179200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2753924400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2766628800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2785374000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2798683200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2816823600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2830132800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2848273200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2861582400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2879722800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2893032000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2911172400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2924481600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2943226800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2955931200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (2974676400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (2987985600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3006126000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3019435200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3037575600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3050884800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3069025200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3082334400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3101079600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3113784000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3132529200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3145838400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3163978800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3177288000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3195428400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3208737600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3226878000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3240187200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3258327600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3271636800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3290382000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3303086400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3321831600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3335140800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3353281200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3366590400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3384730800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3398040000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3416180400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3429489600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3447630000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3460939200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3479684400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3492993600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3511134000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3524443200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3542583600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3555892800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3574033200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3587342400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3605482800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3618792000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3637537200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3650241600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3668986800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3682296000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3700436400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3713745600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3731886000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3745195200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3763335600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3776644800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3794785200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3808094400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3826839600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3839544000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3858289200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3871598400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3889738800, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3903048000, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3921188400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3934497600, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3952638000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3965947200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (3984692400, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (3997396800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (4016142000, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (4029451200, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (4047591600, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (4060900800, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                    (4079041200, FixedTimespan { utc_offset: -14400, dst_offset: 0, name: "-04" }),
                    (4092350400, FixedTimespan { utc_offset: -14400, dst_offset: 3600, name: "-03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -16965,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Chile__EasterIsland => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524495352, FixedTimespan { utc_offset: -26248, dst_offset: 0, name: "EMT" }),
                    (-1178124152, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (-36619200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (-23922000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (-3355200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (7527600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (24465600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (37767600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (55915200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (69217200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (87969600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (100666800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (118209600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (132116400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (150868800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (163566000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (182318400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (195620400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (213768000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (227070000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (245217600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (258519600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (277272000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (289969200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (308721600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (321418800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (340171200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (353473200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (371620800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (384922800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (384948000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (403070400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (416372400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (434520000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (447822000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (466574400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (479271600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (498024000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (510721200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (529473600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (545194800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (560923200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (574225200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (592372800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (605674800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (624427200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (637124400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (653457600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (668574000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (687326400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (700628400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (718776000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (732078000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (750225600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (763527600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (781675200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (794977200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (813729600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (826426800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (845179200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (859690800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (876628800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (889930800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (906868800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (923194800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (939528000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (952830000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (971582400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (984279600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1003032000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1015729200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1034481600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1047178800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1065931200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1079233200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1097380800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1110682800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1128830400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1142132400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1160884800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1173582000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1192334400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1206846000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1223784000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1237086000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1255233600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1270350000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1286683200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1304823600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1313899200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1335668400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1346558400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1367118000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1378612800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1398567600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1410062400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1463281200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1471147200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1494730800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1502596800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1526180400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1534046400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1554606000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1567915200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1586055600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1599364800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1617505200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1630814400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1648954800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1662868800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1680404400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1693713600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1712458800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1725768000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1743908400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1757217600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1775358000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1788667200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1806807600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1820116800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1838257200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1851566400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1870311600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1883016000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1901761200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1915070400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1933210800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1946520000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1964660400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1977969600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1996110000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2009419200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2027559600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2040868800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2059614000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2072318400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2091063600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2104372800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2122513200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2135822400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2153962800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2167272000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2185412400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2198721600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2217466800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2230171200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2248916400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2262225600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2280366000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2293675200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2311815600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2325124800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2343265200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2356574400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2374714800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2388024000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2406769200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2419473600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2438218800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2451528000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2469668400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2482977600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2501118000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2514427200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2532567600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2545876800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2564017200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2577326400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2596071600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2609380800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2627521200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2640830400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2658970800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2672280000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2690420400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2703729600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2721870000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2735179200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2753924400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2766628800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2785374000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2798683200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2816823600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2830132800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2848273200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2861582400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2879722800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2893032000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2911172400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2924481600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2943226800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2955931200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2974676400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2987985600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3006126000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3019435200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3037575600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3050884800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3069025200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3082334400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3101079600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3113784000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3132529200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3145838400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3163978800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3177288000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3195428400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3208737600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3226878000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3240187200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3258327600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3271636800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3290382000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3303086400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3321831600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3335140800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3353281200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3366590400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3384730800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3398040000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3416180400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3429489600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3447630000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3460939200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3479684400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3492993600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3511134000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3524443200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3542583600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3555892800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3574033200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3587342400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3605482800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3618792000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3637537200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3650241600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3668986800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3682296000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3700436400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3713745600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3731886000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3745195200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3763335600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3776644800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3794785200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3808094400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3826839600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3839544000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3858289200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3871598400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3889738800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3903048000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3921188400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3934497600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3952638000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3965947200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3984692400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3997396800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (4016142000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (4029451200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (4047591600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (4060900800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (4079041200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (4092350400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -26248,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Cuba => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524501832, FixedTimespan { utc_offset: -19776, dst_offset: 0, name: "HMT" }),
                    (-1402813824, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-1311534000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-1300996800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-933534000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-925675200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-902084400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-893620800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-870030000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-862171200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-775681200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-767822400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-744231600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-736372800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-144702000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-134251200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-113425200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-102542400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-86295600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-72907200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-54154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-41457600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (-21495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (-5774400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (9954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (25675200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (41403600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (57729600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (73458000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (87364800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (104907600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (118900800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (136357200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (150436800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (167806800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (183528000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (199256400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (215582400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (230706000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (247032000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (263365200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (276667200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (290581200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (308721600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (322030800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (340171200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (358318800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (371620800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (389768400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (403070400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (421218000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (434520000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (452667600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (466574400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (484117200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (498024000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (511333200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (529473600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (542782800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (560923200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (574837200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (592372800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (606286800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (623822400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (638946000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (655876800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (671000400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (687330000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (702450000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (718779600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (733899600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (750229200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (765349200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (781678800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (796798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (813128400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (828853200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (844578000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (860302800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (876632400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (891147600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (909291600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (922597200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (941346000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (954651600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (972795600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (986101200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1004245200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1018155600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1035694800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1049605200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1067144400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1080450000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1162098000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1173589200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1193547600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1205643600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1224997200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1236488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1256446800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1268542800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1288501200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1300597200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1321160400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1333256400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1352005200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1362891600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1383454800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1394341200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1414904400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1425790800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1446354000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1457845200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1478408400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1489294800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1509858000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1520744400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1541307600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1552194000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1572757200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1583643600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1604206800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1615698000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1636261200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1647147600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1667710800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1678597200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1699160400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1710046800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1730610000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1741496400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1762059600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1772946000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1793509200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1805000400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1825563600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1836450000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1857013200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1867899600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1888462800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1899349200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1919912400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1930798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1951362000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1962853200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (1983416400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (1994302800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2014866000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2025752400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2046315600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2057202000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2077765200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2088651600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2109214800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2120101200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2140664400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2152155600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2172718800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2183605200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2204168400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2215054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2235618000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2246504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2267067600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2277954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2298517200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2309403600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2329966800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2341458000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2362021200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2372907600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2393470800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2404357200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2424920400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2435806800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2456370000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2467256400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2487819600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2499310800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2519874000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2530760400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2551323600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2562210000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2582773200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2593659600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2614222800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2625109200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2645672400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2656558800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2677122000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2688613200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2709176400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2720062800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2740626000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2751512400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2772075600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2782962000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2803525200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2814411600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2834974800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2846466000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2867029200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2877915600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2898478800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2909365200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2929928400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2940814800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2961378000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (2972264400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (2992827600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3003714000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3024277200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3035768400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3056331600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3067218000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3087781200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3098667600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3119230800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3130117200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3150680400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3161566800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3182130000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3193016400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3213579600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3225070800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3245634000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3256520400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3277083600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3287970000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3308533200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3319419600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3339982800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3350869200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3371432400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3382923600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3403486800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3414373200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3434936400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3445822800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3466386000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3477272400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3497835600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3508722000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3529285200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3540171600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3560734800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3572226000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3592789200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3603675600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3624238800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3635125200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3655688400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3666574800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3687138000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3698024400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3718587600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3730078800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3750642000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3761528400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3782091600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3792978000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3813541200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3824427600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3844990800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3855877200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3876440400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3887326800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3907890000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3919381200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3939944400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3950830800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (3971394000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (3982280400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (4002843600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (4013730000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (4034293200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (4045179600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (4065742800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                    (4076629200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "CDT" }),
                    (4097192400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -19768,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::EET => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (228877200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (243997200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (260326800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (276051600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (291776400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (307501200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (323830800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (338950800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (354675600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (370400400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (386125200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (401850000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (417574800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (433299600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (449024400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (465354000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (481078800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (496803600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (512528400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (528253200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (543978000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (559702800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (575427600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (591152400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (606877200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622602000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638326800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654656400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670381200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686106000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701830800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717555600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733280400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749005200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764730000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780454800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796179600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811904400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828234000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846378000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859683600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877827600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891133200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954032400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985482000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490490000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7200,
                        dst_offset: 0,
                        name: "EET",
                    },
                    rest: REST
                }
            },

            Tz::EST => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -18000,
                        dst_offset: 0,
                        name: "EST",
                    },
                    rest: REST
                }
            },

            Tz::EST5EDT => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1633280400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1615140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1601830800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1583690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-880218000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EWT" }),
                    (-769395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EPT" }),
                    (-765396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-84387600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-68666400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-52938000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-37216800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (41410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (57736800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (73465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (89186400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (126687600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (152085600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (162370800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (954658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (972799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -18000,
                        dst_offset: 0,
                        name: "EST",
                    },
                    rest: REST
                }
            },

            Tz::Egypt => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2185409109, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-929844000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-923108400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-906170400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-892868400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-875844000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-857790000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-844308000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-825822000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-812685600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-794199600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-779853600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-762663600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-399088800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-386650800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-368330400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-355114800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-336790800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-323654400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-305168400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-292032000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-273632400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-260496000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-242096400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-228960000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-210560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-197424000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-178938000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-165801600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-147402000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-134265600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-115866000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-102643200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-84330000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-71107200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-52707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-39484800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-21171600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-7948800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (10364400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (23587200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (41900400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (55123200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (73522800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (86745600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (105058800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (118281600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (136594800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (149817600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (168130800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (181353600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (199753200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (212976000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (231289200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (244512000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (262825200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (276048000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (294361200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (307584000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (325983600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (339206400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (357519600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (370742400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (396399600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (402278400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (426812400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (433814400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (452214000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (465436800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (483750000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (496972800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (515286000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (528508800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (546822000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (560044800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (578444400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (591667200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (610412400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (623203200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (641516400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654739200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (673052400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686275200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (704674800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717897600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (736210800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749433600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (767746800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780969600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (799020000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (812322000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (830469600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (843771600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (861919200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (875221200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (893368800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (906670800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (925423200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (938725200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (956872800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (970174800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (988322400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1001624400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1019772000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1033074000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1051221600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1064523600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1083276000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1096578000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1114725600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1128027600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1146175200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1158872400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1177624800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1189112400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1209074400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1219957200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1240524000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1250802000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1272578400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1281474000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1284069600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1285880400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1400191200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1403816400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1406844000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1411678800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1682632800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698354000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1714082400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1730408400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1745532000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761858000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1776981600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1793307600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1809036000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824757200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1840485600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856206800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1871935200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887656400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1903384800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919710800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1934834400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1951160400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1966888800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982610000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1998338400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014059600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2029788000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045509200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2061237600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2076958800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2092687200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2109013200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2124136800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140462800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2156191200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2171912400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2187640800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203362000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2219090400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234811600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2250540000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266866000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2281989600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2298315600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2313439200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329765200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2345493600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361214800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2376943200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392664400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2408392800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424114000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2439842400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2456168400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2471292000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487618000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2503346400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519067600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2534796000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550517200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2566245600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2581966800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2597695200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2614021200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2629144800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645470800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2660594400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676920400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2692648800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708370000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2724098400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2739819600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2755548000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771269200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2786997600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2803323600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2818447200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834773200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2850501600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866222800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2881951200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897672400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2913400800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929122000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2944850400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960571600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2976300000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992626000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3007749600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3024075600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3039804000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055525200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3071253600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3086974800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3102703200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118424400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3134152800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150478800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3165602400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181928400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3197052000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3213378000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3229106400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3244827600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3260556000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3292005600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307726800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3323455200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3354904800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3371230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3386959200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3418408800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3449858400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465579600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3481308000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3512757600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3529083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3544207200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3576261600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3591982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3607711200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3639160800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3654882000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3670610400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3702060000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3718386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3734114400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3749835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3765564000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3797013600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3828463200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3859912800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3876238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3891362400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3923416800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3954866400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3986316000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002037200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4017765600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4034091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4049215200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4080664800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7509,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Eire => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2821649679, FixedTimespan { utc_offset: -1521, dst_offset: 0, name: "DMT" }),
                    (-1691962479, FixedTimespan { utc_offset: -1521, dst_offset: 3600, name: "IST" }),
                    (-1680474879, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1680472800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1664143200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1650146400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1633903200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1617487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1601848800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1586037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1570399200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1552168800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1538344800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1522533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1507500000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1490565600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1473631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1460930400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1442786400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1428876000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1410732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1396216800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1379282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1364767200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1348437600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1333317600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1315778400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1301263200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1284328800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1269813600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1253484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1238364000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1221429600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1206914400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1189980000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1175464800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1159135200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1143410400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1126476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1111960800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1095631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1080511200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1063576800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1049061600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1032127200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1017612000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1001282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-986162400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-969228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-950479200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-942012000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-733359600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-719445600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-699490800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-684972000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-668037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-654732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-636588000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-622072800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-605743200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-590623200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-574293600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-558568800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-542239200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-527119200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-512604000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-496274400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-481154400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-464220000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-449704800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-432165600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-417650400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-401320800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-386200800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-369266400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-354751200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-337816800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-323301600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-306972000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-291852000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-276732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-257983200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-245282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-226533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-213228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-195084000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-182383200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-163634400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-150933600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-132184800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-119484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-100735200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-88034400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-68680800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-59004000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-37242000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (57722400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (69818400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (89172000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (101268000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (120621600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (132717600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (152071200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (164167200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (183520800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (196221600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (214970400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (227671200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (246420000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (259120800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (278474400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (290570400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (309924000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (322020000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (341373600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (372819600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (404269200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (435718800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (467773200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (499222800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (530672400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (562122000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (593571600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (625626000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (657075600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (688525200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (719974800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (751424400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (782874000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (814323600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -1521,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMT => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "GMT",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTPlus0 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "GMT",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTPlus1 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -3600,
                        dst_offset: 0,
                        name: "-01",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTPlus10 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -36000,
                        dst_offset: 0,
                        name: "-10",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTPlus11 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -39600,
                        dst_offset: 0,
                        name: "-11",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTPlus12 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -43200,
                        dst_offset: 0,
                        name: "-12",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTPlus2 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -7200,
                        dst_offset: 0,
                        name: "-02",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTPlus3 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -10800,
                        dst_offset: 0,
                        name: "-03",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTPlus4 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -14400,
                        dst_offset: 0,
                        name: "-04",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTPlus5 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -18000,
                        dst_offset: 0,
                        name: "-05",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTPlus6 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -21600,
                        dst_offset: 0,
                        name: "-06",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTPlus7 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25200,
                        dst_offset: 0,
                        name: "-07",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTPlus8 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -28800,
                        dst_offset: 0,
                        name: "-08",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTPlus9 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -32400,
                        dst_offset: 0,
                        name: "-09",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTMinus0 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "GMT",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTMinus1 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3600,
                        dst_offset: 0,
                        name: "+01",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTMinus10 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 36000,
                        dst_offset: 0,
                        name: "+10",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTMinus11 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 39600,
                        dst_offset: 0,
                        name: "+11",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTMinus12 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 43200,
                        dst_offset: 0,
                        name: "+12",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTMinus13 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 46800,
                        dst_offset: 0,
                        name: "+13",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTMinus14 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 50400,
                        dst_offset: 0,
                        name: "+14",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTMinus2 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7200,
                        dst_offset: 0,
                        name: "+02",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTMinus3 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 10800,
                        dst_offset: 0,
                        name: "+03",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTMinus4 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 14400,
                        dst_offset: 0,
                        name: "+04",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTMinus5 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 18000,
                        dst_offset: 0,
                        name: "+05",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTMinus6 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 21600,
                        dst_offset: 0,
                        name: "+06",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTMinus7 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 25200,
                        dst_offset: 0,
                        name: "+07",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTMinus8 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 28800,
                        dst_offset: 0,
                        name: "+08",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMTMinus9 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 32400,
                        dst_offset: 0,
                        name: "+09",
                    },
                    rest: REST
                }
            },

            Tz::Etc__GMT0 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "GMT",
                    },
                    rest: REST
                }
            },

            Tz::Etc__Greenwich => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "GMT",
                    },
                    rest: REST
                }
            },

            Tz::Etc__UCT => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "UTC",
                    },
                    rest: REST
                }
            },

            Tz::Etc__UTC => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "UTC",
                    },
                    rest: REST
                }
            },

            Tz::Etc__Universal => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "UTC",
                    },
                    rest: REST
                }
            },

            Tz::Etc__Zulu => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "UTC",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Amsterdam => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840141850, FixedTimespan { utc_offset: 1050, dst_offset: 0, name: "BMT" }),
                    (-2450995200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1740355200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693702800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1613829600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1604278800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1585530000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1574038800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1552266000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1539997200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1520557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1507510800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1490576400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1473642000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1459126800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1444006800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1427677200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1411952400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1396227600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1379293200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1364778000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1348448400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1333328400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1316394000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1301263200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1284328800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1269813600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1253484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1238364000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1221429600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1206914400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1191189600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1175464800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1160344800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1143410400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1127685600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1111960800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1096840800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1080511200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1063576800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1049061600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1033336800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1017612000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1002492000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-986162400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-969228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-950479200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-942012000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-934671600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-798073200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766623600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-745455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733273200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (228877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (243997200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (260326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (276051600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (291776400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (307501200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 1050,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Andorra => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177453164, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-733881600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 364,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Astrakhan => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1441249932, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (-1247540400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (354916800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (370724400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (386452800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (402260400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (417988800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (433796400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (449611200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (465343200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (481068000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (496792800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (512517600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (528242400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (543967200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (559692000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (575416800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (591141600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (606866400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (638319600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (654649200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (670374000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (701820000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (717548400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (733273200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (748998000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (764722800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (780447600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (796172400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (811897200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (828226800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (846370800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (859676400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (877820400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (891126000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (909270000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (922575600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (941324400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (954025200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (972774000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (985474800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1004223600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1017529200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1035673200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1048978800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1067122800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1080428400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1099177200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1111878000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1130626800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1143327600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1162076400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1174777200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1193526000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1206831600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1224975600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1238281200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1256425200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1269730800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1288479600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1301180400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1414274400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1459033200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 11532,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Athens => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2344642492, FixedTimespan { utc_offset: 5692, dst_offset: 0, name: "AMT" }),
                    (-1686101632, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1182996000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1178161200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-906861600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-904878000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844477200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828237600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812422800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-552362400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-541652400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (166485600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (186184800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (198028800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (213753600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (228873600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (244080000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (260323200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (275446800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (291798000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (307407600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (323388000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (338936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (354675600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (370400400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (386125200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (401850000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (417574800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (433299600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (449024400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (465354000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (481078800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (496803600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (512528400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (528253200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (543978000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (559702800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (575427600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (591152400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (606877200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622602000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638326800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654656400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670381200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686106000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701830800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717555600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733280400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749005200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764730000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780454800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796179600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811904400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828234000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846378000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859683600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877827600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891133200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954032400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985482000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490490000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 5692,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Belfast => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3852662325, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1691964000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1680472800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1664143200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1650146400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1633903200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1617487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1601848800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1586037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1570399200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1552168800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1538344800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1522533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1507500000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1490565600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1473631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1460930400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1442786400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1428876000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1410732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1396216800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1379282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1364767200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1348437600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1333317600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1315778400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1301263200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1284328800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1269813600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1253484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1238364000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1221429600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1206914400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1189980000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1175464800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1159135200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1143410400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1126476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1111960800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1095631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1080511200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1063576800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1049061600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1032127200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1017612000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1001282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-986162400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-969228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-950479200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-942012000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-904518000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-896050800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-875487600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-864601200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-844038000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-832546800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-812588400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-798073200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-781052400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-772066800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-764805600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-748476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-733356000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-719445600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-717030000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-706748400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-699487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-687996000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-668037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-654732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-636588000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-622072800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-605743200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-590623200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-574293600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-558568800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-542239200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-527119200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-512604000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-496274400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-481154400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-464220000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-449704800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-432165600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-417650400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-401320800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-386200800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-369266400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-354751200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-337816800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-323301600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-306972000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-291852000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-276732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-257983200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-245282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-226533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-213228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-195084000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-182383200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-163634400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-150933600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-132184800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-119484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-100735200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-88034400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-68680800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-59004000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-37242000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "BST" }),
                    (57718800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (69818400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (89172000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (101268000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (120621600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (132717600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (152071200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (164167200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (183520800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (196221600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (214970400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (227671200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (246420000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (259120800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (278474400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (290570400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (309924000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (322020000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (341373600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (354675600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (372819600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (386125200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (404269200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (417574800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (435718800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (449024400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (467773200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (481078800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (499222800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (512528400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (530672400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (543978000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (562122000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (575427600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (593571600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (606877200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (625626000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (638326800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (657075600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (670381200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (688525200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (701830800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (719974800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (733280400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (751424400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (764730000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (782874000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (796179600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (814323600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (828234000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (846378000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (859683600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (877827600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (891133200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (909277200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (922582800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (941331600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (954032400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (972781200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (985482000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1004230800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1017536400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1035680400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1048986000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1067130000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1080435600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1099184400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1111885200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1130634000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1143334800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1162083600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1174784400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1193533200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1206838800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1224982800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1238288400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1256432400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1269738000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1288486800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1301187600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1319936400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1332637200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1351386000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1364691600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1382835600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1396141200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1414285200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1427590800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1445734800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1459040400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1477789200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1490490000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1509238800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1521939600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1540688400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1553994000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1572138000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1585443600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1603587600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1616893200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1635642000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1648342800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1667091600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1679792400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1698541200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1711846800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1729990800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1743296400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1761440400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1774746000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1792890000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1806195600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1824944400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1837645200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1856394000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1869094800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1887843600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1901149200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1919293200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1932598800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1950742800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1964048400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1982797200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1995498000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2014246800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2026947600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2045696400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2058397200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2077146000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2090451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2108595600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2121901200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2140045200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2153350800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2172099600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2184800400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2203549200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2216250000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2234998800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2248304400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2266448400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2279754000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2297898000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2311203600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2329347600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2342653200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2361402000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2374102800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2392851600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2405552400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2424301200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2437606800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2455750800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2469056400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2487200400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2500506000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2519254800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2531955600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2550704400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2563405200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2582154000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2595459600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2613603600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2626909200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2645053200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2658358800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2676502800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2689808400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2708557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2721258000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2740006800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2752707600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2771456400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2784762000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2802906000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2816211600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2834355600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2847661200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2866410000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2879110800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2897859600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2910560400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2929309200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2942010000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2960758800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2974064400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2992208400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3005514000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3023658000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3036963600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3055712400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3068413200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3087162000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3099862800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3118611600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3131917200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3150061200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3163366800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3181510800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3194816400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3212960400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3226266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3245014800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3257715600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3276464400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3289165200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3307914000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3321219600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3339363600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3352669200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3370813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3384118800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3402867600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3415568400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3434317200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3447018000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3465766800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3479072400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3497216400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3510522000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3528666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3541971600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3560115600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3573421200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3592170000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3604870800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3623619600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3636320400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3655069200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3668374800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3686518800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3699824400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3717968400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3731274000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3750022800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3762723600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3781472400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3794173200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3812922000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3825622800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3844371600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3857677200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3875821200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3889126800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3907270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3920576400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3939325200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3952026000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3970774800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3983475600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4002224400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4015530000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4033674000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4046979600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4065123600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4078429200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4096573200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -75,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Belgrade => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713915320, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-905824800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-777942000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766627200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 4920,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Berlin => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2422054408, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693706400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-938905200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-776563200, FixedTimespan { utc_offset: 3600, dst_offset: 7200, name: "CEMT" }),
                    (-765936000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-761180400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-748479600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733273200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-717631200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-714610800, FixedTimespan { utc_offset: 3600, dst_offset: 7200, name: "CEMT" }),
                    (-710380800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701910000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-684975600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-654130800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-639010800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3208,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Bratislava => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3786829064, FixedTimespan { utc_offset: 3464, dst_offset: 0, name: "PMT" }),
                    (-2469401864, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693706400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-938905200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-765327600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-746578800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733359600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-728517600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (-721260000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-716425200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701910000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-684975600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-654217200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-639010800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (291776400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (307501200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3464,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Brussels => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840141850, FixedTimespan { utc_offset: 1050, dst_offset: 0, name: "BMT" }),
                    (-2450995200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1740355200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693702800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1613829600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1604278800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1585530000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1574038800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1552266000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1539997200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1520557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1507510800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1490576400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1473642000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1459126800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1444006800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1427677200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1411952400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1396227600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1379293200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1364778000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1348448400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1333328400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1316394000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1301263200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1284328800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1269813600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1253484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1238364000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1221429600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1206914400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1191189600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1175464800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1160344800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1143410400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1127685600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1111960800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1096840800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1080511200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1063576800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1049061600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1033336800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1017612000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1002492000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-986162400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-969228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-950479200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-942012000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-934671600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-798073200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766623600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-745455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733273200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (228877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (243997200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (260326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (276051600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (291776400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (307501200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 1050,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Bucharest => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2469404664, FixedTimespan { utc_offset: 6264, dst_offset: 0, name: "BMT" }),
                    (-1213148664, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1187056800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1175479200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1159754400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1144029600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1127700000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1111975200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1096250400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1080525600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1064800800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1049076000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1033351200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1017626400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1001901600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-986176800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-970452000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-954727200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (296604000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (307486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (323816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (338940000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (354672000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (370396800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (386121600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (401846400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (417571200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (433296000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (449020800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (465350400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (481075200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (496800000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (512524800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (528249600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (543974400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (559699200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (575424000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (591148800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (606873600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622598400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638323200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654652800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670370400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686095200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701820000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717544800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733269600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (748994400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764719200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796168800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828223200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859683600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877827600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891133200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954032400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985482000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490490000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 6264,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Budapest => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2498260580, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693706400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1600470000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1587250800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1569711600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1555196400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-906775200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-778471200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-762656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-749689200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733276800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-717634800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701910000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-686185200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-654130800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-639010800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-492656400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-481168800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-461199600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-449708400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-428540400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-418258800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-397090800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-386809200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323823600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338943600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354668400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370393200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386118000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401842800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417567600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433292400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 4580,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Busingen => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3675198848, FixedTimespan { utc_offset: 1786, dst_offset: 0, name: "BMT" }),
                    (-2385246586, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-904435200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-891129600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-872985600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-859680000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 2048,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Chisinau => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840147720, FixedTimespan { utc_offset: 6900, dst_offset: 0, name: "CMT" }),
                    (-1637114100, FixedTimespan { utc_offset: 6264, dst_offset: 0, name: "BMT" }),
                    (-1213148664, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1187056800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1175479200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1159754400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1144029600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1127700000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1111975200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1096250400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1080525600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1064800800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1049076000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1033351200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1017626400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1001901600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-986176800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-970452000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-954727200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-927165600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-898138800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-800157600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (354920400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (370728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (386456400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (402264000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (417992400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (433800000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (449614800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (465346800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (481071600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (496796400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (512521200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (528246000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (543970800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (559695600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (575420400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (606870000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (638319600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (641944800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654652800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670377600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686102400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701820000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733269600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (748990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764719200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796168800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828223200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859680000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877824000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891129600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909273600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922579200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941328000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954028800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972777600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985478400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004227200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017532800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035676800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048982400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067126400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080432000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099180800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111881600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130630400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143331200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162080000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174780800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193529600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206835200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224979200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238284800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256428800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269734400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288483200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301184000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319932800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332633600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351382400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364688000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382832000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396137600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414281600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427587200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445731200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459036800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477785600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490486400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509235200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521936000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540684800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553990400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572134400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585440000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603584000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616889600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635638400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648339200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667088000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679788800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698537600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711843200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729987200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743292800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761436800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774742400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792886400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806192000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824940800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837641600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856390400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869091200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887840000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901145600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919289600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932595200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950739200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964044800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982793600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995494400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014243200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026944000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045692800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058393600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077142400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090448000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108592000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121897600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140041600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153347200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172096000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184796800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203545600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216246400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234995200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248300800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266444800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279750400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297894400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311200000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329344000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342649600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361398400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374099200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392848000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405548800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424297600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437603200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455747200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469052800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487196800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500502400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519251200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531952000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550700800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563401600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582150400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595456000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613600000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626905600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645049600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658355200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676499200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689804800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708553600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721254400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740003200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752704000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771452800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784758400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802902400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816208000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834352000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847657600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866406400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879107200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897856000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910556800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929305600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942006400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960755200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974060800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992204800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005510400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023654400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036960000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055708800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068409600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087158400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099859200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118608000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131913600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150057600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163363200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181507200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194812800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212956800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226262400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245011200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257712000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276460800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289161600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307910400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321216000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339360000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352665600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370809600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384115200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402864000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415564800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434313600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447014400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465763200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479068800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497212800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510518400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528662400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541968000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560112000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573417600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592166400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604867200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623616000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636316800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655065600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668371200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686515200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699820800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717964800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731270400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750019200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762720000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781468800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794169600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812918400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825619200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844368000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857673600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875817600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889123200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907267200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920572800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939321600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952022400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970771200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983472000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002220800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015526400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033670400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046976000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065120000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078425600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096569600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 6920,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Copenhagen => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2422054408, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693706400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-938905200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-776563200, FixedTimespan { utc_offset: 3600, dst_offset: 7200, name: "CEMT" }),
                    (-765936000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-761180400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-748479600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733273200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-717631200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-714610800, FixedTimespan { utc_offset: 3600, dst_offset: 7200, name: "CEMT" }),
                    (-710380800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701910000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-684975600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-654130800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-639010800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3208,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Dublin => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2821649679, FixedTimespan { utc_offset: -1521, dst_offset: 0, name: "DMT" }),
                    (-1691962479, FixedTimespan { utc_offset: -1521, dst_offset: 3600, name: "IST" }),
                    (-1680474879, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1680472800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1664143200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1650146400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1633903200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1617487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1601848800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1586037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1570399200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1552168800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1538344800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1522533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1507500000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1490565600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1473631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1460930400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1442786400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1428876000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1410732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1396216800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1379282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1364767200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1348437600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1333317600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1315778400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1301263200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1284328800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1269813600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1253484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1238364000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1221429600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1206914400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1189980000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1175464800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1159135200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1143410400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1126476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1111960800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1095631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1080511200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1063576800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1049061600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1032127200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-1017612000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1001282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-986162400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-969228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-950479200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-942012000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-733359600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-719445600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-699490800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-684972000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-668037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-654732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-636588000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-622072800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-605743200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-590623200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-574293600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-558568800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-542239200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-527119200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-512604000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-496274400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-481154400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-464220000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-449704800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-432165600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-417650400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-401320800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-386200800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-369266400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-354751200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-337816800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-323301600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-306972000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-291852000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-276732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-257983200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-245282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-226533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-213228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-195084000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-182383200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-163634400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-150933600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-132184800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-119484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-100735200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-88034400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-68680800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-59004000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "IST" }),
                    (-37242000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (57722400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (69818400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (89172000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (101268000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (120621600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (132717600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (152071200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (164167200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (183520800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (196221600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (214970400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (227671200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (246420000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (259120800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (278474400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (290570400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (309924000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (322020000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (341373600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (372819600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (404269200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (435718800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (467773200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (499222800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (530672400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (562122000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (593571600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (625626000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (657075600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (688525200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (719974800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (751424400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (782874000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (814323600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "IST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -1521,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Gibraltar => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2821649916, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1691964000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1680472800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1664143200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1650146400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1633903200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1617487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1601848800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1586037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1570399200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1552168800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1538344800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1522533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1507500000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1490565600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1473631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1460930400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1442786400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1428876000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1410732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1396216800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1379282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1364767200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1348437600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1333317600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1315778400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1301263200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1284328800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1269813600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1253484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1238364000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1221429600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1206914400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1189980000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1175464800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1159135200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1143410400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1126476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1111960800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1095631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1080511200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1063576800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1049061600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1032127200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1017612000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1001282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-986162400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-969228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-950479200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-942012000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-904518000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-896050800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-875487600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-864601200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-844038000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-832546800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-812588400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-798073200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-781052400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-772066800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-764805600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-748476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-733356000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-719445600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-717030000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-706748400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-699487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-687996000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-668037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-654732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-636588000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-622072800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-605743200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-590623200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-574293600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-558568800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-542239200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-527119200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-512604000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-496274400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-481154400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-464220000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-449704800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-432165600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-417650400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-401320800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -1284,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Guernsey => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3852662325, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1691964000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1680472800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1664143200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1650146400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1633903200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1617487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1601848800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1586037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1570399200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1552168800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1538344800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1522533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1507500000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1490565600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1473631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1460930400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1442786400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1428876000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1410732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1396216800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1379282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1364767200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1348437600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1333317600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1315778400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1301263200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1284328800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1269813600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1253484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1238364000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1221429600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1206914400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1189980000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1175464800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1159135200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1143410400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1126476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1111960800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1095631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1080511200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1063576800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1049061600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1032127200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1017612000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1001282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-986162400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-969228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-950479200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-942012000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-904518000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-896050800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-875487600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-864601200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-844038000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-832546800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-812588400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-798073200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-781052400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-772066800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-764805600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-748476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-733356000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-719445600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-717030000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-706748400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-699487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-687996000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-668037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-654732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-636588000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-622072800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-605743200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-590623200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-574293600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-558568800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-542239200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-527119200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-512604000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-496274400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-481154400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-464220000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-449704800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-432165600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-417650400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-401320800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-386200800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-369266400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-354751200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-337816800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-323301600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-306972000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-291852000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-276732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-257983200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-245282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-226533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-213228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-195084000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-182383200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-163634400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-150933600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-132184800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-119484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-100735200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-88034400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-68680800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-59004000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-37242000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "BST" }),
                    (57718800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (69818400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (89172000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (101268000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (120621600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (132717600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (152071200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (164167200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (183520800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (196221600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (214970400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (227671200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (246420000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (259120800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (278474400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (290570400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (309924000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (322020000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (341373600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (354675600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (372819600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (386125200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (404269200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (417574800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (435718800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (449024400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (467773200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (481078800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (499222800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (512528400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (530672400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (543978000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (562122000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (575427600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (593571600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (606877200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (625626000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (638326800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (657075600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (670381200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (688525200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (701830800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (719974800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (733280400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (751424400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (764730000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (782874000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (796179600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (814323600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (828234000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (846378000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (859683600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (877827600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (891133200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (909277200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (922582800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (941331600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (954032400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (972781200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (985482000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1004230800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1017536400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1035680400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1048986000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1067130000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1080435600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1099184400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1111885200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1130634000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1143334800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1162083600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1174784400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1193533200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1206838800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1224982800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1238288400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1256432400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1269738000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1288486800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1301187600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1319936400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1332637200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1351386000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1364691600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1382835600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1396141200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1414285200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1427590800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1445734800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1459040400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1477789200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1490490000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1509238800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1521939600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1540688400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1553994000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1572138000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1585443600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1603587600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1616893200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1635642000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1648342800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1667091600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1679792400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1698541200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1711846800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1729990800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1743296400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1761440400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1774746000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1792890000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1806195600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1824944400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1837645200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1856394000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1869094800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1887843600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1901149200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1919293200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1932598800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1950742800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1964048400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1982797200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1995498000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2014246800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2026947600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2045696400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2058397200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2077146000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2090451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2108595600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2121901200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2140045200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2153350800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2172099600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2184800400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2203549200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2216250000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2234998800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2248304400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2266448400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2279754000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2297898000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2311203600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2329347600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2342653200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2361402000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2374102800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2392851600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2405552400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2424301200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2437606800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2455750800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2469056400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2487200400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2500506000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2519254800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2531955600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2550704400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2563405200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2582154000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2595459600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2613603600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2626909200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2645053200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2658358800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2676502800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2689808400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2708557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2721258000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2740006800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2752707600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2771456400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2784762000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2802906000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2816211600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2834355600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2847661200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2866410000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2879110800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2897859600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2910560400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2929309200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2942010000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2960758800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2974064400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2992208400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3005514000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3023658000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3036963600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3055712400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3068413200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3087162000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3099862800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3118611600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3131917200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3150061200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3163366800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3181510800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3194816400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3212960400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3226266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3245014800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3257715600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3276464400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3289165200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3307914000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3321219600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3339363600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3352669200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3370813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3384118800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3402867600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3415568400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3434317200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3447018000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3465766800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3479072400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3497216400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3510522000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3528666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3541971600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3560115600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3573421200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3592170000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3604870800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3623619600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3636320400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3655069200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3668374800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3686518800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3699824400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3717968400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3731274000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3750022800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3762723600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3781472400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3794173200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3812922000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3825622800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3844371600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3857677200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3875821200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3889126800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3907270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3920576400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3939325200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3952026000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3970774800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3983475600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4002224400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4015530000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4033674000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4046979600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4065123600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4078429200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4096573200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -75,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Helsinki => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2890258789, FixedTimespan { utc_offset: 5989, dst_offset: 0, name: "HMT" }),
                    (-1535938789, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-875671200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-859773600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (354672000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (370396800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (386121600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (401846400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (417574800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (433299600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (449024400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (465354000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (481078800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (496803600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (512528400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (528253200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (543978000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (559702800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (575427600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (591152400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (606877200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622602000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638326800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654656400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670381200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686106000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701830800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717555600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733280400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749005200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764730000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780454800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796179600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811904400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828234000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846378000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859683600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877827600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891133200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954032400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985482000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490490000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 5989,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Isle_of_Man => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3852662325, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1691964000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1680472800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1664143200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1650146400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1633903200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1617487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1601848800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1586037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1570399200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1552168800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1538344800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1522533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1507500000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1490565600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1473631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1460930400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1442786400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1428876000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1410732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1396216800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1379282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1364767200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1348437600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1333317600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1315778400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1301263200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1284328800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1269813600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1253484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1238364000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1221429600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1206914400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1189980000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1175464800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1159135200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1143410400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1126476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1111960800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1095631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1080511200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1063576800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1049061600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1032127200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1017612000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1001282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-986162400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-969228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-950479200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-942012000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-904518000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-896050800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-875487600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-864601200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-844038000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-832546800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-812588400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-798073200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-781052400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-772066800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-764805600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-748476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-733356000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-719445600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-717030000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-706748400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-699487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-687996000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-668037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-654732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-636588000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-622072800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-605743200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-590623200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-574293600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-558568800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-542239200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-527119200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-512604000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-496274400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-481154400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-464220000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-449704800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-432165600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-417650400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-401320800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-386200800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-369266400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-354751200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-337816800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-323301600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-306972000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-291852000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-276732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-257983200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-245282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-226533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-213228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-195084000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-182383200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-163634400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-150933600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-132184800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-119484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-100735200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-88034400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-68680800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-59004000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-37242000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "BST" }),
                    (57718800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (69818400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (89172000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (101268000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (120621600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (132717600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (152071200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (164167200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (183520800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (196221600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (214970400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (227671200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (246420000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (259120800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (278474400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (290570400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (309924000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (322020000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (341373600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (354675600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (372819600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (386125200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (404269200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (417574800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (435718800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (449024400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (467773200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (481078800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (499222800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (512528400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (530672400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (543978000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (562122000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (575427600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (593571600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (606877200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (625626000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (638326800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (657075600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (670381200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (688525200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (701830800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (719974800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (733280400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (751424400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (764730000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (782874000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (796179600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (814323600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (828234000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (846378000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (859683600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (877827600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (891133200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (909277200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (922582800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (941331600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (954032400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (972781200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (985482000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1004230800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1017536400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1035680400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1048986000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1067130000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1080435600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1099184400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1111885200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1130634000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1143334800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1162083600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1174784400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1193533200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1206838800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1224982800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1238288400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1256432400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1269738000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1288486800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1301187600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1319936400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1332637200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1351386000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1364691600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1382835600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1396141200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1414285200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1427590800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1445734800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1459040400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1477789200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1490490000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1509238800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1521939600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1540688400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1553994000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1572138000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1585443600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1603587600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1616893200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1635642000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1648342800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1667091600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1679792400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1698541200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1711846800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1729990800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1743296400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1761440400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1774746000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1792890000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1806195600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1824944400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1837645200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1856394000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1869094800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1887843600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1901149200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1919293200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1932598800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1950742800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1964048400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1982797200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1995498000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2014246800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2026947600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2045696400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2058397200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2077146000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2090451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2108595600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2121901200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2140045200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2153350800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2172099600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2184800400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2203549200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2216250000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2234998800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2248304400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2266448400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2279754000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2297898000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2311203600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2329347600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2342653200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2361402000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2374102800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2392851600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2405552400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2424301200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2437606800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2455750800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2469056400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2487200400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2500506000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2519254800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2531955600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2550704400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2563405200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2582154000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2595459600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2613603600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2626909200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2645053200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2658358800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2676502800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2689808400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2708557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2721258000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2740006800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2752707600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2771456400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2784762000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2802906000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2816211600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2834355600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2847661200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2866410000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2879110800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2897859600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2910560400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2929309200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2942010000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2960758800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2974064400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2992208400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3005514000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3023658000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3036963600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3055712400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3068413200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3087162000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3099862800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3118611600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3131917200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3150061200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3163366800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3181510800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3194816400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3212960400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3226266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3245014800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3257715600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3276464400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3289165200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3307914000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3321219600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3339363600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3352669200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3370813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3384118800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3402867600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3415568400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3434317200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3447018000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3465766800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3479072400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3497216400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3510522000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3528666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3541971600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3560115600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3573421200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3592170000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3604870800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3623619600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3636320400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3655069200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3668374800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3686518800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3699824400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3717968400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3731274000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3750022800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3762723600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3781472400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3794173200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3812922000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3825622800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3844371600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3857677200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3875821200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3889126800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3907270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3920576400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3939325200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3952026000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3970774800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3983475600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4002224400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4015530000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4033674000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4046979600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4065123600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4078429200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4096573200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -75,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Istanbul => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840147752, FixedTimespan { utc_offset: 7016, dst_offset: 0, name: "IMT" }),
                    (-1869875816, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1693706400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1680490800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1570413600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1552186800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1538359200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1522551600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1507514400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1490583600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1440208800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1428030000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1409709600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1396494000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-931053600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-922676400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-917834400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-892436400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-875844000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-764737200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-744343200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-733806000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-716436000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-701924400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-684986400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-670474800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-654141600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-639025200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-622087200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-606970800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-590032800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-575521200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-235620000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-194842800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-177732000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-165726000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (107910000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (121215600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (133920000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (152665200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (164678400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (184114800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (196214400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (215564400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (228873600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (245804400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (260323200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (267915600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (428454000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (433893600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (468111600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (482799600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (496710000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (512521200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (528246000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (543970800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (559695600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (575420400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (591145200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (606870000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622594800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638319600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654649200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670374000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686098800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701823600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717548400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733273200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (748998000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764118000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780447600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796172400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811897200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828226800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846370800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859676400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877820400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891126000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909270000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922575600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941324400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954025200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972774000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985474800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004223600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017529200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035673200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048978800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067122800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080428400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099177200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111878000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130626800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143327600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162076400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301266800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396220400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1446933600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1473195600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 6952,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Jersey => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3852662325, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1691964000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1680472800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1664143200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1650146400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1633903200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1617487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1601848800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1586037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1570399200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1552168800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1538344800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1522533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1507500000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1490565600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1473631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1460930400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1442786400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1428876000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1410732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1396216800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1379282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1364767200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1348437600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1333317600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1315778400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1301263200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1284328800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1269813600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1253484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1238364000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1221429600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1206914400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1189980000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1175464800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1159135200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1143410400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1126476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1111960800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1095631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1080511200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1063576800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1049061600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1032127200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1017612000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1001282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-986162400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-969228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-950479200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-942012000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-904518000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-896050800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-875487600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-864601200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-844038000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-832546800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-812588400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-798073200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-781052400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-772066800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-764805600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-748476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-733356000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-719445600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-717030000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-706748400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-699487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-687996000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-668037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-654732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-636588000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-622072800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-605743200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-590623200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-574293600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-558568800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-542239200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-527119200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-512604000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-496274400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-481154400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-464220000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-449704800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-432165600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-417650400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-401320800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-386200800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-369266400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-354751200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-337816800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-323301600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-306972000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-291852000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-276732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-257983200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-245282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-226533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-213228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-195084000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-182383200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-163634400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-150933600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-132184800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-119484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-100735200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-88034400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-68680800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-59004000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-37242000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "BST" }),
                    (57718800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (69818400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (89172000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (101268000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (120621600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (132717600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (152071200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (164167200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (183520800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (196221600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (214970400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (227671200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (246420000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (259120800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (278474400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (290570400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (309924000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (322020000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (341373600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (354675600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (372819600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (386125200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (404269200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (417574800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (435718800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (449024400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (467773200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (481078800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (499222800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (512528400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (530672400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (543978000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (562122000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (575427600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (593571600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (606877200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (625626000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (638326800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (657075600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (670381200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (688525200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (701830800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (719974800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (733280400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (751424400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (764730000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (782874000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (796179600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (814323600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (828234000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (846378000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (859683600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (877827600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (891133200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (909277200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (922582800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (941331600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (954032400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (972781200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (985482000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1004230800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1017536400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1035680400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1048986000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1067130000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1080435600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1099184400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1111885200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1130634000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1143334800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1162083600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1174784400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1193533200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1206838800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1224982800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1238288400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1256432400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1269738000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1288486800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1301187600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1319936400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1332637200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1351386000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1364691600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1382835600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1396141200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1414285200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1427590800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1445734800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1459040400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1477789200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1490490000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1509238800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1521939600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1540688400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1553994000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1572138000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1585443600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1603587600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1616893200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1635642000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1648342800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1667091600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1679792400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1698541200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1711846800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1729990800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1743296400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1761440400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1774746000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1792890000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1806195600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1824944400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1837645200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1856394000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1869094800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1887843600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1901149200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1919293200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1932598800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1950742800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1964048400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1982797200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1995498000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2014246800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2026947600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2045696400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2058397200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2077146000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2090451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2108595600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2121901200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2140045200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2153350800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2172099600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2184800400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2203549200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2216250000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2234998800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2248304400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2266448400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2279754000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2297898000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2311203600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2329347600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2342653200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2361402000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2374102800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2392851600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2405552400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2424301200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2437606800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2455750800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2469056400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2487200400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2500506000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2519254800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2531955600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2550704400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2563405200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2582154000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2595459600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2613603600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2626909200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2645053200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2658358800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2676502800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2689808400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2708557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2721258000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2740006800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2752707600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2771456400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2784762000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2802906000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2816211600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2834355600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2847661200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2866410000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2879110800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2897859600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2910560400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2929309200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2942010000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2960758800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2974064400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2992208400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3005514000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3023658000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3036963600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3055712400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3068413200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3087162000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3099862800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3118611600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3131917200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3150061200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3163366800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3181510800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3194816400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3212960400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3226266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3245014800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3257715600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3276464400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3289165200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3307914000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3321219600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3339363600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3352669200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3370813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3384118800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3402867600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3415568400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3434317200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3447018000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3465766800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3479072400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3497216400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3510522000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3528666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3541971600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3560115600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3573421200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3592170000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3604870800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3623619600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3636320400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3655069200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3668374800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3686518800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3699824400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3717968400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3731274000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3750022800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3762723600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3781472400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3794173200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3812922000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3825622800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3844371600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3857677200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3875821200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3889126800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3907270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3920576400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3939325200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3952026000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3970774800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3983475600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4002224400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4015530000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4033674000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4046979600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4065123600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4078429200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4096573200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -75,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Kaliningrad => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2422056120, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693706400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-938905200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-780372000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-778730400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-762663600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-749095200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (354920400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (370728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (386456400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (402264000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (417992400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (433800000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (449614800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (465346800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (481071600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (496796400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (512521200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (528246000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (543970800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (559695600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (575420400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (606870000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622598400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638323200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654652800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670377600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686102400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701827200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717552000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733276800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749001600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764726400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780451200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796176000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811900800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828230400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846374400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859680000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877824000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891129600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909273600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922579200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941328000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954028800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972777600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985478400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004227200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017532800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035676800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048982400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067126400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080432000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099180800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111881600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130630400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143331200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162080000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174780800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193529600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206835200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224979200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238284800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256428800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269734400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288483200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301184000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1414278000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 4920,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Kiev => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840148124, FixedTimespan { utc_offset: 7324, dst_offset: 0, name: "KMT" }),
                    (-1441159324, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1247536800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (-892522800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-825382800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (354920400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (370728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (386456400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (402264000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (417992400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (433800000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (449614800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (465346800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (481071600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (496796400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (512521200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (528246000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (543970800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (559695600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (575420400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (606870000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (638319600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (646783200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686102400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701827200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717552000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733276800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749001600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764726400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780451200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796176000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811900800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828230400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846378000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859683600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877827600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891133200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954032400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985482000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490490000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7324,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Kirov => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1593832728, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (-1247540400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (354916800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (370724400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (386452800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (402260400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (417988800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (433796400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (449611200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (465343200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (481068000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (496792800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (512517600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (528242400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (543967200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (559692000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (575416800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (591141600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (606866400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (638319600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (654649200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (670374000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (701820000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (717548400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (733273200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (748998000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (764722800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (780447600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (796172400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (811897200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (828226800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (846370800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (859676400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (877820400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (891126000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (909270000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (922575600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (941324400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (954025200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (972774000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (985474800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1004223600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1017529200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1035673200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1048978800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1067122800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1080428400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1099177200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1111878000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1130626800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1143327600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1162076400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1174777200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1193526000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1206831600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1224975600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1238281200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1256425200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1269730800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1288479600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1301180400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "MSK" }),
                    (1414274400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 11928,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Kyiv => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840148124, FixedTimespan { utc_offset: 7324, dst_offset: 0, name: "KMT" }),
                    (-1441159324, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1247536800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (-892522800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-825382800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (354920400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (370728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (386456400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (402264000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (417992400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (433800000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (449614800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (465346800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (481071600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (496796400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (512521200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (528246000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (543970800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (559695600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (575420400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (606870000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (638319600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (646783200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686102400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701827200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717552000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733276800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749001600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764726400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780451200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796176000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811900800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828230400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846378000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859683600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877827600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891133200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954032400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985482000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490490000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7324,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Lisbon => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830381795, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1689555600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1677801600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1667437200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1647738000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1635814800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1616202000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1604365200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1584666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1572742800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1553043600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1541206800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1521507600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1442451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1426813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1379293200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1364778000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1348448400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1333328400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1316394000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1301274000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1284339600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1269824400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1221440400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1206925200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1191200400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1175475600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1127696400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1111971600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1096851600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1080522000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1063587600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1049072400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1033347600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1017622800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1002502800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-986173200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-969238800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-950490000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-942022800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-922669200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-906944400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-891133200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-877309200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-873684000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "WEMT" }),
                    (-864007200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-857955600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-845859600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-842839200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "WEMT" }),
                    (-831348000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-825901200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-814410000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-810784800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "WEMT" }),
                    (-799898400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-794451600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-782960400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-779335200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "WEMT" }),
                    (-768448800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-763002000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-749091600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-733366800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-717631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-701906400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-686181600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-670456800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-654732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-639007200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-623282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-607557600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-591832800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-575503200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-559778400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-544053600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-528328800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-512604000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-496879200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-481154400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-465429600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-449704800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-433980000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-417650400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-401925600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-386200800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-370476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-354751200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-339026400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-323301600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-307576800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-291852000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-276127200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-260402400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-244677600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-228348000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-212623200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-196898400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-181173600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-165448800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-149724000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-133999200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-118274400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (212544000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (228268800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (243993600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (260323200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (276048000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (291772800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (307501200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (323222400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (338950800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (354675600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (370400400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (386125200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (401850000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (417578400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (433299600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (449024400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (465354000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (481078800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (496803600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (512528400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (528253200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (543978000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (559702800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (575427600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (591152400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (606877200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (622602000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (638326800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (654656400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (670381200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (686106000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (701830800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (717552000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828230400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (846378000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (859683600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (877827600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (891133200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (909277200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (922582800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (941331600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (954032400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (972781200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (985482000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1004230800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1017536400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1035680400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1048986000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1067130000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1080435600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1099184400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1111885200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1130634000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1143334800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1162083600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1174784400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1193533200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1206838800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1224982800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1238288400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1256432400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1269738000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1288486800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1301187600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1319936400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1332637200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1351386000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1364691600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1382835600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1396141200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1414285200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1427590800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1445734800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1459040400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1477789200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1490490000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1509238800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1521939600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1540688400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1553994000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1572138000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1585443600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1603587600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1616893200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1635642000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1648342800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1667091600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1679792400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1698541200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1711846800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1729990800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1743296400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1761440400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1774746000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1792890000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1806195600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1824944400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1837645200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1856394000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1869094800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1887843600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1901149200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1919293200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1932598800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1950742800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1964048400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1982797200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1995498000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2014246800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2026947600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2045696400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2058397200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2077146000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2090451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2108595600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2121901200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2140045200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2153350800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2172099600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2184800400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2203549200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2216250000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2234998800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2248304400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2266448400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2279754000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2297898000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2311203600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2329347600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2342653200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2361402000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2374102800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2392851600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2405552400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2424301200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2437606800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2455750800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2469056400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2487200400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2500506000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2519254800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2531955600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2550704400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2563405200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2582154000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2595459600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2613603600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2626909200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2645053200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2658358800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2676502800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2689808400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2708557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2721258000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2740006800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2752707600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2771456400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2784762000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2802906000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2816211600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2834355600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2847661200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2866410000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2879110800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2897859600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2910560400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2929309200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2942010000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2960758800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2974064400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2992208400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3005514000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3023658000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3036963600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3055712400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3068413200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3087162000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3099862800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3118611600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3131917200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3150061200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3163366800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3181510800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3194816400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3212960400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3226266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3245014800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3257715600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3276464400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3289165200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3307914000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3321219600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3339363600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3352669200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3370813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3384118800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3402867600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3415568400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3434317200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3447018000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3465766800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3479072400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3497216400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3510522000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3528666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3541971600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3560115600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3573421200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3592170000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3604870800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3623619600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3636320400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3655069200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3668374800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3686518800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3699824400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3717968400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3731274000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3750022800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3762723600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3781472400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3794173200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3812922000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3825622800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3844371600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3857677200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3875821200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3889126800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3907270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3920576400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3939325200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3952026000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3970774800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3983475600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4002224400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4015530000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4033674000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4046979600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4065123600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4078429200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4096573200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -2205,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Ljubljana => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713915320, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-905824800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-777942000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766627200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 4920,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__London => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3852662325, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1691964000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1680472800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1664143200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1650146400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1633903200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1617487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1601848800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1586037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1570399200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1552168800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1538344800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1522533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1507500000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1490565600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1473631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1460930400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1442786400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1428876000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1410732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1396216800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1379282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1364767200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1348437600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1333317600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1315778400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1301263200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1284328800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1269813600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1253484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1238364000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1221429600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1206914400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1189980000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1175464800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1159135200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1143410400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1126476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1111960800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1095631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1080511200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1063576800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1049061600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1032127200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1017612000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1001282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-986162400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-969228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-950479200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-942012000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-904518000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-896050800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-875487600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-864601200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-844038000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-832546800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-812588400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-798073200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-781052400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-772066800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-764805600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-748476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-733356000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-719445600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-717030000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-706748400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-699487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-687996000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-668037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-654732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-636588000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-622072800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-605743200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-590623200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-574293600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-558568800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-542239200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-527119200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-512604000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-496274400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-481154400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-464220000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-449704800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-432165600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-417650400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-401320800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-386200800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-369266400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-354751200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-337816800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-323301600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-306972000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-291852000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-276732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-257983200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-245282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-226533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-213228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-195084000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-182383200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-163634400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-150933600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-132184800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-119484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-100735200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-88034400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-68680800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-59004000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-37242000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "BST" }),
                    (57718800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (69818400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (89172000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (101268000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (120621600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (132717600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (152071200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (164167200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (183520800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (196221600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (214970400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (227671200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (246420000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (259120800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (278474400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (290570400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (309924000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (322020000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (341373600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (354675600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (372819600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (386125200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (404269200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (417574800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (435718800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (449024400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (467773200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (481078800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (499222800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (512528400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (530672400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (543978000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (562122000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (575427600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (593571600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (606877200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (625626000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (638326800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (657075600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (670381200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (688525200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (701830800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (719974800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (733280400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (751424400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (764730000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (782874000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (796179600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (814323600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (828234000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (846378000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (859683600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (877827600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (891133200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (909277200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (922582800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (941331600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (954032400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (972781200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (985482000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1004230800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1017536400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1035680400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1048986000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1067130000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1080435600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1099184400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1111885200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1130634000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1143334800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1162083600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1174784400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1193533200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1206838800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1224982800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1238288400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1256432400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1269738000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1288486800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1301187600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1319936400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1332637200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1351386000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1364691600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1382835600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1396141200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1414285200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1427590800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1445734800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1459040400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1477789200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1490490000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1509238800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1521939600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1540688400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1553994000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1572138000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1585443600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1603587600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1616893200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1635642000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1648342800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1667091600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1679792400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1698541200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1711846800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1729990800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1743296400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1761440400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1774746000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1792890000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1806195600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1824944400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1837645200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1856394000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1869094800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1887843600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1901149200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1919293200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1932598800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1950742800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1964048400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1982797200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1995498000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2014246800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2026947600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2045696400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2058397200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2077146000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2090451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2108595600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2121901200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2140045200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2153350800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2172099600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2184800400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2203549200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2216250000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2234998800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2248304400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2266448400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2279754000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2297898000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2311203600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2329347600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2342653200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2361402000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2374102800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2392851600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2405552400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2424301200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2437606800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2455750800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2469056400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2487200400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2500506000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2519254800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2531955600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2550704400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2563405200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2582154000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2595459600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2613603600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2626909200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2645053200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2658358800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2676502800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2689808400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2708557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2721258000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2740006800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2752707600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2771456400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2784762000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2802906000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2816211600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2834355600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2847661200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2866410000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2879110800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2897859600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2910560400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2929309200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2942010000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2960758800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2974064400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2992208400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3005514000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3023658000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3036963600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3055712400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3068413200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3087162000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3099862800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3118611600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3131917200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3150061200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3163366800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3181510800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3194816400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3212960400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3226266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3245014800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3257715600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3276464400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3289165200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3307914000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3321219600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3339363600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3352669200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3370813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3384118800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3402867600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3415568400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3434317200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3447018000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3465766800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3479072400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3497216400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3510522000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3528666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3541971600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3560115600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3573421200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3592170000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3604870800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3623619600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3636320400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3655069200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3668374800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3686518800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3699824400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3717968400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3731274000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3750022800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3762723600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3781472400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3794173200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3812922000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3825622800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3844371600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3857677200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3875821200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3889126800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3907270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3920576400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3939325200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3952026000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3970774800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3983475600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4002224400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4015530000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4033674000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4046979600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4065123600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4078429200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4096573200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -75,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Luxembourg => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840141850, FixedTimespan { utc_offset: 1050, dst_offset: 0, name: "BMT" }),
                    (-2450995200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1740355200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693702800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1613829600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1604278800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1585530000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1574038800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1552266000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1539997200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1520557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1507510800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1490576400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1473642000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1459126800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1444006800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1427677200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1411952400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1396227600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1379293200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1364778000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1348448400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1333328400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1316394000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1301263200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1284328800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1269813600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1253484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1238364000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1221429600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1206914400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1191189600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1175464800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1160344800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1143410400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1127685600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1111960800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1096840800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1080511200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1063576800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1049061600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1033336800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1017612000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1002492000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-986162400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-969228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-950479200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-942012000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-934671600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-798073200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766623600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-745455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733273200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (228877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (243997200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (260326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (276051600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (291776400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (307501200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 1050,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Madrid => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177451916, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1631926800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1616889600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1601168400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1585353600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1442451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1427673600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1379293200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1364774400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1348448400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1333324800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1316390400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1301270400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1284339600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1269820800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1026954000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1017619200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1001898000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-999482400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "WEMT" }),
                    (-986090400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-954115200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-940208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-873079200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-862621200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-842839200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828320400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-811389600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796870800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-779940000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-765421200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-748490400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733971600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-652327200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-639018000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (135122400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (150246000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (166572000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (181695600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (196812000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (212540400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (228866400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (243990000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (260326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (276051600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (291776400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (307501200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -884,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Malta => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2403478684, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1690765200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680487200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1664758800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1648951200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1635123600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1616896800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1604278800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1585533600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1571014800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1555293600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-932432400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812588400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-798073200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766717200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-750898800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733359600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-719456400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701917200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-689209200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-114051600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-103168800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-81997200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-71715600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-50547600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-40266000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-18493200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-8211600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (12956400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (23238000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (43801200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (54687600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (75855600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (86742000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (102380400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (118105200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (135730800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (148518000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (167187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (180489600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (198637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (211939200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (230086800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (243388800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (261536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (274838400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (292986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (306288000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323312400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338342400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3484,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Mariehamn => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2890258789, FixedTimespan { utc_offset: 5989, dst_offset: 0, name: "HMT" }),
                    (-1535938789, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-875671200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-859773600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (354672000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (370396800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (386121600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (401846400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (417574800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (433299600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (449024400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (465354000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (481078800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (496803600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (512528400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (528253200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (543978000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (559702800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (575427600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (591152400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (606877200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622602000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638326800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654656400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670381200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686106000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701830800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717555600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733280400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749005200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764730000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780454800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796179600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811904400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828234000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846378000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859683600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877827600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891133200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954032400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985482000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490490000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 5989,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Minsk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840147416, FixedTimespan { utc_offset: 6600, dst_offset: 0, name: "MMT" }),
                    (-1441158600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1247536800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (-899780400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-804650400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (354920400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (370728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (386456400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (402264000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (417992400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (433800000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (449614800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (465346800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (481071600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (496796400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (512521200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (528246000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (543970800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (559695600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (575420400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (606870000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (670374000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686102400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701827200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717552000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733276800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749001600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764726400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780451200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796176000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811900800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828230400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846374400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859680000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877824000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891129600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909273600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922579200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941328000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954028800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972777600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985478400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004227200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017532800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035676800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048982400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067126400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080432000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099180800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111881600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130630400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143331200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162080000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174780800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193529600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206835200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224979200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238284800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256428800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269734400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288483200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301184000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 6616,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Monaco => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2486592561, FixedTimespan { utc_offset: 561, dst_offset: 0, name: "PMT" }),
                    (-1855958961, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1689814800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1680397200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1665363600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1648342800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1635123600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1616893200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1604278800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1585443600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1574038800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1552266000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1539997200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1520557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1507510800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1490576400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1470618000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1459126800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1444006800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1427677200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1411952400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1396227600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1379293200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1364778000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1348448400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1333328400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1316394000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1301274000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1284339600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1269824400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1253494800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1238374800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1221440400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1206925200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1191200400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1175475600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1160355600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1143421200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1127696400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1111971600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1096851600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1080522000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1063587600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1049072400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1033347600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1017622800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1002502800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-986173200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-969238800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-950490000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-942012000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-932436000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-800071200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "WEMT" }),
                    (-796266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-781052400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "WEMT" }),
                    (-766623600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (196819200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (212540400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (228877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (243997200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (260326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (276051600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (291776400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (307501200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 561,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Moscow => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840149817, FixedTimespan { utc_offset: 9017, dst_offset: 0, name: "MMT" }),
                    (-1688265017, FixedTimespan { utc_offset: 9079, dst_offset: 0, name: "MMT" }),
                    (-1656819079, FixedTimespan { utc_offset: 9079, dst_offset: 3600, name: "MST" }),
                    (-1641353479, FixedTimespan { utc_offset: 9079, dst_offset: 0, name: "MMT" }),
                    (-1627965079, FixedTimespan { utc_offset: 9079, dst_offset: 7200, name: "MDST" }),
                    (-1618716679, FixedTimespan { utc_offset: 9079, dst_offset: 3600, name: "MST" }),
                    (-1596429079, FixedTimespan { utc_offset: 9079, dst_offset: 7200, name: "MDST" }),
                    (-1593837079, FixedTimespan { utc_offset: 10800, dst_offset: 7200, name: "MDST" }),
                    (-1593820800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (-1589860800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (-1542427200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (-1539493200, FixedTimespan { utc_offset: 10800, dst_offset: 7200, name: "+05" }),
                    (-1525323600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (-1522728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (-1491188400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1247536800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (354920400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (370728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (386456400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (402264000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (417992400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (433800000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (449614800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (465346800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (481071600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (496796400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (512521200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (528246000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (543970800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (559695600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (575420400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (606870000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (638319600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (654649200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (670374000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686102400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (695779200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (701823600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (717548400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (733273200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (748998000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (764722800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (780447600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (796172400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (811897200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (828226800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (846370800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (859676400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (877820400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (891126000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (909270000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (922575600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (941324400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (954025200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (972774000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (985474800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1004223600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1017529200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1035673200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1048978800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1067122800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1080428400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1099177200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1111878000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1130626800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1143327600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1162076400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1174777200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1193526000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1206831600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1224975600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1238281200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1256425200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1269730800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1288479600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1301180400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "MSK" }),
                    (1414274400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 9017,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Nicosia => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1518920008, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (166572000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (182293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (200959200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (213829200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (228866400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (243982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (260316000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (276123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (291765600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (307486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (323820000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (338936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (354664800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (370386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (386114400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (401835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (417564000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (433285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (449013600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (465339600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (481068000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (496789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (512517600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (528238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (543967200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (559688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (575416800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (591138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (606866400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638316000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670370400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701820000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733269600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (748990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764719200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796168800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828223200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (843944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859672800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (875394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891122400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954032400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985482000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490490000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8008,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Oslo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2422054408, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693706400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-938905200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-776563200, FixedTimespan { utc_offset: 3600, dst_offset: 7200, name: "CEMT" }),
                    (-765936000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-761180400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-748479600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733273200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-717631200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-714610800, FixedTimespan { utc_offset: 3600, dst_offset: 7200, name: "CEMT" }),
                    (-710380800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701910000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-684975600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-654130800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-639010800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3208,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Paris => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2486592561, FixedTimespan { utc_offset: 561, dst_offset: 0, name: "PMT" }),
                    (-1855958961, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1689814800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1680397200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1665363600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1648342800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1635123600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1616893200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1604278800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1585443600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1574038800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1552266000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1539997200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1520557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1507510800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1490576400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1470618000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1459126800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1444006800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1427677200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1411952400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1396227600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1379293200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1364778000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1348448400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1333328400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1316394000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1301274000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1284339600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1269824400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1253494800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1238374800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1221440400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1206925200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1191200400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1175475600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1160355600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1143421200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1127696400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1111971600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1096851600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1080522000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1063587600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1049072400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1033347600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1017622800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1002502800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-986173200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-969238800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-950490000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-942012000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-932436000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-800071200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "WEMT" }),
                    (-796266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-781052400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "WEMT" }),
                    (-766623600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (196819200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (212540400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (228877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (243997200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (260326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (276051600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (291776400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (307501200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 561,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Podgorica => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713915320, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-905824800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-777942000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766627200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 4920,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Prague => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3786829064, FixedTimespan { utc_offset: 3464, dst_offset: 0, name: "PMT" }),
                    (-2469401864, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693706400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-938905200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-765327600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-746578800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733359600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-728517600, FixedTimespan { utc_offset: 3600, dst_offset: -3600, name: "GMT" }),
                    (-721260000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-716425200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701910000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-684975600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-654217200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-639010800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (291776400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (307501200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3464,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Riga => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840146594, FixedTimespan { utc_offset: 5794, dst_offset: 0, name: "RMT" }),
                    (-1632008194, FixedTimespan { utc_offset: 5794, dst_offset: 3600, name: "LST" }),
                    (-1618702594, FixedTimespan { utc_offset: 5794, dst_offset: 0, name: "RMT" }),
                    (-1601681794, FixedTimespan { utc_offset: 5794, dst_offset: 3600, name: "LST" }),
                    (-1597275394, FixedTimespan { utc_offset: 5794, dst_offset: 0, name: "RMT" }),
                    (-1377308194, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-928029600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (-899521200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-795834000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (354920400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (370728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (386456400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (402264000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (417992400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (433800000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (449614800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (465346800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (481071600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (496796400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (512521200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (528246000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (543970800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (559695600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (575420400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (606870000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622598400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638323200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654652800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670377600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686102400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701827200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717552000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733276800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749001600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764726400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780451200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796176000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811900800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828230400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (843955200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859683600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877827600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891133200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985482000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490490000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 5794,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Rome => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3252098996, FixedTimespan { utc_offset: 2996, dst_offset: 0, name: "RMT" }),
                    (-2403568196, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1690765200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680487200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1664758800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1648951200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1635123600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1616896800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1604278800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1585533600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1571014800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1555293600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-932432400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-798073200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766717200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-750898800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733359600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-719456400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701917200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-689209200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-114051600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-103168800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-81997200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-71715600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-50547600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-40266000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-18493200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-8211600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (12956400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (23238000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (43801200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (54687600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (75855600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (86742000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (107910000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (118191600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (138754800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (149641200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (170809200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (181090800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (202258800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (212540400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (233103600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (243990000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (265158000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (276044400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (296607600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (307494000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 2996,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Samara => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1593832820, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (-1247540400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (354916800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (370724400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (386452800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (402260400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (417988800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (433796400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (449611200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (465343200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (481068000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (496792800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (512517600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (528242400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (543967200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (559692000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (575416800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (591141600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (606866400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (638319600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (654649200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (670374000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "+03" }),
                    (686098800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (687916800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (701820000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (717544800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (733269600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (748994400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (764719200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (780444000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (796168800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (811893600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (828223200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (846367200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (859672800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (877816800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (891122400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (909266400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (922572000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (941320800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (954021600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (972770400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (985471200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1004220000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1017525600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1035669600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1048975200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1067119200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1080424800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1099173600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1111874400, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1130623200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1143324000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1162072800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1174773600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1193522400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1206828000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1224972000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1238277600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1256421600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1269727200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1288479600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1301180400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 12020,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__San_Marino => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3252098996, FixedTimespan { utc_offset: 2996, dst_offset: 0, name: "RMT" }),
                    (-2403568196, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1690765200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680487200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1664758800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1648951200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1635123600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1616896800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1604278800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1585533600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1571014800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1555293600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-932432400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-798073200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766717200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-750898800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733359600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-719456400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701917200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-689209200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-114051600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-103168800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-81997200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-71715600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-50547600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-40266000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-18493200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-8211600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (12956400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (23238000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (43801200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (54687600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (75855600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (86742000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (107910000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (118191600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (138754800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (149641200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (170809200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (181090800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (202258800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (212540400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (233103600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (243990000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (265158000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (276044400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (296607600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (307494000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 2996,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Sarajevo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713915320, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-905824800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-777942000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766627200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 4920,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Saratov => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1593831858, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (-1247540400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (354916800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (370724400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (386452800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (402260400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (417988800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (433796400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (449611200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (465343200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (481068000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (496792800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (512517600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (528242400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (543967200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (559692000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (575416800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (606870000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (638319600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (654649200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (670374000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (701820000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (717548400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (733273200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (748998000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (764722800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (780447600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (796172400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (811897200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (828226800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (846370800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (859676400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (877820400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (891126000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (909270000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (922575600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (941324400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (954025200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (972774000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (985474800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1004223600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1017529200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1035673200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1048978800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1067122800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1080428400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1099177200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1111878000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1130626800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1143327600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1162076400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1174777200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1193526000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1206831600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1224975600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1238281200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1256425200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1269730800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1288479600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1301180400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1414274400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1480806000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 11058,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Simferopol => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840148984, FixedTimespan { utc_offset: 8160, dst_offset: 0, name: "SMT" }),
                    (-1441160160, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1247536800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (-888894000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-811648800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (354920400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (370728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (386456400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (402264000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (417992400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (433800000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (449614800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (465346800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (481071600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (496796400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (512521200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (528246000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (543970800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (559695600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (575420400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (606870000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (646786800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701827200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717552000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733276800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749001600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764726400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (767739600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (780447600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (796172400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (811897200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (828219600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (846370800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (859672800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859683600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877827600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891133200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954032400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985482000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396137600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "MSK" }),
                    (1414274400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8184,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Skopje => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713915320, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-905824800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-777942000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766627200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 4920,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Sofia => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840146396, FixedTimespan { utc_offset: 7016, dst_offset: 0, name: "IMT" }),
                    (-2369527016, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781048800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (291762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (307576800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (323816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (339026400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (355266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (370393200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (386715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (401846400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (417571200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (433296000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (449020800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (465350400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (481075200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (496800000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (512524800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (528249600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (543974400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (559699200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (575424000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (591148800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (606873600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622598400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638323200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654652800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670370400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701820000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733269600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (748990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764719200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796168800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828223200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859683600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877827600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891133200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954032400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985482000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490490000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 5596,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Stockholm => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2422054408, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693706400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-938905200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-776563200, FixedTimespan { utc_offset: 3600, dst_offset: 7200, name: "CEMT" }),
                    (-765936000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-761180400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-748479600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733273200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-717631200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-714610800, FixedTimespan { utc_offset: 3600, dst_offset: 7200, name: "CEMT" }),
                    (-710380800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701910000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-684975600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-654130800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-639010800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3208,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Tallinn => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840146740, FixedTimespan { utc_offset: 5940, dst_offset: 0, name: "TMT" }),
                    (-1638322740, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1593824400, FixedTimespan { utc_offset: 5940, dst_offset: 0, name: "TMT" }),
                    (-1535938740, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-927943200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (-892954800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-797652000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (354920400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (370728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (386456400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (402264000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (417992400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (433800000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (449614800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (465346800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (481071600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (496796400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (512521200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (528246000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (543970800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (559695600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (575420400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (606870000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622598400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638323200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654652800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670377600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686102400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701827200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717552000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733276800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749001600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764726400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780451200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796176000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811900800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828230400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846374400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859680000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877824000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891129600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490490000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 5940,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Tirane => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1767230360, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-932346000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-843519600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (136854000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (149896800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (168130800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (181432800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (199839600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (213141600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (231894000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (244591200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (263257200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (276040800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (294706800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (307490400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (326156400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (339458400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (357087600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370389600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (389142000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (402444000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (419468400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433807200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449622000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 4760,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Tiraspol => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840147720, FixedTimespan { utc_offset: 6900, dst_offset: 0, name: "CMT" }),
                    (-1637114100, FixedTimespan { utc_offset: 6264, dst_offset: 0, name: "BMT" }),
                    (-1213148664, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1187056800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1175479200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1159754400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1144029600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1127700000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1111975200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1096250400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1080525600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1064800800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1049076000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1033351200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1017626400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1001901600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-986176800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-970452000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-954727200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-927165600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-898138800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-800157600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (354920400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (370728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (386456400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (402264000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (417992400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (433800000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (449614800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (465346800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (481071600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (496796400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (512521200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (528246000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (543970800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (559695600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (575420400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (606870000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (638319600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (641944800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654652800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670377600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686102400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701820000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733269600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (748990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764719200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796168800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828223200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859680000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877824000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891129600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909273600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922579200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941328000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954028800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972777600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985478400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004227200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017532800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035676800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048982400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067126400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080432000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099180800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111881600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130630400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143331200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162080000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174780800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193529600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206835200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224979200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238284800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256428800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269734400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288483200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301184000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319932800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332633600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351382400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364688000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382832000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396137600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414281600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427587200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445731200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459036800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477785600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490486400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509235200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521936000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540684800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553990400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572134400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585440000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603584000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616889600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635638400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648339200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667088000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679788800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698537600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711843200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729987200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743292800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761436800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774742400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792886400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806192000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824940800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837641600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856390400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869091200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887840000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901145600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919289600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932595200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950739200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964044800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982793600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995494400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014243200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026944000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045692800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058393600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077142400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090448000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108592000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121897600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140041600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153347200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172096000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184796800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203545600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216246400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234995200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248300800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266444800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279750400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297894400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311200000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329344000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342649600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361398400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374099200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392848000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405548800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424297600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437603200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455747200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469052800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487196800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500502400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519251200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531952000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550700800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563401600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582150400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595456000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613600000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626905600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645049600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658355200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676499200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689804800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708553600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721254400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740003200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752704000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771452800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784758400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802902400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816208000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834352000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847657600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866406400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879107200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897856000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910556800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929305600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942006400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960755200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974060800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992204800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005510400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023654400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036960000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055708800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068409600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087158400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099859200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118608000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131913600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150057600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163363200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181507200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194812800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212956800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226262400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245011200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257712000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276460800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289161600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307910400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321216000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339360000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352665600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370809600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384115200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402864000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415564800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434313600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447014400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465763200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479068800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497212800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510518400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528662400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541968000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560112000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573417600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592166400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604867200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623616000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636316800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655065600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668371200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686515200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699820800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717964800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731270400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750019200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762720000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781468800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794169600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812918400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825619200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844368000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857673600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875817600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889123200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907267200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920572800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939321600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952022400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970771200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983472000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002220800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015526400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033670400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046976000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065120000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078425600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096569600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 6920,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Ulyanovsk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1593832416, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (-1247540400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (354916800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (370724400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (386452800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (402260400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (417988800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (433796400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (449611200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (465343200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (481068000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (496792800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (512517600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (528242400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (543967200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (559692000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (575416800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (591141600, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (606866400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (638319600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (654649200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (670374000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "+03" }),
                    (686102400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "+02" }),
                    (695779200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (701823600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (717548400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (733273200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (748998000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (764722800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (780447600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (796172400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (811897200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (828226800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (846370800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (859676400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (877820400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (891126000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (909270000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (922575600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (941324400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (954025200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (972774000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (985474800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1004223600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1017529200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1035673200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1048978800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1067122800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1080428400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1099177200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1111878000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1130626800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1143327600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1162076400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1174777200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1193526000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1206831600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1224975600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1238281200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1256425200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1269730800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (1288479600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1301180400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1414274400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (1459033200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 11616,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Uzhgorod => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840148124, FixedTimespan { utc_offset: 7324, dst_offset: 0, name: "KMT" }),
                    (-1441159324, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1247536800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (-892522800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-825382800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (354920400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (370728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (386456400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (402264000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (417992400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (433800000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (449614800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (465346800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (481071600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (496796400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (512521200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (528246000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (543970800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (559695600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (575420400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (606870000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (638319600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (646783200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686102400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701827200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717552000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733276800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749001600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764726400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780451200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796176000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811900800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828230400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846378000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859683600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877827600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891133200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954032400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985482000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490490000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7324,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Vaduz => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3675198848, FixedTimespan { utc_offset: 1786, dst_offset: 0, name: "BMT" }),
                    (-2385246586, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-904435200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-891129600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-872985600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-859680000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 2048,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Vatican => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3252098996, FixedTimespan { utc_offset: 2996, dst_offset: 0, name: "RMT" }),
                    (-2403568196, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1690765200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680487200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1664758800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1648951200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1635123600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1616896800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1604278800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1585533600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1571014800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1555293600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-932432400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-798073200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766717200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-750898800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733359600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-719456400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701917200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-689209200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-114051600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-103168800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-81997200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-71715600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-50547600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-40266000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-18493200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-8211600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (12956400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (23238000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (43801200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (54687600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (75855600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (86742000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (107910000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (118191600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (138754800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (149641200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (170809200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (181090800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (202258800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (212540400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (233103600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (243990000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (265158000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (276044400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (296607600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (307494000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 2996,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Vienna => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2422055121, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693706400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1569711600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1555801200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-938905200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-780192000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-748479600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733273200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-717634800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701910000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-684975600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323823600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338940000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3921,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Vilnius => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840146876, FixedTimespan { utc_offset: 5040, dst_offset: 0, name: "WMT" }),
                    (-1672536240, FixedTimespan { utc_offset: 5736, dst_offset: 0, name: "KMT" }),
                    (-1585100136, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1561251600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1553565600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-928198800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (-900126000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-802144800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (354920400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (370728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (386456400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (402264000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (417992400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (433800000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (449614800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (465346800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (481071600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (496796400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (512521200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (528246000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (543970800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (559695600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (575420400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (606870000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622598400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638323200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654652800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670377600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686102400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701827200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717552000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733276800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749001600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764726400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780451200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796176000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811900800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828230400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846374400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859680000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877824000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891126000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941324400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490490000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 6076,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Volgograd => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577761060, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (-1247540400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (354916800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (370724400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (386452800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (402260400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (417988800, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (433796400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (449611200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (465343200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (481068000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (496792800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (512517600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (528242400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (543967200, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (559692000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (575416800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (606870000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (638319600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (654649200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (670374000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (701820000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (717548400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (733273200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (748998000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (764722800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (780447600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (796172400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (811897200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (828226800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (846370800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (859676400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (877820400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (891126000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (909270000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (922575600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (941324400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (954025200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (972774000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (985474800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1004223600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1017529200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1035673200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1048978800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1067122800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1080428400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1099177200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1111878000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1130626800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1143327600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1162076400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1174777200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1193526000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1206831600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1224975600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1238281200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1256425200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1269730800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1288479600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1301180400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "MSK" }),
                    (1414274400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1540681200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1609020000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 10660,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Warsaw => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840145840, FixedTimespan { utc_offset: 5040, dst_offset: 0, name: "WMT" }),
                    (-1717032240, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693706400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1600473600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1587168000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1501725600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-931734000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796608000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-778726800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-762660000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-748486800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733273200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-715215600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701910000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-684975600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-654130800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-639010800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-397094400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-386812800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-371088000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-355363200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-334195200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-323308800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-307584000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-291859200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-271296000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-260409600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-239846400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-228960000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-208396800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-197510400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-176342400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-166060800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (228873600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (243993600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (260323200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (276048000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (291772800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (307497600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323827200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338947200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354672000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370396800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386121600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401846400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417571200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433296000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449020800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465350400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481075200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496800000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512524800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528249600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543974400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559699200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 5040,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Zagreb => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2713915320, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-905824800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-777942000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-766627200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 4920,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Zaporozhye => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840148124, FixedTimespan { utc_offset: 7324, dst_offset: 0, name: "KMT" }),
                    (-1441159324, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1247536800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (-892522800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-825382800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (354920400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (370728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (386456400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (402264000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (417992400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (433800000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (449614800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (465346800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (481071600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (496796400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (512521200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (528246000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (543970800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (559695600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (575420400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (606870000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (638319600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (646783200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686102400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701827200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717552000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733276800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (749001600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764726400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780451200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796176000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811900800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828230400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846378000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859683600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877827600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891133200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909277200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922582800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941331600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954032400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972781200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985482000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004230800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017536400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035680400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048986000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067130000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080435600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111885200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130634000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143334800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162083600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301187600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396141200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1445734800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1477789200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1490490000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1509238800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1521939600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1540688400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1553994000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1572138000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1585443600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1603587600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1616893200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1635642000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1648342800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1667091600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1679792400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1698541200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1711846800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1729990800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1743296400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1761440400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1774746000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1792890000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1806195600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1824944400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1837645200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1856394000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1869094800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1887843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1901149200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1919293200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1932598800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1950742800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1964048400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1982797200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1995498000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2014246800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2026947600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2045696400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2058397200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2077146000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2090451600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2108595600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2121901200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2140045200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2153350800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2172099600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2184800400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2203549200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2216250000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2234998800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2248304400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2266448400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2279754000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2297898000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2311203600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2329347600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2342653200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2361402000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2374102800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2392851600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2405552400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2424301200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2437606800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2455750800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2469056400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2487200400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2500506000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2519254800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2531955600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2550704400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2563405200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2582154000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2595459600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2613603600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2626909200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2645053200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2658358800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2676502800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2689808400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2708557200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2721258000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2740006800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2752707600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2771456400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2784762000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2802906000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2816211600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2834355600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2847661200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2866410000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2879110800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2897859600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2910560400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2929309200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2942010000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2960758800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (2974064400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (2992208400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3005514000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3023658000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3036963600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3055712400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3068413200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3087162000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3099862800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3118611600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3131917200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3150061200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3163366800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3181510800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3194816400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3212960400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3226266000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3245014800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3257715600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3276464400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3289165200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3307914000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3321219600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3339363600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3352669200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3370813200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3384118800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3402867600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3415568400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3434317200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3447018000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3465766800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3479072400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3497216400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3510522000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3528666000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3541971600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3560115600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3573421200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3592170000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3604870800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3623619600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3636320400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3655069200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3668374800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3686518800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3699824400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3717968400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3731274000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3750022800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3762723600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3781472400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3794173200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3812922000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3825622800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3844371600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3857677200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3875821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3889126800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3907270800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3920576400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3939325200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3952026000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (3970774800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (3983475600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4002224400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4015530000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4033674000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4046979600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4065123600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (4078429200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (4096573200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 7324,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Europe__Zurich => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3675198848, FixedTimespan { utc_offset: 1786, dst_offset: 0, name: "BMT" }),
                    (-2385246586, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-904435200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-891129600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-872985600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-859680000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 2048,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::GB => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3852662325, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1691964000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1680472800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1664143200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1650146400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1633903200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1617487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1601848800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1586037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1570399200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1552168800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1538344800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1522533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1507500000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1490565600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1473631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1460930400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1442786400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1428876000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1410732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1396216800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1379282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1364767200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1348437600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1333317600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1315778400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1301263200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1284328800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1269813600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1253484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1238364000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1221429600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1206914400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1189980000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1175464800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1159135200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1143410400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1126476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1111960800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1095631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1080511200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1063576800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1049061600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1032127200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1017612000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1001282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-986162400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-969228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-950479200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-942012000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-904518000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-896050800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-875487600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-864601200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-844038000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-832546800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-812588400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-798073200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-781052400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-772066800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-764805600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-748476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-733356000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-719445600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-717030000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-706748400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-699487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-687996000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-668037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-654732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-636588000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-622072800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-605743200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-590623200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-574293600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-558568800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-542239200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-527119200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-512604000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-496274400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-481154400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-464220000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-449704800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-432165600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-417650400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-401320800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-386200800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-369266400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-354751200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-337816800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-323301600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-306972000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-291852000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-276732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-257983200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-245282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-226533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-213228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-195084000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-182383200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-163634400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-150933600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-132184800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-119484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-100735200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-88034400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-68680800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-59004000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-37242000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "BST" }),
                    (57718800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (69818400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (89172000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (101268000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (120621600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (132717600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (152071200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (164167200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (183520800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (196221600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (214970400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (227671200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (246420000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (259120800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (278474400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (290570400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (309924000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (322020000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (341373600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (354675600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (372819600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (386125200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (404269200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (417574800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (435718800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (449024400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (467773200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (481078800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (499222800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (512528400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (530672400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (543978000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (562122000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (575427600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (593571600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (606877200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (625626000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (638326800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (657075600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (670381200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (688525200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (701830800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (719974800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (733280400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (751424400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (764730000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (782874000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (796179600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (814323600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (828234000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (846378000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (859683600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (877827600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (891133200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (909277200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (922582800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (941331600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (954032400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (972781200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (985482000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1004230800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1017536400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1035680400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1048986000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1067130000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1080435600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1099184400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1111885200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1130634000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1143334800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1162083600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1174784400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1193533200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1206838800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1224982800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1238288400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1256432400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1269738000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1288486800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1301187600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1319936400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1332637200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1351386000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1364691600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1382835600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1396141200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1414285200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1427590800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1445734800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1459040400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1477789200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1490490000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1509238800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1521939600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1540688400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1553994000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1572138000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1585443600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1603587600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1616893200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1635642000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1648342800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1667091600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1679792400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1698541200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1711846800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1729990800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1743296400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1761440400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1774746000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1792890000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1806195600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1824944400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1837645200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1856394000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1869094800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1887843600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1901149200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1919293200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1932598800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1950742800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1964048400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1982797200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1995498000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2014246800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2026947600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2045696400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2058397200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2077146000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2090451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2108595600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2121901200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2140045200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2153350800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2172099600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2184800400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2203549200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2216250000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2234998800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2248304400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2266448400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2279754000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2297898000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2311203600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2329347600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2342653200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2361402000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2374102800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2392851600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2405552400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2424301200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2437606800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2455750800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2469056400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2487200400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2500506000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2519254800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2531955600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2550704400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2563405200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2582154000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2595459600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2613603600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2626909200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2645053200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2658358800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2676502800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2689808400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2708557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2721258000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2740006800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2752707600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2771456400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2784762000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2802906000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2816211600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2834355600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2847661200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2866410000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2879110800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2897859600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2910560400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2929309200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2942010000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2960758800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2974064400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2992208400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3005514000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3023658000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3036963600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3055712400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3068413200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3087162000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3099862800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3118611600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3131917200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3150061200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3163366800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3181510800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3194816400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3212960400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3226266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3245014800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3257715600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3276464400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3289165200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3307914000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3321219600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3339363600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3352669200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3370813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3384118800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3402867600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3415568400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3434317200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3447018000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3465766800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3479072400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3497216400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3510522000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3528666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3541971600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3560115600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3573421200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3592170000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3604870800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3623619600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3636320400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3655069200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3668374800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3686518800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3699824400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3717968400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3731274000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3750022800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3762723600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3781472400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3794173200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3812922000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3825622800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3844371600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3857677200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3875821200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3889126800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3907270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3920576400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3939325200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3952026000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3970774800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3983475600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4002224400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4015530000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4033674000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4046979600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4065123600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4078429200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4096573200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -75,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::GBEire => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3852662325, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1691964000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1680472800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1664143200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1650146400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1633903200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1617487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1601848800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1586037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1570399200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1552168800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1538344800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1522533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1507500000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1490565600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1473631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1460930400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1442786400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1428876000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1410732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1396216800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1379282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1364767200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1348437600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1333317600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1315778400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1301263200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1284328800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1269813600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1253484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1238364000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1221429600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1206914400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1189980000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1175464800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1159135200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1143410400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1126476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1111960800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1095631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1080511200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1063576800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1049061600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1032127200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-1017612000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-1001282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-986162400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-969228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-950479200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-942012000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-904518000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-896050800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-875487600, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-864601200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-844038000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-832546800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-812588400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-798073200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-781052400, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-772066800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-764805600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-748476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-733356000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-719445600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-717030000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "BDST" }),
                    (-706748400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-699487200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-687996000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-668037600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-654732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-636588000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-622072800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-605743200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-590623200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-574293600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-558568800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-542239200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-527119200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-512604000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-496274400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-481154400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-464220000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-449704800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-432165600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-417650400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-401320800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-386200800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-369266400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-354751200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-337816800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-323301600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-306972000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-291852000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-276732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-257983200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-245282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-226533600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-213228000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-195084000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-182383200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-163634400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-150933600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-132184800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-119484000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-100735200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-88034400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-68680800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (-59004000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (-37242000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "BST" }),
                    (57718800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (69818400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (89172000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (101268000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (120621600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (132717600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (152071200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (164167200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (183520800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (196221600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (214970400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (227671200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (246420000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (259120800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (278474400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (290570400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (309924000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (322020000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (341373600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (354675600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (372819600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (386125200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (404269200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (417574800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (435718800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (449024400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (467773200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (481078800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (499222800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (512528400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (530672400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (543978000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (562122000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (575427600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (593571600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (606877200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (625626000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (638326800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (657075600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (670381200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (688525200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (701830800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (719974800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (733280400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (751424400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (764730000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (782874000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (796179600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (814323600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (828234000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (846378000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (859683600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (877827600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (891133200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (909277200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (922582800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (941331600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (954032400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (972781200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (985482000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1004230800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1017536400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1035680400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1048986000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1067130000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1080435600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1099184400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1111885200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1130634000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1143334800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1162083600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1174784400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1193533200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1206838800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1224982800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1238288400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1256432400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1269738000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1288486800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1301187600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1319936400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1332637200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1351386000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1364691600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1382835600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1396141200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1414285200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1427590800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1445734800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1459040400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1477789200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1490490000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1509238800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1521939600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1540688400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1553994000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1572138000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1585443600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1603587600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1616893200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1635642000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1648342800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1667091600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1679792400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1698541200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1711846800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1729990800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1743296400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1761440400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1774746000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1792890000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1806195600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1824944400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1837645200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1856394000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1869094800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1887843600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1901149200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1919293200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1932598800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1950742800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1964048400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (1982797200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (1995498000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2014246800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2026947600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2045696400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2058397200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2077146000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2090451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2108595600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2121901200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2140045200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2153350800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2172099600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2184800400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2203549200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2216250000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2234998800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2248304400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2266448400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2279754000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2297898000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2311203600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2329347600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2342653200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2361402000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2374102800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2392851600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2405552400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2424301200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2437606800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2455750800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2469056400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2487200400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2500506000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2519254800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2531955600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2550704400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2563405200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2582154000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2595459600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2613603600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2626909200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2645053200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2658358800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2676502800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2689808400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2708557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2721258000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2740006800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2752707600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2771456400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2784762000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2802906000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2816211600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2834355600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2847661200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2866410000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2879110800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2897859600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2910560400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2929309200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2942010000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2960758800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (2974064400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (2992208400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3005514000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3023658000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3036963600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3055712400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3068413200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3087162000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3099862800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3118611600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3131917200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3150061200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3163366800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3181510800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3194816400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3212960400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3226266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3245014800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3257715600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3276464400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3289165200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3307914000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3321219600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3339363600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3352669200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3370813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3384118800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3402867600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3415568400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3434317200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3447018000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3465766800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3479072400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3497216400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3510522000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3528666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3541971600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3560115600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3573421200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3592170000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3604870800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3623619600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3636320400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3655069200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3668374800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3686518800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3699824400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3717968400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3731274000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3750022800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3762723600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3781472400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3794173200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3812922000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3825622800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3844371600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3857677200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3875821200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3889126800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3907270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3920576400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3939325200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3952026000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (3970774800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (3983475600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4002224400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4015530000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4033674000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4046979600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4065123600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                    (4078429200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "BST" }),
                    (4096573200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -75,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::GMT => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "GMT",
                    },
                    rest: REST
                }
            },

            Tz::GMTPlus0 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "GMT",
                    },
                    rest: REST
                }
            },

            Tz::GMTMinus0 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "GMT",
                    },
                    rest: REST
                }
            },

            Tz::GMT0 => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "GMT",
                    },
                    rest: REST
                }
            },

            Tz::Greenwich => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "GMT",
                    },
                    rest: REST
                }
            },

            Tz::HST => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -36000,
                        dst_offset: 0,
                        name: "HST",
                    },
                    rest: REST
                }
            },

            Tz::Hongkong => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2056718202, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-900910800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-891579600, FixedTimespan { utc_offset: 28800, dst_offset: 1800, name: "HKWT" }),
                    (-884248200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-761209200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-747907200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-728541000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-717049800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-697091400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-683785800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-668061000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-654755400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-636611400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-623305800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-605161800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-591856200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-573712200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-559801800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-541657800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-528352200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-510211800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-498112200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-478762200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-466662600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-446707800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-435213000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-415258200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-403158600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-383808600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-371709000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-352359000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-340259400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-320909400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-308809800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-288855000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-277360200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-257405400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-245910600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-225955800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-213856200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-194506200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-182406600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-163056600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-148537800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-132816600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-117088200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-101367000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-85638600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-69312600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-53584200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-37863000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (-22134600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (-6413400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (9315000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (25036200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (40764600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (56485800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (72214200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (88540200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (104268600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (119989800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (126041400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (151439400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (167167800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (182889000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (198617400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (214338600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                    (295385400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "HKST" }),
                    (309292200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "HKT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 27402,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Iceland => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830383032, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "GMT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -968,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Indian__Antananarivo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1946168836, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1309746600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                    (-1261969200, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1041388200, FixedTimespan { utc_offset: 9900, dst_offset: 0, name: "+0245" }),
                    (-865305900, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8836,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Indian__Chagos => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1988167780, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                    (820436400, FixedTimespan { utc_offset: 21600, dst_offset: 0, name: "+06" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 17380,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Indian__Christmas => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840164924, FixedTimespan { utc_offset: 24124, dst_offset: 0, name: "BMT" }),
                    (-1570084924, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 24124,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Indian__Cocos => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840163887, FixedTimespan { utc_offset: 23087, dst_offset: 0, name: "RMT" }),
                    (-1577946287, FixedTimespan { utc_offset: 23400, dst_offset: 0, name: "+0630" }),
                    (-873268200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-778410000, FixedTimespan { utc_offset: 23400, dst_offset: 0, name: "+0630" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 23087,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Indian__Comoro => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1946168836, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1309746600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                    (-1261969200, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1041388200, FixedTimespan { utc_offset: 9900, dst_offset: 0, name: "+0245" }),
                    (-865305900, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8836,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Indian__Kerguelen => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840158440, FixedTimespan { utc_offset: 17640, dst_offset: 0, name: "MMT" }),
                    (-315636840, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 17640,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Indian__Mahe => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577936472, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 13272,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Indian__Maldives => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840158440, FixedTimespan { utc_offset: 17640, dst_offset: 0, name: "MMT" }),
                    (-315636840, FixedTimespan { utc_offset: 18000, dst_offset: 0, name: "+05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 17640,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Indian__Mauritius => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1988164200, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (403041600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (417034800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (1224972000, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (1238274000, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 13800,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Indian__Mayotte => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1946168836, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1309746600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                    (-1261969200, FixedTimespan { utc_offset: 9000, dst_offset: 0, name: "+0230" }),
                    (-1041388200, FixedTimespan { utc_offset: 9900, dst_offset: 0, name: "+0245" }),
                    (-865305900, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "EAT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8836,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Indian__Reunion => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577936472, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 13272,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Iran => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1704165944, FixedTimespan { utc_offset: 12344, dst_offset: 0, name: "TMT" }),
                    (-1090466744, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (227820600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (246223800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (259617600, FixedTimespan { utc_offset: 14400, dst_offset: 3600, name: "+05" }),
                    (271108800, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "+04" }),
                    (283982400, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (296598600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (306531000, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (322432200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (338499000, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (673216200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (685481400, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (701209800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (717103800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (732745800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (748639800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (764281800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (780175800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (795817800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (811711800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (827353800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (843247800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (858976200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (874870200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (890512200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (906406200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (922048200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (937942200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (953584200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (969478200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (985206600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1001100600, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1016742600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1032636600, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1048278600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1064172600, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1079814600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1095708600, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1111437000, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1127331000, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1206045000, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1221939000, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1237667400, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1253561400, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1269203400, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1285097400, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1300739400, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1316633400, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1332275400, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1348169400, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1363897800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1379791800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1395433800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1411327800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1426969800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1442863800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1458505800, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1474399800, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1490128200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1506022200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1521664200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1537558200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1553200200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1569094200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1584736200, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1600630200, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1616358600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1632252600, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                    (1647894600, FixedTimespan { utc_offset: 12600, dst_offset: 3600, name: "+0430" }),
                    (1663788600, FixedTimespan { utc_offset: 12600, dst_offset: 0, name: "+0330" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 12344,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Israel => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840149254, FixedTimespan { utc_offset: 8440, dst_offset: 0, name: "JMT" }),
                    (-1641003640, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-933638400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-923097600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-919036800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-857347200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-844300800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-825811200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-812678400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-794188800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-779846400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-762652800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-748310400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-731116800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-681955200, FixedTimespan { utc_offset: 7200, dst_offset: 7200, name: "IDDT" }),
                    (-673228800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-667958400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-652320000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-636422400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-622080000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-608947200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-591840000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-572486400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-558576000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-542851200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-527731200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-514425600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-490838400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-482976000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-459388800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-451526400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-428544000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-418262400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (-400118400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (-387417600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (142380000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (150843600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (167176800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (178664400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (334101600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (337730400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (452642400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (462319200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (482277600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (494370000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (516751200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (526424400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (545436000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (558478800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (576626400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (589323600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (609890400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (620773200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (638316000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (651618000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (669765600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (683672400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (701820000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (715726800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (733701600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (747176400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (765151200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (778021200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (796600800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (810075600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (826840800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (842821200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (858895200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (874184400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (890344800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (905029200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (923011200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (936313200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (955670400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (970783200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (986770800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1001282400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1017356400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1033941600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1048806000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1065132000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1081292400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1095804000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1112313600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1128812400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1143763200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1159657200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1175212800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1189897200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1206662400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1223161200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1238112000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1254006000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1269561600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1284246000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1301616000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1317510000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1333065600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1348354800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1364515200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1382828400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1395964800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1414278000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1427414400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1445727600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1458864000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1477782000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1490313600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1509231600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1521763200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1540681200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1553817600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1572130800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1585267200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1603580400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1616716800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1635634800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1648166400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1667084400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1679616000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1698534000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1711670400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1729983600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1743120000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1761433200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1774569600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1792882800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1806019200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1824937200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1837468800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1856386800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1868918400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1887836400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1900972800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1919286000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1932422400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1950735600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1963872000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (1982790000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (1995321600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2014239600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2026771200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2045689200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2058220800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2077138800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2090275200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2108588400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2121724800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2140038000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2153174400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2172092400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2184624000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2203542000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2216073600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2234991600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2248128000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2266441200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2279577600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2297890800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2311027200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2329340400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2342476800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2361394800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2373926400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2392844400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2405376000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2424294000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2437430400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2455743600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2468880000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2487193200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2500329600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2519247600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2531779200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2550697200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2563228800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2582146800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2595283200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2613596400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2626732800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2645046000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2658182400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2676495600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2689632000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2708550000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2721081600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2739999600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2752531200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2771449200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2784585600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2802898800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2816035200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2834348400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2847484800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2866402800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2878934400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2897852400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2910384000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2929302000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2941833600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2960751600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (2973888000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (2992201200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3005337600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3023650800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3036787200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3055705200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3068236800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3087154800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3099686400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3118604400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3131740800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3150054000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3163190400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3181503600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3194640000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3212953200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3226089600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3245007600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3257539200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3276457200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3288988800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3307906800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3321043200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3339356400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3352492800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3370806000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3383942400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3402860400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3415392000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3434310000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3446841600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3465759600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3478896000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3497209200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3510345600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3528658800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3541795200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3560108400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3573244800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3592162800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3604694400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3623612400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3636144000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3655062000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3668198400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3686511600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3699648000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3717961200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3731097600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3750015600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3762547200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3781465200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3793996800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3812914800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3825446400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3844364400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3857500800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3875814000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3888950400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3907263600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3920400000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3939318000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3951849600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (3970767600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (3983299200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (4002217200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (4015353600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (4033666800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (4046803200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (4065116400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                    (4078252800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "IDT" }),
                    (4096566000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "IST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 8454,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Jamaica => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524503170, FixedTimespan { utc_offset: -18430, dst_offset: 0, name: "KMT" }),
                    (-1827687170, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (126687600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (152085600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (162370800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -18430,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Japan => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2587745939, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-683802000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "JDT" }),
                    (-672310800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-654771600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "JDT" }),
                    (-640861200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-620298000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "JDT" }),
                    (-609411600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-588848400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "JDT" }),
                    (-577962000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 33539,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Kwajalein => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177492960, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (-1041418800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (-907408800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-817462800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (-7988400, FixedTimespan { utc_offset: -43200, dst_offset: 0, name: "-12" }),
                    (745934400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 40160,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Libya => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1577926364, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-574902000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-568087200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-512175600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-504928800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-449888400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-441856800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-347158800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (378684000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386463600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (402271200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417999600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433807200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449622000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465429600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481590000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496965600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512953200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528674400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (544230000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (560037600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575852400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591660000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (607388400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (623196000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (641775600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (844034400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (860108400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (875916000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1352505600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364515200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382659200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3164,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::MET => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1693706400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (-938905200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (-796777200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (-781052400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (-766623600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (228877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (243997200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (260326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (276051600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (291776400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (307501200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (323830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (338950800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (354675600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (370400400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (386125200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (401850000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (417574800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (433299600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (449024400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (465354000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (481078800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (496803600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (512528400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (528253200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (543978000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (559702800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "MEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "MET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 3600,
                        dst_offset: 0,
                        name: "MET",
                    },
                    rest: REST
                }
            },

            Tz::MST => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25200,
                        dst_offset: 0,
                        name: "MST",
                    },
                    rest: REST
                }
            },

            Tz::MST7MDT => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1633273200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1601823600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1583683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-84380400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-68659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-52930800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-37209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-21481200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-5760000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (9968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (25689600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (41418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (57744000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (73472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (89193600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (104922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (120643200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (126694800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (152092800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (162378000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (183542400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (199270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (215596800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (230720400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (247046400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (262774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (278496000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (294224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (309945600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (325674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (341395200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (357123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (372844800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (388573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (404899200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (420022800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (436348800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (452077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (467798400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (483526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (499248000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (514976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (530697600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (544611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (562147200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (576061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (594201600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (607510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (625651200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (638960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (657100800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (671014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (688550400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (702464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (720000000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (733914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (752054400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (765363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (783504000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (796813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (814953600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (986115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1004256000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1173603600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1194163200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1205053200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225612800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1236502800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1257062400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1268557200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1289116800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1300006800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1320566400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1331456400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1352016000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1362906000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1383465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1394355600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414915200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1425805200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1446364800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1457859600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1478419200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1489309200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509868800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1520758800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1541318400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1552208400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572768000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1583658000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1604217600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1615712400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1636272000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1647162000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667721600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1678611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1699171200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1710061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1730620800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1741510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1762070400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1772960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1793520000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1805014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1825574400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1836464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1857024000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1867914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1888473600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1899363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1919923200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1930813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1951372800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1962867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1983427200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1994317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2014876800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2025766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2046326400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2057216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2077776000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2088666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2109225600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2120115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2140675200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2152170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2172729600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2183619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2204179200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2215069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2235628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2246518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2267078400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2277968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2298528000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2309418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2329977600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2341472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2362032000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2372922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2393481600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2404371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2424931200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2435821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2456380800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2467270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2487830400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2499325200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2519884800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2530774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2551334400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2562224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2582784000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2593674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2614233600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2625123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2645683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2656573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2677132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2688627600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2709187200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2720077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2740636800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2751526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2772086400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2782976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2803536000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2814426000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2834985600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2846480400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2867040000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2877930000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2898489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2909379600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2929939200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2940829200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2961388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2972278800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2992838400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3003728400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3024288000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3035782800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3056342400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3067232400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3087792000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3098682000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3119241600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3130131600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3150691200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3161581200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3182140800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3193030800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3213590400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3225085200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3245644800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3256534800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3277094400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3287984400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3308544000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3319434000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3339993600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3350883600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3371443200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3382938000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3403497600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3414387600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3434947200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3445837200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3466396800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3477286800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3497846400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3508736400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3529296000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3540186000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3560745600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3572240400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3592800000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3603690000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3624249600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3635139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3655699200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3666589200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3687148800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3698038800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3718598400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3730093200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3750652800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3761542800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3782102400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3792992400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3813552000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3824442000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3845001600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3855891600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3876451200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3887341200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3907900800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3919395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3939955200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3950845200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3971404800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3982294800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4002854400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4013744400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4034304000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4045194000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4065753600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4076643600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4097203200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25200,
                        dst_offset: 0,
                        name: "MST",
                    },
                    rest: REST
                }
            },

            Tz::Mexico__BajaNorte => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514711516, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1451667600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1343062800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1234803600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1222963200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1207242000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-873820800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PWT" }),
                    (-769370400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PPT" }),
                    (-761677200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-686073600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-661539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-495039600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-481734000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-463590000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-450284400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-431535600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-418230000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-400086000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-386780400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-368636400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-355330800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-337186800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-323881200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-305737200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-292431600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (199274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (215600400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (230724000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (247050000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (262778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (278499600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (294228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (309949200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (341398800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (452080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (467802000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (483530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (499251600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (514980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (530701200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (544615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (562150800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (576064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (594205200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (607514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (625654800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (638964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (657104400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (671018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (688554000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (702468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (720003600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (733917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (752058000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (765367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (783507600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (796816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (814957200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (828871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (846406800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (860320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (877856400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (891770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (909306000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (923220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (941360400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (954669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (972810000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (986119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1004259600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1018173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1035709200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1049623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1067158800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1081072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1099213200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1112522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1130662800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1143972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1162112400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1175421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1193562000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1207476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1225011600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1238925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1256461200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1268560800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1289120400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1300010400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1320570000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1331460000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1352019600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1362909600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1383469200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1394359200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1414918800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1425808800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1446368400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1457863200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1478422800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1489312800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1509872400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1520762400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1541322000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1552212000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1572771600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1583661600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1604221200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1615716000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1636275600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1647165600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1667725200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1678615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1699174800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1710064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1730624400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1741514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1762074000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1772964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1793523600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1805018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1825578000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1836468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1857027600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1867917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1888477200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1899367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1919926800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1930816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1951376400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1962871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1983430800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1994320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2014880400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2025770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2046330000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2057220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2077779600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2088669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2109229200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2120119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2140678800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2152173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2172733200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2183623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2204182800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2215072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2235632400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2246522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2267082000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2277972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2298531600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2309421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2329981200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2341476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2362035600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2372925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2393485200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2404375200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2424934800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2435824800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2456384400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2467274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2487834000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2499328800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2519888400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2530778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2551338000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2562228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2582787600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2593677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2614237200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2625127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2645686800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2656576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2677136400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2688631200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2709190800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2720080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2740640400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2751530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2772090000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2782980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2803539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2814429600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2834989200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2846484000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2867043600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2877933600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2898493200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2909383200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2929942800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2940832800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2961392400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2972282400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2992842000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3003732000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3024291600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3035786400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3056346000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3067236000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3087795600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3098685600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3119245200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3130135200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3150694800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3161584800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3182144400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3193034400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3213594000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3225088800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3245648400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3256538400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3277098000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3287988000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3308547600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3319437600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3339997200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3350887200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3371446800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3382941600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3403501200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3414391200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3434950800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3445840800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3466400400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3477290400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3497850000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3508740000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3529299600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3540189600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3560749200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3572244000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3592803600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3603693600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3624253200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3635143200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3655702800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3666592800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3687152400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3698042400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3718602000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3730096800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3750656400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3761546400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3782106000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3792996000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3813555600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3824445600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3845005200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3855895200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3876454800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3887344800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3907904400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3919399200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3939958800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3950848800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3971408400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3982298400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4002858000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4013748000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4034307600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4045197600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4065757200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4076647200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4097206800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -28084,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Mexico__BajaSur => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514714060, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1343066400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1234807200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1220292000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1207159200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1191344400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-873828000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-661539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (28800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (989139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1001836800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1175418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1193558400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1207472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225008000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1238922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1256457600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1270371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1288512000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1301821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1319961600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1333270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1351411200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1365325200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1382860800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1396774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414310400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1428224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1445760000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1459674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1477814400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1491123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509264000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1522573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1540713600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1554627600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572163200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1586077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1603612800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1617526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1635667200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1648976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667116800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25540,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Mexico__General => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1514715804, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1343066400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1234807200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1220292000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1207159200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1191344400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-975261600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-963169200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-917114400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-907354800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-821901600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-810068400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-627501600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-612990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909298800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941353200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (989136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1001833200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035702000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067151600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099206000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130655600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1175414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1193554800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1207468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225004400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1238918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1256454000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1270368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1288508400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1301817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1319958000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1333267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1351407600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1365321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1382857200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1396771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414306800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1428220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1445756400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1459670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1477810800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1491120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509260400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1522569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1540710000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1554624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1586073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1603609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1617523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1635663600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1648972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -23796,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::NZ => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3192435544, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1330335000, FixedTimespan { utc_offset: 41400, dst_offset: 3600, name: "NZST" }),
                    (-1320057000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1300699800, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1287396000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1269250200, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1255946400, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1237800600, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1224496800, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1206351000, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1192442400, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1174901400, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1160992800, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1143451800, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1125914400, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1112607000, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1094464800, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1081157400, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1063015200, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1049707800, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1031565600, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1018258200, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1000116000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-986808600, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-968061600, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-955359000, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-936612000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-923304600, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-757425600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (152632800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (162309600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (183477600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (194968800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (215532000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (226418400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (246981600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (257868000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (278431200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (289317600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (309880800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (320767200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (341330400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (352216800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (372780000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (384271200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (404834400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (415720800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (436284000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (447170400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (467733600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (478620000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (499183200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (510069600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (530632800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (541519200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (562082400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (573573600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (594136800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (605023200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (623772000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (637682400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (655221600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (669132000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (686671200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (700581600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (718120800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (732636000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (749570400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (764085600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (781020000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (795535200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (812469600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (826984800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (844524000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (858434400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (875973600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (889884000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (907423200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (921938400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (938872800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (953388000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (970322400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (984837600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1002376800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1016287200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1033826400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1047736800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1065276000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1079791200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1096725600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1111240800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1128175200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1142690400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1159624800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1174140000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1191074400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1207404000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1222524000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1238853600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1253973600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1270303200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1285423200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1301752800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1316872800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1333202400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1348927200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1365256800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1380376800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1396706400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1411826400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1428156000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1443276000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1459605600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1474725600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1491055200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1506175200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1522504800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1538229600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1554559200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1569679200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1586008800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1601128800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1617458400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1632578400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1648908000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1664028000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1680357600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1695477600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1712412000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1727532000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1743861600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1758981600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1775311200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1790431200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1806760800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1821880800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1838210400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1853330400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1869660000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1885384800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1901714400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1916834400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1933164000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1948284000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1964613600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1979733600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1996063200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2011183200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2027512800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2042632800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2058962400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2074687200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2091016800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2106136800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2122466400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2137586400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2153916000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2169036000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2185365600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2200485600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2216815200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2232540000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2248869600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2263989600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2280319200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2295439200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2311768800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2326888800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2343218400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2358338400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2374668000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2389788000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2406117600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2421842400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2438172000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2453292000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2469621600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2484741600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2501071200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2516191200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2532520800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2547640800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2563970400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2579090400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2596024800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2611144800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2627474400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2642594400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2658924000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2674044000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2690373600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2705493600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2721823200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2736943200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2753272800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2768997600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2785327200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2800447200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2816776800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2831896800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2848226400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2863346400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2879676000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2894796000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2911125600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2926245600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2942575200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2958300000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2974629600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2989749600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3006079200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3021199200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3037528800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3052648800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3068978400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3084098400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3100428000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3116152800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3132482400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3147602400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3163932000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3179052000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3195381600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3210501600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3226831200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3241951200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3258280800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3273400800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3289730400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3305455200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3321784800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3336904800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3353234400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3368354400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3384684000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3399804000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3416133600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3431253600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3447583200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3462703200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3479637600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3494757600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3511087200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3526207200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3542536800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3557656800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3573986400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3589106400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3605436000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3620556000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3636885600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3652610400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3668940000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3684060000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3700389600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3715509600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3731839200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3746959200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3763288800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3778408800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3794738400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3809858400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3826188000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3841912800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3858242400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3873362400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3889692000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3904812000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3921141600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3936261600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3952591200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3967711200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3984040800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3999765600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (4016095200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (4031215200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (4047544800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (4062664800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (4078994400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (4094114400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 41944,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::NZCHAT => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3192437628, FixedTimespan { utc_offset: 44100, dst_offset: 0, name: "+1215" }),
                    (-757426500, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (152632800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (162309600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (183477600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (194968800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (215532000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (226418400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (246981600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (257868000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (278431200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (289317600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (309880800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (320767200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (341330400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (352216800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (372780000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (384271200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (404834400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (415720800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (436284000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (447170400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (467733600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (478620000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (499183200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (510069600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (530632800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (541519200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (562082400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (573573600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (594136800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (605023200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (623772000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (637682400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (655221600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (669132000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (686671200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (700581600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (718120800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (732636000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (749570400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (764085600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (781020000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (795535200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (812469600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (826984800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (844524000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (858434400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (875973600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (889884000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (907423200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (921938400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (938872800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (953388000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (970322400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (984837600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1002376800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1016287200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1033826400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1047736800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1065276000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1079791200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1096725600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1111240800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1128175200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1142690400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1159624800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1174140000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1191074400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1207404000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1222524000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1238853600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1253973600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1270303200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1285423200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1301752800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1316872800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1333202400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1348927200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1365256800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1380376800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1396706400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1411826400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1428156000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1443276000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1459605600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1474725600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1491055200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1506175200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1522504800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1538229600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1554559200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1569679200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1586008800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1601128800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1617458400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1632578400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1648908000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1664028000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1680357600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1695477600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1712412000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1727532000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1743861600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1758981600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1775311200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1790431200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1806760800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1821880800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1838210400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1853330400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1869660000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1885384800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1901714400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1916834400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1933164000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1948284000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1964613600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1979733600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1996063200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2011183200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2027512800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2042632800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2058962400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2074687200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2091016800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2106136800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2122466400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2137586400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2153916000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2169036000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2185365600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2200485600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2216815200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2232540000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2248869600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2263989600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2280319200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2295439200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2311768800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2326888800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2343218400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2358338400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2374668000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2389788000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2406117600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2421842400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2438172000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2453292000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2469621600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2484741600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2501071200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2516191200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2532520800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2547640800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2563970400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2579090400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2596024800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2611144800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2627474400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2642594400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2658924000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2674044000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2690373600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2705493600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2721823200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2736943200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2753272800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2768997600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2785327200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2800447200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2816776800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2831896800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2848226400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2863346400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2879676000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2894796000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2911125600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2926245600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2942575200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2958300000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2974629600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2989749600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3006079200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3021199200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3037528800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3052648800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3068978400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3084098400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3100428000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3116152800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3132482400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3147602400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3163932000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3179052000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3195381600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3210501600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3226831200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3241951200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3258280800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3273400800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3289730400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3305455200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3321784800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3336904800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3353234400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3368354400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3384684000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3399804000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3416133600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3431253600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3447583200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3462703200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3479637600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3494757600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3511087200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3526207200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3542536800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3557656800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3573986400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3589106400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3605436000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3620556000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3636885600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3652610400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3668940000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3684060000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3700389600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3715509600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3731839200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3746959200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3763288800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3778408800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3794738400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3809858400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3826188000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3841912800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3858242400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3873362400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3889692000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3904812000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3921141600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3936261600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3952591200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3967711200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3984040800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3999765600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (4016095200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (4031215200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (4047544800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (4062664800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (4078994400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (4094114400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 44028,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Navajo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717618404, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1633273200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1601823600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1583683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1570374000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1551628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1538924400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1534089600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-147884400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-131558400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-116434800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-100108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-84380400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-68659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-52930800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-37209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-21481200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-5760000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (9968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (25689600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (41418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (57744000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (73472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (89193600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (104922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (120643200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (126694800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (152092800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (162378000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (183542400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (199270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (215596800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (230720400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (247046400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (262774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (278496000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (294224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (309945600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (325674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (341395200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (357123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (372844800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (388573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (404899200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (420022800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (436348800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (452077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (467798400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (483526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (499248000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (514976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (530697600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (544611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (562147200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (576061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (594201600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (607510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (625651200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (638960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (657100800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (671014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (688550400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (702464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (720000000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (733914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (752054400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (765363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (783504000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (796813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (814953600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (986115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1004256000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1173603600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1194163200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1205053200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225612800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1236502800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1257062400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1268557200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1289116800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1300006800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1320566400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1331456400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1352016000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1362906000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1383465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1394355600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414915200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1425805200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1446364800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1457859600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1478419200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1489309200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509868800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1520758800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1541318400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1552208400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572768000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1583658000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1604217600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1615712400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1636272000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1647162000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667721600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1678611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1699171200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1710061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1730620800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1741510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1762070400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1772960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1793520000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1805014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1825574400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1836464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1857024000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1867914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1888473600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1899363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1919923200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1930813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1951372800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1962867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1983427200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1994317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2014876800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2025766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2046326400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2057216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2077776000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2088666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2109225600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2120115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2140675200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2152170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2172729600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2183619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2204179200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2215069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2235628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2246518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2267078400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2277968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2298528000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2309418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2329977600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2341472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2362032000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2372922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2393481600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2404371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2424931200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2435821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2456380800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2467270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2487830400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2499325200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2519884800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2530774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2551334400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2562224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2582784000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2593674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2614233600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2625123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2645683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2656573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2677132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2688627600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2709187200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2720077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2740636800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2751526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2772086400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2782976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2803536000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2814426000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2834985600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2846480400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2867040000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2877930000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2898489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2909379600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2929939200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2940829200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2961388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2972278800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2992838400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3003728400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3024288000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3035782800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3056342400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3067232400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3087792000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3098682000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3119241600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3130131600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3150691200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3161581200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3182140800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3193030800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3213590400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3225085200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3245644800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3256534800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3277094400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3287984400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3308544000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3319434000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3339993600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3350883600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3371443200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3382938000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3403497600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3414387600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3434947200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3445837200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3466396800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3477286800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3497846400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3508736400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3529296000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3540186000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3560745600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3572240400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3592800000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3603690000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3624249600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3635139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3655699200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3666589200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3687148800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3698038800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3718598400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3730093200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3750652800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3761542800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3782102400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3792992400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3813552000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3824442000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3845001600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3855891600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3876451200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3887341200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3907900800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3919395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3939955200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3950845200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3971404800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3982294800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4002854400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4013744400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4034304000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4045194000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4065753600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4076643600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4097203200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25196,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::PRC => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177481943, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-1600675200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-1585904400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-933667200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-922093200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-908870400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-888829200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-881049600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-767869200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-745833600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-733827600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-716889600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-699613200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-683884800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-670669200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-652348800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-650019600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (515527200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (527014800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (545162400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (558464400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (577216800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (589914000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (608666400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (621968400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (640116000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (653418000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (671565600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (684867600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 29143,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::PST8PDT => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1633269600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1615129200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1601820000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1583679600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-880207200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PWT" }),
                    (-769395600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PPT" }),
                    (-765385200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-84376800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-68655600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-52927200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-37206000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-21477600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-5756400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (9972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (25693200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (41421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (57747600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (73476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (89197200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (104925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (120646800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (126698400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (152096400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (162381600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (183546000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (199274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (215600400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (230724000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (247050000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (262778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (278499600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (294228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (309949200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (341398800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (452080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (467802000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (483530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (499251600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (514980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (530701200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (544615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (562150800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (576064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (594205200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (607514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (625654800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (638964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (657104400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (671018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (688554000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (702468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (720003600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (733917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (752058000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (765367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (783507600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (796816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (814957200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (828871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (846406800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (860320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (877856400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (891770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (909306000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (923220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (941360400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (954669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (972810000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (986119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1004259600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1018173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1035709200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1049623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1067158800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1081072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1099213200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1112522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1130662800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1143972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1162112400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1173607200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1194166800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1205056800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1225616400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1236506400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1257066000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1268560800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1289120400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1300010400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1320570000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1331460000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1352019600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1362909600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1383469200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1394359200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1414918800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1425808800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1446368400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1457863200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1478422800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1489312800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1509872400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1520762400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1541322000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1552212000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1572771600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1583661600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1604221200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1615716000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1636275600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1647165600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1667725200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1678615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1699174800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1710064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1730624400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1741514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1762074000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1772964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1793523600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1805018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1825578000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1836468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1857027600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1867917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1888477200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1899367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1919926800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1930816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1951376400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1962871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1983430800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1994320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2014880400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2025770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2046330000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2057220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2077779600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2088669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2109229200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2120119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2140678800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2152173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2172733200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2183623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2204182800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2215072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2235632400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2246522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2267082000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2277972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2298531600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2309421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2329981200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2341476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2362035600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2372925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2393485200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2404375200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2424934800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2435824800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2456384400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2467274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2487834000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2499328800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2519888400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2530778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2551338000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2562228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2582787600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2593677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2614237200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2625127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2645686800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2656576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2677136400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2688631200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2709190800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2720080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2740640400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2751530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2772090000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2782980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2803539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2814429600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2834989200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2846484000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2867043600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2877933600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2898493200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2909383200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2929942800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2940832800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2961392400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2972282400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2992842000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3003732000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3024291600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3035786400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3056346000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3067236000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3087795600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3098685600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3119245200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3130135200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3150694800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3161584800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3182144400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3193034400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3213594000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3225088800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3245648400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3256538400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3277098000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3287988000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3308547600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3319437600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3339997200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3350887200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3371446800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3382941600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3403501200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3414391200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3434950800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3445840800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3466400400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3477290400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3497850000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3508740000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3529299600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3540189600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3560749200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3572244000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3592803600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3603693600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3624253200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3635143200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3655702800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3666592800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3687152400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3698042400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3718602000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3730096800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3750656400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3761546400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3782106000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3792996000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3813555600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3824445600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3845005200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3855895200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3876454800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3887344800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3907904400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3919399200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3939958800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3950848800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3971408400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3982298400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4002858000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4013748000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4034307600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4045197600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4065757200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4076647200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4097206800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -28800,
                        dst_offset: 0,
                        name: "PST",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Apia => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2445424384, FixedTimespan { utc_offset: -41216, dst_offset: 0, name: "LMT" }),
                    (-1861878784, FixedTimespan { utc_offset: -41400, dst_offset: 0, name: "-1130" }),
                    (-631110600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "-11" }),
                    (1285498800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "-10" }),
                    (1301752800, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "-11" }),
                    (1316872800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "-10" }),
                    (1325239200, FixedTimespan { utc_offset: 46800, dst_offset: 3600, name: "+14" }),
                    (1333202400, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                    (1348927200, FixedTimespan { utc_offset: 46800, dst_offset: 3600, name: "+14" }),
                    (1365256800, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                    (1380376800, FixedTimespan { utc_offset: 46800, dst_offset: 3600, name: "+14" }),
                    (1396706400, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                    (1411826400, FixedTimespan { utc_offset: 46800, dst_offset: 3600, name: "+14" }),
                    (1428156000, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                    (1443276000, FixedTimespan { utc_offset: 46800, dst_offset: 3600, name: "+14" }),
                    (1459605600, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                    (1474725600, FixedTimespan { utc_offset: 46800, dst_offset: 3600, name: "+14" }),
                    (1491055200, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                    (1506175200, FixedTimespan { utc_offset: 46800, dst_offset: 3600, name: "+14" }),
                    (1522504800, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                    (1538229600, FixedTimespan { utc_offset: 46800, dst_offset: 3600, name: "+14" }),
                    (1554559200, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                    (1569679200, FixedTimespan { utc_offset: 46800, dst_offset: 3600, name: "+14" }),
                    (1586008800, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                    (1601128800, FixedTimespan { utc_offset: 46800, dst_offset: 3600, name: "+14" }),
                    (1617458400, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 45184,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Auckland => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3192435544, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1330335000, FixedTimespan { utc_offset: 41400, dst_offset: 3600, name: "NZST" }),
                    (-1320057000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1300699800, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1287396000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1269250200, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1255946400, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1237800600, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1224496800, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1206351000, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1192442400, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1174901400, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1160992800, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1143451800, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1125914400, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1112607000, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1094464800, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1081157400, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1063015200, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1049707800, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1031565600, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-1018258200, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-1000116000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-986808600, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-968061600, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-955359000, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-936612000, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "NZMT" }),
                    (-923304600, FixedTimespan { utc_offset: 41400, dst_offset: 1800, name: "NZST" }),
                    (-757425600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (152632800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (162309600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (183477600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (194968800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (215532000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (226418400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (246981600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (257868000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (278431200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (289317600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (309880800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (320767200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (341330400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (352216800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (372780000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (384271200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (404834400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (415720800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (436284000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (447170400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (467733600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (478620000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (499183200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (510069600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (530632800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (541519200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (562082400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (573573600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (594136800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (605023200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (623772000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (637682400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (655221600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (669132000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (686671200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (700581600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (718120800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (732636000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (749570400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (764085600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (781020000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (795535200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (812469600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (826984800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (844524000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (858434400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (875973600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (889884000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (907423200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (921938400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (938872800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (953388000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (970322400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (984837600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1002376800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1016287200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1033826400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1047736800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1065276000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1079791200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1096725600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1111240800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1128175200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1142690400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1159624800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1174140000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1191074400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1207404000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1222524000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1238853600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1253973600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1270303200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1285423200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1301752800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1316872800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1333202400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1348927200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1365256800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1380376800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1396706400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1411826400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1428156000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1443276000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1459605600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1474725600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1491055200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1506175200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1522504800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1538229600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1554559200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1569679200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1586008800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1601128800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1617458400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1632578400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1648908000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1664028000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1680357600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1695477600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1712412000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1727532000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1743861600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1758981600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1775311200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1790431200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1806760800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1821880800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1838210400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1853330400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1869660000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1885384800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1901714400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1916834400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1933164000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1948284000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1964613600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (1979733600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (1996063200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2011183200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2027512800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2042632800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2058962400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2074687200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2091016800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2106136800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2122466400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2137586400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2153916000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2169036000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2185365600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2200485600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2216815200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2232540000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2248869600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2263989600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2280319200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2295439200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2311768800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2326888800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2343218400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2358338400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2374668000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2389788000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2406117600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2421842400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2438172000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2453292000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2469621600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2484741600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2501071200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2516191200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2532520800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2547640800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2563970400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2579090400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2596024800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2611144800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2627474400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2642594400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2658924000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2674044000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2690373600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2705493600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2721823200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2736943200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2753272800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2768997600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2785327200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2800447200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2816776800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2831896800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2848226400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2863346400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2879676000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2894796000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2911125600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2926245600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2942575200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2958300000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (2974629600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (2989749600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3006079200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3021199200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3037528800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3052648800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3068978400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3084098400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3100428000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3116152800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3132482400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3147602400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3163932000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3179052000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3195381600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3210501600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3226831200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3241951200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3258280800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3273400800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3289730400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3305455200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3321784800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3336904800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3353234400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3368354400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3384684000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3399804000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3416133600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3431253600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3447583200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3462703200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3479637600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3494757600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3511087200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3526207200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3542536800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3557656800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3573986400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3589106400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3605436000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3620556000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3636885600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3652610400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3668940000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3684060000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3700389600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3715509600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3731839200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3746959200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3763288800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3778408800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3794738400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3809858400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3826188000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3841912800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3858242400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3873362400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3889692000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3904812000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3921141600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3936261600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3952591200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3967711200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (3984040800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (3999765600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (4016095200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (4031215200, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (4047544800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (4062664800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                    (4078994400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "NZST" }),
                    (4094114400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "NZDT" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 41944,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Bougainville => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840178136, FixedTimespan { utc_offset: 35312, dst_offset: 0, name: "PMMT" }),
                    (-2366790512, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (-868010400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-768906000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (1419696000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 37336,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Chatham => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3192437628, FixedTimespan { utc_offset: 44100, dst_offset: 0, name: "+1215" }),
                    (-757426500, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (152632800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (162309600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (183477600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (194968800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (215532000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (226418400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (246981600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (257868000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (278431200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (289317600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (309880800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (320767200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (341330400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (352216800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (372780000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (384271200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (404834400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (415720800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (436284000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (447170400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (467733600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (478620000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (499183200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (510069600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (530632800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (541519200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (562082400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (573573600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (594136800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (605023200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (623772000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (637682400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (655221600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (669132000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (686671200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (700581600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (718120800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (732636000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (749570400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (764085600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (781020000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (795535200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (812469600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (826984800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (844524000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (858434400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (875973600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (889884000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (907423200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (921938400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (938872800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (953388000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (970322400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (984837600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1002376800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1016287200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1033826400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1047736800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1065276000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1079791200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1096725600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1111240800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1128175200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1142690400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1159624800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1174140000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1191074400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1207404000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1222524000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1238853600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1253973600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1270303200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1285423200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1301752800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1316872800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1333202400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1348927200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1365256800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1380376800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1396706400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1411826400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1428156000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1443276000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1459605600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1474725600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1491055200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1506175200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1522504800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1538229600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1554559200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1569679200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1586008800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1601128800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1617458400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1632578400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1648908000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1664028000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1680357600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1695477600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1712412000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1727532000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1743861600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1758981600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1775311200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1790431200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1806760800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1821880800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1838210400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1853330400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1869660000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1885384800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1901714400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1916834400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1933164000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1948284000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1964613600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (1979733600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (1996063200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2011183200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2027512800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2042632800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2058962400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2074687200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2091016800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2106136800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2122466400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2137586400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2153916000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2169036000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2185365600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2200485600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2216815200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2232540000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2248869600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2263989600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2280319200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2295439200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2311768800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2326888800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2343218400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2358338400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2374668000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2389788000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2406117600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2421842400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2438172000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2453292000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2469621600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2484741600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2501071200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2516191200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2532520800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2547640800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2563970400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2579090400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2596024800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2611144800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2627474400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2642594400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2658924000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2674044000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2690373600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2705493600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2721823200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2736943200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2753272800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2768997600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2785327200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2800447200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2816776800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2831896800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2848226400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2863346400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2879676000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2894796000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2911125600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2926245600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2942575200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2958300000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (2974629600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (2989749600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3006079200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3021199200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3037528800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3052648800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3068978400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3084098400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3100428000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3116152800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3132482400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3147602400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3163932000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3179052000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3195381600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3210501600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3226831200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3241951200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3258280800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3273400800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3289730400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3305455200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3321784800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3336904800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3353234400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3368354400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3384684000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3399804000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3416133600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3431253600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3447583200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3462703200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3479637600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3494757600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3511087200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3526207200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3542536800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3557656800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3573986400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3589106400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3605436000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3620556000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3636885600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3652610400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3668940000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3684060000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3700389600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3715509600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3731839200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3746959200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3763288800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3778408800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3794738400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3809858400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3826188000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3841912800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3858242400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3873362400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3889692000, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3904812000, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3921141600, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3936261600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3952591200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3967711200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (3984040800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (3999765600, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (4016095200, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (4031215200, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (4047544800, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (4062664800, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                    (4078994400, FixedTimespan { utc_offset: 45900, dst_offset: 0, name: "+1245" }),
                    (4094114400, FixedTimespan { utc_offset: 45900, dst_offset: 3600, name: "+1345" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 44028,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Chuuk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840176120, FixedTimespan { utc_offset: 35312, dst_offset: 0, name: "PMMT" }),
                    (-2366790512, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 35320,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Easter => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2524495352, FixedTimespan { utc_offset: -26248, dst_offset: 0, name: "EMT" }),
                    (-1178124152, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (-36619200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (-23922000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (-3355200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (7527600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (24465600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (37767600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (55915200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (69217200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (87969600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (100666800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (118209600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (132116400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (150868800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (163566000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (182318400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (195620400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (213768000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (227070000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (245217600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (258519600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (277272000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (289969200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (308721600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (321418800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (340171200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (353473200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (371620800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "-06" }),
                    (384922800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "-07" }),
                    (384948000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (403070400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (416372400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (434520000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (447822000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (466574400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (479271600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (498024000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (510721200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (529473600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (545194800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (560923200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (574225200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (592372800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (605674800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (624427200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (637124400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (653457600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (668574000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (687326400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (700628400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (718776000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (732078000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (750225600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (763527600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (781675200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (794977200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (813729600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (826426800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (845179200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (859690800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (876628800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (889930800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (906868800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (923194800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (939528000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (952830000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (971582400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (984279600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1003032000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1015729200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1034481600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1047178800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1065931200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1079233200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1097380800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1110682800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1128830400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1142132400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1160884800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1173582000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1192334400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1206846000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1223784000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1237086000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1255233600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1270350000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1286683200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1304823600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1313899200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1335668400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1346558400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1367118000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1378612800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1398567600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1410062400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1463281200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1471147200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1494730800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1502596800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1526180400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1534046400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1554606000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1567915200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1586055600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1599364800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1617505200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1630814400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1648954800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1662868800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1680404400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1693713600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1712458800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1725768000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1743908400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1757217600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1775358000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1788667200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1806807600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1820116800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1838257200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1851566400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1870311600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1883016000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1901761200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1915070400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1933210800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1946520000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1964660400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (1977969600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (1996110000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2009419200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2027559600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2040868800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2059614000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2072318400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2091063600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2104372800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2122513200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2135822400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2153962800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2167272000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2185412400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2198721600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2217466800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2230171200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2248916400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2262225600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2280366000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2293675200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2311815600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2325124800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2343265200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2356574400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2374714800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2388024000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2406769200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2419473600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2438218800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2451528000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2469668400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2482977600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2501118000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2514427200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2532567600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2545876800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2564017200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2577326400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2596071600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2609380800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2627521200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2640830400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2658970800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2672280000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2690420400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2703729600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2721870000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2735179200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2753924400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2766628800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2785374000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2798683200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2816823600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2830132800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2848273200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2861582400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2879722800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2893032000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2911172400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2924481600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2943226800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2955931200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (2974676400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (2987985600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3006126000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3019435200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3037575600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3050884800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3069025200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3082334400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3101079600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3113784000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3132529200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3145838400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3163978800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3177288000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3195428400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3208737600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3226878000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3240187200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3258327600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3271636800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3290382000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3303086400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3321831600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3335140800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3353281200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3366590400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3384730800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3398040000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3416180400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3429489600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3447630000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3460939200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3479684400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3492993600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3511134000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3524443200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3542583600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3555892800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3574033200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3587342400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3605482800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3618792000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3637537200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3650241600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3668986800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3682296000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3700436400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3713745600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3731886000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3745195200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3763335600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3776644800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3794785200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3808094400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3826839600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3839544000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3858289200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3871598400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3889738800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3903048000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3921188400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3934497600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3952638000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3965947200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (3984692400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (3997396800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (4016142000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (4029451200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (4047591600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (4060900800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (4079041200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (4092350400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -26248,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Efate => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1829387596, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (125409600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (133876800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (433256400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (448977600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (464706000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (480427200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (496760400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (511876800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (528210000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (543931200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (559659600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (575380800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (591109200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (606830400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (622558800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (638280000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (654008400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (669729600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (686062800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (696340800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (719931600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (727790400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 40396,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Enderbury => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1020470400, FixedTimespan { utc_offset: -43200, dst_offset: 0, name: "-12" }),
                    (307627200, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "-11" }),
                    (788871600, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Fakaofo => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177411704, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "-11" }),
                    (1325242800, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -41096,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Fiji => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1709985344, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (909842400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (920124000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (941896800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (951573600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1259416800, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1269698400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1287842400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1299333600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1319292000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1327154400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1350741600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1358604000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1382796000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1390050000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1414850400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1421503200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1446300000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1452952800, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1478354400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1484402400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1509804000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1515852000, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1541253600, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1547301600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1573308000, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1578751200, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (1608386400, FixedTimespan { utc_offset: 43200, dst_offset: 3600, name: "+13" }),
                    (1610805600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 42944,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Funafuti => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177494324, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 41524,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Galapagos => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1230746496, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "-05" }),
                    (504939600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                    (722930400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "-05" }),
                    (728888400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "-06" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -21504,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Gambier => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1806678012, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "-09" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -32388,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Guadalcanal => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1806748788, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 38388,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Guam => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3944626740, FixedTimespan { utc_offset: 34740, dst_offset: 0, name: "LMT" }),
                    (-2177487540, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (-885549600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-802256400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (-331891200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (-281610000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (-73728000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (-29415540, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (-16704000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (-10659600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (9907200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (21394800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (41356800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (52844400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (124819200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (130863600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (201888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (209487660, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (230659200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (241542000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (977493600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "ChST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -51660,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Honolulu => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2334101314, FixedTimespan { utc_offset: -37800, dst_offset: 0, name: "HST" }),
                    (-1157283000, FixedTimespan { utc_offset: -37800, dst_offset: 3600, name: "HDT" }),
                    (-1155436200, FixedTimespan { utc_offset: -37800, dst_offset: 0, name: "HST" }),
                    (-880198200, FixedTimespan { utc_offset: -37800, dst_offset: 3600, name: "HWT" }),
                    (-769395600, FixedTimespan { utc_offset: -37800, dst_offset: 3600, name: "HPT" }),
                    (-765376200, FixedTimespan { utc_offset: -37800, dst_offset: 0, name: "HST" }),
                    (-712150200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -37886,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Johnston => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2334101314, FixedTimespan { utc_offset: -37800, dst_offset: 0, name: "HST" }),
                    (-1157283000, FixedTimespan { utc_offset: -37800, dst_offset: 3600, name: "HDT" }),
                    (-1155436200, FixedTimespan { utc_offset: -37800, dst_offset: 0, name: "HST" }),
                    (-880198200, FixedTimespan { utc_offset: -37800, dst_offset: 3600, name: "HWT" }),
                    (-769395600, FixedTimespan { utc_offset: -37800, dst_offset: 3600, name: "HPT" }),
                    (-765376200, FixedTimespan { utc_offset: -37800, dst_offset: 0, name: "HST" }),
                    (-712150200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -37886,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Kanton => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1020470400, FixedTimespan { utc_offset: -43200, dst_offset: 0, name: "-12" }),
                    (307627200, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "-11" }),
                    (788871600, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "-00",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Kiritimati => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177415040, FixedTimespan { utc_offset: -38400, dst_offset: 0, name: "-1040" }),
                    (307622400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "-10" }),
                    (788868000, FixedTimespan { utc_offset: 50400, dst_offset: 0, name: "+14" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -37760,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Kosrae => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3944631116, FixedTimespan { utc_offset: 39116, dst_offset: 0, name: "LMT" }),
                    (-2177491916, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (-1743678000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-1606813200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (-1041418800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (-907408800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-770634000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (-7988400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                    (915105600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -47284,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Kwajalein => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177492960, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (-1041418800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                    (-907408800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-817462800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (-7988400, FixedTimespan { utc_offset: -43200, dst_offset: 0, name: "-12" }),
                    (745934400, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 40160,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Majuro => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177494324, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 41524,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Marquesas => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1806676920, FixedTimespan { utc_offset: -34200, dst_offset: 0, name: "-0930" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -33480,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Midway => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2445424632, FixedTimespan { utc_offset: -40968, dst_offset: 0, name: "LMT" }),
                    (-1861879032, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "SST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 45432,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Nauru => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1545131260, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "+1130" }),
                    (-862918200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-767350800, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "+1130" }),
                    (287418600, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 40060,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Niue => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-543069620, FixedTimespan { utc_offset: -40800, dst_offset: 0, name: "-1120" }),
                    (-173623200, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "-11" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -40780,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Norfolk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177493112, FixedTimespan { utc_offset: 40320, dst_offset: 0, name: "+1112" }),
                    (-599656320, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "+1130" }),
                    (152029800, FixedTimespan { utc_offset: 41400, dst_offset: 3600, name: "+1230" }),
                    (162912600, FixedTimespan { utc_offset: 41400, dst_offset: 0, name: "+1130" }),
                    (1443882600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1570287600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1586012400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1601737200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1617462000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1633186800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1648911600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1664636400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1680361200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1696086000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1712415600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1728140400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1743865200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1759590000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1775314800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1791039600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1806764400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1822489200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1838214000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1853938800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1869663600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1885993200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1901718000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1917442800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1933167600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1948892400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1964617200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (1980342000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (1996066800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2011791600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2027516400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2043241200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2058966000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2075295600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2091020400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2106745200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2122470000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2138194800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2153919600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2169644400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2185369200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2201094000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2216818800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2233148400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2248873200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2264598000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2280322800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2296047600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2311772400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2327497200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2343222000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2358946800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2374671600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2390396400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2406121200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2422450800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2438175600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2453900400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2469625200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2485350000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2501074800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2516799600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2532524400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2548249200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2563974000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2579698800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2596028400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2611753200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2627478000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2643202800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2658927600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2674652400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2690377200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2706102000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2721826800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2737551600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2753276400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2769606000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2785330800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2801055600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2816780400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2832505200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2848230000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2863954800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2879679600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2895404400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2911129200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2926854000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2942578800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2958908400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (2974633200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (2990358000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3006082800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3021807600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3037532400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3053257200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3068982000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3084706800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3100431600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3116761200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3132486000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3148210800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3163935600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3179660400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3195385200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3211110000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3226834800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3242559600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3258284400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3274009200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3289734000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3306063600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3321788400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3337513200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3353238000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3368962800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3384687600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3400412400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3416137200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3431862000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3447586800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3463311600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3479641200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3495366000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3511090800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3526815600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3542540400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3558265200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3573990000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3589714800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3605439600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3621164400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3636889200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3653218800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3668943600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3684668400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3700393200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3716118000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3731842800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3747567600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3763292400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3779017200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3794742000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3810466800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3826191600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3842521200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3858246000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3873970800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3889695600, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3905420400, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3921145200, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3936870000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3952594800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (3968319600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (3984044400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (4000374000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (4016098800, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (4031823600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (4047548400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (4063273200, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (4078998000, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (4094722800, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 40312,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Noumea => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1829387148, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (250002000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (257342400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (281451600, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (288878400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                    (849366000, FixedTimespan { utc_offset: 39600, dst_offset: 3600, name: "+12" }),
                    (857228400, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 39948,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Pago_Pago => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2445424632, FixedTimespan { utc_offset: -40968, dst_offset: 0, name: "LMT" }),
                    (-1861879032, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "SST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 45432,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Palau => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3944624276, FixedTimespan { utc_offset: 32276, dst_offset: 0, name: "LMT" }),
                    (-2177485076, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -54124,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Pitcairn => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177421580, FixedTimespan { utc_offset: -30600, dst_offset: 0, name: "-0830" }),
                    (893665800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "-08" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -31220,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Pohnpei => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1806748788, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 38388,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Ponape => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1806748788, FixedTimespan { utc_offset: 39600, dst_offset: 0, name: "+11" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 38388,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Port_Moresby => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840176120, FixedTimespan { utc_offset: 35312, dst_offset: 0, name: "PMMT" }),
                    (-2366790512, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 35320,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Rarotonga => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2209555256, FixedTimespan { utc_offset: -38344, dst_offset: 0, name: "LMT" }),
                    (-543072056, FixedTimespan { utc_offset: -37800, dst_offset: 0, name: "-1030" }),
                    (279714600, FixedTimespan { utc_offset: -36000, dst_offset: 1800, name: "-0930" }),
                    (289387800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "-10" }),
                    (309952800, FixedTimespan { utc_offset: -36000, dst_offset: 1800, name: "-0930" }),
                    (320837400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "-10" }),
                    (341402400, FixedTimespan { utc_offset: -36000, dst_offset: 1800, name: "-0930" }),
                    (352287000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "-10" }),
                    (372852000, FixedTimespan { utc_offset: -36000, dst_offset: 1800, name: "-0930" }),
                    (384341400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "-10" }),
                    (404906400, FixedTimespan { utc_offset: -36000, dst_offset: 1800, name: "-0930" }),
                    (415791000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "-10" }),
                    (436356000, FixedTimespan { utc_offset: -36000, dst_offset: 1800, name: "-0930" }),
                    (447240600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "-10" }),
                    (467805600, FixedTimespan { utc_offset: -36000, dst_offset: 1800, name: "-0930" }),
                    (478690200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "-10" }),
                    (499255200, FixedTimespan { utc_offset: -36000, dst_offset: 1800, name: "-0930" }),
                    (510139800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "-10" }),
                    (530704800, FixedTimespan { utc_offset: -36000, dst_offset: 1800, name: "-0930" }),
                    (541589400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "-10" }),
                    (562154400, FixedTimespan { utc_offset: -36000, dst_offset: 1800, name: "-0930" }),
                    (573643800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "-10" }),
                    (594208800, FixedTimespan { utc_offset: -36000, dst_offset: 1800, name: "-0930" }),
                    (605093400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "-10" }),
                    (625658400, FixedTimespan { utc_offset: -36000, dst_offset: 1800, name: "-0930" }),
                    (636543000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "-10" }),
                    (657108000, FixedTimespan { utc_offset: -36000, dst_offset: 1800, name: "-0930" }),
                    (667992600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "-10" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 48056,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Saipan => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3944626740, FixedTimespan { utc_offset: 34740, dst_offset: 0, name: "LMT" }),
                    (-2177487540, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (-885549600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-802256400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (-331891200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (-281610000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (-73728000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (-29415540, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (-16704000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (-10659600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (9907200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (21394800, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (41356800, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (52844400, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (124819200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (130863600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (201888000, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (209487660, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (230659200, FixedTimespan { utc_offset: 36000, dst_offset: 3600, name: "GDT" }),
                    (241542000, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "GST" }),
                    (977493600, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "ChST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -51660,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Samoa => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2445424632, FixedTimespan { utc_offset: -40968, dst_offset: 0, name: "LMT" }),
                    (-1861879032, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "SST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 45432,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Tahiti => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1806674504, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "-10" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -35896,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Tarawa => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177494324, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 41524,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Tongatapu => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-767189952, FixedTimespan { utc_offset: 44400, dst_offset: 0, name: "+1220" }),
                    (-284041200, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                    (939214800, FixedTimespan { utc_offset: 46800, dst_offset: 3600, name: "+14" }),
                    (953384400, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                    (973342800, FixedTimespan { utc_offset: 46800, dst_offset: 3600, name: "+14" }),
                    (980596800, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                    (1004792400, FixedTimespan { utc_offset: 46800, dst_offset: 3600, name: "+14" }),
                    (1012046400, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                    (1478350800, FixedTimespan { utc_offset: 46800, dst_offset: 3600, name: "+14" }),
                    (1484398800, FixedTimespan { utc_offset: 46800, dst_offset: 0, name: "+13" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 44352,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Truk => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840176120, FixedTimespan { utc_offset: 35312, dst_offset: 0, name: "PMMT" }),
                    (-2366790512, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 35320,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Wake => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177494324, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 41524,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Wallis => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177494324, FixedTimespan { utc_offset: 43200, dst_offset: 0, name: "+12" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 41524,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Pacific__Yap => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840176120, FixedTimespan { utc_offset: 35312, dst_offset: 0, name: "PMMT" }),
                    (-2366790512, FixedTimespan { utc_offset: 36000, dst_offset: 0, name: "+10" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 35320,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Poland => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840145840, FixedTimespan { utc_offset: 5040, dst_offset: 0, name: "WMT" }),
                    (-1717032240, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1693706400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1680483600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1663455600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1650150000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-1632006000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-1618700400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1600473600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1587168000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1501725600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-931734000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-857257200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-844556400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-828226800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-812502000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-796608000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-778726800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-762660000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-748486800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-733273200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-715215600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-701910000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-684975600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-670460400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-654130800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-639010800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-397094400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-386812800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-371088000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-355363200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-334195200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-323308800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-307584000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-291859200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-271296000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-260409600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-239846400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-228960000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-208396800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-197510400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (-176342400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (-166060800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (228873600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (243993600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (260323200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (276048000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (291772800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (307497600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (323827200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (338947200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (354672000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (370396800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (386121600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (401846400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (417571200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (433296000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (449020800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (465350400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (481075200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (496800000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (512524800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (528249600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (543974400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (559699200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (575427600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (591152400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (606877200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (622602000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (638326800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (654656400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (670381200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (686106000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (701830800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828234000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (846378000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (859683600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (877827600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (891133200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (909277200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (922582800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (941331600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (954032400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (972781200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (985482000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1004230800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1017536400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1035680400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1048986000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1067130000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1080435600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1099184400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1111885200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1130634000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1143334800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1162083600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1174784400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1193533200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1206838800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1224982800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1238288400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1256432400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1269738000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1288486800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1301187600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1319936400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1332637200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1351386000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1364691600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1382835600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1396141200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1414285200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1427590800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1445734800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1459040400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1477789200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1490490000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1509238800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1521939600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1540688400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1553994000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1572138000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1585443600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1603587600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1616893200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1635642000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1648342800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1667091600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1679792400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1698541200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1711846800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1729990800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1743296400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1761440400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1774746000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1792890000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1806195600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1824944400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1837645200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1856394000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1869094800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1887843600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1901149200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1919293200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1932598800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1950742800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1964048400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (1982797200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (1995498000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2014246800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2026947600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2045696400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2058397200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2077146000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2090451600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2108595600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2121901200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2140045200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2153350800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2172099600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2184800400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2203549200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2216250000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2234998800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2248304400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2266448400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2279754000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2297898000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2311203600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2329347600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2342653200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2361402000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2374102800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2392851600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2405552400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2424301200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2437606800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2455750800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2469056400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2487200400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2500506000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2519254800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2531955600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2550704400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2563405200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2582154000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2595459600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2613603600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2626909200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2645053200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2658358800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2676502800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2689808400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2708557200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2721258000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2740006800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2752707600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2771456400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2784762000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2802906000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2816211600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2834355600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2847661200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2866410000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2879110800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2897859600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2910560400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2929309200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2942010000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2960758800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (2974064400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (2992208400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3005514000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3023658000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3036963600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3055712400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3068413200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3087162000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3099862800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3118611600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3131917200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3150061200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3163366800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3181510800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3194816400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3212960400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3226266000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3245014800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3257715600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3276464400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3289165200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3307914000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3321219600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3339363600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3352669200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3370813200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3384118800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3402867600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3415568400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3434317200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3447018000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3465766800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3479072400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3497216400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3510522000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3528666000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3541971600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3560115600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3573421200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3592170000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3604870800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3623619600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3636320400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3655069200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3668374800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3686518800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3699824400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3717968400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3731274000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3750022800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3762723600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3781472400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3794173200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3812922000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3825622800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3844371600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3857677200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3875821200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3889126800, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3907270800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3920576400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3939325200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3952026000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (3970774800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (3983475600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4002224400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4015530000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4033674000, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4046979600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4065123600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (4078429200, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (4096573200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 5040,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Portugal => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1830381795, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1689555600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1677801600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1667437200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1647738000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1635814800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1616202000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1604365200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1584666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1572742800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1553043600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1541206800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1521507600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1442451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1426813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1379293200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1364778000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1348448400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1333328400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1316394000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1301274000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1284339600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1269824400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1221440400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1206925200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1191200400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1175475600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1127696400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1111971600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1096851600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1080522000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1063587600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1049072400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1033347600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-1017622800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-1002502800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-986173200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-969238800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-950490000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-942022800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-922669200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-906944400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-891133200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-877309200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-873684000, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "WEMT" }),
                    (-864007200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-857955600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-845859600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-842839200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "WEMT" }),
                    (-831348000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-825901200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-814410000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-810784800, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "WEMT" }),
                    (-799898400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-794451600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-782960400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-779335200, FixedTimespan { utc_offset: 0, dst_offset: 7200, name: "WEMT" }),
                    (-768448800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-763002000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-749091600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-733366800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-717631200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-701906400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-686181600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-670456800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-654732000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-639007200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-623282400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-607557600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-591832800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-575503200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-559778400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-544053600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-528328800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-512604000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-496879200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-481154400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-465429600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-449704800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-433980000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-417650400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-401925600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-386200800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-370476000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-354751200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-339026400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-323301600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-307576800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-291852000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-276127200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-260402400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-244677600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-228348000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-212623200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-196898400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-181173600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-165448800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-149724000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (-133999200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (-118274400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (212544000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (228268800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (243993600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (260323200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (276048000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (291772800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (307501200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (323222400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (338950800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (354675600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (370400400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (386125200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (401850000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (417578400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (433299600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (449024400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (465354000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (481078800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (496803600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (512528400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (528253200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (543978000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (559702800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (575427600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (591152400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (606877200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (622602000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (638326800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (654656400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (670381200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (686106000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (701830800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (717552000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (717555600, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (733280400, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (749005200, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (764730000, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (780454800, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (796179600, FixedTimespan { utc_offset: 3600, dst_offset: 3600, name: "CEST" }),
                    (811904400, FixedTimespan { utc_offset: 3600, dst_offset: 0, name: "CET" }),
                    (828230400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (846378000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (859683600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (877827600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (891133200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (909277200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (922582800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (941331600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (954032400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (972781200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (985482000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1004230800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1017536400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1035680400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1048986000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1067130000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1080435600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1099184400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1111885200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1130634000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1143334800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1162083600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1174784400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1193533200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1206838800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1224982800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1238288400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1256432400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1269738000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1288486800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1301187600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1319936400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1332637200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1351386000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1364691600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1382835600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1396141200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1414285200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1427590800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1445734800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1459040400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1477789200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1490490000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1509238800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1521939600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1540688400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1553994000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1572138000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1585443600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1603587600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1616893200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1635642000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1648342800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1667091600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1679792400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1698541200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1711846800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1729990800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1743296400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1761440400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1774746000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1792890000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1806195600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1824944400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1837645200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1856394000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1869094800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1887843600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1901149200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1919293200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1932598800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1950742800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1964048400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1982797200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1995498000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2014246800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2026947600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2045696400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2058397200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2077146000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2090451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2108595600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2121901200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2140045200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2153350800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2172099600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2184800400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2203549200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2216250000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2234998800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2248304400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2266448400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2279754000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2297898000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2311203600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2329347600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2342653200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2361402000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2374102800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2392851600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2405552400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2424301200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2437606800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2455750800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2469056400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2487200400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2500506000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2519254800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2531955600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2550704400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2563405200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2582154000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2595459600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2613603600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2626909200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2645053200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2658358800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2676502800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2689808400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2708557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2721258000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2740006800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2752707600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2771456400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2784762000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2802906000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2816211600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2834355600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2847661200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2866410000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2879110800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2897859600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2910560400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2929309200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2942010000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2960758800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2974064400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2992208400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3005514000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3023658000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3036963600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3055712400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3068413200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3087162000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3099862800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3118611600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3131917200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3150061200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3163366800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3181510800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3194816400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3212960400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3226266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3245014800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3257715600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3276464400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3289165200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3307914000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3321219600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3339363600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3352669200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3370813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3384118800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3402867600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3415568400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3434317200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3447018000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3465766800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3479072400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3497216400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3510522000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3528666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3541971600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3560115600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3573421200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3592170000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3604870800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3623619600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3636320400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3655069200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3668374800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3686518800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3699824400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3717968400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3731274000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3750022800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3762723600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3781472400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3794173200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3812922000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3825622800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3844371600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3857677200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3875821200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3889126800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3907270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3920576400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3939325200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3952026000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3970774800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3983475600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4002224400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4015530000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4033674000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4046979600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4065123600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4078429200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4096573200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -2205,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::ROC => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2335248360, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-1017820800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-766224000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-745833600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-733827600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-716889600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-699613200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-683884800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-670669200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-652348800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-639133200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-620812800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-607597200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-589276800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-576061200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-562924800, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-541760400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-528710400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-510224400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-497174400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-478688400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-465638400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-449830800, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-434016000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-418208400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-402480000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-386672400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-370944000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-355136400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-339408000, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-323600400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-302515200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-291978000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (-270979200, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (-260442000, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (133977600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (149785200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (165513600, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (181321200, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                    (299606400, FixedTimespan { utc_offset: 28800, dst_offset: 3600, name: "CDT" }),
                    (307551600, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 29160,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::ROK => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-1948782472, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-1830414600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "JST" }),
                    (-767350800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                    (-681210000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "KDT" }),
                    (-672228000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                    (-654771600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "KDT" }),
                    (-640864800, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                    (-623408400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "KDT" }),
                    (-609415200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                    (-588848400, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "KDT" }),
                    (-577965600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                    (-498128400, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-462702600, FixedTimespan { utc_offset: 30600, dst_offset: 3600, name: "KDT" }),
                    (-451733400, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-429784200, FixedTimespan { utc_offset: 30600, dst_offset: 3600, name: "KDT" }),
                    (-418296600, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-399544200, FixedTimespan { utc_offset: 30600, dst_offset: 3600, name: "KDT" }),
                    (-387451800, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-368094600, FixedTimespan { utc_offset: 30600, dst_offset: 3600, name: "KDT" }),
                    (-356002200, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-336645000, FixedTimespan { utc_offset: 30600, dst_offset: 3600, name: "KDT" }),
                    (-324552600, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-305195400, FixedTimespan { utc_offset: 30600, dst_offset: 3600, name: "KDT" }),
                    (-293103000, FixedTimespan { utc_offset: 30600, dst_offset: 0, name: "KST" }),
                    (-264933000, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                    (547578000, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "KDT" }),
                    (560883600, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                    (579027600, FixedTimespan { utc_offset: 32400, dst_offset: 3600, name: "KDT" }),
                    (592333200, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "KST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 30472,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Singapore => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2177477725, FixedTimespan { utc_offset: 24925, dst_offset: 0, name: "SMT" }),
                    (-2038200925, FixedTimespan { utc_offset: 25200, dst_offset: 0, name: "+07" }),
                    (-1167634800, FixedTimespan { utc_offset: 25200, dst_offset: 1200, name: "+0720" }),
                    (-1073028000, FixedTimespan { utc_offset: 26400, dst_offset: 0, name: "+0720" }),
                    (-894180000, FixedTimespan { utc_offset: 27000, dst_offset: 0, name: "+0730" }),
                    (-879665400, FixedTimespan { utc_offset: 32400, dst_offset: 0, name: "+09" }),
                    (-767005200, FixedTimespan { utc_offset: 27000, dst_offset: 0, name: "+0730" }),
                    (378635400, FixedTimespan { utc_offset: 28800, dst_offset: 0, name: "+08" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 24925,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::Turkey => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840147752, FixedTimespan { utc_offset: 7016, dst_offset: 0, name: "IMT" }),
                    (-1869875816, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1693706400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1680490800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1570413600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1552186800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1538359200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1522551600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1507514400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1490583600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1440208800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1428030000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1409709600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-1396494000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-931053600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-922676400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-917834400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-892436400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-875844000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-764737200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-744343200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-733806000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-716436000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-701924400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-684986400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-670474800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-654141600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-639025200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-622087200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-606970800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-590032800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-575521200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-235620000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-194842800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-177732000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (-165726000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (107910000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (121215600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (133920000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (152665200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (164678400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (184114800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (196214400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (215564400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (228873600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (245804400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (260323200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (267915600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (428454000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "+04" }),
                    (433893600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                    (468111600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (482799600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (496710000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (512521200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (528246000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (543970800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (559695600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (575420400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (591145200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (606870000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (622594800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (638319600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (654649200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (670374000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686098800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (701823600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (717548400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (733273200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (748998000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (764118000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (780447600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (796172400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (811897200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (828226800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (846370800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (859676400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (877820400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (891126000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (909270000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (922575600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (941324400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (954025200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (972774000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (985474800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1004223600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1017529200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1035673200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1048978800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1067122800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1080428400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1099177200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1111878000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1130626800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1143327600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1162076400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1174784400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1193533200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1206838800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1224982800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1238288400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1256432400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1269738000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1288486800, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1301266800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1319936400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1332637200, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1351386000, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1364691600, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1382835600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1396220400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1414285200, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1427590800, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1446933600, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (1459040400, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (1473195600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "+03" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 6952,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::UCT => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "UTC",
                    },
                    rest: REST
                }
            },

            Tz::US__Alaska => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3225223727, FixedTimespan { utc_offset: -35976, dst_offset: 0, name: "LMT" }),
                    (-2188951224, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AST" }),
                    (-880200000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AWT" }),
                    (-769395600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "APT" }),
                    (-765378000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AST" }),
                    (-86882400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (-21470400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (-5749200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (9979200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (25700400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (41428800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (57754800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (73483200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (89204400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (104932800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (120654000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (126705600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (152103600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (162388800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (183553200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (199281600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (215607600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (230731200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (247057200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (262785600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (278506800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (294235200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (309956400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (325684800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (341406000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (357134400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (372855600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (388584000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (404910000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (420033600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "AHDT" }),
                    (436359600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "YST" }),
                    (439030800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (452084400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (467805600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (483534000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (499255200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (514983600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (530704800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (544618800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (562154400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (576068400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (594208800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (607518000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (625658400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (638967600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (657108000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (671022000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (688557600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (702471600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (720007200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (733921200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (752061600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (765370800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (783511200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (796820400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (814960800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (828874800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (846410400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (860324400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (877860000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (891774000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (909309600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (923223600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (941364000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (954673200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (972813600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (986122800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1004263200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1018177200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1035712800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1049626800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1067162400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1081076400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1099216800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1112526000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1130666400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1143975600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1162116000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1173610800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1194170400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1205060400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1225620000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1236510000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1257069600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1268564400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1289124000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1300014000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1320573600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1331463600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1352023200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1362913200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1383472800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1394362800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1414922400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1425812400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1446372000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1457866800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1478426400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1489316400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1509876000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1520766000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1541325600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1552215600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1572775200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1583665200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1604224800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1615719600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1636279200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1647169200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1667728800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1678618800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1699178400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1710068400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1730628000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1741518000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1762077600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1772967600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1793527200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1805022000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1825581600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1836471600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1857031200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1867921200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1888480800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1899370800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1919930400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1930820400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1951380000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1962874800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (1983434400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (1994324400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2014884000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2025774000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2046333600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2057223600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2077783200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2088673200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2109232800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2120122800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2140682400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2152177200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2172736800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2183626800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2204186400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2215076400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2235636000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2246526000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2267085600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2277975600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2298535200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2309425200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2329984800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2341479600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2362039200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2372929200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2393488800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2404378800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2424938400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2435828400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2456388000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2467278000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2487837600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2499332400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2519892000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2530782000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2551341600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2562231600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2582791200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2593681200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2614240800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2625130800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2645690400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2656580400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2677140000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2688634800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2709194400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2720084400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2740644000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2751534000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2772093600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2782983600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2803543200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2814433200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2834992800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2846487600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2867047200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2877937200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2898496800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2909386800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2929946400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2940836400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2961396000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (2972286000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (2992845600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3003735600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3024295200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3035790000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3056349600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3067239600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3087799200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3098689200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3119248800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3130138800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3150698400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3161588400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3182148000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3193038000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3213597600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3225092400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3245652000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3256542000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3277101600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3287991600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3308551200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3319441200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3340000800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3350890800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3371450400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3382945200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3403504800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3414394800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3434954400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3445844400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3466404000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3477294000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3497853600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3508743600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3529303200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3540193200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3560752800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3572247600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3592807200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3603697200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3624256800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3635146800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3655706400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3666596400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3687156000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3698046000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3718605600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3730100400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3750660000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3761550000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3782109600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3792999600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3813559200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3824449200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3845008800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3855898800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3876458400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3887348400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3907908000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3919402800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3939962400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3950852400, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (3971412000, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (3982302000, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4002861600, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4013751600, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4034311200, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4045201200, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4065760800, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                    (4076650800, FixedTimespan { utc_offset: -32400, dst_offset: 3600, name: "AKDT" }),
                    (4097210400, FixedTimespan { utc_offset: -32400, dst_offset: 0, name: "AKST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 50424,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::US__Aleutian => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-3225223727, FixedTimespan { utc_offset: -42398, dst_offset: 0, name: "LMT" }),
                    (-2188944802, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "NST" }),
                    (-880196400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "NWT" }),
                    (-769395600, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "NPT" }),
                    (-765374400, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "NST" }),
                    (-86878800, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (-21466800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (-5745600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (9982800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (25704000, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (41432400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (57758400, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (73486800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (89208000, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (104936400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (120657600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (126709200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (152107200, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (162392400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (183556800, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (199285200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (215611200, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (230734800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (247060800, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (262789200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (278510400, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (294238800, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (309960000, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (325688400, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (341409600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (357138000, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (372859200, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (388587600, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (404913600, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "BST" }),
                    (420037200, FixedTimespan { utc_offset: -39600, dst_offset: 3600, name: "BDT" }),
                    (436363200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "AHST" }),
                    (439034400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (452088000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (467809200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (483537600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (499258800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (514987200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (530708400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (544622400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (562158000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (576072000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (594212400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (607521600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (625662000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (638971200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (657111600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (671025600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (688561200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (702475200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (720010800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (733924800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (752065200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (765374400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (783514800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (796824000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (814964400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (828878400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (846414000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (860328000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (877863600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (891777600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (909313200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (923227200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (941367600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (954676800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (972817200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (986126400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1004266800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1018180800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1035716400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1049630400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1067166000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1081080000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1099220400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1112529600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1130670000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1143979200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1162119600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1173614400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1194174000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1205064000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1225623600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1236513600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1257073200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1268568000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1289127600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1300017600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1320577200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1331467200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1352026800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1362916800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1383476400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1394366400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1414926000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1425816000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1446375600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1457870400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1478430000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1489320000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1509879600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1520769600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1541329200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1552219200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1572778800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1583668800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1604228400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1615723200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1636282800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1647172800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1667732400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1678622400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1699182000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1710072000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1730631600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1741521600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1762081200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1772971200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1793530800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1805025600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1825585200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1836475200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1857034800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1867924800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1888484400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1899374400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1919934000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1930824000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1951383600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1962878400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (1983438000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (1994328000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2014887600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2025777600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2046337200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2057227200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2077786800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2088676800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2109236400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2120126400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2140686000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2152180800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2172740400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2183630400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2204190000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2215080000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2235639600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2246529600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2267089200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2277979200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2298538800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2309428800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2329988400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2341483200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2362042800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2372932800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2393492400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2404382400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2424942000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2435832000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2456391600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2467281600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2487841200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2499336000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2519895600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2530785600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2551345200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2562235200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2582794800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2593684800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2614244400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2625134400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2645694000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2656584000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2677143600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2688638400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2709198000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2720088000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2740647600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2751537600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2772097200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2782987200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2803546800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2814436800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2834996400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2846491200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2867050800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2877940800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2898500400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2909390400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2929950000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2940840000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2961399600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (2972289600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (2992849200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3003739200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3024298800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3035793600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3056353200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3067243200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3087802800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3098692800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3119252400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3130142400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3150702000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3161592000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3182151600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3193041600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3213601200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3225096000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3245655600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3256545600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3277105200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3287995200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3308554800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3319444800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3340004400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3350894400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3371454000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3382948800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3403508400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3414398400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3434958000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3445848000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3466407600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3477297600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3497857200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3508747200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3529306800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3540196800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3560756400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3572251200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3592810800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3603700800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3624260400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3635150400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3655710000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3666600000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3687159600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3698049600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3718609200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3730104000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3750663600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3761553600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3782113200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3793003200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3813562800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3824452800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3845012400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3855902400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3876462000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3887352000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3907911600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3919406400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3939966000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3950856000, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (3971415600, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (3982305600, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (4002865200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (4013755200, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (4034314800, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (4045204800, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (4065764400, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                    (4076654400, FixedTimespan { utc_offset: -36000, dst_offset: 3600, name: "HDT" }),
                    (4097214000, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 44002,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::US__Arizona => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717616702, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1633273200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1601823600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1583683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-820519140, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-812653140, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-796845540, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-84380400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-68659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -26898,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::US__Central => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626164, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1563724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1551632400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1538928000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1520182800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1504454400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1491757200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1473004800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1459702800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1441555200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1428253200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1410105600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1396803600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1378656000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1365354000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1347206400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1333904400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1315152000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1301850000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1283702400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1270400400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1252252800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1238950800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1220803200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1207501200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1189353600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1176051600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1157299200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1144602000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1125849600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1112547600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1094400000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1081098000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1067788800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1045414800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1031500800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1018198800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1000051200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-986749200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-967996800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-955299600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-936547200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-923245200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-905097600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-891795600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-747244800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-733942800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-715795200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-702493200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-684345600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-671043600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-652896000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-639594000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-620841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-608144400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-557942400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-544640400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-447267600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-431539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-415818000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-400089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-384368400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-352918800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-337190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-321469200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-305740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-289414800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-273686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-257965200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-242236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-226515600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-210787200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-195066000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-179337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-163616400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-147888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-131562000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-116438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-100112400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-84384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-68662800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-52934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-37213200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-21484800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-5763600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (9964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (25686000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (41414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (57740400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (73468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (89190000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (104918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (120639600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (126691200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152089200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (162374400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (183538800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (199267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (215593200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (230716800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (247042800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (262771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (278492400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (294220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (309942000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (325670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (341391600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (357120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (372841200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (388569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (404895600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (420019200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (436345200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (452073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (467794800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (483523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (499244400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (514972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (530694000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (544608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (562143600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (576057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (594198000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (607507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (625647600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (638956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (657097200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (671011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (688546800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (702460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (719996400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (733910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (752050800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (765360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (783500400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (796809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (814950000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (828864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (846399600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (860313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (877849200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (891763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (909298800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (923212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (941353200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (954662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (972802800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (986112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1004252400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1018166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1035702000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1049616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1067151600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1081065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1099206000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1112515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1130655600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1143964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1205049600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1236499200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1257058800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -21036,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::US__EastIndiana => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626522, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-900259200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-891795600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-747244800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-733942800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-715795200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-702493200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-684345600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-671043600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-652896000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-639594000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-620841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-608144400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-557942400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-544640400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-386787600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20678,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::US__Eastern => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717633038, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1633280400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1615140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1601830800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1583690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1570381200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1551636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1536512400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1523210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1504458000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1491760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1473008400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1459706400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1441558800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1428256800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1410109200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1396807200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1378659600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1365357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1347210000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1333908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1315155600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1301853600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1283706000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1270404000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1252256400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1238954400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1220806800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1207504800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1189357200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1176055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1157302800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1144605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1125853200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1112551200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1094403600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1081101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1062954000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1049652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1031504400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-1018202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-1000054800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-986752800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-968000400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-955303200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-936550800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-923248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-905101200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-891799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-880218000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EWT" }),
                    (-769395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EPT" }),
                    (-765396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-747248400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-733946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-715798800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-702496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-684349200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-671047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-652899600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-639597600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-620845200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-608148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-589395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-576093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-557946000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-544644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-526496400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-513194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-495046800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-481744800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-463597200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-447271200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-431542800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-415821600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-400093200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-384372000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-368643600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-352922400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-337194000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-321472800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-305744400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-289418400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-273690000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-257968800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-242240400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-226519200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-210790800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-195069600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-179341200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-163620000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-147891600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-131565600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-116442000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-100116000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-84387600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-68666400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-52938000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-37216800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-21488400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-5767200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (9961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (25682400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (41410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (57736800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (73465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (89186400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (126687600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (152085600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (162370800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (954658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (972799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -17762,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::US__Hawaii => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2334101314, FixedTimespan { utc_offset: -37800, dst_offset: 0, name: "HST" }),
                    (-1157283000, FixedTimespan { utc_offset: -37800, dst_offset: 3600, name: "HDT" }),
                    (-1155436200, FixedTimespan { utc_offset: -37800, dst_offset: 0, name: "HST" }),
                    (-880198200, FixedTimespan { utc_offset: -37800, dst_offset: 3600, name: "HWT" }),
                    (-769395600, FixedTimespan { utc_offset: -37800, dst_offset: 3600, name: "HPT" }),
                    (-765376200, FixedTimespan { utc_offset: -37800, dst_offset: 0, name: "HST" }),
                    (-712150200, FixedTimespan { utc_offset: -36000, dst_offset: 0, name: "HST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -37886,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::US__IndianaStarke => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717626410, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1633276800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1615136400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1601827200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-1583686800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-880214400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CWT" }),
                    (-769395600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CPT" }),
                    (-765392400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-715795200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-702493200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-684345600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-671043600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-652896000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-639594000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-620841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-608144400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-589392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-576090000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-557942400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-544640400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-526492800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-513190800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-495043200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-481741200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-463593600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-447267600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-431539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-415818000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-400089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-386787600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-368640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-355338000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-337190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-321469200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-305740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-289414800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-273686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-257965200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-242236800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-195066000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-84384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-68662800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-52934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-37213200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-21484800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (-5763600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (9964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (25686000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (41414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (57740400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (73468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (89190000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (104918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (120639600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (126691200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (152089200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (162374400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (183538800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (199267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (215593200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (230716800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (247042800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (262771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (278492400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (294220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (309942000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (325670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (341391600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (357120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (372841200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (388569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (404895600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (420019200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (436345200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (452073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (467794800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (483523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (499244400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (514972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (530694000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (544608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (562143600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (576057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (594198000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (607507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (625647600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (638956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (657097200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (671011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (688546800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1162105200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1173600000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1194159600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1205049600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1225609200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1236499200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1257058800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1268553600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1289113200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1300003200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1320562800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1331452800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1352012400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1362902400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1383462000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1394352000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1414911600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1425801600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1446361200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1457856000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1478415600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1489305600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1509865200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1520755200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1541314800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1552204800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1572764400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1583654400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1604214000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1615708800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1636268400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1647158400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1667718000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1678608000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1699167600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1710057600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1730617200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1741507200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1762066800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1772956800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1793516400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1805011200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1825570800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1836460800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1857020400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1867910400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1888470000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1899360000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1919919600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1930809600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1951369200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1962864000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (1983423600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (1994313600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2014873200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2025763200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2046322800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2057212800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2077772400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2088662400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2109222000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2120112000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2140671600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2152166400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2172726000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2183616000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2204175600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2215065600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2235625200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2246515200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2267074800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2277964800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2298524400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2309414400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2329974000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2341468800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2362028400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2372918400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2393478000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2404368000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2424927600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2435817600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2456377200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2467267200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2487826800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2499321600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2519881200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2530771200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2551330800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2562220800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2582780400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2593670400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2614230000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2625120000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2645679600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2656569600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2677129200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2688624000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2709183600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2720073600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2740633200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2751523200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2772082800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2782972800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2803532400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2814422400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2834982000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2846476800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2867036400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2877926400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2898486000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2909376000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2929935600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2940825600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2961385200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (2972275200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (2992834800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3003724800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3024284400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3035779200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3056338800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3067228800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3087788400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3098678400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3119238000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3130128000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3150687600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3161577600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3182137200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3193027200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3213586800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3225081600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3245641200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3256531200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3277090800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3287980800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3308540400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3319430400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3339990000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3350880000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3371439600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3382934400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3403494000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3414384000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3434943600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3445833600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3466393200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3477283200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3497842800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3508732800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3529292400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3540182400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3560742000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3572236800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3592796400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3603686400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3624246000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3635136000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3655695600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3666585600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3687145200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3698035200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3718594800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3730089600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3750649200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3761539200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3782098800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3792988800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3813548400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3824438400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3844998000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3855888000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3876447600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3887337600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3907897200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3919392000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3939951600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3950841600, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (3971401200, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (3982291200, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4002850800, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4013740800, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4034300400, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4045190400, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4065750000, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (4076640000, FixedTimespan { utc_offset: -21600, dst_offset: 3600, name: "CDT" }),
                    (4097199600, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -20790,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::US__Michigan => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2051202469, FixedTimespan { utc_offset: -21600, dst_offset: 0, name: "CST" }),
                    (-1724083200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-880218000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EWT" }),
                    (-769395600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EPT" }),
                    (-765396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-684349200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-671047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-80506740, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-68666400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (-52938000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (-37216800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (104914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (120636000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (126687600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (152085600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (167814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (183535200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (199263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (215589600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (230713200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (247039200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (262767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (278488800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (294217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (309938400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (325666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (341388000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (357116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (372837600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (388566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (404892000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (420015600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (436341600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (452070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (467791200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (483519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (499240800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (514969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (530690400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (544604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (562140000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (576054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (594194400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (607503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (625644000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (638953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (657093600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (671007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (688543200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (702457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (719992800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (733906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (752047200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (765356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (783496800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (796806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (814946400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (828860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (846396000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (860310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (877845600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (891759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (909295200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (923209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (941349600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (954658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (972799200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (986108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1004248800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1018162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1035698400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1049612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1067148000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1081062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1099202400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1112511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1130652000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1143961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1162101600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1173596400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1194156000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1205046000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1225605600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1236495600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1257055200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1268550000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1289109600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1299999600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1320559200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1331449200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1352008800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1362898800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1383458400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1394348400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1414908000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1425798000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1446357600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1457852400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1478412000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1489302000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1509861600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1520751600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1541311200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1552201200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1572760800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1583650800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1604210400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1615705200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1636264800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1647154800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1667714400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1678604400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1699164000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1710054000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1730613600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1741503600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1762063200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1772953200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1793512800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1805007600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1825567200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1836457200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1857016800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1867906800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1888466400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1899356400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1919916000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1930806000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1951365600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1962860400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (1983420000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (1994310000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2014869600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2025759600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2046319200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2057209200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2077768800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2088658800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2109218400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2120108400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2140668000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2152162800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2172722400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2183612400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2204172000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2215062000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2235621600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2246511600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2267071200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2277961200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2298520800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2309410800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2329970400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2341465200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2362024800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2372914800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2393474400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2404364400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2424924000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2435814000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2456373600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2467263600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2487823200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2499318000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2519877600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2530767600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2551327200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2562217200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2582776800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2593666800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2614226400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2625116400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2645676000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2656566000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2677125600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2688620400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2709180000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2720070000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2740629600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2751519600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2772079200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2782969200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2803528800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2814418800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2834978400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2846473200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2867032800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2877922800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2898482400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2909372400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2929932000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2940822000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2961381600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (2972271600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (2992831200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3003721200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3024280800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3035775600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3056335200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3067225200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3087784800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3098674800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3119234400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3130124400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3150684000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3161574000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3182133600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3193023600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3213583200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3225078000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3245637600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3256527600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3277087200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3287977200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3308536800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3319426800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3339986400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3350876400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3371436000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3382930800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3403490400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3414380400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3434940000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3445830000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3466389600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3477279600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3497839200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3508729200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3529288800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3540178800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3560738400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3572233200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3592792800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3603682800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3624242400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3635132400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3655692000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3666582000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3687141600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3698031600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3718591200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3730086000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3750645600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3761535600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3782095200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3792985200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3813544800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3824434800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3844994400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3855884400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3876444000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3887334000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3907893600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3919388400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3939948000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3950838000, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (3971397600, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (3982287600, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4002847200, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4013737200, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4034296800, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4045186800, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4065746400, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                    (4076636400, FixedTimespan { utc_offset: -18000, dst_offset: 3600, name: "EDT" }),
                    (4097196000, FixedTimespan { utc_offset: -18000, dst_offset: 0, name: "EST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -19931,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::US__Mountain => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717618404, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1633273200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1615132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1601823600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1583683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1570374000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1551628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-1538924400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-1534089600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-880210800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MWT" }),
                    (-769395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MPT" }),
                    (-765388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-147884400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-131558400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-116434800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-100108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-84380400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-68659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-52930800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-37209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (-21481200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (-5760000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (9968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (25689600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (41418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (57744000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (73472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (89193600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (104922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (120643200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (126694800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (152092800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (162378000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (183542400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (199270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (215596800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (230720400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (247046400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (262774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (278496000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (294224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (309945600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (325674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (341395200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (357123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (372844800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (388573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (404899200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (420022800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (436348800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (452077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (467798400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (483526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (499248000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (514976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (530697600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (544611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (562147200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (576061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (594201600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (607510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (625651200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (638960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (657100800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (671014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (688550400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (702464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (720000000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (733914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (752054400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (765363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (783504000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (796813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (814953600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (828867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (846403200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (860317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (877852800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (891766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (909302400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (923216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (941356800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (954666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (972806400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (986115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1004256000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1018170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1035705600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1049619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1067155200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1081069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1099209600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1112518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1130659200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1143968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1162108800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1173603600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1194163200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1205053200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1225612800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1236502800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1257062400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1268557200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1289116800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1300006800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1320566400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1331456400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1352016000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1362906000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1383465600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1394355600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1414915200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1425805200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1446364800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1457859600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1478419200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1489309200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1509868800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1520758800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1541318400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1552208400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1572768000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1583658000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1604217600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1615712400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1636272000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1647162000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1667721600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1678611600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1699171200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1710061200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1730620800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1741510800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1762070400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1772960400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1793520000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1805014800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1825574400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1836464400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1857024000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1867914000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1888473600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1899363600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1919923200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1930813200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1951372800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1962867600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (1983427200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (1994317200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2014876800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2025766800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2046326400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2057216400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2077776000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2088666000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2109225600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2120115600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2140675200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2152170000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2172729600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2183619600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2204179200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2215069200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2235628800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2246518800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2267078400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2277968400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2298528000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2309418000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2329977600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2341472400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2362032000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2372922000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2393481600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2404371600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2424931200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2435821200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2456380800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2467270800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2487830400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2499325200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2519884800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2530774800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2551334400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2562224400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2582784000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2593674000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2614233600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2625123600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2645683200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2656573200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2677132800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2688627600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2709187200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2720077200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2740636800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2751526800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2772086400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2782976400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2803536000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2814426000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2834985600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2846480400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2867040000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2877930000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2898489600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2909379600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2929939200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2940829200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2961388800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (2972278800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (2992838400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3003728400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3024288000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3035782800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3056342400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3067232400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3087792000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3098682000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3119241600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3130131600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3150691200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3161581200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3182140800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3193030800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3213590400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3225085200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3245644800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3256534800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3277094400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3287984400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3308544000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3319434000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3339993600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3350883600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3371443200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3382938000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3403497600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3414387600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3434947200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3445837200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3466396800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3477286800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3497846400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3508736400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3529296000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3540186000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3560745600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3572240400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3592800000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3603690000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3624249600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3635139600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3655699200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3666589200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3687148800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3698038800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3718598400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3730093200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3750652800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3761542800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3782102400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3792992400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3813552000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3824442000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3845001600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3855891600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3876451200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3887341200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3907900800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3919395600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3939955200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3950845200, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (3971404800, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (3982294800, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4002854400, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4013744400, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4034304000, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4045194000, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4065753600, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                    (4076643600, FixedTimespan { utc_offset: -25200, dst_offset: 3600, name: "MDT" }),
                    (4097203200, FixedTimespan { utc_offset: -25200, dst_offset: 0, name: "MST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -25196,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::US__Pacific => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2717611622, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1633269600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1615129200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-1601820000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-1583679600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-880207200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PWT" }),
                    (-769395600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PPT" }),
                    (-765385200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-687967140, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-662655600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-620838000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-608137200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-589388400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-576082800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-557938800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-544633200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-526489200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-513183600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-495039600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-481734000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-463590000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-450284400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-431535600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-418230000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-400086000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-386780400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-368636400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-355330800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-337186800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-323881200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-305737200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-292431600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-273682800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-260982000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-242233200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-226508400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-210783600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-195058800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-179334000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-163609200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-147884400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-131554800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-116434800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-100105200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-84376800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-68655600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-52927200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-37206000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (-21477600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (-5756400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (9972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (25693200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (41421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (57747600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (73476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (89197200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (104925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (120646800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (126698400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (152096400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (162381600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (183546000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (199274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (215600400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (230724000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (247050000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (262778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (278499600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (294228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (309949200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (325677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (341398800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (357127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (372848400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (388576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (404902800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (420026400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (436352400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (452080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (467802000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (483530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (499251600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (514980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (530701200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (544615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (562150800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (576064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (594205200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (607514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (625654800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (638964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (657104400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (671018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (688554000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (702468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (720003600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (733917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (752058000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (765367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (783507600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (796816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (814957200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (828871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (846406800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (860320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (877856400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (891770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (909306000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (923220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (941360400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (954669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (972810000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (986119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1004259600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1018173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1035709200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1049623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1067158800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1081072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1099213200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1112522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1130662800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1143972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1162112400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1173607200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1194166800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1205056800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1225616400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1236506400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1257066000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1268560800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1289120400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1300010400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1320570000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1331460000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1352019600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1362909600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1383469200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1394359200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1414918800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1425808800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1446368400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1457863200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1478422800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1489312800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1509872400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1520762400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1541322000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1552212000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1572771600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1583661600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1604221200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1615716000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1636275600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1647165600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1667725200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1678615200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1699174800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1710064800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1730624400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1741514400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1762074000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1772964000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1793523600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1805018400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1825578000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1836468000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1857027600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1867917600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1888477200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1899367200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1919926800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1930816800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1951376400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1962871200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (1983430800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (1994320800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2014880400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2025770400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2046330000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2057220000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2077779600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2088669600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2109229200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2120119200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2140678800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2152173600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2172733200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2183623200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2204182800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2215072800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2235632400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2246522400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2267082000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2277972000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2298531600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2309421600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2329981200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2341476000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2362035600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2372925600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2393485200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2404375200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2424934800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2435824800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2456384400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2467274400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2487834000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2499328800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2519888400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2530778400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2551338000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2562228000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2582787600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2593677600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2614237200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2625127200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2645686800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2656576800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2677136400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2688631200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2709190800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2720080800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2740640400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2751530400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2772090000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2782980000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2803539600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2814429600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2834989200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2846484000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2867043600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2877933600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2898493200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2909383200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2929942800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2940832800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2961392400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (2972282400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (2992842000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3003732000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3024291600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3035786400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3056346000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3067236000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3087795600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3098685600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3119245200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3130135200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3150694800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3161584800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3182144400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3193034400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3213594000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3225088800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3245648400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3256538400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3277098000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3287988000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3308547600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3319437600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3339997200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3350887200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3371446800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3382941600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3403501200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3414391200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3434950800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3445840800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3466400400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3477290400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3497850000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3508740000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3529299600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3540189600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3560749200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3572244000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3592803600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3603693600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3624253200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3635143200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3655702800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3666592800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3687152400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3698042400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3718602000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3730096800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3750656400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3761546400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3782106000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3792996000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3813555600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3824445600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3845005200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3855895200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3876454800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3887344800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3907904400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3919399200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3939958800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3950848800, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (3971408400, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (3982298400, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4002858000, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4013748000, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4034307600, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4045197600, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4065757200, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                    (4076647200, FixedTimespan { utc_offset: -28800, dst_offset: 3600, name: "PDT" }),
                    (4097206800, FixedTimespan { utc_offset: -28800, dst_offset: 0, name: "PST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: -28378,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::US__Samoa => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2445424632, FixedTimespan { utc_offset: -40968, dst_offset: 0, name: "LMT" }),
                    (-1861879032, FixedTimespan { utc_offset: -39600, dst_offset: 0, name: "SST" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 45432,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::UTC => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "UTC",
                    },
                    rest: REST
                }
            },

            Tz::Universal => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "UTC",
                    },
                    rest: REST
                }
            },

            Tz::WSU => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (-2840149817, FixedTimespan { utc_offset: 9017, dst_offset: 0, name: "MMT" }),
                    (-1688265017, FixedTimespan { utc_offset: 9079, dst_offset: 0, name: "MMT" }),
                    (-1656819079, FixedTimespan { utc_offset: 9079, dst_offset: 3600, name: "MST" }),
                    (-1641353479, FixedTimespan { utc_offset: 9079, dst_offset: 0, name: "MMT" }),
                    (-1627965079, FixedTimespan { utc_offset: 9079, dst_offset: 7200, name: "MDST" }),
                    (-1618716679, FixedTimespan { utc_offset: 9079, dst_offset: 3600, name: "MST" }),
                    (-1596429079, FixedTimespan { utc_offset: 9079, dst_offset: 7200, name: "MDST" }),
                    (-1593837079, FixedTimespan { utc_offset: 10800, dst_offset: 7200, name: "MDST" }),
                    (-1593820800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (-1589860800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (-1542427200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (-1539493200, FixedTimespan { utc_offset: 10800, dst_offset: 7200, name: "+05" }),
                    (-1525323600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (-1522728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (-1491188400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (-1247536800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (354920400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (370728000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (386456400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (402264000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (417992400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (433800000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (449614800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (465346800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (481071600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (496796400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (512521200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (528246000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (543970800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (559695600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (575420400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (591145200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (606870000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (622594800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (638319600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (654649200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (670374000, FixedTimespan { utc_offset: 7200, dst_offset: 3600, name: "EEST" }),
                    (686102400, FixedTimespan { utc_offset: 7200, dst_offset: 0, name: "EET" }),
                    (695779200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (701823600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (717548400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (733273200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (748998000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (764722800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (780447600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (796172400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (811897200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (828226800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (846370800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (859676400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (877820400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (891126000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (909270000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (922575600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (941324400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (954025200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (972774000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (985474800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1004223600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1017529200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1035673200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1048978800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1067122800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1080428400, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1099177200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1111878000, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1130626800, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1143327600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1162076400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1174777200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1193526000, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1206831600, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1224975600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1238281200, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1256425200, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1269730800, FixedTimespan { utc_offset: 10800, dst_offset: 3600, name: "MSD" }),
                    (1288479600, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                    (1301180400, FixedTimespan { utc_offset: 14400, dst_offset: 0, name: "MSK" }),
                    (1414274400, FixedTimespan { utc_offset: 10800, dst_offset: 0, name: "MSK" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 9017,
                        dst_offset: 0,
                        name: "LMT",
                    },
                    rest: REST
                }
            },

            Tz::WET => {
                const REST: &[(i64, FixedTimespan)] = &[
                    (228877200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (243997200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (260326800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (276051600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (291776400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (307501200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (323830800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (338950800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (354675600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (370400400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (386125200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (401850000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (417574800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (433299600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (449024400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (465354000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (481078800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (496803600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (512528400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (528253200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (543978000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (559702800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (575427600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (591152400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (606877200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (622602000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (638326800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (654656400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (670381200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (686106000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (701830800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (717555600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (733280400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (749005200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (764730000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (780454800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (796179600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (811904400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (828234000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (846378000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (859683600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (877827600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (891133200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (909277200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (922582800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (941331600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (954032400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (972781200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (985482000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1004230800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1017536400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1035680400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1048986000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1067130000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1080435600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1099184400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1111885200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1130634000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1143334800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1162083600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1174784400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1193533200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1206838800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1224982800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1238288400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1256432400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1269738000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1288486800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1301187600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1319936400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1332637200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1351386000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1364691600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1382835600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1396141200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1414285200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1427590800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1445734800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1459040400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1477789200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1490490000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1509238800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1521939600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1540688400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1553994000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1572138000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1585443600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1603587600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1616893200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1635642000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1648342800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1667091600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1679792400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1698541200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1711846800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1729990800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1743296400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1761440400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1774746000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1792890000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1806195600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1824944400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1837645200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1856394000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1869094800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1887843600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1901149200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1919293200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1932598800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1950742800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1964048400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (1982797200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (1995498000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2014246800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2026947600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2045696400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2058397200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2077146000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2090451600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2108595600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2121901200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2140045200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2153350800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2172099600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2184800400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2203549200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2216250000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2234998800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2248304400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2266448400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2279754000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2297898000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2311203600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2329347600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2342653200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2361402000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2374102800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2392851600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2405552400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2424301200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2437606800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2455750800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2469056400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2487200400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2500506000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2519254800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2531955600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2550704400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2563405200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2582154000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2595459600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2613603600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2626909200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2645053200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2658358800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2676502800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2689808400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2708557200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2721258000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2740006800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2752707600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2771456400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2784762000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2802906000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2816211600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2834355600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2847661200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2866410000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2879110800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2897859600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2910560400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2929309200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2942010000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2960758800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (2974064400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (2992208400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3005514000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3023658000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3036963600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3055712400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3068413200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3087162000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3099862800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3118611600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3131917200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3150061200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3163366800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3181510800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3194816400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3212960400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3226266000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3245014800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3257715600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3276464400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3289165200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3307914000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3321219600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3339363600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3352669200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3370813200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3384118800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3402867600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3415568400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3434317200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3447018000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3465766800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3479072400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3497216400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3510522000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3528666000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3541971600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3560115600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3573421200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3592170000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3604870800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3623619600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3636320400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3655069200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3668374800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3686518800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3699824400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3717968400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3731274000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3750022800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3762723600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3781472400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3794173200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3812922000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3825622800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3844371600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3857677200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3875821200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3889126800, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3907270800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3920576400, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3939325200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3952026000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (3970774800, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (3983475600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4002224400, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4015530000, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4033674000, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4046979600, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4065123600, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                    (4078429200, FixedTimespan { utc_offset: 0, dst_offset: 3600, name: "WEST" }),
                    (4096573200, FixedTimespan { utc_offset: 0, dst_offset: 0, name: "WET" }),
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "WET",
                    },
                    rest: REST
                }
            },

            Tz::Zulu => {
                const REST: &[(i64, FixedTimespan)] = &[
                ];
                FixedTimespanSet {
                    first: FixedTimespan {
                        utc_offset: 0,
                        dst_offset: 0,
                        name: "UTC",
                    },
                    rest: REST
                }
            },

         }
    }
}
/// An array of every known variant
///
/// Useful for iterating over known timezones:
///
/// ```
/// use chrono_tz::{TZ_VARIANTS, Tz};
/// assert!(TZ_VARIANTS.iter().any(|v| *v == Tz::UTC));
/// ```
pub static TZ_VARIANTS: [Tz; 596] = [
    Tz::Africa__Abidjan,
    Tz::Africa__Accra,
    Tz::Africa__Addis_Ababa,
    Tz::Africa__Algiers,
    Tz::Africa__Asmara,
    Tz::Africa__Asmera,
    Tz::Africa__Bamako,
    Tz::Africa__Bangui,
    Tz::Africa__Banjul,
    Tz::Africa__Bissau,
    Tz::Africa__Blantyre,
    Tz::Africa__Brazzaville,
    Tz::Africa__Bujumbura,
    Tz::Africa__Cairo,
    Tz::Africa__Casablanca,
    Tz::Africa__Ceuta,
    Tz::Africa__Conakry,
    Tz::Africa__Dakar,
    Tz::Africa__Dar_es_Salaam,
    Tz::Africa__Djibouti,
    Tz::Africa__Douala,
    Tz::Africa__El_Aaiun,
    Tz::Africa__Freetown,
    Tz::Africa__Gaborone,
    Tz::Africa__Harare,
    Tz::Africa__Johannesburg,
    Tz::Africa__Juba,
    Tz::Africa__Kampala,
    Tz::Africa__Khartoum,
    Tz::Africa__Kigali,
    Tz::Africa__Kinshasa,
    Tz::Africa__Lagos,
    Tz::Africa__Libreville,
    Tz::Africa__Lome,
    Tz::Africa__Luanda,
    Tz::Africa__Lubumbashi,
    Tz::Africa__Lusaka,
    Tz::Africa__Malabo,
    Tz::Africa__Maputo,
    Tz::Africa__Maseru,
    Tz::Africa__Mbabane,
    Tz::Africa__Mogadishu,
    Tz::Africa__Monrovia,
    Tz::Africa__Nairobi,
    Tz::Africa__Ndjamena,
    Tz::Africa__Niamey,
    Tz::Africa__Nouakchott,
    Tz::Africa__Ouagadougou,
    Tz::Africa__PortoNovo,
    Tz::Africa__Sao_Tome,
    Tz::Africa__Timbuktu,
    Tz::Africa__Tripoli,
    Tz::Africa__Tunis,
    Tz::Africa__Windhoek,
    Tz::America__Adak,
    Tz::America__Anchorage,
    Tz::America__Anguilla,
    Tz::America__Antigua,
    Tz::America__Araguaina,
    Tz::America__Argentina__Buenos_Aires,
    Tz::America__Argentina__Catamarca,
    Tz::America__Argentina__ComodRivadavia,
    Tz::America__Argentina__Cordoba,
    Tz::America__Argentina__Jujuy,
    Tz::America__Argentina__La_Rioja,
    Tz::America__Argentina__Mendoza,
    Tz::America__Argentina__Rio_Gallegos,
    Tz::America__Argentina__Salta,
    Tz::America__Argentina__San_Juan,
    Tz::America__Argentina__San_Luis,
    Tz::America__Argentina__Tucuman,
    Tz::America__Argentina__Ushuaia,
    Tz::America__Aruba,
    Tz::America__Asuncion,
    Tz::America__Atikokan,
    Tz::America__Atka,
    Tz::America__Bahia,
    Tz::America__Bahia_Banderas,
    Tz::America__Barbados,
    Tz::America__Belem,
    Tz::America__Belize,
    Tz::America__BlancSablon,
    Tz::America__Boa_Vista,
    Tz::America__Bogota,
    Tz::America__Boise,
    Tz::America__Buenos_Aires,
    Tz::America__Cambridge_Bay,
    Tz::America__Campo_Grande,
    Tz::America__Cancun,
    Tz::America__Caracas,
    Tz::America__Catamarca,
    Tz::America__Cayenne,
    Tz::America__Cayman,
    Tz::America__Chicago,
    Tz::America__Chihuahua,
    Tz::America__Ciudad_Juarez,
    Tz::America__Coral_Harbour,
    Tz::America__Cordoba,
    Tz::America__Costa_Rica,
    Tz::America__Creston,
    Tz::America__Cuiaba,
    Tz::America__Curacao,
    Tz::America__Danmarkshavn,
    Tz::America__Dawson,
    Tz::America__Dawson_Creek,
    Tz::America__Denver,
    Tz::America__Detroit,
    Tz::America__Dominica,
    Tz::America__Edmonton,
    Tz::America__Eirunepe,
    Tz::America__El_Salvador,
    Tz::America__Ensenada,
    Tz::America__Fort_Nelson,
    Tz::America__Fort_Wayne,
    Tz::America__Fortaleza,
    Tz::America__Glace_Bay,
    Tz::America__Godthab,
    Tz::America__Goose_Bay,
    Tz::America__Grand_Turk,
    Tz::America__Grenada,
    Tz::America__Guadeloupe,
    Tz::America__Guatemala,
    Tz::America__Guayaquil,
    Tz::America__Guyana,
    Tz::America__Halifax,
    Tz::America__Havana,
    Tz::America__Hermosillo,
    Tz::America__Indiana__Indianapolis,
    Tz::America__Indiana__Knox,
    Tz::America__Indiana__Marengo,
    Tz::America__Indiana__Petersburg,
    Tz::America__Indiana__Tell_City,
    Tz::America__Indiana__Vevay,
    Tz::America__Indiana__Vincennes,
    Tz::America__Indiana__Winamac,
    Tz::America__Indianapolis,
    Tz::America__Inuvik,
    Tz::America__Iqaluit,
    Tz::America__Jamaica,
    Tz::America__Jujuy,
    Tz::America__Juneau,
    Tz::America__Kentucky__Louisville,
    Tz::America__Kentucky__Monticello,
    Tz::America__Knox_IN,
    Tz::America__Kralendijk,
    Tz::America__La_Paz,
    Tz::America__Lima,
    Tz::America__Los_Angeles,
    Tz::America__Louisville,
    Tz::America__Lower_Princes,
    Tz::America__Maceio,
    Tz::America__Managua,
    Tz::America__Manaus,
    Tz::America__Marigot,
    Tz::America__Martinique,
    Tz::America__Matamoros,
    Tz::America__Mazatlan,
    Tz::America__Mendoza,
    Tz::America__Menominee,
    Tz::America__Merida,
    Tz::America__Metlakatla,
    Tz::America__Mexico_City,
    Tz::America__Miquelon,
    Tz::America__Moncton,
    Tz::America__Monterrey,
    Tz::America__Montevideo,
    Tz::America__Montreal,
    Tz::America__Montserrat,
    Tz::America__Nassau,
    Tz::America__New_York,
    Tz::America__Nipigon,
    Tz::America__Nome,
    Tz::America__Noronha,
    Tz::America__North_Dakota__Beulah,
    Tz::America__North_Dakota__Center,
    Tz::America__North_Dakota__New_Salem,
    Tz::America__Nuuk,
    Tz::America__Ojinaga,
    Tz::America__Panama,
    Tz::America__Pangnirtung,
    Tz::America__Paramaribo,
    Tz::America__Phoenix,
    Tz::America__PortauPrince,
    Tz::America__Port_of_Spain,
    Tz::America__Porto_Acre,
    Tz::America__Porto_Velho,
    Tz::America__Puerto_Rico,
    Tz::America__Punta_Arenas,
    Tz::America__Rainy_River,
    Tz::America__Rankin_Inlet,
    Tz::America__Recife,
    Tz::America__Regina,
    Tz::America__Resolute,
    Tz::America__Rio_Branco,
    Tz::America__Rosario,
    Tz::America__Santa_Isabel,
    Tz::America__Santarem,
    Tz::America__Santiago,
    Tz::America__Santo_Domingo,
    Tz::America__Sao_Paulo,
    Tz::America__Scoresbysund,
    Tz::America__Shiprock,
    Tz::America__Sitka,
    Tz::America__St_Barthelemy,
    Tz::America__St_Johns,
    Tz::America__St_Kitts,
    Tz::America__St_Lucia,
    Tz::America__St_Thomas,
    Tz::America__St_Vincent,
    Tz::America__Swift_Current,
    Tz::America__Tegucigalpa,
    Tz::America__Thule,
    Tz::America__Thunder_Bay,
    Tz::America__Tijuana,
    Tz::America__Toronto,
    Tz::America__Tortola,
    Tz::America__Vancouver,
    Tz::America__Virgin,
    Tz::America__Whitehorse,
    Tz::America__Winnipeg,
    Tz::America__Yakutat,
    Tz::America__Yellowknife,
    Tz::Antarctica__Casey,
    Tz::Antarctica__Davis,
    Tz::Antarctica__DumontDUrville,
    Tz::Antarctica__Macquarie,
    Tz::Antarctica__Mawson,
    Tz::Antarctica__McMurdo,
    Tz::Antarctica__Palmer,
    Tz::Antarctica__Rothera,
    Tz::Antarctica__South_Pole,
    Tz::Antarctica__Syowa,
    Tz::Antarctica__Troll,
    Tz::Antarctica__Vostok,
    Tz::Arctic__Longyearbyen,
    Tz::Asia__Aden,
    Tz::Asia__Almaty,
    Tz::Asia__Amman,
    Tz::Asia__Anadyr,
    Tz::Asia__Aqtau,
    Tz::Asia__Aqtobe,
    Tz::Asia__Ashgabat,
    Tz::Asia__Ashkhabad,
    Tz::Asia__Atyrau,
    Tz::Asia__Baghdad,
    Tz::Asia__Bahrain,
    Tz::Asia__Baku,
    Tz::Asia__Bangkok,
    Tz::Asia__Barnaul,
    Tz::Asia__Beirut,
    Tz::Asia__Bishkek,
    Tz::Asia__Brunei,
    Tz::Asia__Calcutta,
    Tz::Asia__Chita,
    Tz::Asia__Choibalsan,
    Tz::Asia__Chongqing,
    Tz::Asia__Chungking,
    Tz::Asia__Colombo,
    Tz::Asia__Dacca,
    Tz::Asia__Damascus,
    Tz::Asia__Dhaka,
    Tz::Asia__Dili,
    Tz::Asia__Dubai,
    Tz::Asia__Dushanbe,
    Tz::Asia__Famagusta,
    Tz::Asia__Gaza,
    Tz::Asia__Harbin,
    Tz::Asia__Hebron,
    Tz::Asia__Ho_Chi_Minh,
    Tz::Asia__Hong_Kong,
    Tz::Asia__Hovd,
    Tz::Asia__Irkutsk,
    Tz::Asia__Istanbul,
    Tz::Asia__Jakarta,
    Tz::Asia__Jayapura,
    Tz::Asia__Jerusalem,
    Tz::Asia__Kabul,
    Tz::Asia__Kamchatka,
    Tz::Asia__Karachi,
    Tz::Asia__Kashgar,
    Tz::Asia__Kathmandu,
    Tz::Asia__Katmandu,
    Tz::Asia__Khandyga,
    Tz::Asia__Kolkata,
    Tz::Asia__Krasnoyarsk,
    Tz::Asia__Kuala_Lumpur,
    Tz::Asia__Kuching,
    Tz::Asia__Kuwait,
    Tz::Asia__Macao,
    Tz::Asia__Macau,
    Tz::Asia__Magadan,
    Tz::Asia__Makassar,
    Tz::Asia__Manila,
    Tz::Asia__Muscat,
    Tz::Asia__Nicosia,
    Tz::Asia__Novokuznetsk,
    Tz::Asia__Novosibirsk,
    Tz::Asia__Omsk,
    Tz::Asia__Oral,
    Tz::Asia__Phnom_Penh,
    Tz::Asia__Pontianak,
    Tz::Asia__Pyongyang,
    Tz::Asia__Qatar,
    Tz::Asia__Qostanay,
    Tz::Asia__Qyzylorda,
    Tz::Asia__Rangoon,
    Tz::Asia__Riyadh,
    Tz::Asia__Saigon,
    Tz::Asia__Sakhalin,
    Tz::Asia__Samarkand,
    Tz::Asia__Seoul,
    Tz::Asia__Shanghai,
    Tz::Asia__Singapore,
    Tz::Asia__Srednekolymsk,
    Tz::Asia__Taipei,
    Tz::Asia__Tashkent,
    Tz::Asia__Tbilisi,
    Tz::Asia__Tehran,
    Tz::Asia__Tel_Aviv,
    Tz::Asia__Thimbu,
    Tz::Asia__Thimphu,
    Tz::Asia__Tokyo,
    Tz::Asia__Tomsk,
    Tz::Asia__Ujung_Pandang,
    Tz::Asia__Ulaanbaatar,
    Tz::Asia__Ulan_Bator,
    Tz::Asia__Urumqi,
    Tz::Asia__UstNera,
    Tz::Asia__Vientiane,
    Tz::Asia__Vladivostok,
    Tz::Asia__Yakutsk,
    Tz::Asia__Yangon,
    Tz::Asia__Yekaterinburg,
    Tz::Asia__Yerevan,
    Tz::Atlantic__Azores,
    Tz::Atlantic__Bermuda,
    Tz::Atlantic__Canary,
    Tz::Atlantic__Cape_Verde,
    Tz::Atlantic__Faeroe,
    Tz::Atlantic__Faroe,
    Tz::Atlantic__Jan_Mayen,
    Tz::Atlantic__Madeira,
    Tz::Atlantic__Reykjavik,
    Tz::Atlantic__South_Georgia,
    Tz::Atlantic__St_Helena,
    Tz::Atlantic__Stanley,
    Tz::Australia__ACT,
    Tz::Australia__Adelaide,
    Tz::Australia__Brisbane,
    Tz::Australia__Broken_Hill,
    Tz::Australia__Canberra,
    Tz::Australia__Currie,
    Tz::Australia__Darwin,
    Tz::Australia__Eucla,
    Tz::Australia__Hobart,
    Tz::Australia__LHI,
    Tz::Australia__Lindeman,
    Tz::Australia__Lord_Howe,
    Tz::Australia__Melbourne,
    Tz::Australia__NSW,
    Tz::Australia__North,
    Tz::Australia__Perth,
    Tz::Australia__Queensland,
    Tz::Australia__South,
    Tz::Australia__Sydney,
    Tz::Australia__Tasmania,
    Tz::Australia__Victoria,
    Tz::Australia__West,
    Tz::Australia__Yancowinna,
    Tz::Brazil__Acre,
    Tz::Brazil__DeNoronha,
    Tz::Brazil__East,
    Tz::Brazil__West,
    Tz::CET,
    Tz::CST6CDT,
    Tz::Canada__Atlantic,
    Tz::Canada__Central,
    Tz::Canada__Eastern,
    Tz::Canada__Mountain,
    Tz::Canada__Newfoundland,
    Tz::Canada__Pacific,
    Tz::Canada__Saskatchewan,
    Tz::Canada__Yukon,
    Tz::Chile__Continental,
    Tz::Chile__EasterIsland,
    Tz::Cuba,
    Tz::EET,
    Tz::EST,
    Tz::EST5EDT,
    Tz::Egypt,
    Tz::Eire,
    Tz::Etc__GMT,
    Tz::Etc__GMTPlus0,
    Tz::Etc__GMTPlus1,
    Tz::Etc__GMTPlus10,
    Tz::Etc__GMTPlus11,
    Tz::Etc__GMTPlus12,
    Tz::Etc__GMTPlus2,
    Tz::Etc__GMTPlus3,
    Tz::Etc__GMTPlus4,
    Tz::Etc__GMTPlus5,
    Tz::Etc__GMTPlus6,
    Tz::Etc__GMTPlus7,
    Tz::Etc__GMTPlus8,
    Tz::Etc__GMTPlus9,
    Tz::Etc__GMTMinus0,
    Tz::Etc__GMTMinus1,
    Tz::Etc__GMTMinus10,
    Tz::Etc__GMTMinus11,
    Tz::Etc__GMTMinus12,
    Tz::Etc__GMTMinus13,
    Tz::Etc__GMTMinus14,
    Tz::Etc__GMTMinus2,
    Tz::Etc__GMTMinus3,
    Tz::Etc__GMTMinus4,
    Tz::Etc__GMTMinus5,
    Tz::Etc__GMTMinus6,
    Tz::Etc__GMTMinus7,
    Tz::Etc__GMTMinus8,
    Tz::Etc__GMTMinus9,
    Tz::Etc__GMT0,
    Tz::Etc__Greenwich,
    Tz::Etc__UCT,
    Tz::Etc__UTC,
    Tz::Etc__Universal,
    Tz::Etc__Zulu,
    Tz::Europe__Amsterdam,
    Tz::Europe__Andorra,
    Tz::Europe__Astrakhan,
    Tz::Europe__Athens,
    Tz::Europe__Belfast,
    Tz::Europe__Belgrade,
    Tz::Europe__Berlin,
    Tz::Europe__Bratislava,
    Tz::Europe__Brussels,
    Tz::Europe__Bucharest,
    Tz::Europe__Budapest,
    Tz::Europe__Busingen,
    Tz::Europe__Chisinau,
    Tz::Europe__Copenhagen,
    Tz::Europe__Dublin,
    Tz::Europe__Gibraltar,
    Tz::Europe__Guernsey,
    Tz::Europe__Helsinki,
    Tz::Europe__Isle_of_Man,
    Tz::Europe__Istanbul,
    Tz::Europe__Jersey,
    Tz::Europe__Kaliningrad,
    Tz::Europe__Kiev,
    Tz::Europe__Kirov,
    Tz::Europe__Kyiv,
    Tz::Europe__Lisbon,
    Tz::Europe__Ljubljana,
    Tz::Europe__London,
    Tz::Europe__Luxembourg,
    Tz::Europe__Madrid,
    Tz::Europe__Malta,
    Tz::Europe__Mariehamn,
    Tz::Europe__Minsk,
    Tz::Europe__Monaco,
    Tz::Europe__Moscow,
    Tz::Europe__Nicosia,
    Tz::Europe__Oslo,
    Tz::Europe__Paris,
    Tz::Europe__Podgorica,
    Tz::Europe__Prague,
    Tz::Europe__Riga,
    Tz::Europe__Rome,
    Tz::Europe__Samara,
    Tz::Europe__San_Marino,
    Tz::Europe__Sarajevo,
    Tz::Europe__Saratov,
    Tz::Europe__Simferopol,
    Tz::Europe__Skopje,
    Tz::Europe__Sofia,
    Tz::Europe__Stockholm,
    Tz::Europe__Tallinn,
    Tz::Europe__Tirane,
    Tz::Europe__Tiraspol,
    Tz::Europe__Ulyanovsk,
    Tz::Europe__Uzhgorod,
    Tz::Europe__Vaduz,
    Tz::Europe__Vatican,
    Tz::Europe__Vienna,
    Tz::Europe__Vilnius,
    Tz::Europe__Volgograd,
    Tz::Europe__Warsaw,
    Tz::Europe__Zagreb,
    Tz::Europe__Zaporozhye,
    Tz::Europe__Zurich,
    Tz::GB,
    Tz::GBEire,
    Tz::GMT,
    Tz::GMTPlus0,
    Tz::GMTMinus0,
    Tz::GMT0,
    Tz::Greenwich,
    Tz::HST,
    Tz::Hongkong,
    Tz::Iceland,
    Tz::Indian__Antananarivo,
    Tz::Indian__Chagos,
    Tz::Indian__Christmas,
    Tz::Indian__Cocos,
    Tz::Indian__Comoro,
    Tz::Indian__Kerguelen,
    Tz::Indian__Mahe,
    Tz::Indian__Maldives,
    Tz::Indian__Mauritius,
    Tz::Indian__Mayotte,
    Tz::Indian__Reunion,
    Tz::Iran,
    Tz::Israel,
    Tz::Jamaica,
    Tz::Japan,
    Tz::Kwajalein,
    Tz::Libya,
    Tz::MET,
    Tz::MST,
    Tz::MST7MDT,
    Tz::Mexico__BajaNorte,
    Tz::Mexico__BajaSur,
    Tz::Mexico__General,
    Tz::NZ,
    Tz::NZCHAT,
    Tz::Navajo,
    Tz::PRC,
    Tz::PST8PDT,
    Tz::Pacific__Apia,
    Tz::Pacific__Auckland,
    Tz::Pacific__Bougainville,
    Tz::Pacific__Chatham,
    Tz::Pacific__Chuuk,
    Tz::Pacific__Easter,
    Tz::Pacific__Efate,
    Tz::Pacific__Enderbury,
    Tz::Pacific__Fakaofo,
    Tz::Pacific__Fiji,
    Tz::Pacific__Funafuti,
    Tz::Pacific__Galapagos,
    Tz::Pacific__Gambier,
    Tz::Pacific__Guadalcanal,
    Tz::Pacific__Guam,
    Tz::Pacific__Honolulu,
    Tz::Pacific__Johnston,
    Tz::Pacific__Kanton,
    Tz::Pacific__Kiritimati,
    Tz::Pacific__Kosrae,
    Tz::Pacific__Kwajalein,
    Tz::Pacific__Majuro,
    Tz::Pacific__Marquesas,
    Tz::Pacific__Midway,
    Tz::Pacific__Nauru,
    Tz::Pacific__Niue,
    Tz::Pacific__Norfolk,
    Tz::Pacific__Noumea,
    Tz::Pacific__Pago_Pago,
    Tz::Pacific__Palau,
    Tz::Pacific__Pitcairn,
    Tz::Pacific__Pohnpei,
    Tz::Pacific__Ponape,
    Tz::Pacific__Port_Moresby,
    Tz::Pacific__Rarotonga,
    Tz::Pacific__Saipan,
    Tz::Pacific__Samoa,
    Tz::Pacific__Tahiti,
    Tz::Pacific__Tarawa,
    Tz::Pacific__Tongatapu,
    Tz::Pacific__Truk,
    Tz::Pacific__Wake,
    Tz::Pacific__Wallis,
    Tz::Pacific__Yap,
    Tz::Poland,
    Tz::Portugal,
    Tz::ROC,
    Tz::ROK,
    Tz::Singapore,
    Tz::Turkey,
    Tz::UCT,
    Tz::US__Alaska,
    Tz::US__Aleutian,
    Tz::US__Arizona,
    Tz::US__Central,
    Tz::US__EastIndiana,
    Tz::US__Eastern,
    Tz::US__Hawaii,
    Tz::US__IndianaStarke,
    Tz::US__Michigan,
    Tz::US__Mountain,
    Tz::US__Pacific,
    Tz::US__Samoa,
    Tz::UTC,
    Tz::Universal,
    Tz::WSU,
    Tz::WET,
    Tz::Zulu,
];